/**
 * \file MatrixSolver.cpp
 *
 * \brief Making model matrixes, solving model matrixes.
 *
 * Matrix form of linear equations: A*X[t+1] = B*X[t],
 * where A - model matrix, B - RHS, X[t] - known values of function (PSD), X[t+1] - unknown values of function.
 *
 * In that file there are procedures for making model matrix for 1d-diffusion, 2d-diffusion, some  ideas of 3d-diffusion and mixed terms.
 * Solver for tridiagonal matrix, solver by gauss method and iteration method (upper relaxation).
 *
 * This file is under development, and has a lot of commented code, unfinished etc code.
 *
 * There is a checked version for 1d diffusion (or split method of 2d, 3d diffusions) - 1d_universal_solver.
 *
 * \author Developed under supervision of the PI Yuri Shprits
 */

//#define CPPL_DEBUG
//#define CPPL_VERBOSE

// Memory leaks
#if defined(_WIN32) || defined(_WIN64)
#define _CRTDBG_MAP_ALLOC
#include <crtdbg.h>
#endif

#include "MatrixSolver.h"
#include <math.h>
#include <malloc/malloc.h>
#include <string>
#include <ctime>

// Lapack debugging
//#define CPPL_VERBOSE
//#define CPPL_DEBUG

using namespace std;

// parameters class
#include "../Parameters/Parameters.h"

//#include "PSD.h"
#include "../VariousFunctions/variousFunctions.h"
#include "../Exceptions/error.h"

// Lapack
#include "../Lapack/Include/cpplapack.h"

// gmres
extern "C" {
	#include "../GMRES/itlin.h"
}

/// Used for gmres
void matvec(int n, double *x, double *b);
void preconr(int n, double *x, double *b);
extern "C" double gmres_norm2(int n, double *v);

extern "C" void gmres(int n, double *y, MATVEC *matvec,
                  PRECON *preconr, PRECON *preconl, double *b,
                  struct ITLIN_OPT *opt, struct ITLIN_INFO *info);

#include <iostream>

#include "../GMRES2/GMRES2.h"


/**
 * Make matrix for 1D diffusion.
 */
bool MakeMatrix(double *A,
				double *B1,
				double *C,
				double *R,
				double	*x,							// one dimensional grid
				double	tau,						// lifetime
				double	n,							// power of x
				double	f_lower,					// lower boundary conditions
				double	f_upper,					// upper boundary conditions
				int		nx,							// number of grid points
				double	dt,							// time step
				double	*Dxx,						// diffusion coefficient
				double	taulc,						// lifetime in loss cone
				double	alc,						// loss cone size
				int		g_flag,						// diffusion flag
				string		lower_border_condition_type,// type of lower boundary conditions: 0 - for value, 1 - for derivative
				string		upper_border_condition_type,// type of upper boundary conditions: 0 - for value, 1 - for derivative
				string approximationMethod)
{

	// local variables
	double Coef1, Dm_05, Dp_05, tau_cur;

	// initialisation of temp matrix_arrays
//	double *A = new double[nx];
	double *B = new double[nx];
//	double *C = new double[nx];
//	double *B1 = new double[nx];
	double *h = new double[nx];
	double *h_bar = new double[nx];

	// iterators i, j
	int i, j;
	double coef_05m, coef_05p;

	// making matrix
	for (j = 1; j <= nx-1; j++) {
		h[j] = x[j] - x[j-1];
	}
	for (j = 0; j <= nx-2; j++) {
		h_bar[j] = (h[j] + h[j+1])/2;
	}

	for (j = 1; j <= nx-2; j++) {
		// making model matrix, according to the approximation method
        /**
         * \brief \b AM_Split_LR
         *
         * \f$ Coef1 = \frac{x^n}{\hbar} \f$
         *
         * \f$ coef\_05m = 0.5 * (Dxx * x^{-n} * G(x)) + Dxx_{j-1} * G(x_{j-1}) * {x_{j-1}}^{-n} \f$ \n
         * \f$ coef\_05p = 0.5 * (Dxx * x^{-n} * G(x)) + Dxx_{j+1} * G(x_{j+1}) * {x_{j+1}}^{-n} \f$
         * where VF::G() is used if g_flag is set
         *
         * \f$ A = \frac{Coef1 * coef\_05m}{h} \f$
         *
         * \f$ C = \frac{Coef1 * coef\_05p}{h+1} \f$
         */
		if (approximationMethod == "AM_Split_LR") {
				//Ver1
				Coef1 = pow(x[j], n)/h_bar[j];
				if (g_flag == 1) {
					Coef1 = Coef1/VF::G(x[j]);
					coef_05m = 0.5*(Dxx[j]*VF::G(x[j])*pow(x[j], -n) + Dxx[j-1]*VF::G(x[j-1])*pow(x[j-1], -n));
					coef_05p = 0.5*(Dxx[j]*VF::G(x[j])*pow(x[j], -n) + Dxx[j+1]*VF::G(x[j+1])*pow(x[j+1], -n));
				} else {
					coef_05m = 0.5*(Dxx[j]*pow(x[j], -n) + Dxx[j-1]*pow(x[j-1], -n));
					coef_05p = 0.5*(Dxx[j]*pow(x[j], -n) + Dxx[j+1]*pow(x[j+1], -n));
				}
				A[j] = Coef1 * coef_05m / h[j];
				C[j] = Coef1 * coef_05p / h[j+1];
				// end ver1
		} else if (approximationMethod == "AM_Split_LR_oldway") {
				//Ver1
				Coef1 = pow(x[j], n);
				if (g_flag == 1) {
					Coef1 = Coef1/VF::G(x[j]);
					coef_05m = 0.5*(Dxx[j]*VF::G(x[j])*pow(x[j], -n) + Dxx[j-1]*VF::G(x[j-1])*pow(x[j-1], -n));
					coef_05p = 0.5*(Dxx[j]*VF::G(x[j])*pow(x[j], -n) + Dxx[j+1]*VF::G(x[j+1])*pow(x[j+1], -n));
				} else {
					coef_05m = 0.5*(Dxx[j]*pow(x[j], -n) + Dxx[j-1]*pow(x[j-1], -n));
					coef_05p = 0.5*(Dxx[j]*pow(x[j], -n) + Dxx[j+1]*pow(x[j+1], -n));
				}
				A[j] = Coef1 * coef_05m / h[j]      / h[j+1];
				C[j] = Coef1 * coef_05p / h[j+1]    / h[j];
				// end ver1
		}
        /**
         * \brief \b AM_Split_C
         *
         * \f$ Coef1 = \frac{x^n}{\hbar} \f$
         * 
         * \f$ Dm\_05 = 0.5 * (Dxx_j + Dxx_{j-1}) \f$
         * \f$ Dp\_05 = 0.5 * (Dxx_j + Dxx_{j+1}) \f$
         * 
         * \f$ A = Coef1 * \frac{Dm\_05 * G(\frac{x-h}{2}) * (\frac{x-h}{2})^{-n} }{h} \f$
         *
         * \f$ C = Coef1 * \frac{Dp\_05 * G(\frac{x-(h+1)}{2}) * (\frac{x-(h+1)}{2})^{-n} }{h_{j+1}} \f$
         */
        else if (approximationMethod == "AM_Split_C") {
				//Ver2
				Coef1 = pow(x[j], n)/h_bar[j];
                //diffusion flag on
				if (g_flag == 1) {
					Coef1 = Coef1/VF::G(x[j]);
				}
				Dm_05 = 0.5*(Dxx[j] + Dxx[j-1]);
				Dp_05 = 0.5*(Dxx[j] + Dxx[j+1]);
                //diffusion flag off
				if (g_flag == 0) {
					A[j] = Coef1 * (Dm_05                          * pow(x[j] - h[j]/2, -n) / h[j]);
					C[j] = Coef1 * (Dp_05                          * pow(x[j] + h[j+1]/2, -n) / h[j+1]);
				} else {
					A[j] = Coef1 * (Dm_05 * VF::G(x[j] - h[j]/2)   * pow(x[j] - h[j]/2, -n) / h[j]);
					C[j] = Coef1 * (Dp_05 * VF::G(x[j] + h[j+1]/2) * pow(x[j] + h[j+1]/2, -n) / h[j+1]);
				}
				// end ver2
		} else if (approximationMethod == "AM_Split_C_oldway") {
				//Ver2
				Coef1 = pow(x[j], n)/h_bar[j];
				if (g_flag == 1) {
					Coef1 = Coef1/VF::G(x[j]);
				}
				Dm_05 = 0.5*(Dxx[j] + Dxx[j-1]);
				Dp_05 = 0.5*(Dxx[j] + Dxx[j+1]);
				if (g_flag == 0) {
					A[j] = Coef1 * (Dm_05						   * pow(x[j] - h[j]/2, -n) / h[j]);
					C[j] = Coef1 * (Dp_05						   * pow(x[j] + h[j]/2, -n) / h[j+1]);
				} else {
					A[j] = Coef1 * (Dm_05 * VF::G(x[j] - h[j]/2)   * pow(x[j] - h[j]/2, -n) / h[j]);
					C[j] = Coef1 * (Dp_05 * VF::G(x[j] + h[j]/2)   * pow(x[j] + h[j]/2, -n) / h[j+1]);
				}
				// end ver2
		} else {
			throw error_msg("APPRERR", "Unknown approximation type: %d", approximationMethod.c_str());
		}

		B[j] = -A[j] - C[j];

		//if (x[j] < (alc * pi / 180) && (g_flag == 1)) {
		// Loss cone close to 0 and 180 deg pitch-angle
		if ((x[j] < alc || x[j] > VC::pi - alc) && (g_flag == 1)) {
			tau_cur = taulc;
		} else {
			tau_cur = tau;
		}

		B1[j] = B[j] - (1.0/dt + 1/tau_cur);
	}

    //calculate inverse of timesteps
	for (i = 1; i <= nx-2; i++) {
		R[i] = - 1.0 / dt;
	}

	// Using boundary conditions.
	// lower boundary
	i = 0;
	R[i] = f_lower;
	if (lower_border_condition_type == "BCT_CONSTANT_VALUE") { // for condition on value
		A[i] = 0; B1[i] = 1; C[i] = 0;
	} else if (lower_border_condition_type == "BCT_CONSTANT_DERIVATIVE") { // for condition on derivative
		A[i] = 0; B1[i] = -1; C[i] = 1;
	} else {
		throw error_msg("1D_DIFF_BOUNDARY", "1d_universal_solver: unknown lower boundary type: %d", lower_border_condition_type.c_str());
		//log1::fcout << "1d_universal_solver: unknown lower boundary type: " << lower_border_condition_type << "\n";
		//return -1;
	}

	// upper boundary
	i = nx-1;
	R[i] = f_upper;
	if (upper_border_condition_type == "BCT_CONSTANT_VALUE") {
		A[i] = 0; B1[i] = 1; C[i] = 0;
	} else if (upper_border_condition_type == "BCT_CONSTANT_DERIVATIVE") {
		A[i] = -1; B1[i] = 1; C[i] = 0;
	} else {
		throw error_msg("1D_DIFF_BOUNDARY", "1d_universal_solver: unknown lower boundary type: %d", upper_border_condition_type.c_str());
		//log1::fcout << "1d_universal_solver: unknown upper boundary type: " << upper_border_condition_type << "\n";
		//return -1;
	}

	// solving matrix
//	tridag(A, B1, C, R, f, nx);

	// deleting matrix_arrays
//	delete A;
	delete B;
//	delete C;
//	delete B1;
	delete h;
	delete h_bar;

	return true;
}

/**
 *  Solver for 1d-diffusion matrix (tridiagonal).
 */
bool SolveMatrix(double	*f,							// phase space density function
				 double *A,
				 double *B1,
				 double *C,
				 double *R,
				 double	f_lower,					// lower boundary conditions
				 double	f_upper,					// upper boundary conditions
				 int	nx,							// number of grid points
				 double	dt)							// time step
{
	// initialisation of temp matrix_arrays
	double *R1 = new double[nx];

	// iterator i
	int i;

	for (i = 1; i <= nx-2; i++) {
		R1[i] = R[i] * f[i];
	}

	// Using borders conditions.
	// lower boundary
	i = 0;
	R1[i] = R[i];

	// upper boundary
	i = nx-1;
	R1[i] = R[i];

	// solving matrix
	tridag(A, B1, C, R1, f, nx);

	// deleting matrix_arrays
	delete R1;

	return true;
}


/** Supportive sub-function to add boundary conditions to model matrix
 * \param &Matrix_A - model matrix to be updated
 * \param type - type of boundary condition
 * \param in - second index of matrix
 * \param id1 - upper boundary index
 * \param dh - derivative value?
 */
void AddBoundary(DiagMatrix &Matrix_A, string type, int in, int id1, double dh) {
	int id0 = 0;
				if (type == "BCT_CONSTANT_VALUE") { // for condition on value
					Matrix_A[id0][in] = 1;
					Matrix_A[id1][in] = 0;
				} else if (type == "BCT_CONSTANT_DERIVATIVE") { // for condition on derivative
					Matrix_A[id0][in] = -1/dh;
					Matrix_A[id1][in] = 1/dh;
				} else {
					throw error_msg("2D_DIFF_BOUNDARY", "1d_universal_solver: unknown boundary type: %d", type.c_str());
				}
}

/**
 * Create matrix form of the Fokker-Planck equation: matr_A * PSD(t+1) = matr_B * PSD(t) + matr_C
 *
 * Calculation matr_A, matr_B, and matr_C, i.e. numerical approximation of the derivatives
 * Takes boundary conditions and diffusion coefficients as an input.
 * L, pc, alpha should be orthogonal for 3D!!!
 *
 * \param &matr_A - matrix A
 * \param &matr_B - matrix B
 * \param &matr_C - matrix C
 * \param &L - matrix for L
 * \param &pc - matrix for pc
 * \param &alpha - matrix for alpha
 * \param L_size - size of L
 * \param pc_size - size of pc
 * \param alpha_size - size of alpha
 * \param &L_lowerBoundaryCondition - lower boundary condition for L
 * \param &L_upperBoundaryCondition - upper boundary condition for L
 * \param &pc_lowerBoundaryCondition - lower boundary condition for pc
 * \param &pc_upperBoundaryCondition - upper boundary condition for pc
 * \param &alpha_lowerBoundaryCondition - lower boundary condition for alpha
 * \param &alpha_upperBoundaryCondition - upper boundary condition for alpha
 * \param L_lowerBoundaryCondition_calculationType - calculation type of L's lower boundary condition
 * \param L_upperBoundaryCondition_calculationType - calculation type of L's upper boundary condition
 * \param pc_lowerBoundaryCondition_calculationType - calculation type of pc's lower boundary condition
 * \param pc_upperBoundaryCondition_calculationType - calculation type of pc's upper boundary condition
 * \param alpha_lowerBoundaryCondition_calculationType - calculation type of alpha's lower boundary condition
 * \param alpha_upperBoundaryCondition_calculationType - calculation type of alpha's upper boundary condition
 * \param &DLL - matrix of diffusion coefficients for L
 * \param &Dpcpc - diffusion coefficients group for Dpcpc
 * \param &DpcpcLpp - diffusion coefficients group for DpcpcLpp
 * \param &Daa - diffusion coefficients group for Daa
 * \param &DaaLpp - diffusion coefficients group for DaaLpp
 * \param &Dpca - diffusion coefficients group for Dpca
 * \param &DpcaLpp - diffusion coefficients group for DpcaLpp
 * \param &Jacobian - Jacobian values
 * \param dt - diffusion time(?)
 * \param Lpp - plasma pause location
 * \param tau - lifetime outside of plasmasphere
 * \param tauLpp - lifetime inside of plasmasphere
 */
bool MakeModelMatrix_3D(
				  CalculationMatrix &matr_A, CalculationMatrix &matr_B, CalculationMatrix &matr_C,
				  Matrix3D<double> &L, Matrix3D<double> &pc, Matrix3D<double> &alpha,
				  int L_size, int pc_size, int alpha_size,
				  Matrix2D<double> &L_lowerBoundaryCondition,
				  Matrix2D<double> &L_upperBoundaryCondition,
				  Matrix2D<double> &pc_lowerBoundaryCondition,
				  Matrix2D<double> &pc_upperBoundaryCondition,
				  Matrix2D<double> &alpha_lowerBoundaryCondition,
				  Matrix2D<double> &alpha_upperBoundaryCondition,
				  string L_lowerBoundaryCondition_calculationType,
				  string L_upperBoundaryCondition_calculationType,
				  string pc_lowerBoundaryCondition_calculationType,
				  string pc_upperBoundaryCondition_calculationType,
				  string alpha_lowerBoundaryCondition_calculationType,
				  string alpha_upperBoundaryCondition_calculationType,
				  Matrix3D<double> &DLL,
				  Matrix3D<double> &Dpcpc, Matrix3D<double> &DpcpcLpp,
				  Matrix3D<double> &Daa, Matrix3D<double> &DaaLpp,
				  Matrix3D<double> &Dpca, Matrix3D<double> &DpcaLpp,
				  Matrix3D<double> &Jacobian, double dt, double Lpp,
				  double tau, double tauLpp) {

	// variables, for future use
	// il - L-index, im - pc-index, ia - alpha-index,
	// id - index of a diagonal of the model matrix (counting from main diagonal, main is zero),
	// in - model matrix line number
	int il, im, ia, id, in;

	// create A-matrix, B-matrix, C-matrix
	// Output::echo("Recalculating model matrix... ");

	// make everything zero
	// TODO: make zero not everything, but existing diagonals
	/* !!! matr_C[0] = 0;
	for (il = -1; il <= 1; il++) {
		for (im = -1; im <= 1; im++) {
			for (ia = -1; ia <= 1; ia++) {
				// calculating a diagonal number (id)
				id = matr_A.index1d(ia, im, il);
				// make the diagonal zero
				matr_A[id] = 0;
				matr_B[id] = 0;
			}
		}
	}*/

	// Make diagonals to be equal to zero
	DiagMatrix::iterator it;
	for (it = matr_A.begin(); it != matr_A.end(); it++) it->second = 0;
	for (it = matr_B.begin(); it != matr_B.end(); it++) it->second = 0;
	for (it = matr_C.begin(); it != matr_C.end(); it++) it->second = 0;

	// create a new model matrix
	// (f^{t+1} - f^{t})/dt = L1(f^{t+1}) + L2(f^{t+1}) + L3(f^{t+1})[main equation, losses are calculated separately]
	// L1, L2, L3 - diffusion operators
	//  + need to add there the equations for boundary conditions also
	double dh;
	for (il = 0; il < L_size; il++) {
		for (im = 0; im < pc_size; im++) {
			for (ia = 0; ia < alpha_size; ia++) {
				// calculating current line number (in)
				in = matr_A.index1d(ia, im, il);

				// Describe all boundaries first. The 'inner' area will be approximated later.
				// If L_size > 1 we should use L-boundaries, right?
				// If not, then just calculate diffusion
				if (il == 0 && L_size >= 3) {
					// Boundary conditions
					// Points at the corners are used only at one of the boundaries
					// L-bottom boundary
					matr_C[0][in] = L_lowerBoundaryCondition[im][ia];
					id = matr_A.index1d(ia, im, il+1) - in;
					dh = L[il+1][im][ia] - L[il][im][ia];
					AddBoundary(matr_A, L_lowerBoundaryCondition_calculationType, in, id, dh);

				} else if (il == L_size-1 && L_size >= 3) {
					// L-top boundary
					matr_C[0][in] = L_upperBoundaryCondition[im][ia];
					id = matr_A.index1d(ia, im, il-1) - in;
					dh = L[il][im][ia] - L[il-1][im][ia];
					AddBoundary(matr_A, L_upperBoundaryCondition_calculationType, in, id, dh);

				} else if (im == 0 && pc_size >= 3) {
					// pc-low boundary
					matr_C[0][in] = pc_lowerBoundaryCondition[il][ia];
					id = matr_A.index1d(ia, im+1, il) - in;
					dh = pc[il][im+1][ia] - pc[il][im][ia];
					AddBoundary(matr_A, pc_lowerBoundaryCondition_calculationType, in, id, dh);

				} else if (im == pc_size-1 && pc_size >= 3) {
					// pc-high boundary
					matr_C[0][in] = pc_upperBoundaryCondition[il][ia];
					id = matr_A.index1d(ia, im-1, il) - in;
					dh = pc[il][im][ia] - pc[il][im-1][ia];
					AddBoundary(matr_A, pc_upperBoundaryCondition_calculationType, in, id, dh);
				} else if (ia == 0 && alpha_size >= 3) {
					// alpha-loss cone boundary
					matr_C[0][in] = alpha_lowerBoundaryCondition[il][im];
					id = matr_A.index1d(ia+1, im, il) - in;
					dh = alpha[il][im][ia+1] - alpha[il][im][ia];
					AddBoundary(matr_A, alpha_lowerBoundaryCondition_calculationType, in, id, dh);

				} else if (ia == alpha_size-1 && alpha_size >= 3) {
					// alpha-equatorial boundary
					matr_C[0][in] = alpha_upperBoundaryCondition[il][im];
					id = matr_A.index1d(ia-1, im, il) - in;
					dh = alpha[il][im][ia] - alpha[il][im][ia-1];
					AddBoundary(matr_A, alpha_upperBoundaryCondition_calculationType, in, id, dh);

				} else {

					// now we are sure we are not on a boundary, can do the Fokker-Planck equation approximation in the inner area

					// f^{t+1}/dt
					matr_A[0][in] += 1.0/dt;
					// f^{t}/dt
					matr_B[0][in] += 1.0/dt;

					// calculate radial diffusion, if L_size >= 3
					if (L_size >=3) {
						// L term, add approximated derivatives into the calculation matrix
						SecondDerivativeApproximation_3D(matr_A, il, im, ia,
							"DT_L_left", "DT_L_right",
							L, pc, alpha, DLL, Jacobian, -0.5);
						SecondDerivativeApproximation_3D(matr_A, il, im, ia,
							"DT_L_right", "DT_L_left",
							L, pc, alpha, DLL, Jacobian, -0.5);
					}

					if (pc_size >=3) {
						// Energy term
						if (L[il][im][ia] >= Lpp) {
							SecondDerivativeApproximation_3D(matr_A, il, im, ia,
								"DT_PC_left", "DT_PC_right",
								L, pc, alpha, Dpcpc, Jacobian, -0.5);
							SecondDerivativeApproximation_3D(matr_A, il, im, ia,
								"DT_PC_right", "DT_PC_left",
								L, pc, alpha, Dpcpc, Jacobian, -0.5);
						}  else  {
							SecondDerivativeApproximation_3D(matr_A, il, im, ia,
								"DT_PC_left", "DT_PC_right",
								L, pc, alpha, DpcpcLpp, Jacobian, -0.5);
							SecondDerivativeApproximation_3D(matr_A, il, im, ia,
								"DT_PC_right", "DT_PC_left",
								L, pc, alpha, DpcpcLpp, Jacobian, -0.5);
						}
					}

					if (alpha_size >=3) {
						// Pitch angle term(s)

						// Loss cone.
						// Warning! Can cause the loss cone to be added during each diffusion, but
						// now it's ok, cause alpha_size >= 3 only for pitch-angle diffusion
						// f^{n+1}/taulc


						if ( (alpha[il][im][ia] <= VF::alc(L[il][im][0]) || alpha[il][im][ia] >= VC::pi - VF::alc(L[il][im][0]))) {
							double taulc;
							//double taulc = VF::bounce_time(VF::Kfunc(pc[il][im][ia]), L[il][im][ia])/60./60./24.;
							// Quarter bounce time
							taulc = 0.25 * VF::bounce_time_new(L[il][im][ia], pc[il][im][ia], alpha[il][im][ia]);
							matr_A[0][in] += 1.0/taulc;  // Implicit
							//matr_B[0][in] -= 1.0/taulc; // Explicit
						}

						if (L[il][im][ia] >= Lpp) {
							SecondDerivativeApproximation_3D(matr_A, il, im, ia,
								"DT_ALPHA_left", "DT_ALPHA_right",
								L, pc, alpha, Daa, Jacobian, -0.5);
							SecondDerivativeApproximation_3D(matr_A, il, im, ia,
								"DT_ALPHA_right", "DT_ALPHA_left",
								L, pc, alpha, Daa, Jacobian, -0.5);
						}  else  {
							SecondDerivativeApproximation_3D(matr_A, il, im, ia,
								"DT_ALPHA_left", "DT_ALPHA_right",
								L, pc, alpha, DaaLpp, Jacobian, -0.5);
							SecondDerivativeApproximation_3D(matr_A, il, im, ia,
								"DT_ALPHA_right", "DT_ALPHA_left",
								L, pc, alpha, DaaLpp, Jacobian, -0.5);
						}
					}

					if (alpha_size >= 3 && pc_size >= 3) {
						// Mixed pitch angle and energy term
						// There are several possibilities for approximation
						if (L[il][im][ia] >= Lpp) {

							// d/dp D d/da f
							SecondDerivativeApproximation_3D(matr_A, il, im, ia,
								"DT_ALPHA_left", "DT_PC_right",
								L, pc, alpha, Dpca, Jacobian, -0.25);
							SecondDerivativeApproximation_3D(matr_A, il, im, ia,
								"DT_ALPHA_right", "DT_PC_left",
								L, pc, alpha, Dpca, Jacobian, -0.25);

							// d/da D d/dp f
							SecondDerivativeApproximation_3D(matr_A, il, im, ia,
								"DT_PC_left", "DT_ALPHA_right",
								L, pc, alpha, Dpca, Jacobian, -0.25);
							SecondDerivativeApproximation_3D(matr_A, il, im, ia,
								"DT_PC_right", "DT_ALPHA_left",
								L, pc, alpha, Dpca, Jacobian, -0.25);

							// d/dp D d/da f
							SecondDerivativeApproximation_3D(matr_A, il, im, ia,
								"DT_ALPHA_left", "DT_PC_left",
								L, pc, alpha, Dpca, Jacobian, -0.25);
							SecondDerivativeApproximation_3D(matr_A, il, im, ia,
								"DT_ALPHA_right", "DT_PC_right",
								L, pc, alpha, Dpca, Jacobian, -0.25);

							// d/da D d/dp f
							SecondDerivativeApproximation_3D(matr_A, il, im, ia,
								"DT_PC_left", "DT_ALPHA_left",
								L, pc, alpha, Dpca, Jacobian, -0.25);
							SecondDerivativeApproximation_3D(matr_A, il, im, ia,
								"DT_PC_right", "DT_ALPHA_right",
								L, pc, alpha, Dpca, Jacobian, -0.25);

						}  else  {

							// d/dp D d/da f
							SecondDerivativeApproximation_3D(matr_A, il, im, ia,
								"DT_ALPHA_left", "DT_PC_right",
								L, pc, alpha, DpcaLpp, Jacobian, -0.25);
							SecondDerivativeApproximation_3D(matr_A, il, im, ia,
								"DT_ALPHA_right", "DT_PC_left",
								L, pc, alpha, DpcaLpp, Jacobian, -0.25);
							SecondDerivativeApproximation_3D(matr_A, il, im, ia,
								"DT_ALPHA_left", "DT_PC_left",
								L, pc, alpha, DpcaLpp, Jacobian, -0.25);
							SecondDerivativeApproximation_3D(matr_A, il, im, ia,
								"DT_ALPHA_right", "DT_PC_right",
								L, pc, alpha, DpcaLpp, Jacobian, -0.25);


							// d/da D d/dp f
							SecondDerivativeApproximation_3D(matr_A, il, im, ia,
								"DT_PC_left", "DT_ALPHA_right",
								L, pc, alpha, DpcaLpp, Jacobian, -0.25);
							SecondDerivativeApproximation_3D(matr_A, il, im, ia,
								"DT_PC_right", "DT_ALPHA_left",
								L, pc, alpha, DpcaLpp, Jacobian, -0.25);
							SecondDerivativeApproximation_3D(matr_A, il, im, ia,
								"DT_PC_left", "DT_ALPHA_left",
								L, pc, alpha, DpcaLpp, Jacobian, -0.25);
							SecondDerivativeApproximation_3D(matr_A, il, im, ia,
								"DT_PC_right", "DT_ALPHA_right",
								L, pc, alpha, DpcaLpp, Jacobian, -0.25);

						}
					}
				}
			}
		}
	}
	// Output::echo("recalculated.\n");

	// save the time of matrix change
	matr_A.change_ind = clock();
	matr_B.change_ind = clock();
	matr_C.change_ind = clock();

	return true;
}

// GMRES stuff starts from here

//static double diag = 6.01;
static CalculationMatrix A_copy;

/**
 * Vector to matrix multiplication for GMRES
 */
void matvec(int m_size, double *x, double *b)
{
	DiagMatrix::iterator it;
	int in, ik;
	for (in = 0; in < m_size; in++) {
		// cout << in << ": ";
		b[in] = 0;
		// for (it = A_copy.begin(); ( (ik = in + it->first) < m_size && it != A_copy.end()); it++) { // 1 - 43 sec
		for (it = A_copy.begin(); (it != A_copy.end()); it++) { // 2 - 46 sec
			// it->first is diagonal number
			// it->second is the pointer to diagonal array
			// First I check, if the element at line (modelMatrixLineNumber) and diagonal (it->first) is inside the matrix.
			// Cause it can be outside, if line number is 1 and diagonal number is -5, for example.
			ik = in + it->first; // column number
			if (ik >= 0  && ik < m_size) {
			// if (in + it->first >= 0  && in + it->first < m_size) {  // 2 - 46 sec
			// if (in >= -it->first) {
			// if (ik >= 0 ) { // 1 - 43 sec
				// multiply
				// cout << it->second[in] << " * " << x[in + it->first] << " + ";
				b[in] += it->second[in] * x[ik];
				// b[in] += A_copy[ik - in][in] * x[ik];
			}
		}
		// cout << " = " << b[in] << endl;
	}
}

/*
 * A pointer to the right preconditioner user routine, which computes
 * w=B_r*z, where B_r should be an approximation of the inverse of the
 * matrix A. If a null pointer is supplied, then a dummy preconditioner
 * routine will be used which realizes the preconditioner matrix
 * B_r=identity.
 *   int     n     input  Number of vector components.
 *   double *z     input  Preconditioned iterate, of size n .
 *   double *w     output Vector which holds the matrix-vector product B_r*z,
 *                       i.e. the original iterate. */

/**
 * Jacobi (or diagonal) preconditioner:
 */
void jacobi(int m_size, double *z, double *w)
{
	int in;
	for (in = 0; in < m_size; in++) {
		if (A_copy[0][in] != 0) {
			w[in] = z[in] / A_copy[0][in];
		} else {
			w[in] = z[in];
		}
	}
}

/**
 * SOR preconditioner:
 *
 * P = (D/rel + L)
 * w = P^-1 * z;
 * P * w = z
 * (D/rel + L) * w = z
 */
void SOR(int m_size, double *z, double *w)
{
	int in;
	DiagMatrix::iterator it;
	double diag;
	double rel = 1; // relaxation parameter - between >0 and <2
	for (in = 0; in < m_size; in++) {
		diag = A_copy[0][in] / rel;
		w[in] = z[in] / diag;
		// Lower triangular part of A
		for (it = A_copy.begin(); (it->first < 0 && it != A_copy.end()); it++) {
			if (in + it->first >= 0 && it->first < 0) { // double (redundant) check
				// cout << (in + it->first) << ", " << it->second[in] << ", " << w[in + it->first] << endl;
				w[in] -= w[in + it->first] * it->second[in] / diag;
			}
		}
	}
}

/**
 * Simple preconditioner
 */
void for_norm_A(int n, double *x, double *b)
{
	int j;
	for (j=0;j<n;j++) b[j] = x[j];
}


/**
 * GMRES inversion.
 * Potentially with bugs.
 *
 * A * X = B - equation
 */
//void gmres_wrapout(CalculationMatrix &A, Matrix1D<double> &B, Matrix1D<double> &X, int maxiter, int i_max, double max_err) {
void gmres_wrapout(CalculationMatrix &A, Matrix1D<double> &B, Matrix1D<double> &X, Parameters_structure::PSD::GMRES_parameters_structure GMRES_parameters) {
	int m_size = A[0].size_x;
	// int maxiter = 1000;

	int in;
	//double sum = 0;
	//double b_norm;
	DiagMatrix::iterator it;

	// variables to store time
	std::clock_t start_time, current_time;

	double err_max = 0.;
	double error;
/*
	for (in = 0; in < m_size; in++) {
		// multiplying vector to matrix and calculating the difference
		error = B[in];
		for (it = A.begin(); it != A.end(); it++) {
			if (in + it->first >= 0 && in + it->first < m_size) {
				error -= it->second[in] * X[in + it->first];
			}
		}
		err_max = max(err_max, fabs(error));
	}
	Output::echo(0, "Error_norm = %e \n", err_max);
	err_max = 0;
*/
	A_copy = A;
	static Matrix1D<double> B_copy;
	B_copy = B;
/*
	A_copy.writeToFile("./Debug/A_copy.dat");
	B_copy.writeToFile("./Debug/B_copy.dat");
*/
	// normalization A
	if (GMRES_parameters.use_normalization) {
		double A0;
		for (in = 0; in < m_size; in++) {
			A0 = A_copy[0][in];
			B_copy[in] = B_copy[in] / A0;
			for (it = A_copy.begin(); it != A_copy.end(); it++) {
				it->second[in] = it->second[in]/A0;
			}
		}
	}

	// normalization of B
	/* for (in = 0; in < m_size; in++) {
		sum += B_copy[in] * B_copy[in];
	}
	b_norm = sqrt(sum);
	cout << b_norm << endl;

	for (in = 0; in < m_size; in++) {
		B_copy[in] = B_copy[in]/b_norm;
		for (it = A_copy.begin(); it != A_copy.end(); it++) {
			it->second[in] = it->second[in] / b_norm;
		}
	}

	A_copy.writeToFile("./Debug/A_copy.dat");
	B_copy.writeToFile("./Debug/B_copy.dat");

	A.writeToFile("./Debug/A_check.dat"); */

	// struct ITLIN_OPT   *opt = malloc(sizeof(struct ITLIN_OPT));
	// struct ITLIN_INFO *info = malloc(sizeof(struct ITLIN_INFO));

	struct ITLIN_OPT   *opt   = new struct ITLIN_OPT;
	struct ITLIN_INFO *info   = new struct ITLIN_INFO;

	TERM_CHECK termcheck = CheckOnRestart;//CheckEachIter;

	/* initialize solver options, see gmres.c for help */
	opt->tol = GMRES_parameters.SOL_max_iter_err; //max_err;
	opt->i_max = GMRES_parameters.SOL_i_max;//i_max;
	opt->termcheck = termcheck;
	opt->maxiter = GMRES_parameters.SOL_maxiter;//maxiter;
	opt->errorlevel = None;//Verbose;
	opt->monitorlevel = None;//Verbose;
	opt->datalevel = None;//Verbose;
	opt->errorfile = stdout;
	opt->monitorfile = stdout;
	opt->datafile = NULL; //fopen("test_gmres.data","w");
	/*if (!opt->datafile)
		fprintf(stdout, "\n open of test_gmres.data failed!\n");*/
	opt->iterfile = NULL;//fopen("test_gmres_iter.data","w");
	opt->resfile  = NULL;//fopen("test_gmres_res.data","w");
	opt->miscfile = NULL;//fopen("test_gmres_misc.data","w");

	start_time = std::clock();
	Output::echo(5, "Inverting... \n");

	if (GMRES_parameters.preconditioner_type == "none")
		gmres(m_size, &X[0], &matvec, NULL, NULL, &B_copy[0], opt, info);
	else if (GMRES_parameters.preconditioner_type == "jacobi")
		gmres(m_size, &X[0], &matvec, &jacobi, NULL, &B_copy[0], opt, info);
	else if (GMRES_parameters.preconditioner_type == "SOR")
		gmres(m_size, &X[0], &matvec, &SOR, NULL, &B_copy[0], opt, info);
	else if (GMRES_parameters.preconditioner_type == "for_norm_A")
		gmres(m_size, &X[0], &matvec, &for_norm_A, NULL, &B_copy[0], opt, info);
	else
		throw error_msg("GMRES", "Unknown preconditioner type: %s", GMRES_parameters.preconditioner_type.c_str());

	current_time = std::clock();
	Output::echo(5, "Done (%.2f sec)\n", (( current_time - start_time ) / (double)CLOCKS_PER_SEC ));


	//fclose(opt->datafile);
	//fclose(opt->iterfile);
	//fclose(opt->miscfile);
	fprintf(stdout,"\n Return code:                  %6i\n",info->rcode);
	fprintf(stdout," Iterations:                   %6i\n",info->iter);
	fprintf(stdout," Matvec calls:                 %6i\n",info->nomatvec);
	fprintf(stdout," Right Precon calls:           %6i\n",info->noprecr);
	fprintf(stdout," Left Precon calls:            %6i\n",info->noprecl);
	fprintf(stdout," Estimated residual reduction: %e\n",info->precision);

	error = 0; err_max = 0;
	for (in = 0; in < m_size; in++) {
		// multiplying vector to matrix and calculating the difference
		error = B[in];
		for (it = A.begin(); it != A.end(); it++) {
			if (in + it->first >= 0 && in + it->first < m_size) {
				error -= it->second[in] * X[in + it->first];
			}
		}
		err_max = max(err_max, fabs(error));
	}
	Output::echo(2, "Error_norm = %e \n", err_max);
	err_max = 0;

	delete opt;
	delete info;

}

/**
 * \brief Used for calculating matrix inversion with GMRES2
 */
class Preconditioner_jacobi {
	const CPPL::dgbmatrix *A;
public:

    /// constructor with matrix
	Preconditioner_jacobi(CPPL::dgbmatrix *A){
		//this->A.copy(A);
		this->A = A;
	}

	/// APPROXIMATELY solves M * X = RHS for X.
	//const Matrix1D<double>& solve(const Matrix1D<double> &RHS) const;
	CPPL::dcovector solve(const CPPL::dcovector &RHS) const {
		int in, m_size = RHS.l;
		CPPL::dcovector res(m_size);
		for (in = 0; in < m_size; in++) {
			if ((*A)(in, in) != 0) {
				// divide to diagonal element
				res(in) = RHS(in) / (*A)(in, in);
			} else {
				res(in) = RHS(in);
			}
		}
		return res;
	}
};

/**
 * \brief Used for calculating matrix inversion with GMRES2
 */
class Preconditioner_SOR {
	const CPPL::dgbmatrix *A;
public:

    /// constructor with matrix
	Preconditioner_SOR(CPPL::dgbmatrix *A){
		//this->A.copy(A);
		this->A = A;
	}

	/// APPROXIMATELY solves M * X = RHS for X.
	CPPL::dcovector solve(const CPPL::dcovector &RHS) const {
		int in, it, m_size = RHS.l;
		CPPL::dcovector res(m_size);
		double diag;
		double rel = 1; // relaxation parameter - between >0 and <2
		for (in = 0; in < m_size; in++) {
			diag = (*A)(in, in) / rel;
			res(in) = RHS(in) / diag;
			// Lower triangular part of A
			//for (it = A_copy.begin(); (it->first < 0 && it != A_copy.end()); it++) {
			for (it = (*A).kl; it <= (*A).ku; it++) {
				//if (in + it->first >= 0 && it->first < 0) { // ???
					res(in) -= res(it) * (*A)(in, it) / diag;
				//}
			}
		}
		return res;
	}
};

/**
 * \brief Used for calculating matrix inversion with GMRES2
 */
void gmres2_wrapout(CalculationMatrix &A, Matrix1D<double> &B, Matrix1D<double> &X, Parameters_structure::PSD::GMRES_parameters_structure GMRES_parameters) {
	int i;

	// Matrix2D<double> H(GMRES_parameters.SOL_i_max, GMRES_parameters.SOL_i_max);
	CPPL::dgematrix H_lp(GMRES_parameters.SOL_i_max+1, GMRES_parameters.SOL_i_max+1);

	CPPL::dcovector X_lp(X.size_x);
	// copy X
	for (i = 0; i < X.size_x; i++) X_lp(i) = X(i);

	CPPL::dcovector B_lp(B.size_x);
	// copy B
	for (i = 0; i < B.size_x; i++) B_lp(i) = B(i);

	long m_size = A[0].size_x;
	DiagMatrix::iterator it;
	it = A.begin();
	long kl = -it->first; // first diagonal
	it = A.end();
	it--;
	long ku = it->first;  // last diagonal
	CPPL::dgbmatrix A_lp(m_size, m_size, kl, ku); // it's our matrix A

	// zero A
	int modelMatrixLineNumber;
	for (modelMatrixLineNumber = 0; modelMatrixLineNumber < m_size; modelMatrixLineNumber++) {
		for (int modelMatrixDiagNumber = -kl; modelMatrixDiagNumber <= ku; modelMatrixDiagNumber++) {
			if (modelMatrixDiagNumber + modelMatrixLineNumber >= 0 && modelMatrixDiagNumber + modelMatrixLineNumber < m_size) {
				A_lp(modelMatrixLineNumber, modelMatrixDiagNumber + modelMatrixLineNumber) = 0;
			}
		}
	}

	// Fast-Copy values into the lapack library matrices
	Output::echo(5, "Copying values...\n");
	for (modelMatrixLineNumber = 0; modelMatrixLineNumber < m_size; modelMatrixLineNumber++) {
		for (it = A.begin(); it != A.end(); it++) {
			// check, if the element at line (modelMatrixLineNumber) and diagonal (it->first) is inside the matrix.
			if (modelMatrixLineNumber + it->first >= 0 && modelMatrixLineNumber + it->first < m_size) {
				// converting matrix, stored as diagonals, into lapack matrix (also diagonal)
				A_lp(modelMatrixLineNumber, modelMatrixLineNumber + it->first) = it->second[modelMatrixLineNumber];
			}
		}
	}

	Output::echo(3, "Inverting...\n");
	int iter = GMRES_parameters.SOL_maxiter;
	double tol = GMRES_parameters.SOL_max_iter_err;

	if (GMRES_parameters.preconditioner_type == "jacobi") {
		Preconditioner_jacobi M(&A_lp);
		GMRES(A_lp, X_lp, B_lp,
				  M, H_lp, GMRES_parameters.SOL_i_max, iter,
				  tol);
	} else if (GMRES_parameters.preconditioner_type == "SOR") {
		Preconditioner_SOR M(&A_lp);
		GMRES(A_lp, X_lp, B_lp,
				  M, H_lp, GMRES_parameters.SOL_i_max, iter,
				  tol);
	} else {
		throw error_msg("GMRES", "Unknown preconditioner type: %s", GMRES_parameters.preconditioner_type.c_str());
	}
	Output::echo(3, "Done: iter %d, err %e\n", iter, tol);

	/*GMRES(A, X, B,
	      M, H, GMRES_parameters.SOL_i_max, GMRES_parameters.SOL_maxiter,
	      GMRES_parameters.SOL_max_iter_err);*/

	// save result into X
	Output::echo(5, "Coping back... ");
	for (modelMatrixLineNumber = 0; modelMatrixLineNumber < m_size; modelMatrixLineNumber++) {
		X[modelMatrixLineNumber] = X_lp(modelMatrixLineNumber);
	}

}


/**
 * Lapack inversion.
 *
 * A * X = B - equation
 */
void Lapack(DiagMatrix &A, Matrix1D<double> &B, Matrix1D<double> &X) {

	// variables to store time
	std::clock_t start_time, current_time;

	// iterator for diagonals of the diagonal matrix
	DiagMatrix::iterator it;

	// model matrix line number
	int modelMatrixLineNumber;

	long i;
	long m_size = A[0].size_x;
	it = A.begin();
	long kl = -it->first; // first diagonal
	it = A.end();
	it--;
	long ku = it->first;  // last diagonal

	// Lapack library diagonal matrices
	Output::echo(5, "Creating matrixes...\n");
	CPPL::dgbmatrix lap_AB(m_size, m_size, kl, ku); // it's our matrix A
	CPPL::dgbmatrix lap_AB_res(m_size, m_size, kl, ku);

	// !!! Make zero
	for (modelMatrixLineNumber = 0; modelMatrixLineNumber < m_size; modelMatrixLineNumber++) {
		for (int modelMatrixDiagNumber = -kl; modelMatrixDiagNumber <= ku; modelMatrixDiagNumber++) {
			if (modelMatrixDiagNumber + modelMatrixLineNumber >= 0 && modelMatrixDiagNumber + modelMatrixLineNumber < m_size) {
				lap_AB(modelMatrixLineNumber, modelMatrixDiagNumber + modelMatrixLineNumber) = 0;
			}
		}
	}


	// Copy values into the lapack library matrices
	Output::echo(5, "Copying values...\n");
	for (modelMatrixLineNumber = 0; modelMatrixLineNumber < m_size; modelMatrixLineNumber++) {
		for (it = A.begin(); it != A.end(); it++) {
			// it->first is diagonal number
			// it->second is the pointer to diagonal array
			// First I check, if the element at line (modelMatrixLineNumber) and diagonal (it->first) is inside the matrix.
			// Cause it can be outside, if line number is 1 and diagonal number is -5, for example.
			if (modelMatrixLineNumber + it->first >= 0 && modelMatrixLineNumber + it->first < m_size) {
				// converting matrix, stored as diagonals, into lapack matrix (also diagonal)
				lap_AB(modelMatrixLineNumber, modelMatrixLineNumber + it->first) = it->second[modelMatrixLineNumber];
			}
		}
	}

	// solution with lapack
	Output::echo(5, "matr --> res\n");
	lap_AB_res = lap_AB;

	// lapack vectors
	CPPL::dcovector lap_B(m_size);
	CPPL::dcovector lap_B_res(m_size);
	for(i = 0; i < lap_B.l; i++){
		lap_B(i) = B[i];
	}
	lap_B_res = lap_B;

	start_time = std::clock();
	Output::echo(5, "inverting... ");

	// lapack solution
	lap_AB.dgbsv(lap_B);

	current_time = std::clock();
	Output::echo(5, "done (%.2f sec)\n", (( current_time - start_time ) / (double)CLOCKS_PER_SEC ));

	Output::echo(5, "Checking... ");
	lap_B_res = lap_AB_res*lap_B - lap_B_res; // should be zero

	// calculate max error
	double max = 0;
	for (i = 0; i < lap_B_res.l; i++) {
		max = (max>lap_B_res(i))?max:lap_B_res(i);
	}
	Output::echo(5, " Max error: %e.\n", max);

	// save result into X
	Output::echo(5, "Coping back... ");
	for (modelMatrixLineNumber = 0; modelMatrixLineNumber < m_size; modelMatrixLineNumber++) {
		X[modelMatrixLineNumber] = lap_B(modelMatrixLineNumber);
	}
	Output::echo(5, "done. \n");

}


/**
 * Over relaxation iteration method. 
 * Uses 0th diagonal row for A.
 * 
 * \param &A - diagonal matrix A
 * \param &B - array B
 * \param &X - output matrix
 * \param max_steps - maximum steps of iteration
 * \param EE - error value
 */
void over_relaxation_diag(DiagMatrix &A, Matrix1D<double> &B, Matrix1D<double> &X, int max_steps, double EE) {
	int i, j;
	double  omega, sum;
	double err_max = 1e99;
	int n = A[0].size_x;

	omega = 1.75;
	//omega = 1.1;

    // iterates through key, value pairs
	DiagMatrix::iterator it;
	int nsteps = 0;

	while(err_max > EE && nsteps < max_steps) {
		nsteps++;
		for (i = 0; i < n; i++) {
			sum = 0;
			for (it = A.begin(); it != A.end(); it++) {
				j = i + it->first;
                // if j is a valid diagonal row index
				if (j >= 0 && j < n && j != i) {
                    // A[0][i] works because DiagMatrix is map<int,Matrix1D>
                    // A[0] gets the diagonal row, A[0][i] gets the ith element in diagonal row
					sum = sum + it->second[i]/A[0][i] * X[j];
				}
			}
			X[i] = (1 - omega) * X[i] - omega * ( sum - B[i]/A[0][i]);
		}

		err_max = 0.;
		// DiagMatrix::iterator it;
		double error;
		for (j = 0; j < n; j++) {
			// multiplying vector to matrix and calculating the difference
			error = B[j];
			for (it = A.begin(); it != A.end(); it++) {
				if (j + it->first >= 0 && j + it->first < n) {
					error -= it->second[j] * X[j + it->first];
				}
			}
			err_max = max(err_max, fabs(error));
		}
	}
	Output::echo(0, "[%d]Error_norm = %e \n", nsteps, err_max);
}


/**
 * Get change in indexes according to the derivatives direction.
 *
 * Used in approximation.
 * If it gets derivativeType == DT_ALPHA_left, for example, that means we have alpha-left-derivative, which is ( f(alpha) - f(alpha-1) ) / ( delta alpha ). So it returns dAlpha = -1, as a derivative direction vector.
 */
void GetDerivativeVector(string derivativeType, int &dL, int &dPc, int &dAlpha) {
	if (derivativeType == "DT_ALPHA_left") {
			dAlpha = -1;
	} else if (derivativeType == "DT_ALPHA_right") {
			dAlpha = 1;
	} else if (derivativeType == "DT_PC_left") {
			dPc = -1;
	} else if (derivativeType == "DT_PC_right") {
			dPc = 1;
	} else if (derivativeType == "DT_L_left") {
			dL = -1;
	} else if (derivativeType == "DT_L_right") {
			dL = 1;
	} else {
		throw error_msg("DERIVATIVE_TYPE", "Unknown derivative approximation: %s\n", derivativeType.c_str());
	}
}


/**
 * Second derivative approximation, returns coefficients to be putted into the model matrix.
 *  \f$L_{\alpha \beta}(y) = (D_{\alpha \beta} \cdot y_{\bar{x}_\alpha})_{x_{\beta}}\f$
 *
 * Samarskiy, page 261
 *
 * Returns coefficients to be put into model matrix for an approximation of a second derivative.
 */
void SecondDerivativeApproximation_3D(CalculationMatrix &matr_A,
								   int il, int im, int ia,
								   string FirstDerivative, string SecondDerivative,
								   Matrix3D<double> &L, Matrix3D<double> &pc, Matrix3D<double> &alpha,
								   Matrix3D<double> &Dxx, Matrix3D<double> &Jacobian,
								   double multiplicator) {

	int dL1 = 0, dPc1 = 0, dAlpha1 = 0;
	int dL2 = 0, dPc2 = 0, dAlpha2 = 0;
	// calculating derivative directions according derivative types
	// each second derivative has two directions, like
	// (alpha, alpha) - is a second derivative in alpha direction
	// (pc, pc) - second in pc direction
	// (pc, alpha) - mixed term
	// here we are getting that directions for derivative
	GetDerivativeVector(FirstDerivative, dL1, dPc1, dAlpha1);
	GetDerivativeVector(SecondDerivative, dL2, dPc2, dAlpha2);

	// matr_A line number
	int in = matr_A.index1d(ia, im, il);

	double dh1, dh11, dh12;
	// The following trick work only for orthogonal grid
	// first and second dh in derivative calculation

	// TODO: Rewrite in some normal way
	dh1 = (L[il][im][ia] + pc[il][im][ia] + alpha[il][im][ia]) - (L[il+dL1][im+dPc1][ia+dAlpha1] + pc[il+dL1][im+dPc1][ia+dAlpha1] + alpha[il+dL1][im+dPc1][ia+dAlpha1]);
	dh11 = (L[il][im][ia] + pc[il][im][ia] + alpha[il][im][ia]) - (L[il+dL2][im+dPc2][ia+dAlpha2] + pc[il+dL2][im+dPc2][ia+dAlpha2] + alpha[il+dL2][im+dPc2][ia+dAlpha2]);
	dh12 = (L[il+dL1][im+dPc1][ia+dAlpha1] + pc[il+dL1][im+dPc1][ia+dAlpha1] + alpha[il+dL1][im+dPc1][ia+dAlpha1]) - (L[il+dL1+dL2][im+dPc1+dPc2][ia+dAlpha1+dAlpha2] + pc[il+dL1+dL2][im+dPc1+dPc2][ia+dAlpha1+dAlpha2] + alpha[il+dL1+dL2][im+dPc1+dPc2][ia+dAlpha1+dAlpha2]);

	// for 2D
	//dh1 = (pc[il][im][ia] + alpha[il][im][ia]) - (pc[il][im+dPc1][ia+dAlpha1] + alpha[il][im+dPc1][ia+dAlpha1]);
	//dh11 = (pc[il][im][ia] + alpha[il][im][ia]) - (pc[il][im+dPc2][ia+dAlpha2] + alpha[il][im+dPc2][ia+dAlpha2]);
	//dh12 = (pc[il][im+dPc1][ia+dAlpha1] + alpha[il][im+dPc1][ia+dAlpha1]) - (pc[il][im+dPc1+dPc2][ia+dAlpha1+dAlpha2] + alpha[il][im+dPc1+dPc2][ia+dAlpha1+dAlpha2]);

	int id;

	// The same for all four coefficients, what's stending befire derivatives pretty much
	double common_part = multiplicator / Jacobian[il][im][ia] / dh1 ;

	// getting model matrix diagonal number according to derivative
	id = matr_A.index1d(ia, im, il) - in;
	// add corresponding coefficient for corresponding matrix cell
	matr_A[id][in] += + common_part / dh11 * Dxx[il][im][ia] * Jacobian[il][im][ia];
	// test - delete: matr_A[id][in] += + common_part / dh11 * Dxx[il][im][ia] * Jacobian[il][im][ia];

	id = matr_A.index1d(ia+dAlpha2, im+dPc2, il+dL2) - in;
	matr_A[id][in] += - common_part / dh11 * Dxx[il][im][ia] * Jacobian[il][im][ia];
	// test - delete: matr_A[id][in] += - common_part / dh12 * Dxx[il][im][ia] * Jacobian[il][im][ia];
	//matr_A[id][in] += -multiplicator / Jacobian[il][im][ia] / dh1 / dh11 * Dxx[il+dL2][im+dPc2][ia+dAlpha2] * Jacobian[il+dL2][im+dPc2][ia+dAlpha2];

	id = matr_A.index1d(ia+dAlpha1, im+dPc1, il+dL1) - in;
	matr_A[id][in] += - common_part / dh12 * Dxx[il+dL1][im+dPc1][ia+dAlpha1] * Jacobian[il+dL1][im+dPc1][ia+dAlpha1];
	// test - delete: matr_A[id][in] += - common_part / dh11 * Dxx[il+dL1][im+dPc1][ia+dAlpha1] * Jacobian[il+dL1][im+dPc1][ia+dAlpha1];

	id = matr_A.index1d(ia+dAlpha1+dAlpha2, im+dPc1+dPc2, il+dL1+dL2) - in;
	matr_A[id][in] += + common_part / dh12 * Dxx[il+dL1][im+dPc1][ia+dAlpha1] * Jacobian[il+dL1][im+dPc1][ia+dAlpha1];
	// test - delete: matr_A[id][in] += + common_part / dh12 * Dxx[il+dL1][im+dPc1][ia+dAlpha1] * Jacobian[il+dL1][im+dPc1][ia+dAlpha1];
	//matr_A[id][in] += +multiplicator / Jacobian[il][im][ia] / dh1 / dh12 * Dxx[il+dL1+dL2][im+dPc1+dPc2][ia+dAlpha1+dAlpha2] * Jacobian[il+dL1+dL2][im+dPc1+dPc2][ia+dAlpha1+dAlpha2];

}


/// Error value
#define EE       1.e-19
/// Specified Index [l][k]
#define I(l,k)	(l)*n+(k)

/// Function returns maximum between a and b.
double max2(double a, double b) {
	return (a>b)?a:b;
}
/// Multiplication between vector and matrix.
void mult_vect2(double *af, double *vf, double *wf, int nf) {
	int k,l,n;
	n = nf;
	for (k = 0; k < nf; k++) {
		wf[k] = 0.;
		for(l=0; l < nf; l++)
			wf[k]+=af[( (k)*n+(l) )]*vf[l];
	}
}

/// Gauss method - added to move out of if statements
void gauss(DiagMatrix &A, Matrix1D<double> &B, Matrix1D<double> &X, int n) {
    // Solving model matrix
    // need A and B for gauss method
    static Matrix2D<double> m_A;
    static Matrix1D<double> m_B;
    // iterators for matrix
    int modelMatrixLineNumber;
    int k, l, in;
    DiagMatrix::iterator it;
    double sum, error, error_max;
    static bool recalculate_m_A = true;
    
    if (recalculate_m_A) {
        m_B = B;
        // if A isn't initialized, allocate nxn memory
        if (!m_A.initialized) {
            m_A.AllocateMemory(n, n);
        }
        m_A = 0;
        //				Output::echo("Writing A0 matrix... ");
        //				A.writeToFile("A0.dat");
        //				Output::echo("done.\n");
        
        Output::echo(5, "Inverting matrix m_A (gauss)... ");
        // Gauss method
        for (it = A.begin(); it != A.end(); it++) {
            for (modelMatrixLineNumber = 0; modelMatrixLineNumber < n; modelMatrixLineNumber++) {
                // check, if the element at line (modelMatrixLineNumber) and diagonal (it->first) is inside the matrix.
                // cause it can be outside, if line number is 1 and diagonal number is -5, for example.
                if (modelMatrixLineNumber + it->first >= 0 && modelMatrixLineNumber + it->first < n) {
                    // converting matrix, stored as diagonals, into normal 2D matrix
                    // cause we need 2D matrix to solve by Gauss method
                    //							cout << it->first << " - " << it->second[modelMatrixLineNumber] << endl;
                    m_A[modelMatrixLineNumber][modelMatrixLineNumber + it->first] = it->second[modelMatrixLineNumber];
                }
            }
        }
        
        // inversion of A
        // sending address of the first element of the plane 2d array
        // it's not actually matrix inversion
        // it just makes A-triangle, so it's easy to solve A*X = B
        gauss_solve(&m_A[0][0], &m_B[0], n);
        recalculate_m_A = true;
        Output::echo(5, "done.\n");
				}
    
				Output::echo(5, "Calculation of new PSD... ");
				X[n-1] = m_B[n-1] / m_A[n-1][n-1];
				for (k = n-2; k >= 0; k--) {
                    sum = 0;
                    for (l = k+1; l < n; l++)
                        sum += m_A[k][l] * X[l];
                    X[k] = m_B[k] - sum;
                }
				/*   end calc  */
    
				/*   control  */
				//mult_vect2(&A[0], &X[0], &RHS[0], n);
				error_max = 0.;
				error = 0;
				for (in = 0; in < n; in++) {
                    error = B[in];
                    for (it = A.begin(); it != A.end(); it++) {
                        // multiplying all diagonals to PSD to check error = RHS - A*PSD = 0
                        if (in + it->first >= 0 && in + it->first < n) {
                            error -= it->second[in] * X[in + it->first];
                        }
                    }
                    error_max = VF::max( error_max, error );
                }
				
    Output::echo(5, "Error_norm = %e \n", error_max);

}


/// Gauss inversion.
void gauss_solve(double *a, double *b, int n) {
	int j, k, l, l_max;
	double a_max;
	double *dum;

	dum = (double*) malloc(n*sizeof(double));

	for (k = 0; k < (n-1); k++) {
		/* find max diag el  */

		a_max = fabs ( a[I(k,k)] );
		l_max = k;

		for(l = k + 1; l < n; l++)

			if( fabs( a[I(l,k)] ) > a_max ) {
				a_max = fabs( a[I(l,k)] );
				l_max = l;
			}
			if( a_max < EE ) {
				printf("GAUSS: zero element string=%d size=%d \n", k, n);
				printf("   a(k,k)=%e  \n",a[I(k,k)]);
				return;
			}
			if( l_max != k) {
				memcpy( dum, &(a[I(k,k)]), (n-k)*sizeof(double) );
				memcpy( &(a[I(k,k)]), &(a[I(l_max,k)]), (n-k)*sizeof(double) );
				memcpy( &(a[I(l_max,k)]), dum, (n-k)*sizeof(double) );
				a_max = b[k];
				b[k] = b[l_max];
				b[l_max] = a_max;
			}

			/*  k-elimination */
			b[k] = b[k] / a[I(k,k)];
			for (l = k + 1; l < n; l++)
				a[( (k)*n+(l) )] = a[( (k)*n+(l) )] / a[I(k,k)];
			a[I(k,k)] = 1.;
			for (l = k + 1; l < n; l++) {
				for(j = k + 1; j < n; j++)
					a[I(l,j)] = a[I(l,j)] - a[I(k,j)] * a[I(l,k)];
				b[l] = b[l] - b[k] * a[I(l,k)];
				a[I(l,k)] = 0.;
			}
	}

	free(dum);
}

/// Solve the AU=R system of equations, where A - tridiagonal matrix nxn with diagonals a[], b[], c[].
bool tridag(double a[], double b[], double c[], double r[], double u[], long n) {
	long j;
	double bet, *gam; //beta, gamma
	gam = new double[n];

	if (b[0] == 0.0) {
		throw error_msg("TRIDAG_MATRIX_ERROR", "tridag: error, b[0] = 0");
		//log1::fcout << "error 1 in tridag\n";
		//return -1;
	}

	u[0]=r[0]/(bet=b[0]);

	for (j=1;j<=n-1;j++) {
		gam[j]=c[j-1]/bet;
		bet=b[j]-a[j]*gam[j];

		if (bet == 0.0) {
			throw error_msg("TRIDAG_MATRIX_ERROR", "tridag: error, bet = 0");
			//log1::fcout << "tridag(: Error 2 in tridag\n";
			//return -1;
		}

		u[j]=(r[j]-a[j]*u[j-1])/bet;
	}

	for (j=(n-2);j>=0;j--)
		u[j] -= gam[j+1]*u[j+1];

	delete gam;

	return true;
}


/** kckim test **/
/**
 * Create matrix form of the Fokker-Planck equation: matr_A * PSD(t+1) = matr_B * PSD(t) + matr_C
 *
 * Calculation matr_A, matr_B, and matr_C, i.e. numerical approximation of the derivatives
 * Takes boundary conditions and diffusion coefficients as an input.
 * L, pc, alpha should be orthogonal for 3D!!!
 */
bool MakeModelMatrix_3D_KC (
				  CalculationMatrix &matr_A, CalculationMatrix &matr_B, CalculationMatrix &matr_C,
				  Matrix3D<double> &L, Matrix3D<double> &pc, Matrix3D<double> &alpha,
				  int L_size, int pc_size, int alpha_size,
				  Matrix2D<double> &L_lowerBoundaryCondition,
				  Matrix2D<double> &L_upperBoundaryCondition,
				  Matrix2D<double> &pc_lowerBoundaryCondition,
				  Matrix2D<double> &pc_upperBoundaryCondition,
				  Matrix2D<double> &alpha_lowerBoundaryCondition,
				  Matrix2D<double> &alpha_upperBoundaryCondition,
				  string L_lowerBoundaryCondition_calculationType,
				  string L_upperBoundaryCondition_calculationType,
				  string pc_lowerBoundaryCondition_calculationType,
				  string pc_upperBoundaryCondition_calculationType,
				  string alpha_lowerBoundaryCondition_calculationType,
				  string alpha_upperBoundaryCondition_calculationType,
				  Matrix3D<double> &DLL,
				  Matrix3D<double> &Dpcpc, Matrix3D<double> &DpcpcLpp,
				  Matrix3D<double> &Daa, Matrix3D<double> &DaaLpp,
				  Matrix3D<double> &Dpca, Matrix3D<double> &DpcaLpp,
				  Matrix3D<double> &Jacobian, double dt, double Lpp,
				  double tau, double tauLpp) {

	// variables, for future use
	// il - L-index, im - pc-index, ia - alpha-index,
	// id - index of a diagonal of the model matrix (counting from main diagonal, main is zero),
	// in - model matrix line number
	int il, im, ia, id, in;

	// create A-matrix, B-matrix, C-matrix
	// Output::echo("Recalculating model matrix... ");

	// make everything zero
	// TODO: make zero not everything, but existing diagonals
	/* !!! matr_C[0] = 0;
	for (il = -1; il <= 1; il++) {
		for (im = -1; im <= 1; im++) {
			for (ia = -1; ia <= 1; ia++) {
				// calculating a diagonal number (id)
				id = matr_A.index1d(ia, im, il);
				// make the diagonal zero
				matr_A[id] = 0;
				matr_B[id] = 0;
			}
		}
	}*/

	// Make diagonals to be equal to zero
	DiagMatrix::iterator it;
	for (it = matr_A.begin(); it != matr_A.end(); it++) it->second = 0;
	for (it = matr_B.begin(); it != matr_B.end(); it++) it->second = 0;
	for (it = matr_C.begin(); it != matr_C.end(); it++) it->second = 0;

	// create a new model matrix
	// (f^{t+1} - f^{t})/dt = L1(f^{t+1}) + L2(f^{t+1}) + L3(f^{t+1})[main equation, losses are calculated separately]
	// L1, L2, L3 - diffusion operators
	//  + need to add there the equations for boundary conditions also
	double dh;
	for (il = 0; il < L_size; il++) {
		for (im = 0; im < pc_size; im++) {
			for (ia = 0; ia < alpha_size; ia++) {
				// calculating current line number (in)
				in = matr_A.index1d(ia, im, il);

				// Describe all boundaries first. The 'inner' area will be approximated later.
				// If L_size > 1 we should use L-boundaries, right?
				// If not, then just calculate diffusion
				if (il == 0 && L_size >= 3) {
					// Boundary conditions
					// Points at the corners are used only at one of the boundaries
					// L-bottom boundary
					matr_C[0][in] = L_lowerBoundaryCondition[im][ia];
					id = matr_A.index1d(ia, im, il+1) - in;
					dh = L[il+1][im][ia] - L[il][im][ia];
					AddBoundary(matr_A, L_lowerBoundaryCondition_calculationType, in, id, dh);

				} else if (il == L_size-1 && L_size >= 3) {
					// L-top boundary
					matr_C[0][in] = L_upperBoundaryCondition[im][ia];
					id = matr_A.index1d(ia, im, il-1) - in;
					dh = L[il][im][ia] - L[il-1][im][ia];
					AddBoundary(matr_A, L_upperBoundaryCondition_calculationType, in, id, dh);

				} else if (im == 0 && pc_size >= 3) {
					// pc-low boundary
					matr_C[0][in] = pc_lowerBoundaryCondition[il][ia];
					id = matr_A.index1d(ia, im+1, il) - in;
					dh = pc[il][im+1][ia] - pc[il][im][ia];
					AddBoundary(matr_A, pc_lowerBoundaryCondition_calculationType, in, id, dh);

				} else if (im == pc_size-1 && pc_size >= 3) {
					// pc-high boundary
					matr_C[0][in] = pc_upperBoundaryCondition[il][ia];
					id = matr_A.index1d(ia, im-1, il) - in;
					dh = pc[il][im][ia] - pc[il][im-1][ia];
					AddBoundary(matr_A, pc_upperBoundaryCondition_calculationType, in, id, dh);
				} else if (ia == 0 && alpha_size >= 3) {
					// alpha-loss cone boundary
					matr_C[0][in] = alpha_lowerBoundaryCondition[il][im];
					id = matr_A.index1d(ia+1, im, il) - in;
					dh = alpha[il][im][ia+1] - alpha[il][im][ia];
					AddBoundary(matr_A, alpha_lowerBoundaryCondition_calculationType, in, id, dh);

				} else if (ia == alpha_size-1 && alpha_size >= 3) {
					// alpha-equatorial boundary
					matr_C[0][in] = alpha_upperBoundaryCondition[il][im];
					id = matr_A.index1d(ia-1, im, il) - in;
					dh = alpha[il][im][ia] - alpha[il][im][ia-1];
					AddBoundary(matr_A, alpha_upperBoundaryCondition_calculationType, in, id, dh);
					//cout<<in<<"  "<<ia<<" "<<id<<"  "<<dh<<"  "<<alpha_upperBoundaryCondition[il][im]<<endl;
				} else {

					// now we are sure we are not on a boundary, can do the Fokker-Planck equation approximation in the inner area

					// f^{t+1}/dt
					matr_A[0][in] += 1.0/dt;
					// f^{t}/dt
					matr_B[0][in] += 1.0/dt;

					// calculate radial diffusion, if L_size >= 3
					if (L_size >=3) {
						// L term, add approximated derivatives into the calculation matrix
						SecondDerivativeApproximation_3D_KC_Diagonal(matr_A, il, im, ia,
							"DT_L_left", "DT_L_right",
							L, pc, alpha, DLL, Jacobian, -0.5);
						SecondDerivativeApproximation_3D_KC_Diagonal(matr_A, il, im, ia,
							"DT_L_right", "DT_L_left",
							L, pc, alpha, DLL, Jacobian, -0.5);
					}

					if (pc_size >=3) {
						// Energy term
						if (L[il][im][ia] >= Lpp) {
							/*
							SecondDerivativeApproximation_3D(matr_A, il, im, ia,
								"DT_PC_left", "DT_PC_right",
								L, pc, alpha, Dpcpc, Jacobian, -0.5);
							SecondDerivativeApproximation_3D(matr_A, il, im, ia,
								"DT_PC_right", "DT_PC_left",
								L, pc, alpha, Dpcpc, Jacobian, -0.5);
							*/
							SecondDerivativeApproximation_3D_KC_Diagonal(matr_A, il, im, ia,
								"DT_PC_right", "DT_PC_left",
								L, pc, alpha, Dpcpc, Jacobian, -1.0);

						}  else  {

							/*
							SecondDerivativeApproximation_3D(matr_A, il, im, ia,
								"DT_PC_left", "DT_PC_right",
								L, pc, alpha, DpcpcLpp, Jacobian, -0.5);

							SecondDerivativeApproximation_3D(matr_A, il, im, ia,
								"DT_PC_right", "DT_PC_left",
								L, pc, alpha, DpcpcLpp, Jacobian, -0.5);
							*/

							SecondDerivativeApproximation_3D_KC_Diagonal(matr_A, il, im, ia,
								"DT_PC_right", "DT_PC_left",
								L, pc, alpha, DpcpcLpp, Jacobian, -1.0);

						}
					}

					if (alpha_size >=3) {
						// Pitch angle term(s)

						// Loss cone.
						// Warning! Can cause the loss cone to be added during each diffusion, but
						// now it's ok, cause alpha_size >= 3 only for pitch-angle diffusion
						// f^{n+1}/taulc


						/*
						if ( (alpha[il][im][ia] <= VF::alc(L[il][im][0]) || alpha[il][im][ia] >= VC::pi - VF::alc(L[il][im][0]))) {
							double taulc = VF::bounce_time(VF::Kfunc(pc[il][im][ia]), L[il][im][ia])/60./60./24.;
							matr_A[0][in] += 1.0/taulc;  // Implicit
							//matr_B[0][in] -= 1.0/taulc; // Explicit
						}*/


						if (L[il][im][ia] >= Lpp) {
							/*
							SecondDerivativeApproximation_3D(matr_A, il, im, ia,
								"DT_ALPHA_left", "DT_ALPHA_right",
								L, pc, alpha, Daa, Jacobian, -0.5);
							SecondDerivativeApproximation_3D(matr_A, il, im, ia,
								"DT_ALPHA_right", "DT_ALPHA_left",
								L, pc, alpha, Daa, Jacobian, -0.5);
							*/
							SecondDerivativeApproximation_3D_KC_Diagonal(matr_A, il, im, ia,
								"DT_ALPHA_right", "DT_ALPHA_left",
								L, pc, alpha, Daa, Jacobian, -1.0);

						}  else  {
							/*
							SecondDerivativeApproximation_3D(matr_A, il, im, ia,
								"DT_ALPHA_left", "DT_ALPHA_right",
								L, pc, alpha, DaaLpp, Jacobian, -0.5);
							SecondDerivativeApproximation_3D(matr_A, il, im, ia,
								"DT_ALPHA_right", "DT_ALPHA_left",
								L, pc, alpha, DaaLpp, Jacobian, -0.5);
							*/

							SecondDerivativeApproximation_3D_KC_Diagonal(matr_A, il, im, ia,
								"DT_ALPHA_right", "DT_ALPHA_left",
								L, pc, alpha, DaaLpp, Jacobian, -1.0);

						}
					}

					if (alpha_size >= 3 && pc_size >= 3) {
						// Mixed pitch angle and energy term
						// There are several possibilities for approximation
						if (L[il][im][ia] >= Lpp) {
							/*
							// d/dp D d/da f
							SecondDerivativeApproximation_3D_Mixed(matr_A, il, im, ia,
								"DT_ALPHA_right", "DT_PC_right",
								L, pc, alpha, Dpca, Jacobian, -1.0);
							SecondDerivativeApproximation_3D_Mixed(matr_A, il, im, ia,
								"DT_ALPHA_right", "DT_PC_left",
								L, pc, alpha, Dpca, Jacobian, +1.0);


							// d/da D d/dp f
							SecondDerivativeApproximation_3D_Mixed(matr_A, il, im, ia,
								"DT_PC_right", "DT_ALPHA_right",
								L, pc, alpha, Dpca, Jacobian, -1.0);
							SecondDerivativeApproximation_3D_Mixed(matr_A, il, im, ia,
								"DT_PC_right", "DT_ALPHA_left",
								L, pc, alpha, Dpca, Jacobian, +1.0);
							*/


							// d/dp D d/da f
							SecondDerivativeApproximation_3D_KC_Mixed(matr_A, il, im, ia,
								"DT_ALPHA_right", "DT_PC_right",
								L, pc, alpha, Dpca, Jacobian, -1.0);
							SecondDerivativeApproximation_3D_KC_Mixed(matr_A, il, im, ia,
								"DT_ALPHA_left", "DT_PC_right",
								L, pc, alpha, Dpca, Jacobian, +1.0);
							SecondDerivativeApproximation_3D_KC_Mixed(matr_A, il, im, ia,
								"DT_ALPHA_right", "DT_PC_left",
								L, pc, alpha, Dpca, Jacobian, +1.0);
							SecondDerivativeApproximation_3D_KC_Mixed(matr_A, il, im, ia,
								"DT_ALPHA_left", "DT_PC_left",
								L, pc, alpha, Dpca, Jacobian, -1.0);


							// d/da D d/dp f

							SecondDerivativeApproximation_3D_KC_Mixed(matr_A, il, im, ia,
								"DT_PC_right", "DT_ALPHA_right",
								L, pc, alpha, Dpca, Jacobian, -1.0);
							SecondDerivativeApproximation_3D_KC_Mixed(matr_A, il, im, ia,
								"DT_PC_left", "DT_ALPHA_right",
								L, pc, alpha, Dpca, Jacobian, +1.0);
							SecondDerivativeApproximation_3D_KC_Mixed(matr_A, il, im, ia,
								"DT_PC_right", "DT_ALPHA_left",
								L, pc, alpha, Dpca, Jacobian, +1.0);
							SecondDerivativeApproximation_3D_KC_Mixed(matr_A, il, im, ia,
								"DT_PC_left", "DT_ALPHA_left",
								L, pc, alpha, Dpca, Jacobian, -1.0);


						}  else  {

							// full form
							// d/dp D d/da f
							/*
							SecondDerivativeApproximation_3D_Mixed(matr_A, il, im, ia,
								"DT_ALPHA_right", "DT_PC_right",
								L, pc, alpha, DpcaLpp, Jacobian, -1.0);
							SecondDerivativeApproximation_3D_Mixed(matr_A, il, im, ia,
								"DT_ALPHA_left", "DT_PC_right",
								L, pc, alpha, DpcaLpp, Jacobian, +1.0);
							SecondDerivativeApproximation_3D_Mixed(matr_A, il, im, ia,
								"DT_ALPHA_right", "DT_PC_left",
								L, pc, alpha, DpcaLpp, Jacobian, +1.0);
							SecondDerivativeApproximation_3D_Mixed(matr_A, il, im, ia,
								"DT_ALPHA_left", "DT_PC_left",
								L, pc, alpha, DpcaLpp, Jacobian, -1.0);


							// d/da D d/dp f

							SecondDerivativeApproximation_3D_Mixed(matr_A, il, im, ia,
								"DT_PC_right", "DT_ALPHA_right",
								L, pc, alpha, DpcaLpp, Jacobian, -1.0);
							SecondDerivativeApproximation_3D_Mixed(matr_A, il, im, ia,
								"DT_PC_left", "DT_ALPHA_right",
								L, pc, alpha, DpcaLpp, Jacobian, +1.0);
							SecondDerivativeApproximation_3D_Mixed(matr_A, il, im, ia,
								"DT_PC_right", "DT_ALPHA_left",
								L, pc, alpha, DpcaLpp, Jacobian, +1.0);
							SecondDerivativeApproximation_3D_Mixed(matr_A, il, im, ia,
								"DT_PC_left", "DT_ALPHA_left",
								L, pc, alpha, DpcaLpp, Jacobian, -1.0);
							*/

							//short form
							// d/dp D d/da f
							SecondDerivativeApproximation_3D_KC_Mixed(matr_A, il, im, ia,
								"DT_ALPHA_right", "DT_PC_right",
								L, pc, alpha, DpcaLpp, Jacobian, -1.0);
							SecondDerivativeApproximation_3D_KC_Mixed(matr_A, il, im, ia,
								"DT_ALPHA_right", "DT_PC_left",
								L, pc, alpha, DpcaLpp, Jacobian, +1.0);
							SecondDerivativeApproximation_3D_KC_Mixed(matr_A, il, im, ia,
								"DT_ALPHA_left", "DT_PC_right",
								L, pc, alpha, DpcaLpp, Jacobian, +1.0);
							SecondDerivativeApproximation_3D_KC_Mixed(matr_A, il, im, ia,
								"DT_ALPHA_left", "DT_PC_left",
								L, pc, alpha, DpcaLpp, Jacobian, -1.0);



							// d/da D d/dp f
							SecondDerivativeApproximation_3D_KC_Mixed(matr_A, il, im, ia,
								"DT_PC_right", "DT_ALPHA_right",
								L, pc, alpha, DpcaLpp, Jacobian, -1.0);
							SecondDerivativeApproximation_3D_KC_Mixed(matr_A, il, im, ia,
								"DT_PC_right", "DT_ALPHA_left",
								L, pc, alpha, DpcaLpp, Jacobian, +1.0);
							SecondDerivativeApproximation_3D_KC_Mixed(matr_A, il, im, ia,
								"DT_PC_left", "DT_ALPHA_right",
								L, pc, alpha, DpcaLpp, Jacobian, +1.0);
							SecondDerivativeApproximation_3D_KC_Mixed(matr_A, il, im, ia,
								"DT_PC_left", "DT_ALPHA_left",
								L, pc, alpha, DpcaLpp, Jacobian, -1.0);

						}
					}
				}
			}
		}
	}
	// Output::echo("recalculated.\n");

	// save the time of matrix change
	matr_A.change_ind = clock();
	matr_B.change_ind = clock();
	matr_C.change_ind = clock();

	return true;
}

/** kckim test **/
/**
 * Second derivative approximation, returns coefficients to be put into the model matrix.
 *  \f$L_{\alpha \beta}(y) = (D_{\alpha \beta} \cdot y_{\bar{x}_\alpha})_{x_{\beta}}\f$
 *
 * Samarskiy, page 261
 *
 * Returns coefficients to be put into model matrix for an approximation of a second derivative.
 */

void SecondDerivativeApproximation_3D_KC_Diagonal(CalculationMatrix &matr_A,
								   int il, int im, int ia,
								   string FirstDerivative, string SecondDerivative,
								   Matrix3D<double> &L, Matrix3D<double> &pc, Matrix3D<double> &alpha,
								   Matrix3D<double> &Dxx, Matrix3D<double> &Jacobian,
								   double multiplicator) {

	int dL1 = 0, dPc1 = 0, dAlpha1 = 0;
	int dL2 = 0, dPc2 = 0, dAlpha2 = 0;
	// calculating derivative directions according derivative types
	// each second derivative has two directions, like
	// (alpha, alpha) - is a second derivative in alpha direction
	// (pc, pc) - second in pc direction
	// here we are getting that directions for derivative
	GetDerivativeVector(FirstDerivative, dL1, dPc1, dAlpha1);
	GetDerivativeVector(SecondDerivative, dL2, dPc2, dAlpha2);

	// matr_A line number
	int in = matr_A.index1d(ia, im, il);

	// The following trick work only for orthogonal grid
	// first and second dh in derivative calculation

	double dh_1, dh_2;            // _0: current point, _1: +1, _2: -1
	double Dxx_0, Dxx_1, Dxx_2, Dxx_11, Dxx_21;
	double J_0, J_1, J_2;
	// TODO: Rewrite in some normal way

	dh_1 = abs((L[il][im][ia] + pc[il][im][ia] + alpha[il][im][ia]) - (L[il+dL1][im+dPc1][ia+dAlpha1] + pc[il+dL1][im+dPc1][ia+dAlpha1] + alpha[il+dL1][im+dPc1][ia+dAlpha1]));
	dh_2 = abs((L[il][im][ia] + pc[il][im][ia] + alpha[il][im][ia]) - (L[il+dL2][im+dPc2][ia+dAlpha2] + pc[il+dL2][im+dPc2][ia+dAlpha2] + alpha[il+dL2][im+dPc2][ia+dAlpha2]));

	Dxx_0 = Dxx[il][im][ia];
	Dxx_1 = Dxx[il+dL1][im+dPc1][ia+dAlpha1];
	Dxx_2 = Dxx[il+dL2][im+dPc2][ia+dAlpha2];
	Dxx_21 = Dxx[il+dL1+dL2][im+dPc1+dPc2][ia+dAlpha1+dAlpha2];

	J_0 = Jacobian[il][im][ia];
	J_1 = Jacobian[il+dL1][im+dPc1][ia+dAlpha1];
	J_2 = Jacobian[il+dL2][im+dPc2][ia+dAlpha2];

	int id;
	double common;

	common = multiplicator / J_0 / (dh_1+dh_2) * 2;
	// getting model matrix diagonal number according to derivative
	id = matr_A.index1d(ia, im, il) - in;
	// add corresponding coefficient for corresponding matrix cell
	matr_A[id][in] += -common / dh_1 * (Dxx_0+Dxx_1)/2 * (J_0+J_1)/2;   // f_0

	id = matr_A.index1d(ia+dAlpha1, im+dPc1, il+dL1) - in;
	matr_A[id][in] += +common / dh_1 * (Dxx_0+Dxx_1)/2 * (J_0+J_1)/2;   // f_+1

	id = matr_A.index1d(ia+dAlpha2, im+dPc2, il+dL2) - in;
	matr_A[id][in] += +common / dh_2 * (Dxx_0+Dxx_2)/2 * (J_0+J_2)/2;   // f_-1

	id = matr_A.index1d(ia+dAlpha1+dAlpha2, im+dPc1+dPc2, il+dL1+dL2) - in;
	matr_A[id][in] += -common / dh_2 * (Dxx_0+Dxx_2)/2 * (J_0+J_2)/2;   // f_0 again because dA1+dA2 = 0, dPc1+dPc2 = 0

}



/**
 * short form mixed term approximation
 * takes approximations using alpha and pc left and right side data for derivatives
 * \note SecondDerivativeApproximation_3D_Mixed is deprecated in favor of this function
 * \param &matr_A - calculation matrix
 * \param il - iterator for L
 * \param im - iterator for pc
 * \param ia - iterator for alpha
 * \param FirstDerivative - key term for first derivative approximation
 * \param SecondDerivative - key term for second derivative approximation
 * \param &L - matrix for L
 * \param &pc - matrix for pc
 * \param &alpha - matrix for alpha
 * \param &Dxx - matrix of derivatives
 * \param &Jacobian - jacobian
 * \param multiplicator - positive or negative multiplicator
 */
void SecondDerivativeApproximation_3D_KC_Mixed(CalculationMatrix &matr_A,
								   int il, int im, int ia,
								   string FirstDerivative, string SecondDerivative,
								   Matrix3D<double> &L, Matrix3D<double> &pc, Matrix3D<double> &alpha,
								   Matrix3D<double> &Dxx, Matrix3D<double> &Jacobian,
								   double multiplicator) {

	int dL1 = 0, dPc1 = 0, dAlpha1 = 0;
	int dL2 = 0, dPc2 = 0, dAlpha2 = 0;
	// calculating derivative directions according derivative types
	// each second derivative has two directions, like
	// (alpha, alpha) - is a second derivative in alpha direction
	// (pc, pc) - second in pc direction
	// here we are getting that directions for derivative
	GetDerivativeVector(FirstDerivative, dL1, dPc1, dAlpha1);
	GetDerivativeVector(SecondDerivative, dL2, dPc2, dAlpha2);

	// matr_A line number
	int in = matr_A.index1d(ia, im, il);

	// The following trick work only for orthogonal grid
	// first and second dh in derivatove calculation

	double dh_1, dh_2, dh_11, dh_21;            // _0: current point, _1: +1, _2: -1
	double Dxx_0, Dxx_1, Dxx_2, Dxx_11, Dxx_21;
	double Dxx_c1, Dxx_c2, Dxx_c3, Dxx_c4;
	double J_0, J_1, J_2, J_11, J_21;
	double p_0;
	// TODO: Rewrite in some normal way

	dh_1 = abs((L[il][im][ia] + pc[il][im][ia] + alpha[il][im][ia]) - (L[il+dL1][im+dPc1][ia+dAlpha1] + pc[il+dL1][im+dPc1][ia+dAlpha1] + alpha[il+dL1][im+dPc1][ia+dAlpha1]));
	dh_2 = abs((L[il][im][ia] + pc[il][im][ia] + alpha[il][im][ia]) - (L[il+dL2][im+dPc2][ia+dAlpha2] + pc[il+dL2][im+dPc2][ia+dAlpha2] + alpha[il+dL2][im+dPc2][ia+dAlpha2]));
	dh_11 = abs((L[il][im][ia] + pc[il][im][ia] + alpha[il][im][ia]) - (L[il-dL1][im-dPc1][ia-dAlpha1] + pc[il-dL1][im-dPc1][ia-dAlpha1] + alpha[il-dL1][im-dPc1][ia-dAlpha1]));
	dh_21 = abs((L[il][im][ia] + pc[il][im][ia] + alpha[il][im][ia]) - (L[il-dL2][im-dPc2][ia-dAlpha2] + pc[il-dL2][im-dPc2][ia-dAlpha2] + alpha[il-dL2][im-dPc2][ia-dAlpha2]));

	Dxx_0 = Dxx[il][im][ia];
	Dxx_1 = Dxx[il+dL1][im+dPc1][ia+dAlpha1];
	Dxx_2 = Dxx[il+dL2][im+dPc2][ia+dAlpha2];
	Dxx_11 = Dxx[il-dL1][im-dPc1][ia-dAlpha1];
	Dxx_21 = Dxx[il-dL2][im-dPc2][ia-dAlpha2];

	// four points at the outer corners
	Dxx_c1 = Dxx[il+dL1+dL2][im+dPc1+dPc2][ia+dAlpha1+dAlpha2];
	Dxx_c2 = Dxx[il+dL1-dL2][im+dPc1-dPc2][ia+dAlpha1-dAlpha2];
	Dxx_c3 = Dxx[il-dL1+dL2][im-dPc1+dPc2][ia-dAlpha1+dAlpha2];
	Dxx_c4 = Dxx[il-dL1-dL2][im-dPc1-dPc2][ia-dAlpha1-dAlpha2];


	J_0 = Jacobian[il][im][ia];
	J_1 = Jacobian[il+dL1][im+dPc1][ia+dAlpha1];
	J_2 = Jacobian[il+dL2][im+dPc2][ia+dAlpha2];
	J_11 = Jacobian[il-dL1][im-dPc1][ia-dAlpha1];
	J_21 = Jacobian[il-dL2][im-dPc2][ia-dAlpha2];

	int id;
	double common;

	common = multiplicator / J_0 / (dh_1+dh_11) / (dh_2+dh_21);
	// getting model matrix diagonal number according to derivative
	id = matr_A.index1d(ia+dAlpha1, im+dPc1, il+dL1) - in;
	// add corresponding coefficient for corresponding matrix cell
	matr_A[id][in] += +common * (Dxx_0+Dxx_2)/2 * (J_0+J_2)/2;   // f_0
	id = matr_A.index1d(ia+dAlpha1+dAlpha2, im+dPc1+dPc2, il+dL1+dL2) - in;
	matr_A[id][in] += +common * (Dxx_0+Dxx_2)/2 * (J_0+J_2)/2;   // f_-1

}

