/**
 * \file MatrixSolver.h
 *
 * Making model matrixes, solving model matrixes.
 *
 * Matrix form of linear equations: A*X[t+1] = B*X[t],
 * where A - model matrix, B - RHS, X[t] - known values of function (PSD), X[t+1] - unknown values of function.
 *
 * In that file there are procedures for making model matrix for 1d-diffusion, 2d-diffusion, some  ideas of 3d-diffusion and mixed terms.
 * Solver for tridiagonal matrix, solver by gauss method and iteration method (upper relaxation).
 *
 * This file is under development, and has a lot of commented code, unfinished etc code.
 *
 * There is a checked version for 1d diffusion (or split method of 2d, 3d diffusions) - 1d_universal_solver.
 *
 * \author Developed under supervision of the PI Yuri Shprits
 */

#ifndef MatrixSolver_H
#define MatrixSolver_H

#include "../Grid/Grid.h"
#include "../DiffusionCoefficient/DiffusionCoefficient.h"

#include "../GMRES/itlin.h"



/// Make matrix for 1d diffusion.
bool MakeMatrix(double  *A,                         // arrays of values on left side
				double  *B1,                        // array of values for middle
				double  *C,                         // array of values on right side
				double  *R,                         // inverse values
                double  *x,							// one dimensional grid
                double	tau,						// lifetime
                double	n,							// power of x
                double	f_lower,					// lower boundary conditions
                double	f_upper,					// upper boundary conditions
                int		nx,							// number of grid points
                double	dt,							// time step
                double	*Dxx,						// diffusion coefficient
                double	taulc,						// lifetime in loss cone
                double	alc,						// loss cone size
                int		g_flag,						// diffusion flag
                string	lower_border_condition_type,// type of lower boundary conditions: 0 - for value, 1 - for derivative
                string	upper_border_condition_type,// type of upper boundary conditions: 0 - for value, 1 - for derivative
                string  approximationMethod = "AM_Split_C");


/// Solver for 1d general equation
bool SolveMatrix(double	*f,							// phase space density function
				 double *A,                         // array of values on left side
				 double *B1,                        // array of values for middle
				 double *C,                         // array of values on right side
				 double *R,                         // inverse values
				 double	f_lower,					// lower boundary conditions
				 double	f_upper,					// upper boundary conditions
				 int	nx,							// number of grid points
				 double	dt);						// time step


// Matrix solvers
// Different methods of solving matrices, but ultimately do the same thing
void over_relaxation_diag(DiagMatrix &A, Matrix1D<double> &B, Matrix1D<double> &X, int max_steps = 1e5, double EE = 1e-3);
void Lapack(DiagMatrix &A, Matrix1D<double> &B, Matrix1D<double> &X);
void gauss(DiagMatrix &A, Matrix1D<double> &B, Matrix1D<double> &X, int n);
void gauss_solve(double *a, double *b, int n);

/** Solver for a system of equations with 3-diagonal matrix.
 *
 * Au = r
 * Where	A = diag(a, b, c),
 *
 * \param a[]	- array, diagonal '-1' of the matrix
 * \param b[]	- array, diagonal '0' of the matrix
 * \param c[]	- array, diagonal '+1' of the matrix
 * \param r[]	- array, r-vector
 * \param u[]	- array, result
 * \param n	- size of the matrix
 */
bool tridag(double a[], double b[], double c[], double r[], double u[], long n);

//void gmres_wrapout(CalculationMatrix &A, Matrix1D<double> &B, Matrix1D<double> &X, int maxiter = 100, int i_max = 10, double tol = 1e-10);
void gmres_wrapout(CalculationMatrix &A, Matrix1D<double> &B, Matrix1D<double> &X, Parameters_structure::PSD::GMRES_parameters_structure GMRES_parameters);

// A test
void gmres2_wrapout(CalculationMatrix &A, Matrix1D<double> &B, Matrix1D<double> &X, Parameters_structure::PSD::GMRES_parameters_structure GMRES_parameters);

bool MakeModelMatrix_3D(
				  CalculationMatrix &matr_A, CalculationMatrix &matr_B, CalculationMatrix &matr_C,
				  Matrix3D<double> &L, Matrix3D<double> &pc, Matrix3D<double> &alpha,
				  int L_size, int pc_size, int alpha_size,
				  Matrix2D<double> &L_lowerBoundaryCondition,
				  Matrix2D<double> &L_upperBoundaryCondition,
				  Matrix2D<double> &pc_lowerBoundaryCondition,
				  Matrix2D<double> &pc_upperBoundaryCondition,
				  Matrix2D<double> &alpha_lowerBoundaryCondition,
				  Matrix2D<double> &alpha_upperBoundaryCondition,
				  string L_lowerBoundaryCondition_calculationType,
				  string L_upperBoundaryCondition_calculationType,
				  string pc_lowerBoundaryCondition_calculationType,
				  string pc_upperBoundaryCondition_calculationType,
				  string alpha_lowerBoundaryCondition_calculationType,
				  string alpha_upperBoundaryCondition_calculationType,
				  Matrix3D<double> &DLL,
				  Matrix3D<double> &Dpcpc, Matrix3D<double> &DpcpcLpp,
				  Matrix3D<double> &Daa, Matrix3D<double> &DaaLpp,
				  Matrix3D<double> &Dpca, Matrix3D<double> &DpcaLpp,
				  Matrix3D<double> &Jacobian, double dt, double Lpp,
				  double tau = 1e99, double tauLpp = 1e99);

bool MakeModelMatrix_3D_KC(
				  CalculationMatrix &matr_A, CalculationMatrix &matr_B, CalculationMatrix &matr_C,
				  Matrix3D<double> &L, Matrix3D<double> &pc, Matrix3D<double> &alpha,
				  int L_size, int pc_size, int alpha_size,
				  Matrix2D<double> &L_lowerBoundaryCondition,
				  Matrix2D<double> &L_upperBoundaryCondition,
				  Matrix2D<double> &pc_lowerBoundaryCondition,
				  Matrix2D<double> &pc_upperBoundaryCondition,
				  Matrix2D<double> &alpha_lowerBoundaryCondition,
				  Matrix2D<double> &alpha_upperBoundaryCondition,
				  string L_lowerBoundaryCondition_calculationType,
				  string L_upperBoundaryCondition_calculationType,
				  string pc_lowerBoundaryCondition_calculationType,
				  string pc_upperBoundaryCondition_calculationType,
				  string alpha_lowerBoundaryCondition_calculationType,
				  string alpha_upperBoundaryCondition_calculationType,
				  Matrix3D<double> &DLL,
				  Matrix3D<double> &Dpcpc, Matrix3D<double> &DpcpcLpp,
				  Matrix3D<double> &Daa, Matrix3D<double> &DaaLpp,
				  Matrix3D<double> &Dpca, Matrix3D<double> &DpcaLpp,
				  Matrix3D<double> &Jacobian, double dt, double Lpp,
				  double tau = 1e99, double tauLpp = 1e99);


void SecondDerivativeApproximation_3D(CalculationMatrix &matr_A,
								   int il, int im, int ia,
								   string FirstDerivative, string SecondDerivative,
								   Matrix3D<double> &L, Matrix3D<double> &pc, Matrix3D<double> &alpha,
								   Matrix3D<double> &Dxx, Matrix3D<double> &Jacobian,
								   double multiplicator);

void SecondDerivativeApproximation_3D_KC_Diagonal(CalculationMatrix &matr_A,
								   int il, int im, int ia,
								   string FirstDerivative, string SecondDerivative,
								   Matrix3D<double> &L, Matrix3D<double> &pc, Matrix3D<double> &alpha,
								   Matrix3D<double> &Dxx, Matrix3D<double> &Jacobian,
								   double multiplicator);

void SecondDerivativeApproximation_3D_KC_Mixed(CalculationMatrix &matr_A,
								   int il, int im, int ia,
								   string FirstDerivative, string SecondDerivative,
								   Matrix3D<double> &L, Matrix3D<double> &pc, Matrix3D<double> &alpha,
								   Matrix3D<double> &Dxx, Matrix3D<double> &Jacobian,
								   double multiplicator);


#endif

