/**
* \file PSD.cpp
*
* \brief Makes operations with PSD (Phase Space Density) (like, diffusion).
*
* General view of Fokker-Planck diffusion equation in terms of linearization:
* \f[ A \times x = B \times x + C \f]
* A, B, C refered as MatrixA, MatrixB and MatrixC in the code
* 
* \todo 
* - A lot of corrections should be done in PSD class to make it more logical and less spread.
*
* \author Developed under supervision of the PI Yuri Shprits
*/

//#define DEBUG_MODE

// Header file
#include "PSD.h"

// Mathematical function
#include <math.h>

// Various function requred for Radiation Belts simulation
#include "../VariousFunctions/variousFunctions.h"



//
#include "../Exceptions/error.h"

// General c++ headers
#include <iostream>
#include <string>
#include <ctime>

#include <stdio.h>
#include <stdlib.h>

using namespace std;


/**
* Initializing: storing parameters, loading initial values, making boundary conditions, initializing output parameters
* Simply, it is a creation of the object.
*/
void PSD::Create_Initial_PSD(Parameters_structure::PSD parameters, Grid &grid, BoundaryCondition L_UpperBoundaryCondition)
{

	if (!arr.initialized) {
		arr.AllocateMemory(grid.L.size, grid.pc.size, grid.alpha.size);
	}

	// Copying from parameters to the class variables
	PSD_parameters = parameters;

	// Loading initial values, from time-dependent boundary conditions takes only first time step values
	//LoadInitialValue(parameters, grid, L_UpperBoundaryCondition.xSlice(0));
	// iterators
	int il, im, ia; ///< il - L; im - pc; ia - alpha

	// loading (calculating) initial values
	if (parameters.initial_PSD_Type == "IPSDT_ORBIT_FLUX_2D") {
			// loading from orbit flux (for 2d)
			if (grid.L.size != 1) {
				throw error_msg("INITIAL_PSD_2D_METHOD_FOR_3D_GRID", "Initial PSD error: impossible to use ORBIT_FLUX_2D if L.size > 1");
			}
			Load_initial_f_2d(grid.L, grid.pc, grid.alpha, parameters.initial_PSD_fileName.c_str());
	////////////////////////// kckim (mawellian) /////////////////////////
	} 
	else if (parameters.initial_PSD_Type == "IPSDT_FLUX_2D_MAXWELLIAN") {
			if (grid.L.size != 1) {
				throw error_msg("INITIAL_PSD_2D_METHOD_FOR_3D_GRID", "Initial PSD error: impossible to use FLUX_2D_MAXWELLIAN if L.size > 1");
			}
			cout<<"kim................................"<<endl;
			Load_initial_f_maxwell(grid.L, grid.pc, grid.alpha);

	//////////////////////////////////////////////////////////////////////
	} 
	else if (parameters.initial_PSD_Type == "IPSDT_STEADY_STATE") {
			// calculating as a steady state (for 3d)
			if (grid.L.size == 1) {
				throw error_msg("INITIAL_PSD_3D_METHOD_FOR_2D_GRID", "Initial PSD error: impossible to use STEADY_STATE if L.size == 1");
			}
			// if tauSteadyState == 0 - using 6.0/Kp
			/// moved to parameters.cpp			if (parameters.initial_PSD_tauSteadyState <= 1.e-99) parameters.initial_PSD_tauSteadyState = 4.0/parameters.initial_PSD_Kp0;
			Load_initial_f(grid.L, grid.pc, grid.alpha, parameters.initial_PSD_tauSteadyState, parameters.initial_PSD_Kp0, parameters.initial_PSD_some_constant_value, parameters.initial_PSD_J_L7_function, parameters.initial_PSD_outer_psd, parameters.initial_PSD_inner_psd, false); // false indicating old method
	/////////////////////// kckim /////////////////////////////////////////////////
	}
    // new version of steady state solution, that requires additional boundary conditions using L7 - ADrozdov
    else if (parameters.initial_PSD_Type == "IPSDT_STEADY_STATE_L7") {
        // calculating as a steady state (for 3d)
        if (grid.L.size == 1) {
            throw error_msg("INITIAL_PSD_3D_METHOD_FOR_2D_GRID", "Initial PSD error: impossible to use STEADY_STATE_L7 if L.size == 1");
        }
        // if tauSteadyState == 0 - using 6.0/Kp
        /// moved to parameters.cpp			if (parameters.initial_PSD_tauSteadyState <= 1.e-99) parameters.initial_PSD_tauSteadyState = 4.0/parameters.initial_PSD_Kp0;
        Load_initial_f(grid.L, grid.pc, grid.alpha, parameters.initial_PSD_tauSteadyState, parameters.initial_PSD_Kp0, parameters.initial_PSD_some_constant_value, parameters.initial_PSD_J_L7_function, parameters.initial_PSD_outer_psd, parameters.initial_PSD_inner_psd, true); // true indicating new method
        /////////////////////// kckim /////////////////////////////////////////////////
    }
	else if (parameters.initial_PSD_Type == "IPSDT_STEADY_STATE_TWO_ZONE") {
			// calculating as a steady state (for 3d)
			if (grid.L.size == 1) {
				throw error_msg("INITIAL_PSD_3D_METHOD_FOR_2D_GRID", "Initial PSD error: impossible to use STEADY_STATE_TWO_ZONE if L.size == 1");
			}
			Load_initial_f_two_zone(grid.L, grid.pc, grid.alpha, parameters.initial_PSD_tauSteadyState, parameters.initial_PSD_tauLppSteadyState, parameters.initial_PSD_Kp0, parameters.initial_PSD_some_constant_value, parameters.initial_PSD_J_L7_function, parameters.initial_PSD_outer_psd, parameters.initial_PSD_inner_psd);
	//////////////////////////////////////////////////////////////////////////////////
	} 
	else if (parameters.initial_PSD_Type == "IPSDT_STEADY_STATE_FROM_BOUNDARY") {
			// calculating as a steady state (for 3d)
			if (grid.L.size == 1) {
				throw error_msg("INITIAL_PSD_3D_METHOD_FOR_2D_GRID", "Initial PSD error: impossible to use STEADY_STATE_FROM_BOUNDARY if L.size == 1");
			}
			if (!L_UpperBoundaryCondition.arr.initialized) {
				throw error_msg("INITIAL_PSD_ERROR", "Initial PSD error: upper boundary is not initialized, can not be used for steady state calculation.");
			}
			// if tauSteadyState == 0 - using 6.0/Kp
			/// moved to parameters.cpp			if (parameters.initial_PSD_tauSteadyState <= 1.e-99) parameters.initial_PSD_tauSteadyState = 4.0/parameters.initial_PSD_Kp0;
			Load_initial_f_from_outer_L(grid.L, grid.pc, grid.alpha,
					parameters.initial_PSD_tauSteadyState, parameters.initial_PSD_Kp0,
					L_UpperBoundaryCondition.arr, parameters.initial_PSD_some_constant_value,
					parameters.initial_PSD_outer_psd, parameters.initial_PSD_inner_psd);
	}
    //not used
	//else if (parameters.initial_PSD_Type == "IPSDT_INTERPOLATION") {
			// Initial values for local grid as an interpolation from radial grid, or opposite
			// throw error_msg("INITIAL_PSD_TYPE_ERROR", "Using interpolation initial PSD type without second PSD to interpolate from");
	//}
	else if (parameters.initial_PSD_Type == "IPSDT_CONSTANT") {
			// just constant values
			for (il = 0; il < grid.L.size; il++)
				for (im = 0; im < grid.pc.size; im++)
					for (ia = 0; ia < grid.alpha.size; ia++)
						arr[il][im][ia] = parameters.initial_PSD_some_constant_value;
	} 
	else if (parameters.initial_PSD_Type == "IPSDT_FILE") {
			// loading from orbit flux (for 2d)
			Load_initial_f_file(grid.L, grid.epc, grid.alpha, parameters.initial_PSD_fileName.c_str(), false);
	} 
	else if (parameters.initial_PSD_Type == "IPSDT_FILE_GRID") {
			// loading from orbit flux (for 2d)
			Load_initial_f_file(grid.L, grid.epc, grid.alpha, parameters.initial_PSD_fileName.c_str(), true);
	} 
	else {
		throw error_msg("INITIAL_PSD_TYPE_UNKNOWN", "Unknown initial PSD type");
	}

	// Emptying loss cone
	if (parameters.initial_PSD_Type == "IPSDT_ORBIT_FLUX_2D" ||
		parameters.initial_PSD_Type == "IPSDT_STEADY_STATE" ||
        parameters.initial_PSD_Type == "IPSDT_STEADY_STATE_L7" ||
		parameters.initial_PSD_Type == "IPSDT_STEADY_STATE_FROM_BOUNDARY") {
			for (il = 0; il < grid.L.size; il++)
				for (im = 0; im < grid.pc.size; im++)
					for (ia = 0; ia < grid.alpha.size; ia++) {
                        // lc = loss cone calculation result in rad
                        double lc = VF::alc(grid.L.arr[il][im][ia]);
                        double a = grid.alpha.arr[il][im][ia];
						if (a < lc || a > VC::pi - lc)
							// arr[il][im][ia] = VC::zero_f;
							// alpha dependence inside the loss cone
							//if (grid.alpha.arr[il][im][ia] < lc || grid.alpha.arr[il][im][ia] > VC::pi - VF::alc(grid.L.arr[il][im][ia])) {
								//fLmp[*,i,*]=fLmp[*,i,*]*sinh(i/8.)^2/sinh(1)^2
								// ???????? arr[il][im][ia] = arr[il][im][ia] * pow(sinh(alpha[il][im][ia]/8.),2)/pow(sinh(1.0),2);
								arr[il][im][ia] = arr[il][im][ia] * pow(sinh(a/lc),2)/pow(sinh(1.0),2);
							//}
					}

	}



	// Output - create stream
	output_without_grid_file.open((parameters.output_PSD_folderName + parameters.output_PSD_fileName4D).c_str());
	output_without_grid_file << "VARIABLES = \"Phase_Space_Density\" " << endl;
	output_without_grid_file.setf(ios_base::scientific, ios_base::floatfield);

	// initialized = true;
}


/**
* Scaling initial PSD if changed upper L-boundary (not 7Re)
*/
/*
void PSD::ScaleBoundaryFlux(Parameters_structure::PSD parameters, GridElement &L, GridElement &pc, GridElement &alpha)
{

	Matrix1D<double> L_1d(61),fL(61),tau_0(61),Ke1(61);
	Matrix2D<double> scale_fm(61,pc.size);
	int il,im,ia;
	double L_upper, scale_f, alpha1,Lpp,tau_c,tau_p;
	double Kp = parameters.initial_PSD_Kp0;
	double tau = parameters.initial_PSD_tauSteadyState;
	double tauLpp = parameters.initial_PSD_tauLppSteadyState;
	double fb_in,fb_out,Lmin,Lmax,Lsize,dL;


	if (parameters.initial_PSD_Type == "IPSDT_STEADY_STATE") { // || parameters.initial_PSD_Type == "IPSDT_STEADY_STATE_TWO_ZONE") {
	// solve 1d steady-state eq. from L=1 to L=7

		fb_in = 0;
		fb_out = 1;
		Lmin = 1.0;
		Lmax = 7.0;
		dL = 0.1;
		Lsize = (Lmax-Lmin)/dL+1;

		for (il=0; il<Lsize; il++) L_1d[il] = 1+il*dL;          // make 1d grid
		steady_state(fL, tau, Kp, Lsize, L_1d, fb_out, fb_in);  // solve 1d steady state eq. for L=1 to 7

		L_upper = L[L.size-1][0][0];  // L-upper boundary we want to change, which is given in parameters.ini file

		if(L_upper != Lmax) {         // if L is not equal to 7
			unsigned i=0;             // interpolate PSD using Lagragian
			while (L_1d[++i] <= L_upper);
			scale_f = fL[i-1] * ( (L_upper-L_1d[i])/(L_1d[i-1]-L_1d[i]) )
				    + fL[i] * ( (L_upper-L_1d[i-1])/(L_1d[i]-L_1d[i-1]) );
		} else {                      // if L is equal to 7
			scale_f = fb_out;         // scale_f = 1;
		}


		for (il = 0; il < L.size; il++) {
			for (im = 0; im < pc.size; im++) {
				for (ia = 0; ia < alpha.size; ia++) {
					arr[il][im][ia] = arr[il][im][ia] * scale_f;

				}
			}
		}



	}


///

	if (parameters.initial_PSD_Type == "IPSDT_STEADY_STATE_TWO_ZONE") {

		fb_in = 0;
		fb_out = 1;
		Lmin = 1.0;
		Lmax = 7.0;
		dL = 0.1;
		Lsize = (Lmax-Lmin)/dL+1;
		//Lpp = (5.6 - 0.46*Kp);

		for (il=0; il<Lsize; il++) L_1d[il] = 1+il*dL;          // make 1d grid

		for (im = 0; im < pc.size; im++){
			for (il = 0; il < L.size; il++) {

				Ke1[il] = VF::Kfunc(pc[il][im][alpha.size-1]);   // Energy in MeV
				alpha1 = alpha[il][im][alpha.size-1];            // ~90 deg alpha
			//	cout<<Ke1[il]<<" "<<il<<Lsize<<endl;
			// make a array of tau
			//if(L_1d[il] >= Lpp) {   // calculate tau
			//	tau_0[il] = tau;
			//} else {    // calculate tauLpp
				if(tauLpp == 1e99) {
					tau_0[il] = VF::Coulomb(L_1d[il],Ke1[il]);
				} else if (tauLpp == 2e99) {
					tau_0[il] = VF::Precipitation(Kp,L_1d[il],Ke1[il]);
				} else if (tauLpp == 3e99) {
					tau_c = VF::Coulomb(L_1d[il],Ke1[il]);
					tau_p = VF::Precipitation(Kp,L_1d[il],Ke1[il]);
					tau_0[il] = 1.0 / (1.0/tau_p+1.0/tau_c);
				} else {
					tau_0[il] = tauLpp;
				}
			//cout<<tau_0[il]<<" "<<il<<endl;
			//}
			}
			steady_state_two_zone(fL, tau_0, Kp, alpha1, Ke1, L.size, L_1d, fb_out, fb_in);

			L_upper = L[L.size-1][0][0];  // L-upper boundary we want to change, which is given in parameters.ini file

			if(L_upper != Lmax) {         // if L is not equal to 7
				unsigned i=0;             // interpolate PSD using Lagragian
				while (L_1d[++i] <= L_upper);
				scale_fm[il][im] = fL[i-1] * ( (L_upper-L_1d[i])/(L_1d[i-1]-L_1d[i]) )
					    + fL[i] * ( (L_upper-L_1d[i-1])/(L_1d[i]-L_1d[i-1]) );
			} else {                      // if L is equal to 7
				scale_fm[il][im] = fb_out;         // scale_f = 1;
			}
		}
	// scaling PSD ///////////////////////
		for (il = 0; il < L.size; il++) {
			for (im = 0; im < pc.size; im++) {
				for (ia = 0; ia < alpha.size; ia++) {
					arr[il][im][ia] = arr[il][im][ia] * scale_fm[il][im];
				}
			}
		}
	//////////////////////////////////////
	}


}
*/


/**
 * Sources and losses term from the FP eq.
 *
 * \param parameters - parameters to be used
 * \param &L - L-grid array
 * \param &pc - pc-grid array
 * \param &alpha - alpha-grid array
 * \param &SL - sources/losses 3D matrix
 * \param dt - time step
 * \param Lpp - plasmapause location
 * \param tau - lifetime outside of the plasmapause
 * \param tauLpp - lifetime inside of the plasmapause
 * \param Kp - Kp index
 */
void PSD::SourcesAndLosses(
		Parameters_structure parameters,
		GridElement &L, GridElement &pc, GridElement &alpha,
		AdditionalSourcesAndLosses &SL,
		double dt, double Lpp,
		double tau, double tauLpp, double Kp)
{
    // iterators
	int il, im, ia;
    // lifetime in loss cone
	double taulc;
	///////////////// kckim /////////////////
	double tau_p,tau_c,tauLpp0;
	/////////////////////////////////////////

	for (il = 0; il < L.size; il++) {
		for (im = 0; im < pc.size; im++) {
			for (ia = 0; ia < alpha.size; ia++) {

				// Loss cone. - does not work here, needs too small time step to be split from pitch-angle diffusion
				// df/dt = -f/taulc


/*
				if (parameters.useLossCone && ((alpha.arr[il][im][ia] <= VF::alc(L.arr[il][im][0])) || (alpha.arr[il][im][ia] >= (VC::pi - VF::alc(L.arr[il][im][0])))) ) {
					// taulc = VF::bounce_time_res(VF::Kfunc(pc.arr[il][im][ia]), L.arr[il][im][ia])/60./60./24.;
					// Quarter bounce time
					taulc = 0.25 * VF::bounce_time_new(L.arr[il][im][ia], pc.arr[il][im][ia], alpha.arr[il][im][ia]);
					arr[il][im][ia] = arr[il][im][ia] / dt / (1/dt + 1/taulc); // implicit
					// arr[il][im][ia] = arr[il][im][ia] * exp(-dt/taulc); // some other
					// arr[il][im][ia] = exp(log(arr[il][im][ia]) - dt/taulc); // to not have PSD=0
				}
*/
				// Additional losses, if any
				// df/dt = -f/tau
				if (L.arr[il][im][ia] >= Lpp) {
					// arr[il][im][ia] = arr[il][im][ia] / dt / (1/dt + 1/tau); // implicit
						arr[il][im][ia] = arr[il][im][ia] - arr[il][im][ia] * (dt / tau); // explicit

				}  else  {
					// (*this)[il][im][ia] = (*this)[il][im][ia] / dt / (1/dt + 1/tauLpp); // implicit
                    
                    // kinetic energy computed
                    double k = VF::Kfunc(pc.arr[il][im][ia]);
					if(parameters.usetauLpp == "coulomb") {
                        // tau in days - initial
					    tauLpp0 = VF::Coulomb(L.arr[il][im][ia], k);
					} else if (parameters.usetauLpp == "precipitaion") {
						tauLpp0 = VF::Precipitation(Kp, L.arr[il][im][ia], k);
					} else if (parameters.usetauLpp == "combined") {
						tau_p = VF::Precipitation(Kp, L.arr[il][im][ia], k);
						tau_c = VF::Coulomb(L.arr[il][im][ia], k);
						tauLpp0 = 1.0/(1.0/tau_p+1.0/tau_c);
					} else {
						tauLpp0 = tauLpp;
					}
					arr[il][im][ia] = arr[il][im][ia] - arr[il][im][ia] * (dt / tauLpp0); // explicit, probably should be implicit! I.e.:
					//arr[il][im][ia] = arr[il][im][ia] / dt / (1/dt + 1/tauLpp0); 		// 	<-- implicit
				}



				// Add losses according to Magnetopause shadowing

				if (parameters.SL.useLmp == true)
					arr[il][im][ia] = arr[il][im][ia] * SL.arr[il][im][ia];


			}
		}
	} // end of loop over all grids

}


/**
* Mixed terms calculation of energy and pitch angle by explicit method. Uses split approximation for calculation.
*
* \param dt - time step
* \param Lpp - plasma pause location
* \param &Dpca - pc-alpha diffusion coefficient
* \param &DpcaLpp - pc-alpha diffusion coefficient under plasma pause location
* \param &L - grid element L
* \param &pc - grid element pc
* \param &alpha - grid element alpha
* \param &Jacobian - jacobian
* \param &pc_lowerBoundaryCondition - lower boundary condition on pc
* \param &pc_upperBoundaryCondition - upper boundary condition on pc
* \param &alpha_lowerBoundaryCondition - lower boundary condition on alpha
* \param &alpha_upperBoundaryCondition  - upper boundary condition on alpha
* \param pc_lowerBoundaryCondition_calculationType - pc lower boundary condition type,
* \param pc_upperBoundaryCondition_calculationType - pc upper boundary condition type,
* \param alpha_lowerBoundaryCondition_calculationType - alpha lower boundary condition type,
* \param alpha_upperBoundaryCondition_calculationType - alpha upper boundary condition type.
*/
void PSD::DiffusionMixTermExplicit(double dt, double Lpp,
								   DiffusionCoefficient &Dpca, DiffusionCoefficient &DpcaLpp,
								   GridElement &L, GridElement &pc, GridElement &alpha, Matrix3D<double> Jacobian,
								   Matrix2D<double> pc_lowerBoundaryCondition,
								   Matrix2D<double> pc_upperBoundaryCondition,
								   Matrix2D<double> alpha_lowerBoundaryCondition,
								   Matrix2D<double> alpha_upperBoundaryCondition,
								   string pc_lowerBoundaryCondition_calculationType,
								   string pc_upperBoundaryCondition_calculationType,
								   string alpha_lowerBoundaryCondition_calculationType,
								   string alpha_upperBoundaryCondition_calculationType)
{

	int il, im, ia;

	// Declaration of arrays for derivatives
	static Matrix3D<double> PSD_Der_a_L(L.size, pc.size, alpha.size);
	static Matrix3D<double> PSD_Der_a_R(L.size, pc.size, alpha.size);
	static Matrix3D<double> PSD_Der_m_L(L.size, pc.size, alpha.size);
	static Matrix3D<double> PSD_Der_m_R(L.size, pc.size, alpha.size);

	// Some variables
    // center, alpha left, alpha right, pc left, pc right
	double Dpca_, Dpca_a_L, Dpca_a_R, Dpca_m_L, Dpca_m_R;
	double J_, J_a_L, J_a_R, J_m_L, J_m_R;
    // alpha pc left, pc alpha left, alpha pc right, pc alpha right
	double Laml, Lmal, Lamr, Lmar;
	double PSD_m_r, PSD_m_l, PSD_a_r, PSD_a_l, Lam, Lma;

	// Different way of derivatives approximation
    /**
     * \brief \b AM_Split_LR
     *
     * \f$ LHS = G(GridElement) * Dpca * \frac {PSD[x]-PSD[x-1]}{GridElement[x]-GridElement[x-1]} \f$
     *
     * \f$ RHS = G(GridElement) * Dpca * \frac {PSD[x+1]-PSD[x]}{GridElement[x+1]-GridElement[x]} \f$
     * for VF::G() and x is iteration index and GridElement is pc or alpha
     */
	if (PSD_parameters.approximationMethod == "AM_Split_LR") {

			// Derivatives calculation
			for (il = 0; il < L.size; il++) {
				for (im = 0; im < pc.size; im++) {
					for (ia = 1; ia < alpha.size-1; ia++) {
                        // LHS derivative for alpha
						PSD_Der_a_L[il][im][ia] = VF::G(alpha.arr[il][im][ia]) * Dpca.arr[il][im][ia]*(arr[il][im][ia] - arr[il][im][ia-1])/(alpha.arr[il][im][ia] - alpha.arr[il][im][ia-1]);
                        // RHS derivative for alpha
						PSD_Der_a_R[il][im][ia] = VF::G(alpha.arr[il][im][ia]) * Dpca.arr[il][im][ia]*(arr[il][im][ia+1] - arr[il][im][ia])/(alpha.arr[il][im][ia+1] - alpha.arr[il][im][ia]);
					}
				}
				for (im = 1; im < pc.size-1; im++) {
					for (ia = 0; ia < alpha.size; ia++) {
                        // LHS derivative for pc
						PSD_Der_m_L[il][im][ia] = pc.arr[il][im][ia] * pc.arr[il][im][ia] * Dpca.arr[il][im][ia]*(arr[il][im][ia] - arr[il][im-1][ia])/(pc.arr[il][im][ia] - pc.arr[il][im-1][ia]);
                        // RHS derivative for pc
						PSD_Der_m_R[il][im][ia] = pc.arr[il][im][ia] * pc.arr[il][im][ia] * Dpca.arr[il][im][ia]*(arr[il][im+1][ia] - arr[il][im][ia])/(pc.arr[il][im+1][ia] - pc.arr[il][im][ia]);
					}
				}
			}

			// Approximation and calculation
			for (il = 0; il < L.size; il++) {
				for (im = 1; im < pc.size-1; im++) {
					for (ia = 1; ia < alpha.size-1; ia++) {
						// for debugging...
						Dpca_ = Dpca.arr[il][im][ia];
						Dpca_a_L = Dpca.arr[il][im][ia-1];
						Dpca_a_R = Dpca.arr[il][im][ia+1];
						Dpca_m_L = Dpca.arr[il][im-1][ia];
						Dpca_m_R = Dpca.arr[il][im+1][ia];
						J_ = Jacobian[il][im][ia];
						J_a_L = Jacobian[il][im][ia-1];
						J_a_R = Jacobian[il][im][ia+1];
						J_m_L = Jacobian[il][im-1][ia];
						J_m_R = Jacobian[il][im+1][ia];


						// d/dp D d/da f
                        /**
                         * \f$ Lmar = \frac{1.0}{{pc}^2} * \frac{RHS[x] - RHS[x-1]}{pc[x] - pc[x-1]} \f$
                         * \f$ Lmal = \frac{1.0}{{pc}^2} * \frac{LHS[x+1] - LHS[x]}{pc[x+1] - pc[x]} \f$
                         */
						Lmar = 1.0/(pc.arr[il][im][ia] * pc.arr[il][im][ia]) * (PSD_Der_a_R[il][im][ia] - PSD_Der_a_R[il][im-1][ia])/(pc.arr[il][im][ia] - pc.arr[il][im-1][ia]);
						Lmal = 1.0/(pc.arr[il][im][ia] * pc.arr[il][im][ia]) * (PSD_Der_a_L[il][im+1][ia] - PSD_Der_a_L[il][im][ia])/(pc.arr[il][im+1][ia] - pc.arr[il][im][ia]);

						// d/da D d/dp f
                        /**
                         * \f$ Lamr = \frac{1.0}{G(\alpha)} * \frac{RHS[x]-RHS[x-1]}{\alpha[x] - \alpha[x-1]} \f$
                         * \f$ Laml = \frac{1.0}{G(\alpha)} * \frac{LHS[x+1]-LHS[x]}{\alpha[x+1] - \alpha[x]} \f$
                         */
						Lamr = 1.0/(VF::G(alpha.arr[il][im][ia])) * (PSD_Der_m_R[il][im][ia] - PSD_Der_m_R[il][im][ia-1])/(alpha.arr[il][im][ia] - alpha.arr[il][im][ia-1]);
						Laml = 1.0/(VF::G(alpha.arr[il][im][ia])) * (PSD_Der_m_L[il][im][ia+1] - PSD_Der_m_L[il][im][ia])/(alpha.arr[il][im][ia+1] - alpha.arr[il][im][ia]);

                        // approximation calculation using L's
                        /// \f$ PSD += \frac{dt}{2} * (Lmar + Lamr + Lmal + Laml) \f$ \n
						arr[il][im][ia] = arr[il][im][ia] + dt * 0.5 * (Lmar + Lamr + Lmal + Laml);

						/*						L12p = 1/J*(Jacobian[il][im+1][ia]*Dpca[il][im+1][ia]*PSD_Der_a_L[il][im+1][ia] -
						Jacobian[il][im  ][ia]*Dpca[il][im  ][ia]*PSD_Der_a_L[il][im  ][ia])/(pc[il][im+1][ia] - pc[il][im  ][ia]);
						L12m = 1/J*(Jacobian[il][im  ][ia]*Dpca[il][im  ][ia]*PSD_Der_a_R[il][im  ][ia] -
						Jacobian[il][im-1][ia]*Dpca[il][im-1][ia]*PSD_Der_a_R[il][im-1][ia])/(pc[il][im  ][ia] - pc[il][im-1][ia]);

						L21p = 1/J*(Jacobian[il][im][ia+1]*Dpca[il][im][ia+1]*PSD_Der_m_L[il][im][ia+1] -
						Jacobian[il][im][ia  ]*Dpca[il][im][ia  ]*PSD_Der_m_L[il][im][ia  ])/(alpha[il][im][ia+1] - alpha[il][im][ia]);
						L21m = 1/J*(Jacobian[il][im][ia  ]*Dpca[il][im][ia  ]*PSD_Der_m_R[il][im][ia  ] -
						Jacobian[il][im][ia-1]*Dpca[il][im][ia-1]*PSD_Der_m_R[il][im][ia-1])/(alpha[il][im][ia] - alpha[il][im][ia-1]);*/
					}
				}
			}
	}
    /**
     * \brief \b AM_Split_C
     *
     * \f$ PSD\_m\_r = Dpca[m+1] * Jacobian[m+1] * \frac{PSD[m+1][a+1] - PSD[m+1][a-1]}{\alpha[m+1][a+1] - \alpha[m+1][a-1]} \f$
     * \f$ PSD\_m\_l = Dpca[m-1] * Jacobian[m-1] * \frac{PSD[m-1][a+1] - PSD[m-1][a-1]}{\alpha[m-1][a+1] - \alpha[m-1][a-1]} \f$
     * \f$ Lma = \frac{1.0}{Jacobian} * \frac{PSD\_m\_r - PSD\_m\_l}{pc[m+1] - pc[m-1]} \f$
     *
     * \f$ PSD\_a\_r = Dpca[a+1] * Jacobian[a+1] * \frac{PSD[m+1][a+1] - PSD[m-1][a+1]}{pc[m+1][a+1] - pc[m-1][a+1]} \f$
     * \f$ PSD\_a\_l = Dpca[a-1] * Jacobian[a-1] * \frac{PSD[m+1][a-1] - PSD[m-1][a-1]}{pc[m+1][a-1] - pc[m-1][a-1]} \f$
     * \f$ Lam = \frac{1.0}{Jacobian} * \frac{PSD\_a\_r - PSD\_a\_l}{\alpha[a+1] - \alpha[a-1]} \f$
     *
     * \f$ PSD += dt * (Lma + Lam) \f$
     */
    else if (PSD_parameters.approximationMethod == "AM_Split_C") { // was not fully checked, may not work
			for (il = 0; il < L.size; il++) {
				for (im = 1; im < pc.size-1; im++) {
					for (ia = 1; ia < alpha.size-1; ia++) {
						// for debugging...
						Dpca_ = Dpca.arr[il][im][ia];
						Dpca_a_L = Dpca.arr[il][im][ia-1];
						Dpca_a_R = Dpca.arr[il][im][ia+1];
						Dpca_m_L = Dpca.arr[il][im-1][ia];
						Dpca_m_R = Dpca.arr[il][im+1][ia];
						J_ = Jacobian[il][im][ia];
						J_a_L = Jacobian[il][im][ia-1];
						J_a_R = Jacobian[il][im][ia+1];
						J_m_L = Jacobian[il][im-1][ia];
						J_m_R = Jacobian[il][im+1][ia];

						// d/dp D d/da f
						PSD_m_r = Dpca.arr[il][im+1][ia]*Jacobian[il][im+1][ia]*(arr[il][im+1][ia+1] - arr[il][im+1][ia-1])/(alpha.arr[il][im+1][ia+1] - alpha.arr[il][im+1][ia-1]);
						PSD_m_l = Dpca.arr[il][im-1][ia]*Jacobian[il][im-1][ia]*(arr[il][im-1][ia+1] - arr[il][im-1][ia-1])/(alpha.arr[il][im-1][ia+1] - alpha.arr[il][im-1][ia-1]);
						Lma  = 1.0/Jacobian[il][im][ia] * (PSD_m_r - PSD_m_l) / (pc.arr[il][im+1][ia] - pc.arr[il][im-1][ia]);

						// d/da D d/dm f
						PSD_a_r = Dpca.arr[il][im][ia+1]*Jacobian[il][im][ia+1]*(arr[il][im+1][ia+1] - arr[il][im-1][ia+1])/(pc.arr[il][im+1][ia+1] - pc.arr[il][im-1][ia+1]);
						PSD_a_l = Dpca.arr[il][im][ia-1]*Jacobian[il][im][ia-1]*(arr[il][im+1][ia-1] - arr[il][im-1][ia-1])/(pc.arr[il][im+1][ia-1] - pc.arr[il][im-1][ia-1]);
						Lam  = 1.0/Jacobian[il][im][ia] * (PSD_a_r - PSD_a_l) / (alpha.arr[il][im][ia+1] - alpha.arr[il][im][ia-1]);

						//						cout << ((Lma + Lam) - (Lmar + Lamr + Lmal + Laml))/((Lma + Lam) + (Lmar + Lamr + Lmal + Laml)) << "%" << endl;
                        // approximation calculation using L's
						arr[il][im][ia] = arr[il][im][ia] + dt * (Lma + Lam);
					}
				}
			}

	} else {
		throw error_msg("APPROXIMATION", "Unknown approximation method for mixed terms");
	}

}


/// \todo lot's of repeated code within diffusion functions. maybe find a way to condense them


/** Pitch angle diffusion calculation function.
* Takes diffusion coefficients, boundary conditions and grid elements as input.
* It calculates model matrix by MakeModelMatrix_3D for 3D problem: A_3D * PSD_3D(t+1) = B_3D * PSD_3D(t) + C_3D
* then splits it for 1D problems: A * PSD_1D(t+1) = B * PSD_1D(t) + C
* and solve by tridiag method.
*
* \param dt - time step
* \param Lpp - plasma pause location
* \param &Daa - L diffusion coefficient
* \param &DaaLpp - L diffusion coefficient
* \param &L - grid element L
* \param &pc - grid elelent pc
* \param &alpha - grid element alpha
* \param &Jacobian - jacobian
* \param &alpha_lowerBoundaryCondition - lower boundary condition
* \param &alpha_upperBoundaryCondition - upper boundary condition
* \param &alpha_lowerBoundaryCondition_calculationType - lower boundary condition type (on value/on derivative)
* \param &alpha_upperBoundaryCondition_calculationType - upper boundary condition type (on value/on derivative)
*/
void PSD::Diffusion_alpha(double dt, double Lpp,
							   DiffusionCoefficient &Daa, DiffusionCoefficient &DaaLpp,
							   GridElement &L, GridElement &pc, GridElement &alpha, Matrix3D<double> Jacobian,
							   Matrix2D<double> alpha_lowerBoundaryCondition,
							   Matrix2D<double> alpha_upperBoundaryCondition,
							   string alpha_lowerBoundaryCondition_calculationType,
							   string alpha_upperBoundaryCondition_calculationType)
{

    // seems to be no longer used
	//char out_name_matr_a[256], out_name_matr_b[256], out_name_matr_c[256];

	// Calculation matrices A, B, and C for 1d diffusion problem
	// A * PSD_1D(t+1) = B * PSD_1D(t) + C, for each L and Energy

	// Creation is moved to PSD.h, so it can be accessed from Main.cpp
	//static CalculationMatrix
	//	matr_A(alpha.size, 1, 1, 1),
	//	matr_B(alpha.size, 1, 1, 0),
	//	matr_C(alpha.size, 1, 1, 0);
	matr_A.Initialize(alpha.size, 1, 1, 1);
	matr_B.Initialize(alpha.size, 1, 1, 0),
	matr_C.Initialize(alpha.size, 1, 1, 0);


	// Create zero matrices, temporary matrices to be used later
	static Matrix3D<double> Zero_Matrix(1, 1, alpha.size);
	static Matrix2D<double> Zero_Matrix_2d(1, 1);
	static Matrix1D<double> Zero_Matrix_1d(alpha.size);
	Zero_Matrix = 0;
	Zero_Matrix_2d = 0;
	Zero_Matrix_1d = 0;

	// Create 1d arrays
	// Creating 1d arrays for PSD and right hand side (RHS = B * PSD + C)
	static Matrix1D<double> slicePSD1D(alpha.size), RHS(alpha.size);

	// Creating 1d arrays which will be sent to the MakeMatrix function
	// Creating 1d matrixes, but with type of Matrix3D, cause we need that type for MakeMatrix function
    // Create 3d and 2d matrices instead of 1d and single values so that it can meet parameters in MakeModelMatrix_3D()
	static Matrix3D<double> Daa_1d(1, 1, alpha.size), DaaLpp_1d(1, 1, alpha.size), Jacobian_1d(1, 1, alpha.size), alpha_1d(1, 1, alpha.size), L_1d(1, 1, alpha.size), pc_1d(1, 1, alpha.size);
	// Creating temporary arrays size 1x1 for boundary conditions, which are just values
	static Matrix2D<double> alpha_lowerBoundaryCondition_1d(1, 1), alpha_upperBoundaryCondition_1d(1, 1);

	// iterators
	int il, im, ia, in; // matrix element numbers on 3D grid
	DiagMatrix::iterator it;
	for (il = 0; il < L.size; il++) {
		for (im = 0; im < pc.size; im++) {

			// Copying 3D arrays into 1d arrays
			for (ia = 0; ia < alpha.size; ia++) {
				slicePSD1D[ia] = arr[il][im][ia];

				Daa_1d[0][0][ia] = Daa.arr[il][im][ia];
				DaaLpp_1d[0][0][ia] = DaaLpp.arr[il][im][ia];

				L_1d[0][0][ia] = L.arr[il][im][ia];
				pc_1d[0][0][ia] = pc.arr[il][im][ia];
				alpha_1d[0][0][ia] = alpha.arr[il][im][ia];

				Jacobian_1d[0][0][ia] = Jacobian[il][im][ia];
			}
			// boundary conditions, these are arrays size 1x1, i.e. values
			alpha_upperBoundaryCondition_1d[0][0] = alpha_upperBoundaryCondition[il][im];
			alpha_lowerBoundaryCondition_1d[0][0] = alpha_lowerBoundaryCondition[il][im];

			// Make A, B, C matrices
			MakeModelMatrix_3D(
				matr_A, matr_B, matr_C, // matrices A, B, C
				L_1d, pc_1d, alpha_1d, // L, pc, alpha grids. We use L crossection here
				1, 1, alpha.size, // L-size, pc-size, alpha-size
				Zero_Matrix_2d, Zero_Matrix_2d, // L-boundaries
				Zero_Matrix_2d, Zero_Matrix_2d, // pc-boundaries
				alpha_lowerBoundaryCondition_1d, alpha_upperBoundaryCondition_1d, // alpha-boundaries
				"BCT_CONSTANT_VALUE", "BCT_CONSTANT_VALUE", // Boundary conditions at L-boundaries
				"BCT_CONSTANT_VALUE", "BCT_CONSTANT_VALUE", // Boundary conditions at pc-boundaries
				alpha_lowerBoundaryCondition_calculationType, alpha_upperBoundaryCondition_calculationType, // Boundary conditions at alpha-boundaries
				Zero_Matrix, // DLL - use 0 matrix
				Zero_Matrix, Zero_Matrix, // Dpp, DppLpp - use 0 matrix
				Daa_1d, DaaLpp_1d, // Daa, DaaLpp
				Zero_Matrix, Zero_Matrix, // Dpca, Dpca_Lpp - use 0 matrix
				Jacobian_1d, dt, Lpp);

			/*Output::echo(0, "Writing calculation matrices. ");
			sprintf(out_name_matr_a, "./Debug/alpha_matr_A_%d_%d_0.dat", il, im);
			sprintf(out_name_matr_b, "./Debug/alpha_matr_B_%d_%d_0.dat", il, im);
			sprintf(out_name_matr_c, "./Debug/alpha_matr_C_%d_%d_0.dat", il, im);
			matr_A.writeToFile(out_name_matr_a);
			matr_B.writeToFile(out_name_matr_b);
			matr_C.writeToFile(out_name_matr_c);
			Output::echo(0, "done.\n");*/

			// make RHS = B*f + C
			for (in = 0; in < alpha.size; in++) {
				// If assume matr_B is 1d and has only +1, 0, -1 diagonals, can use this:
				//				RHS[in] = matr_B[-1][in] * slicePSD1D[in-1] + matr_B[0][in] * slicePSD1D[in] + matr_B[+1][in] * slicePSD1D[in+1] + matr_C[0][in];
				RHS[in] = matr_C[0][in];
				for (it = matr_B.begin(); it != matr_B.end(); it++) {
					// multiplication B * f
					if (in + it->first >= 0 && in + it->first < matr_B.total_size) {
						RHS[in] += it->second[in] * slicePSD1D[in + it->first];
					}
				}
			}

			// right hand side to files
			//			RHS.writeToFile("alpha_RHS.dat");

			// Solving model matrix, obtaining PSD
			// Assume matr_A has only +1, 0, -1 diagonals
			// TODO: Check if matr_A has only +1, 0, -1 diagonals, i.e. is tridiagonal
			tridag(
				&matr_A[-1][0], // just a way to get a pointer to first element of an array
				&matr_A[0][0],
				&matr_A[+1][0],
				&RHS[0],
				&slicePSD1D[0],
				alpha.size);

			// copy PSD back, from 1d array to 3d array
			for (ia = 0; ia < alpha.size; ia++) {
				arr[il][im][ia] = slicePSD1D[ia];
			}
		}
	}
}


/** Energy diffusion calculation function.
* Takes diffusion coefficients, boundary conditions and grid elements as input.
* It calculates model matrix by MakeModelMatrix_3D for 3D problem: A_3D * PSD_3D(t+1) = B_3D * PSD_3D(t) + C_3D
* then splits it for 1D problems: A * PSD_1D(t+1) = B * PSD_1D(t) + C
* and solve by tridiag method.
*
* \param dt - time step
* \param Lpp - plasma pause location
* \param &Dpcpc - L diffusion coefficient
* \param &DpcpcLpp - L diffusion coefficient with plasma pause location
* \param &L - grid element L
* \param &pc - grid element pc
* \param &alpha - grid element alpha
* \param &Jacobian - jacobian
* \param &pc_lowerBoundaryCondition - lower boundary condition
* \param &pc_upperBoundaryCondition - upper boundary condition
* \param &pc_lowerBoundaryCondition_calculationType - lower boundary condition type (on value/on derivative)
* \param &pc_upperBoundaryCondition_calculationType - upper boundary condition type (on value/on derivative)
*/
void PSD::Diffusion_pc(double dt, double Lpp,
							DiffusionCoefficient &Dpcpc, DiffusionCoefficient &DpcpcLpp,
							GridElement &L, GridElement &pc, GridElement &alpha, Matrix3D<double> Jacobian,
							Matrix2D<double> pc_lowerBoundaryCondition,
							Matrix2D<double> pc_upperBoundaryCondition,
							string pc_lowerBoundaryCondition_calculationType,
							string pc_upperBoundaryCondition_calculationType)
{
    // no longer used
	// char out_name_matr_a[256], out_name_matr_b[256], out_name_matr_c[256];


	// Calculation matrices A, B, and C for 1d diffusion problem
	// A * PSD_1D(t+1) = B * PSD_1D(t) + C, for each L and Energy

	// Creation is moved to PSD.h, so it can be accessed from Main.cpp
	//static CalculationMatrix
	//	matr_A(1, pc.size, 1, 1),
	//	matr_B(1, pc.size, 1, 0),
	//	matr_C(1, pc.size, 1, 0);
	matr_A.Initialize(1, pc.size, 1, 1);
	matr_B.Initialize(1, pc.size, 1, 0),
	matr_C.Initialize(1, pc.size, 1, 0);

	// Create zero matrices, temporary matrices to be used later
	static Matrix3D<double> Zero_Matrix(1, pc.size, 1);
	static Matrix2D<double> Zero_Matrix_2d(1, 1);
	static Matrix1D<double> Zero_Matrix_1d(alpha.size);
	Zero_Matrix = 0;
	Zero_Matrix_2d = 0;
	Zero_Matrix_1d = 0;

	// Create 1d arrays
	// Creating 1d arrays for PSD and right hand side (RHS = B * PSD + C)
	static Matrix1D<double> slicePSD1D(pc.size), RHS(pc.size);

	// Creating 1d arrays which will be sent to the MakeMatrix function
	// Creating 1d matrixes, but with type of Matrix3D, cause we need that type for MakeMatrix function
    // Create 3d and 2d matrices instead of 1d and single values so that it can meet parameters in MakeModelMatrix_3D()
	static Matrix3D<double> Dpcpc_1d(1, pc.size, 1), DpcpcLpp_1d(1, pc.size, 1), Jacobian_1d(1, pc.size, 1), L_1d(1, pc.size, 1), pc_1d(1, pc.size, 1), alpha_1d(1, pc.size, 1);
	// Creating temporary arrays size 1x1 for boundary conditions, which are just values
	static Matrix2D<double> pc_lowerBoundaryCondition_1d(1, 1), pc_upperBoundaryCondition_1d(1, 1);

	// iterators
	int il, im, ia, in; // matrix element numbers on 3D grid
	DiagMatrix::iterator it;
	for (il = 0; il < L.size; il++) {
		for (ia = 0; ia < alpha.size; ia++) {

			// Copying 3D arrays into 1d arrays
			for (im = 0; im < pc.size; im++) {
				// Copying 3D PSD to 1d PSD
				slicePSD1D[im] = arr[il][im][ia];

				Dpcpc_1d[0][im][0] = Dpcpc.arr[il][im][ia];
				DpcpcLpp_1d[0][im][0] = DpcpcLpp.arr[il][im][ia];

				L_1d[0][im][0] = L.arr[il][im][ia];
				pc_1d[0][im][0] = pc.arr[il][im][ia];
				alpha_1d[0][im][0] = alpha.arr[il][im][ia];

				Jacobian_1d[0][im][0] = Jacobian[il][im][ia];
			}
			// boundary conditions, these are arrays size 1x1, i.e. values
			pc_upperBoundaryCondition_1d[0][0] = pc_upperBoundaryCondition[il][ia];
			pc_lowerBoundaryCondition_1d[0][0] = pc_lowerBoundaryCondition[il][ia];

			// Make A, B, C matrices
			MakeModelMatrix_3D(
				matr_A, matr_B, matr_C, // matrices A, B, C
				L_1d, pc_1d, alpha_1d,  // L, pc, alpha grids.
				1, pc.size, 1, // L-size, pc-size, alpha-size
				Zero_Matrix_2d, Zero_Matrix_2d, // L-boundaries
				pc_lowerBoundaryCondition_1d, pc_upperBoundaryCondition_1d, // pc-boundaries
				Zero_Matrix_2d, Zero_Matrix_2d, // alpha-boundaries
				"BCT_CONSTANT_VALUE", "BCT_CONSTANT_VALUE", // Boundary conditions at L-boundaries
				pc_lowerBoundaryCondition_calculationType, pc_upperBoundaryCondition_calculationType, // Boundary conditions at pc-boundaries
				"BCT_CONSTANT_VALUE", "BCT_CONSTANT_VALUE", // Boundary conditions at alpha-boundaries
				Zero_Matrix,  // DLL - use 0 matrix
				Dpcpc_1d, DpcpcLpp_1d, // Dpp, DppLpp
				Zero_Matrix, Zero_Matrix, // Daa, DaaLpp - use 0 matrix
				Zero_Matrix, Zero_Matrix, // Dpca, Dpca_Lpp - use 0 matrix
				Jacobian_1d, dt, Lpp);

			/*Output::echo(0, "Writing calculation matrices. ");
			sprintf(out_name_matr_a, "./Debug/pc_matr_A_%d_0_%d.dat", il, ia);
			sprintf(out_name_matr_b, "./Debug/pc_matr_B_%d_0_%d.dat", il, ia);
			sprintf(out_name_matr_c, "./Debug/pc_matr_C_%d_0_%d.dat", il, ia);
			matr_A.writeToFile(out_name_matr_a);
			matr_B.writeToFile(out_name_matr_b);
			matr_C.writeToFile(out_name_matr_c);
			Output::echo(0, "done.\n");*/

			// make RHS = B*f + C
			DiagMatrix::iterator it;
			for (in = 0; in < pc.size; in++) {
				// I guess matr_B has only main diagonal in implicit method, so it can be used like this:
				//				RHS[in] = matr_B[0][in] * slicePSD1D[in] + matr_C[0][in];
				RHS[in] = matr_C[0][in];
				for (it = matr_B.begin(); it != matr_B.end(); it++) {
					// multiplication B * f
					if (in + it->first >= 0 && in + it->first < matr_B.total_size) {
						RHS[in] += it->second[in] * slicePSD1D[in + it->first];
					}
				}
			}

			// right hand side to files
			//			RHS.writeToFile("pc_RHS.dat");

			// Solving model matrix, obtaining PSD
			// Assume matr_A has only +1, 0, -1 diagonals
			// TODO: Check if matr_A has only +1, 0, -1 diagonals, i.e. is tridiagonal
			tridag(
				&matr_A[-1][0], // just a way to get a pointer to first element of an array
				&matr_A[0][0],
				&matr_A[+1][0],
				&RHS[0],
				&slicePSD1D[0],
				pc.size);

			// copying PSD back, from 1d array to 3d array
			for (im = 0; im < pc.size; im++) {
				arr[il][im][ia] = slicePSD1D[im];
			}
		}
	}
}


/** Radial diffusion calculation function.
* Takes diffusion coefficients, boundary conditions and grid elements as parameters.
* It caltulates model matrix by makeMatrix and then solve it by SolveMatrix method.
*
* \param dt - time step
* \param Lpp - plasma pause location
* \param &DLL - L diffusion coefficient
* \param &L - grid element L
* \param &pc - grid elelent pc
* \param &alpha - grid element alpha
* \param &Jacobian - jacobian
* \param &L_lowerBoundaryCondition - lower boundary condition on L
* \param &L_upperBoundaryCondition - upper boundary condition on L
* \param &L_lowerBoundaryCondition_calculationType - lower boundary condition on L calculation type
* \param &L_upperBoundaryCondition_calculationType - upper boundary condition on L calculation type
* \param tau - life time upper location of the plasma pause
* \param tauLpp - life time lower location of the plasma pause
*/
void PSD::Diffusion_L(	double dt, double Lpp,
						   DiffusionCoefficient &DLL,
						   GridElement &L, GridElement &pc, GridElement &alpha, Matrix3D<double> Jacobian,
						   Matrix2D<double> L_lowerBoundaryCondition,
						   Matrix2D<double> L_upperBoundaryCondition,
						   string L_lowerBoundaryCondition_calculationType,
						   string L_upperBoundaryCondition_calculationType,
						   double tau, double tauLpp)
{
    // no longer used
	// char out_name_matr_a[256], out_name_matr_b[256], out_name_matr_c[256];


	// Calculation matrices A, B, and C for 1d diffusion problem
	// A * PSD_1D(t+1) = B * PSD_1D(t) + C, for each L and Energy

	// Creation is moved to PSD.h, so it can be accessed from Main.cpp
	//static CalculationMatrix
	//	matr_A(1, 1, L.size, 1),
	//	matr_B(1, 1, L.size, 0),
	//	matr_C(1, 1, L.size, 0);
	matr_A.Initialize(1, 1, L.size, 1);
	matr_B.Initialize(1, 1, L.size, 0),
	matr_C.Initialize(1, 1, L.size, 0);

	// Create zero matrices, temporary matrices to be used later
	static Matrix3D<double> Zero_Matrix(L.size, 1, 1);
	static Matrix2D<double> Zero_Matrix_2d(1, 1);
	static Matrix1D<double> Zero_Matrix_1d(1);
	Zero_Matrix = 0;
	Zero_Matrix_2d = 0;
	Zero_Matrix_1d = 0;

	// Create 1d arrays
	// Creating 1d arrays for PSD and right hand side (RHS = B * PSD + C)
	static Matrix1D<double> slicePSD1D(L.size), RHS(L.size);

	// Creating 1d arrays which will be send to the MakeMatrix function
	// Creating 1d matrixes, but with type of Matrix3D, cause we need that type for MakeMatrix function
    // Create 3d and 2d matrices instead of 1d and single values so that it can meet parameters in MakeModelMatrix_3D()
	static Matrix3D<double> DLL_1d(L.size, 1, 1), L_1d(L.size, 1, 1), Jacobian_1d(L.size, 1, 1); // pc_1d(L.size, 1, 1), alpha_1d(L.size, 1, 1)
	// Creating temporary arrays size 1x1 for boundary conditions, which are just values
	static Matrix2D<double> L_lowerBoundaryCondition_1d(1, 1), L_upperBoundaryCondition_1d(1, 1);

	// iterators
	int il, im, ia, in; // matrix element numbers on 3D grid
	DiagMatrix::iterator it;
	for (ia = 0; ia < alpha.size; ia++) {
		for (im = 0; im < pc.size; im++) {

			// Copying 3D arrays into 1d arrays
			for (il = 0; il < L.size; il++) {
				// Copying 3D PSD to 1d PSD
				slicePSD1D[il] = arr[il][im][ia];

				DLL_1d[il][0][0] = DLL.arr[il][im][ia];

				L_1d[il][0][0] = L.arr[il][im][ia];
				// pc_1d[il][0][0] = pc[il][im][ia];
				// alpha_1d[il][0][0] = alpha[il][im][ia];

				Jacobian_1d[il][0][0] = Jacobian[il][im][ia];
			}
			// boundary conditions, these are arrays size 1x1, i.e. values
			L_upperBoundaryCondition_1d[0][0] = L_upperBoundaryCondition[im][ia];
			L_lowerBoundaryCondition_1d[0][0] = L_lowerBoundaryCondition[im][ia];

			// Make A, B, C matrices
			MakeModelMatrix_3D(
				matr_A, matr_B, matr_C, // matrices A, B, C
				L_1d, Zero_Matrix, Zero_Matrix, // L, pc, alpha grids. We use L grid only here
				L.size, 1, 1, // L-size, pc-size, alpha-size
				L_lowerBoundaryCondition_1d, L_upperBoundaryCondition_1d, // L-boundaries
				Zero_Matrix_2d, Zero_Matrix_2d, // pc-boundaries
				Zero_Matrix_2d, Zero_Matrix_2d, // alpha-boundaries
				L_lowerBoundaryCondition_calculationType, L_upperBoundaryCondition_calculationType, // Boundary conditions at L-boundaries
				"BCT_CONSTANT_VALUE", "BCT_CONSTANT_VALUE", // Boundary conditions at pc-boundaries
				"BCT_CONSTANT_VALUE", "BCT_CONSTANT_VALUE", // Boundary conditions at alpha-boundaries
				DLL_1d, // DLL
				Zero_Matrix, Zero_Matrix, // Dpp, DppLpp - use 0
				Zero_Matrix, Zero_Matrix, // Daa, DaaLpp - use 0
				Zero_Matrix, Zero_Matrix, // Dpca, Dpca_Lpp - use 0
				Jacobian_1d, dt, Lpp,
				tau, tauLpp);

			/*
			Output::echo(0, "Writing calculation matrices. ");
			sprintf(out_name_matr_a, "./Debug/L_matr_A_0_%d_%d.dat", im, ia);
			sprintf(out_name_matr_b, "./Debug/L_matr_B_0_%d_%d.dat", im, ia);
			sprintf(out_name_matr_c, "./Debug/L_matr_C_0_%d_%d.dat", im, ia);
			matr_A.writeToFile(out_name_matr_a);
			matr_B.writeToFile(out_name_matr_b);
			matr_C.writeToFile(out_name_matr_c);
			Output::echo(0, "done.\n");
			*/

			// make RHS = B*f + C
			DiagMatrix::iterator it;
			for (in = 0; in < L.size; in++) {
				// I guess matr_B has only main diagonal in implicit method, so it can be used like this:
				//				RHS[in] = matr_B[0][in] * slicePSD1D[in] + matr_C[0][in];
				RHS[in] = matr_C[0][in];
				for (it = matr_B.begin(); it != matr_B.end(); it++) {
					// multiplication B * f
					if (in + it->first >= 0 && in + it->first < matr_B.total_size) {
						RHS[in] += it->second[in] * slicePSD1D[in + it->first];
					}
				}
			}

			// right hand side to files
			//			RHS.writeToFile("pc_RHS.dat");

			// Solving model matrix, obtaining PSD
			// Assume matr_A has only +1, 0, -1 diagonals
			// TODO: Check if matr_A has only +1, 0, -1 diagonals, i.e. is tridiagonal
			tridag(
				&matr_A[-1][0], // just a way to get a pointer to first element of an array
				&matr_A[0][0],
				&matr_A[+1][0],
				&RHS[0],
				&slicePSD1D[0],
				L.size);

			// copying PSD back, from 1d array to 3d array
			for (il = 0; il < L.size; il++) {
				arr[il][im][ia] = slicePSD1D[il];
			}
		}
	}
}


/** Radial diffusion calculation function.
* Takes diffusion coefficients, boundary conditions and grid elements as parameters.
* It caltulates model matrix by makeMatrix and then solve it by SolveMatrix method.
*
* \param dt - time step
* \param Lpp - plasma pause location
* \param &Dpcpc - diffusion coefficient
* \param &DpcpcLpp - diffusion coefficient
* \param &Daa - diffusion coefficient
* \param &DaaLpp - diffusion coefficient
* \param &Dpca - diffusion coefficient
* \param &DpcaLpp - diffusion coefficient
* \param &L - grid element L
* \param &pc - grid elelent pc
* \param &alpha - grid element alpha
* \param &Jacobian - jacobian
* \param &pc_lowerBoundaryCondition - lower boundary condition
* \param &pc_upperBoundaryCondition - upper boundary condition
* \param &alpha_lowerBoundaryCondition - lower boundary condition
* \param &alpha_upperBoundaryCondition - upper boundary condition
* \param &pc_lowerBoundaryCondition_calculationType - lower boundary condition calculation type
* \param &pc_upperBoundaryCondition_calculationType - upper boundary condition calculation type
* \param &alpha_lowerBoundaryCondition_calculationType - lower boundary condition calculation type
* \param &alpha_upperBoundaryCondition_calculationType - upper boundary condition calculation type
*/
void PSD::Diffusion_pc_alpha(double dt, double Lpp,
								  DiffusionCoefficient &Dpcpc, DiffusionCoefficient &DpcpcLpp,
								  DiffusionCoefficient &Daa, DiffusionCoefficient &DaaLpp,
								  DiffusionCoefficient &Dpca, DiffusionCoefficient &DpcaLpp,
								  GridElement &L, GridElement &pc, GridElement &alpha, Matrix3D<double> Jacobian,
								  Matrix2D<double> pc_lowerBoundaryCondition,
								  Matrix2D<double> pc_upperBoundaryCondition,
								  Matrix2D<double> alpha_lowerBoundaryCondition,
								  Matrix2D<double> alpha_upperBoundaryCondition,
								  string pc_lowerBoundaryCondition_calculationType,
								  string pc_upperBoundaryCondition_calculationType,
								  string alpha_lowerBoundaryCondition_calculationType,
								  string alpha_upperBoundaryCondition_calculationType)
{

	// Calculation matrices A, B, and C
	// A * PSD_1D(t+1) = B * PSD_1D(t) + C, for each L and Energy

	// Creation is moved to PSD.h, so it can be accessed from Main.cpp
	//static CalculationMatrix
	//	matr_A(alpha.size, pc.size, 1, 1),
	//	matr_B(alpha.size, pc.size, 1, 0),
	//	matr_C(alpha.size, pc.size, 1, 0);
	matr_A.Initialize(alpha.size, pc.size, 1, 1);
	matr_B.Initialize(alpha.size, pc.size, 1, 0),
	matr_C.Initialize(alpha.size, pc.size, 1, 0);

	// Create zero matrices, temporary matrices to be used later
	static Matrix3D<double> Zero_Matrix(1, pc.size, alpha.size);
	static Matrix2D<double> Zero_Matrix_2d(1, 1);
	static Matrix1D<double> Zero_Matrix_1d(1);
	Zero_Matrix = 0;
	Zero_Matrix_2d = 0;
	Zero_Matrix_1d = 0;

	// Create 1d arrays
	// Creating 1d arrays for PSD and right hand side (RHS = B * PSD + C)
	static Matrix1D<double> slicePSD1D(matr_A.total_size), RHS(matr_A.total_size);
	// Creating 1d arrays which will be send to the MakeMatrix function
	// Creating 1d matrixes, but with type of Matrix3D, cause we need that type for MakeMatrix function
	static Matrix3D<double> Dpcpc_slice(1, pc.size, alpha.size), DpcpcLpp_slice(1, pc.size, alpha.size),  Daa_slice(1, pc.size, alpha.size), DaaLpp_slice(1, pc.size, alpha.size), Dpca_slice(1, pc.size, alpha.size),  DpcaLpp_slice(1, pc.size, alpha.size);
	static Matrix3D<double> L_slice(1, pc.size, alpha.size), pc_slice(1, pc.size, alpha.size), alpha_slice(1, pc.size, alpha.size), Jacobian_slice(1, pc.size, alpha.size);
	// Creating temporary arrays for boundary conditions
	static Matrix2D<double> pc_lowerBoundaryCondition_slice(1, alpha.size), pc_upperBoundaryCondition_slice(1, alpha.size);
	static Matrix2D<double> alpha_lowerBoundaryCondition_slice(1, pc.size), alpha_upperBoundaryCondition_slice(1, pc.size);

	// iterators
	int il, im, ia, in; // matrix element numbers on 3D grid
	DiagMatrix::iterator it;
	for (il = 0; il < L.size; il++) {
		// Copying 3D arrays into 1d arrays
		for (im = 0; im < pc.size; im++) {
			for (ia = 0; ia < alpha.size; ia++) {
				// Copying 3D PSD to 1d PSD
				in = matr_A.index1d(ia, im);
				slicePSD1D[in] = arr[il][im][ia];

				Dpcpc_slice[0][im][ia] = Dpcpc.arr[il][im][ia];
				DpcpcLpp_slice[0][im][ia] = DpcpcLpp.arr[il][im][ia];
				Daa_slice[0][im][ia] = Daa.arr[il][im][ia];
				DaaLpp_slice[0][im][ia] = DaaLpp.arr[il][im][ia];
				Dpca_slice[0][im][ia] = Dpca.arr[il][im][ia];
				DpcaLpp_slice[0][im][ia] = DpcaLpp.arr[il][im][ia];

				L_slice[0][im][ia] = L.arr[il][im][ia];
				pc_slice[0][im][ia] = pc.arr[il][im][ia];
				alpha_slice[0][im][ia] = alpha.arr[il][im][ia];

				Jacobian_slice[0][im][ia] = Jacobian[il][im][ia];
			}
		}
		// boundary conditions
		for (ia = 0; ia < alpha.size; ia++) {
			pc_upperBoundaryCondition_slice[0][ia] = pc_upperBoundaryCondition[il][ia];
			pc_lowerBoundaryCondition_slice[0][ia] = pc_lowerBoundaryCondition[il][ia];
		}
		for (im = 0; im < pc.size; im++) {
			alpha_upperBoundaryCondition_slice[0][im] = alpha_upperBoundaryCondition[il][im];
			alpha_lowerBoundaryCondition_slice[0][im] = alpha_lowerBoundaryCondition[il][im];
		}

		// Make
		MakeModelMatrix_3D (
			matr_A, matr_B, matr_C,  // matrices A, B, C
			L_slice, pc_slice, alpha_slice, // L, pc, alpha grids. We use pc and alpha crossections here
			1, pc.size, alpha.size,  // L-size, pc-size, alpha-size
			Zero_Matrix_2d, Zero_Matrix_2d, // L-boundaries
			pc_lowerBoundaryCondition_slice, pc_upperBoundaryCondition_slice, // pc-boundaries
			alpha_lowerBoundaryCondition_slice, alpha_upperBoundaryCondition_slice, // alpha-boundaries
			"BCT_CONSTANT_VALUE", "BCT_CONSTANT_VALUE", // Boundary conditions at L-boundaries
			pc_lowerBoundaryCondition_calculationType, pc_upperBoundaryCondition_calculationType, // Boundary conditions at pc-boundaries
			alpha_lowerBoundaryCondition_calculationType, alpha_upperBoundaryCondition_calculationType, // Boundary conditions at alpha-boundaries
			Zero_Matrix, // DLL
			Dpcpc_slice, DpcpcLpp_slice, // Dpp, DppLpp - use 0
			Daa_slice, DaaLpp_slice, // Daa, DaaLpp - use 0
			Dpca_slice, DpcaLpp_slice, // Dpca, Dpca_Lpp - use 0
			Jacobian_slice, dt, Lpp);

		/* Output::echo(0, "Writing calculation matrices for L[%d] = %f... ", il, L[il][0][0]);
		matr_A.writeToFile("./Debug/pc_alpha_matr_A.dat");
		matr_B.writeToFile("./Debug/pc_alpha_matr_B.dat");
		matr_C.writeToFile("./Debug/pc_alpha_matr_C.dat");
		Output::echo(0, "done.\n"); */

		// make RHS = B*f + C
		DiagMatrix::iterator it;
		for (im = 0; im < pc.size; im++) {
			for (ia = 0; ia < alpha.size; ia++) {
				in = matr_B.index1d(ia, im);
				// I guess matr_B has only main diagonal in implicit method, so it can be used like this:
				//				RHS[in] = matr_B[0][in] * slicePSD1D[in] + matr_C[0][in];
				RHS[in] = matr_C[0][in]; // RHS = C
				for (it = matr_B.begin(); it != matr_B.end(); it++) {
					// multiplication B * f
					if (in + it->first >= 0 && in + it->first < matr_B.total_size) {
						RHS[in] += it->second[in] * slicePSD1D[in + it->first]; // RHS = C + Bf
					}
				}

			}
		}

		// right hand side to files
		// RHS.writeToFile("./Debug/pc_alpha_RHS.dat");

		// Solving model matrix
		int n = pc.size * alpha.size;

		if (PSD_parameters.solutionMethod == "SM_Gauss") {
            // external gauss function
            gauss(matr_A, RHS, slicePSD1D, n);

            Output::echo(5, "done.\n");
				
            /* moved this block of code to helper function gauss() located in MatrixSolver */

		} else if (PSD_parameters.solutionMethod == "SM_Relaxation") {
				// Relaxation method

				Output::echo(5, "Inverting A matrix (relaxation)... ");
				// runs the over_relaxation_diag solution method
				over_relaxation_diag(matr_A, RHS, slicePSD1D);
				Output::echo(5, "done.");

		} else if (PSD_parameters.solutionMethod == "SM_Lapack") {
				// Relaxation method

				Output::echo(5, "Inverting A matrix (Lapack)... ");
				// runs the over_relaxation_diag solution method
				Lapack(matr_A, RHS, slicePSD1D);
				Output::echo(5, "done.");

		} else if (PSD_parameters.solutionMethod == "SM_GMRES") {
				// Relaxation method

				Output::echo(5, "Inverting A matrix (GMRES)... ");
				// runs the over_relaxation_diag solution method
				//gmres_wrapout(matr_A, RHS, slicePSD1D, PSD_parameters.SOL_maxiter, PSD_parameters.SOL_i_max, PSD_parameters.SOL_max_iter_err);
				gmres_wrapout(matr_A, RHS, slicePSD1D, PSD_parameters.GMRES_parameters);
				Output::echo(5, "done.");

		} else if (PSD_parameters.solutionMethod == "SM_GMRES2") {
				// Relaxation method

				Output::echo(5, "Inverting A matrix (GMRES2)... ");
				// runs the over_relaxation_diag solution method
				//gmres_wrapout(matr_A, RHS, slicePSD1D, PSD_parameters.SOL_maxiter, PSD_parameters.SOL_i_max, PSD_parameters.SOL_max_iter_err);
				gmres2_wrapout(matr_A, RHS, slicePSD1D, PSD_parameters.GMRES_parameters);
				Output::echo(5, "done.");


		} else if (PSD_parameters.solutionMethod == "SM_Tridiag") {
				// Tridiagonal method
				throw error_msg("SOLUTIONERR", "Tridiagonal method can't be used");
		} else {				// Unknown method
				throw error_msg("SOLUTIONERR", "Unknown solution method");
		}

		// copying PSD back, from 1d array to 3d array
		for (im = 0; im < pc.size; im++) {
			for (ia = 0; ia < alpha.size; ia++) {
				in = matr_A.index1d(ia, im);
				arr[il][im][ia] = slicePSD1D[in];
			}
		}
	}

}


/** Radial diffusion calculation function with Kyungchan's method.
 * Takes diffusion coefficients, boundary conditions and grid elements as parameters.
 * It caltulates model matrix by makeMatrix and then solve it by SolveMatrix method.
 * \todo Literally the only difference between KC version and original version is calling one custom method MakeModelMatrix_3D_KC. Find a way to compress two functions together.
 *
 * \param dt - time step
 * \param Lpp - plasma pause location
 * \param &Dpcpc - diffusion coefficient
 * \param &DpcpcLpp - diffusion coefficient
 * \param &Daa - diffusion coefficient
 * \param &DaaLpp - diffusion coefficient
 * \param &Dpca - diffusion coefficient
 * \param &DpcaLpp - diffusion coefficient
 * \param &L - grid element L
 * \param &pc - grid elelent pc
 * \param &alpha - grid element alpha
 * \param &Jacobian - jacobian
 * \param &pc_lowerBoundaryCondition - lower boundary condition
 * \param &pc_upperBoundaryCondition - upper boundary condition
 * \param &alpha_lowerBoundaryCondition - lower boundary condition
 * \param &alpha_upperBoundaryCondition - upper boundary condition
 * \param &pc_lowerBoundaryCondition_calculationType - lower boundary condition calculation type
 * \param &pc_upperBoundaryCondition_calculationType - upper boundary condition calculation type
 * \param &alpha_lowerBoundaryCondition_calculationType - lower boundary condition calculation type
 * \param &alpha_upperBoundaryCondition_calculationType - upper boundary condition calculation type
 */
void PSD::Diffusion_pc_alpha_KC(double dt, double Lpp,
								  DiffusionCoefficient &Dpcpc, DiffusionCoefficient &DpcpcLpp,
								  DiffusionCoefficient &Daa, DiffusionCoefficient &DaaLpp,
								  DiffusionCoefficient &Dpca, DiffusionCoefficient &DpcaLpp,
								  GridElement &L, GridElement &pc, GridElement &alpha, Matrix3D<double> Jacobian,
								  Matrix2D<double> pc_lowerBoundaryCondition,
								  Matrix2D<double> pc_upperBoundaryCondition,
								  Matrix2D<double> alpha_lowerBoundaryCondition,
								  Matrix2D<double> alpha_upperBoundaryCondition,
								  string pc_lowerBoundaryCondition_calculationType,
								  string pc_upperBoundaryCondition_calculationType,
								  string alpha_lowerBoundaryCondition_calculationType,
								  string alpha_upperBoundaryCondition_calculationType)
{

	// Calculation matrices A, B, and C
	// A * PSD_1D(t+1) = B * PSD_1D(t) + C, for each L and Energy

	// Creation is moved to PSD.h, so it can be accessed from Main.cpp
	//static CalculationMatrix
	//	matr_A(alpha.size, pc.size, 1, 1),
	//	matr_B(alpha.size, pc.size, 1, 0),
	//	matr_C(alpha.size, pc.size, 1, 0);
	matr_A.Initialize(alpha.size, pc.size, 1, 1);
	matr_B.Initialize(alpha.size, pc.size, 1, 0),
	matr_C.Initialize(alpha.size, pc.size, 1, 0);




	// Create zero matrices, temporary matrices to be used later
	static Matrix3D<double> Zero_Matrix(1, pc.size, alpha.size);
	static Matrix2D<double> Zero_Matrix_2d(1, 1);
	static Matrix1D<double> Zero_Matrix_1d(1);
	Zero_Matrix = 0;
	Zero_Matrix_2d = 0;
	Zero_Matrix_1d = 0;

	// Create 1d arrays
	// Creating 1d arrays for PSD and right hand side (RHS = B * PSD + C)
	static Matrix1D<double> slicePSD1D(matr_A.total_size), RHS(matr_A.total_size);
	// Creating 1d arrays which will be send to the MakeMatrix function
	// Creating 1d matrixes, but with type of Matrix3D, cause we need that type for MakeMatrix function
	static Matrix3D<double> Dpcpc_slice(1, pc.size, alpha.size), DpcpcLpp_slice(1, pc.size, alpha.size),  Daa_slice(1, pc.size, alpha.size), DaaLpp_slice(1, pc.size, alpha.size), Dpca_slice(1, pc.size, alpha.size),  DpcaLpp_slice(1, pc.size, alpha.size);
	static Matrix3D<double> L_slice(1, pc.size, alpha.size), pc_slice(1, pc.size, alpha.size), alpha_slice(1, pc.size, alpha.size), Jacobian_slice(1, pc.size, alpha.size);
	// Creating temporary arrays for boundary conditions
	static Matrix2D<double> pc_lowerBoundaryCondition_slice(1, alpha.size), pc_upperBoundaryCondition_slice(1, alpha.size);
	static Matrix2D<double> alpha_lowerBoundaryCondition_slice(1, pc.size), alpha_upperBoundaryCondition_slice(1, pc.size);

	// iterators
	int il, im, ia, in; // matrix element numbers on 3D grid
	DiagMatrix::iterator it;
	for (il = 0; il < L.size; il++) {
		// Copying 3D arrays into 1d arrays
		for (im = 0; im < pc.size; im++) {
			for (ia = 0; ia < alpha.size; ia++) {
				// Copying 3D PSD to 1d PSD
				in = matr_A.index1d(ia, im);
				slicePSD1D[in] = arr[il][im][ia];

				Dpcpc_slice[0][im][ia] = Dpcpc.arr[il][im][ia];
				DpcpcLpp_slice[0][im][ia] = DpcpcLpp.arr[il][im][ia];
				Daa_slice[0][im][ia] = Daa.arr[il][im][ia];
				DaaLpp_slice[0][im][ia] = DaaLpp.arr[il][im][ia];
				Dpca_slice[0][im][ia] = Dpca.arr[il][im][ia];
				DpcaLpp_slice[0][im][ia] = DpcaLpp.arr[il][im][ia];

				L_slice[0][im][ia] = L.arr[il][im][ia];
				pc_slice[0][im][ia] = pc.arr[il][im][ia];
				alpha_slice[0][im][ia] = alpha.arr[il][im][ia];

				Jacobian_slice[0][im][ia] = Jacobian[il][im][ia];
			}
		}
		// boundary conditions
		for (ia = 0; ia < alpha.size; ia++) {
			pc_upperBoundaryCondition_slice[0][ia] = pc_upperBoundaryCondition[il][ia];
			pc_lowerBoundaryCondition_slice[0][ia] = pc_lowerBoundaryCondition[il][ia];
		}
		for (im = 0; im < pc.size; im++) {
			alpha_upperBoundaryCondition_slice[0][im] = alpha_upperBoundaryCondition[il][im];
			alpha_lowerBoundaryCondition_slice[0][im] = alpha_lowerBoundaryCondition[il][im];
		}

		// Make
		MakeModelMatrix_3D_KC (
			matr_A, matr_B, matr_C,  // matrices A, B, C
			L_slice, pc_slice, alpha_slice, // L, pc, alpha grids. We use pc and alpha crossections here
			1, pc.size, alpha.size,  // L-size, pc-size, alpha-size
			Zero_Matrix_2d, Zero_Matrix_2d, // L-boundaries
			pc_lowerBoundaryCondition_slice, pc_upperBoundaryCondition_slice, // pc-boundaries
			alpha_lowerBoundaryCondition_slice, alpha_upperBoundaryCondition_slice, // alpha-boundaries
			"BCT_CONSTANT_VALUE", "BCT_CONSTANT_VALUE", // Boundary conditions at L-boundaries
			pc_lowerBoundaryCondition_calculationType, pc_upperBoundaryCondition_calculationType, // Boundary conditions at pc-boundaries
			alpha_lowerBoundaryCondition_calculationType, alpha_upperBoundaryCondition_calculationType, // Boundary conditions at alpha-boundaries
			Zero_Matrix, // DLL
			Dpcpc_slice, DpcpcLpp_slice, // Dpp, DppLpp - use 0
			Daa_slice, DaaLpp_slice, // Daa, DaaLpp - use 0
			Dpca_slice, DpcaLpp_slice, // Dpca, Dpca_Lpp - use 0
			Jacobian_slice, dt, Lpp);

		/* Output::echo(0, "Writing calculation matrices for L[%d] = %f... ", il, L[il][0][0]);
		matr_A.writeToFile("./Debug/pc_alpha_matr_A.dat");
		matr_B.writeToFile("./Debug/pc_alpha_matr_B.dat");
		matr_C.writeToFile("./Debug/pc_alpha_matr_C.dat");
		Output::echo(0, "done.\n"); */

		// make RHS = B*f + C
		DiagMatrix::iterator it;
		for (im = 0; im < pc.size; im++) {
			for (ia = 0; ia < alpha.size; ia++) {
				in = matr_B.index1d(ia, im);
				// I guess matr_B has only main diagonal in implicit method, so it gan be used like this:
				//				RHS[in] = matr_B[0][in] * slicePSD1D[in] + matr_C[0][in];
				RHS[in] = matr_C[0][in];
				for (it = matr_B.begin(); it != matr_B.end(); it++) {
					// multiplication B * f
					if (in + it->first >= 0 && in + it->first < matr_B.total_size) {
						RHS[in] += it->second[in] * slicePSD1D[in + it->first];
					}
				}

			}
		}

		// right hand side to files
		// RHS.writeToFile("./Debug/pc_alpha_RHS.dat");

		// Solving model matrix
		int n = pc.size * alpha.size;
		// need A and B for gauss method
		static Matrix2D<double> A;
		static Matrix1D<double> B;
		int modelMatrixLineNumber;
		int k, l, in;
		double sum, error, error_max;
		static bool recalculate_A = true;
		if (PSD_parameters.solutionMethod == "SM_Gauss") {
				// the routine does not actually invert matrix, but just solve AX=B eq
				if (recalculate_A) {
					B = RHS;
					if (!A.initialized) {
						A.AllocateMemory(n, n);
					}
					A = 0;
					//				Output::echo("Writing A0 matrix... ");
					//				A.writeToFile("A0.dat");
					//				Output::echo("done.\n");

					Output::echo(5, "Inverting A matrix (gauss)... ");
					// Gauss method
					for (it = matr_A.begin(); it != matr_A.end(); it++) {
						for (modelMatrixLineNumber = 0; modelMatrixLineNumber < n; modelMatrixLineNumber++) {
							// check, if the element at line (modelMatrixLineNumber) and diagonal (it->first) is inside the matrix.
							// cause it can be outside, if line number is 1 and diagonal number is -5, for example.
							if (modelMatrixLineNumber + it->first >= 0 && modelMatrixLineNumber + it->first < n) {
								// converting matrix, stored as diagonals, into normal 2D matrix
								// cause we need 2D matrix to solve by Gauss method
								//							cout << it->first << " - " << it->second[modelMatrixLineNumber] << endl;
								A[modelMatrixLineNumber][modelMatrixLineNumber + it->first] = it->second[modelMatrixLineNumber];
							}
						}
					}

					// inversion of A
					// sending address of the first element of the plane 2d array
					// it's not actually matrix inversion
					// it just makes A-triangle, so it's easy to solve A*X = B
					gauss_solve(&A[0][0], &B[0], n);
					recalculate_A = true;
					Output::echo(5, "done.\n");
				}

				Output::echo(5, "Calculation of new PSD... ");
				slicePSD1D[n-1] = B[n-1] / A[n-1][n-1];
				for (k = n - 2; k >= 0; k--) {
					sum = 0;
					for (l = k + 1; l < n; l++)
						sum += A[k][l] * slicePSD1D[l];
					slicePSD1D[k] = B[k] - sum;
				}
				/*   end calc  */

				/*   control  */
				//mult_vect2(&A[0], &slicePSD1D[0], &RHS[0], n);
				error_max = 0.;
				error = 0;
				for (in = 0; in < n; in++) {
					error = RHS[in];
					for (it = matr_A.begin(); it != matr_A.end(); it++) {
						// multiplying all diagonals to PSD to check error = RHS - A*PSD = 0
						if (in + it->first >= 0 && in + it->first < n) {
							error -= it->second[in] * slicePSD1D[in + it->first];
						}
					}
					error_max = VF::max( error_max, error );
				}
				Output::echo(5, "done.\n");
				Output::echo(5, "Error_norm = %e \n", error_max);

		} else if (PSD_parameters.solutionMethod == "SM_Relaxation") {
				// Relaxation method

				Output::echo(5, "Inverting A matrix (relaxation)... ");
				// runs the over_relaxation_diag solution method
				over_relaxation_diag(matr_A, RHS, slicePSD1D);
				Output::echo(5, "done.");

		} else if (PSD_parameters.solutionMethod == "SM_Lapack") {
				// Relaxation method

				Output::echo(5, "Inverting A matrix (Lapack)... ");
				// runs the over_relaxation_diag solution method
				Lapack(matr_A, RHS, slicePSD1D);
				Output::echo(5, "done.");

		} else if (PSD_parameters.solutionMethod == "SM_GMRES") {
				// Relaxation method

				Output::echo(5, "Inverting A matrix (GMRES)... ");
				// runs the over_relaxation_diag solution method
				//gmres_wrapout(matr_A, RHS, slicePSD1D, PSD_parameters.SOL_maxiter, PSD_parameters.SOL_i_max, PSD_parameters.SOL_max_iter_err);
				gmres_wrapout(matr_A, RHS, slicePSD1D, PSD_parameters.GMRES_parameters);
				Output::echo(5, "done.");

		} else if (PSD_parameters.solutionMethod == "SM_GMRES2") {
				// Relaxation method

				Output::echo(5, "Inverting A matrix (GMRES2)... ");
				// runs the over_relaxation_diag solution method
				//gmres_wrapout(matr_A, RHS, slicePSD1D, PSD_parameters.SOL_maxiter, PSD_parameters.SOL_i_max, PSD_parameters.SOL_max_iter_err);
				gmres2_wrapout(matr_A, RHS, slicePSD1D, PSD_parameters.GMRES_parameters);
				Output::echo(5, "done.");


		} else if (PSD_parameters.solutionMethod == "SM_Tridiag") {
				// Tridiagonal method
				throw error_msg("SOLUTIONERR", "Tridiagonal method can't be used");
		} else {				// Unknown method
				throw error_msg("SOLUTIONERR", "Unknown solution method");
		}

		// copying PSD back, from 1d array to 3d array
		for (im = 0; im < pc.size; im++) {
			for (ia = 0; ia < alpha.size; ia++) {
				in = matr_A.index1d(ia, im);
				arr[il][im][ia] = slicePSD1D[in];
			}
		}
	}

}



/** Checking for infinity value appears during interpolation for 3D arrays.
*
* \param arr - array of values
* \param il - index
* \param im - index
* \param ia - index
* \param maxNum = 1e99 - max number (to compare with)
*/
void checkInf_3D(Matrix3D<double> arr, int il, int im, int ia, double maxNum = 1e99) {
	// Some weird condition... It tries to check for Not-A-Number which gives false for all comparisons and same time for infinite values... But I guess there is a bug here.
	/// first comparison to check for NaN
	if (arr[il][im][ia] != arr[il][im][ia] || arr[il][im][ia] > maxNum || arr[il][im][ia] < -maxNum) {
		throw error_msg("INTERPOLATION_ERROR", "Interpolation error at [%d][%d][%d]", il, im, ia);
	}
}

/** Checking for infinity value appears during interpolation for 1D arrays.
*
* \param arr - array of values
* \param il - index
* \param im - index
* \param ia - index
* \param maxNum = 1e99 - max number (to compare with)
*/
void checkInf_1D(Matrix1D<double> arr, int il, int im, int ia, double maxNum = 1e99) {
	// Some weird condition... It tries to check for Not-A-Number which gives false for all comparisons and same time for infinite values... But I guess there is a bug here.
    /// first comparison to check for NaN
	if ( arr[im] != arr[im] || arr[im] > maxNum || arr[im] < -maxNum) {
		throw error_msg("INTERPOLATION_ERROR", "Interpolation error at [%d][%d][%d]", il, im, ia);
	}
}

///  It does interpolation... for log(function) or for just function, depends on parameters.
void PSD::Interpolate(PSD &otherPSD, Parameters_structure::Interpolation interpolationParameters_structure, Grid &oldGrid, Grid &newGrid, Matrix2D<double> newGrid_pc_lowerBoundaryCondition, Matrix2D<double> newGrid_pc_upperBoundaryCondition) {
	// check...

	int il, im, ia;
	double lowerBoundary1d=0, upperBoundary1d=0;

	// if interpolation method is IT_NONE - exit
	// if grids have the same type, we just copy the values
	if (interpolationParameters_structure.type == "IT_NONE") {
		return;
	} else if (oldGrid.type == newGrid.type) { /// \todo If grid types are the same, it's not necessary that grids are the same
		//		arr.Matrix3D<double>::operator =(otherPSD);

		for (il = 0; il < oldGrid.L.size; il++) {
			for (ia = 0; ia < oldGrid.alpha.size; ia++) {
				for (im = 0; im < oldGrid.pc.size; im++) {
					// copying of the values from one grid to anouther
					arr[il][im][ia] = otherPSD.arr[il][im][ia];
					// some weird bug. Model matrix does not work correctly without this. Don't know why.
					if (arr[il][im][ia] < VC::zero_f) {
						arr[il][im][ia] = VC::zero_f;
					}
				}
			}
		}
		return;
	}

    // matrices to be interpolated
	Matrix1D<double> matrix_array_1d_old(oldGrid.pc.size);
	Matrix1D<double> matrix_array_1d_new(newGrid.pc.size);
	Matrix1D<double> grid_1d_old(oldGrid.pc.size);
	Matrix1D<double> grid_1d_new(newGrid.pc.size);

	for (il = 0; il < oldGrid.L.size; il++) {
		for (ia = 0; ia < oldGrid.alpha.size; ia++) {
			for (im = 0; im < oldGrid.pc.size; im++) {

				if (newGrid.L.arr[il][im][ia] != oldGrid.L.arr[il][im][ia] || newGrid.alpha.arr[il][im][ia] != oldGrid.alpha.arr[il][im][ia]) {
					// if L-size or alpha-size are different, then we can not interpolate by that interpolation function
					throw error_msg("INTERPOLATION_ERROR", "Interpolation error: grids must have the same L and Alpha");
				}
#ifdef DEBUG_MODE
				// check for infinity and NAN
				checkInf_3D(arr, il, im, ia);
#endif

				// some work around boundary conditions
				if (interpolationParameters_structure.useLog == "Log_10") {
					// if we are interpolating log of values, then store log of values to 1d array
					if (otherPSD.arr[il][im][ia] >= VC::zero_f) {
						matrix_array_1d_old[im] = log10(otherPSD.arr[il][im][ia]);
					} else {
						// if PSD lower then lowest possible value (it should not be zero, because we are interpolating log of values for example)
						matrix_array_1d_old[im] = log10(VC::zero_f);
					}
					// theoretically we don't need to make the boundary conditions lower then 10^21
					// because we check that in its constructor
					// But just in case...
					lowerBoundary1d = log10(VF::max(newGrid_pc_lowerBoundaryCondition[il][ia], VC::zero_f));
					upperBoundary1d = log10(VF::max(newGrid_pc_upperBoundaryCondition[il][ia], VC::zero_f));
				} 
				else if (interpolationParameters_structure.useLog == "Log_2") {
					// if we are interpolating log of values, then store log of values to 1d array
					if (otherPSD.arr[il][im][ia] >= VC::zero_f) {
						matrix_array_1d_old[im] = log(otherPSD.arr[il][im][ia])/log(2.0);
					} else {
						// if PSD lower then lowest possible value (it should not be zero, because we are interpolating log of values for example)
						matrix_array_1d_old[im] = log(VC::zero_f)/log(2.0);
					}
					// theoretically we don't need to make the boundary conditions lower then 10^21
					// because we check that in its constructor
					// But just in case...

					// adrozdov: Bug report. log10 replaced by log2
					lowerBoundary1d = log(VF::max(newGrid_pc_lowerBoundaryCondition[il][ia], VC::zero_f)) / log(2.0);
					upperBoundary1d = log(VF::max(newGrid_pc_upperBoundaryCondition[il][ia], VC::zero_f)) / log(2.0);
				} 
				else if (interpolationParameters_structure.useLog == "Log_E") {
					// if we are interpolating log of values, then store log of values to 1d array
					if (otherPSD.arr[il][im][ia] >= VC::zero_f) {
						matrix_array_1d_old[im] = log(otherPSD.arr[il][im][ia]);
					} else {
						// if PSD lower then lowest possible value (it should not be zero, because we are interpolating log of values for example)
						matrix_array_1d_old[im] = log(VC::zero_f);
					}
					// theoretically we don't need to make the boundary conditions lower then 10^21
					// because we check that in its constructor
					// But just in case...
					
					// adrozdov: Bug report. log10 replaced by log
					lowerBoundary1d = log(VF::max(newGrid_pc_lowerBoundaryCondition[il][ia], VC::zero_f));
					upperBoundary1d = log(VF::max(newGrid_pc_upperBoundaryCondition[il][ia], VC::zero_f));
				} 
				else if (interpolationParameters_structure.useLog == "NoLog") {
					matrix_array_1d_old[im] = otherPSD.arr[il][im][ia];
					lowerBoundary1d = VF::max(newGrid_pc_lowerBoundaryCondition[il][ia], VC::zero_f);
					upperBoundary1d = VF::max(newGrid_pc_upperBoundaryCondition[il][ia], VC::zero_f);
				} 
				else {
					throw error_msg("INTERPOLATION_ERROR", "Usage of log(PSD) is not define properly: %s", interpolationParameters_structure.useLog.c_str());
					break;
				}

#ifdef DEBUG_MODE
				// check for infinity and NAN
				checkInf_1D(matrix_array_1d_old, il, im, ia, 99);
#endif

				grid_1d_old[im] = oldGrid.pc.arr[il][im][ia];
				grid_1d_new[im] = newGrid.pc.arr[il][im][ia];
				//cout<<grid_1d_old[im]<<"kim......."<<grid_1d_new[im]<<"..........."<<matrix_array_1d_old[im]<<endl;
			}

            // interpolation based on type
			if (interpolationParameters_structure.type == "IT_LINEAR") {
					// do linear interpolation
					matrix_array_1d_new.Interpolate(matrix_array_1d_old, grid_1d_old, grid_1d_new);
			} else if (interpolationParameters_structure.type == "IT_POLYNOMIAL") {
					// do polynomial - linear interpolation (linear close to the boundaries)
					matrix_array_1d_new.Polilinear(matrix_array_1d_old, grid_1d_old, grid_1d_new, lowerBoundary1d, upperBoundary1d);
			} else if (interpolationParameters_structure.type == "IT_SPLINE") {
					// do spline interpolation
					matrix_array_1d_new.Spline(matrix_array_1d_old, grid_1d_old, grid_1d_new, lowerBoundary1d, upperBoundary1d, interpolationParameters_structure.linearSplineCoef, interpolationParameters_structure.maxSecondDerivative);
			} else if (interpolationParameters_structure.type == "IT_COPY") {
					// just copy
					matrix_array_1d_new = matrix_array_1d_old;
			} else {					// unknown
					throw error_msg("INTERPOLATION_TYPE_ERROR", "Unknown interpolation type");
			}

			for (im = 0; im < newGrid.pc.size; im++) {

#ifdef DEBUG_MODE
				checkInf_1D(matrix_array_1d_new, il, im, ia, 99);
#endif

				// copying values back to 3d arrays
				if (interpolationParameters_structure.useLog == "Log_10") arr[il][im][ia] = pow(10, matrix_array_1d_new[im]);
				else if (interpolationParameters_structure.useLog == "Log_2") arr[il][im][ia] = pow(2, matrix_array_1d_new[im]);
				else if (interpolationParameters_structure.useLog == "Log_E") arr[il][im][ia] = pow(VC::exp, matrix_array_1d_new[im]);
				else arr[il][im][ia] = matrix_array_1d_new[im];

#ifdef DEBUG_MODE
				checkInf_3D(arr, il, im, ia);
#endif

			}
		}
	}

	// check result...
}

/*-------------------------------------------------------------------------------------------------------------------------------
Outputs
-------------------------------------------------------------------------------------------------------------------------------*/

/**
* %PSD output.
*
* \param time - time
*/
void PSD::Output_without_grid(double time) {
	int il, im, ia;
	output_without_grid_file << "ZONE T=\"" << internal << time << "\" I=" << arr.size_z << ", J=" << arr.size_y << ", K=" << arr.size_x << endl;
	for (il = 0; il < arr.size_x; il++) {
		for (im = 0; im < arr.size_y; im++) {
			for (ia = 0; ia < arr.size_z; ia++) {
				output_without_grid_file << arr[il][im][ia] << endl;
			}
		}
	}

}


/**
* Load initial %PSD from file
* \param &L - grid element  L
* \param &epc - grid element pc
* \param &alpha - grid element alpha
* \param filename - filename
* \param withGrid - indicate if the file contains grid values
*/
void PSD::Load_initial_f_file(GridElement &L, GridElement &epc, GridElement &alpha, const char *filename, bool withGrid) {
	int il, im, ia;
	string inBuf;
	double Load_L=-1, Load_epc=-1, Load_alpha=-1;// temporary variables for the grid point
	double err = 1.e-3;

	// Load j from input file
	ifstream input_f(filename);
	if (input_f != NULL && !input_f.eof()) {
		// Skipping first two lines. Should search for 'ZONE' and read from following line better.
		getline(input_f, inBuf);
		getline(input_f, inBuf);
		for (il = 0; il < L.size; il++) {
			for (im = 0; im < epc.size; im++) {
				for (ia = 0; ia < alpha.size; ia++) {
					if (withGrid) input_f >> Load_L >> Load_epc >> Load_alpha;
					input_f >> arr[il][im][ia];
					if (withGrid)
						if (fabs(Load_L - L.arr[il][im][ia]) > err || fabs(Load_epc - epc.arr[il][im][ia]) > err || fabs(Load_alpha - alpha.arr[il][im][ia]) > err) {
							throw error_msg("INITIAL_PSD_LOAD_GRID_ERR", "Loading %s: grid mismatch.\nLoaded: L=%e, epc=%e, alpha=%e\nGrid:  L=%e, epc=%e, alpha=%e\n", filename, Load_L, Load_epc, Load_alpha, L.arr[il][im][ia], epc.arr[il][im][ia], alpha.arr[il][im][ia]);
						}
				}
			}
		}
	} else {
		throw error_msg("INITIAL_PSD_ERROR", "Error reading PSD input file %s.\n", filename);
		//cout << "Error reading initial conditions input file " << filename << endl;
		//getchar();
		//exit(-1);
	}
	input_f.close();
}


/**
 *  Calculate initial %PSD from steady state.
 *
 * \param &L - grid element  L
 * \param &pc - grid element pc
 * \param &alpha - grid element alpha
 * \param tau - life time
 * \param Kp - Kp value
 * \param min_f - minimum of function (should be some positive value)
 * \param J_L7_function - parameter, J_L7 function. Can be J_L7 or J_L7_corrected
 * \param fb_out - value at the outer L boundary
 * \param fb_in - value at the inner L boundary
 * \param using_ia_90 - flag indicating if using ia_90 value for calculating L values
 */
void PSD::Load_initial_f(GridElement &L, GridElement &pc, GridElement &alpha, double tau, double Kp, double min_f, string J_L7_function, double fb_out, double fb_in, bool using_ia_90) {
    int il, im, ia;
    double p1, Lmax, Lmin, scale_factor, Lgrid;
    int Lsize_7, addGrid; // These two numbers should be of integer type.
    Matrix1D<double> L_1d, fL;
    //Matrix2D<double> fLm(L.size, pc.size);
    
    //////////
    // Steady state solution for any outer-L boundary, D. Subbotin implimentation
    // Create 1D grid at min energy and 90 deg alpha
    // First, find a closest index of 90 deg pitch-angle
    int ia_90=alpha.size-1;
    while (alpha.arr[L.size-1][0][ia_90] > VC::pi/2 && ia_90 > 0) ia_90--;
    
    // Make sure we have a point at L=7
    // Our boundary values J_L7() are at L=7, so we need to know the steady state solution at L=7.
    if (L.arr[L.size-1][0][ia_90] < 7)  {
        // If maximum L value is less than 7, we add another points from Lmax to 7
        
        // estimate grid points to add
        if (using_ia_90) {
            // L values set based on ia_90
            Lmax = L.arr[L.size-1][0][ia_90];
            Lmin = L.arr[0][0][ia_90];
        } else {
            Lmax = L.arr[L.size-1][0][0];
            Lmin = L.arr[0][0][0];
        }
        
        Lgrid = (Lmax-Lmin)/(L.size-1);
        Lsize_7 = (7.0-Lmin)/Lgrid + 1;
        addGrid = Lsize_7 - L.size;
        
        // construct new L and psd grid
        L_1d.AllocateMemory(L.size+addGrid);
        fL.AllocateMemory(L.size+addGrid);
        
        for (il = 0; il < L.size; il++) {
            L_1d[il] = L.arr[il][0][ia_90];
        }
        for (il = L.size; il < Lsize_7; il++) {  // add another points from Lmax to 7
            L_1d[il] = Lmin+Lgrid*il;
        }
        
    } else {
        Lsize_7 = L.size;
        L_1d.AllocateMemory(L.size);
        fL.AllocateMemory(L.size);
        for (il = 0; il < L.size; il++) L_1d[il] = L.arr[il][0][ia_90];
        
    }
    
    steady_state(fL, tau, Kp, Lsize_7, L_1d, fb_out, fb_in);
    if (using_ia_90) {
        scale_factor = 1; // Bugfix. In terms of usage bf.txt as Flux/MeanFlux PSD should be 1 at L==7
    } else {
        scale_factor = fb_out/fL[L.size-1]; // scale factor to make psd to be 1 at the outer boundary (=Lmax(<7))
    }
    
    // If max(L)>7, normalize, so the steady state fL(L=7) = 1;
    if (L_1d[L_1d.size_x-1] > 7)  {
        int il_7 = L.size-1;
        while (L_1d[il_7] > 7 && il_7 > 0) il_7--;
        Output::echo(5, "Steady state normalized by fL(L[%d] = %f) = %f.\n", il_7, L_1d[il_7], fL(il_7));
        fL = fL / fL(il_7);
    }
    //
    ///////////////
    
    // if PSD is lower then min_f - make it min_f
    for (int i = 0; i < fL.size_x; i++) if (fL[i] < min_f) fL[i] = min_f;
    
    
    // inner belt simulation
    //	int il_2=L.size-1;
    //	while (L[il_2][0][0] > 2 && il_2 > 0) il_2--;
    //	for (int i = il_2; i >= 0; i--) fL[i] = min_f/pc[i][0][0]/pc[i][0][0]*pc[il_2][0][0]*pc[il_2][0][0];
    
    
    //#ifdef DEBUG_MODE
    fL.writeToFile("./Debug/fL.dat");
    //#endif
    
    for (im = 0; im < pc.size; im++) {
        for (il = 0; il < L.size; il++) {
            for (ia = 0; ia < alpha.size; ia++) {
                // filling for all pc
                // Why do we do it for alpha=90?:
                // Each point is calculated at the equator and multiplied by sin(pitch angle).
                // BUT! The equatorial point should have the SAME ENERGY.
                // So, we shouldn't use pc[...][ia_90], but should just use pc[...][ia]
                // From the other hand, LATER we use this boundary for the ortogonal grid WITHOUD CHANGES,
                // so, calculation with pc[...][ia_90] is surprisingly more correct at this point.
                // p1 = pc[il][im][alpha.size-1] * sqrt( VF::B(7.)/VF::B(L[il][im][alpha.size-1]));
                if (!using_ia_90) {
                    ia_90=alpha.size-1;
                    while (alpha.arr[il][im][ia_90] > VC::pi/2 && ia_90 > 0) ia_90--;
                }
                
                p1 = pc.arr[il][im][ia_90] * sqrt( VF::B(7.)/VF::B(L.arr[il][im][ia_90]));
                
                arr[il][im][ia] = sin(alpha.arr[il][im][ia]) * scale_factor * fL[il] * VF::Outer_Boundary_Flux_at_L7(VF::Kfunc(p1),J_L7_function) * pow(1.0/p1,2);
                
                // alpha dependence inside the loss cone
                if (alpha.arr[il][im][ia] < VF::alc(L.arr[il][im][ia]) || alpha.arr[il][im][ia] > VC::pi - VF::alc(L.arr[il][im][ia])) {
                    //fLmp[*,i,*]=fLmp[*,i,*]*sinh(i/8.)^2/sinh(1)^2
                    // ???????? arr[il][im][ia] = arr[il][im][ia] * pow(sinh(alpha[il][im][ia]/8.),2)/pow(sinh(1.0),2);
                    arr[il][im][ia] = arr[il][im][ia] * pow(sinh(alpha.arr[il][im][ia]/VF::alc(L.arr[il][im][ia])),2)/pow(sinh(1.0),2);
                }
                //       fLm[*,im]=fL[*]*f1L7(epc[im])*(1/pc[im])^2/fL[index7]
                //       fLm[iL,im]=fL[iL]
                //cout << fLm[il][im] << endl;
            }
        }
    }
    
}



/// two zone initial profile (kckim)
/**
 * \brief Calculate initial %PSD from steady state (kckim).
 * Load initial %PSD with two zone
 *
 * \param &L - grid element  L
 * \param &pc - grid element pc
 * \param &alpha - grid element alpha
 * \param tau - lifetime
 * \param tauLpp - lifetime inside plasmasphere
 * \param Kp - scale diffusion
 * \param min_f - minimum value possible
 * \param J_L7_function - name of boundary flux function at L=7
 * \param fb_out - value of outer boundary
 * \param fb_in - value of inner boundary
 */
void PSD::Load_initial_f_two_zone(GridElement &L, GridElement &pc, GridElement &alpha, double tau, double tauLpp, double Kp, double min_f, string J_L7_function, double fb_out, double fb_in) {
	int il, im, ia;
	double Lmax, Lmin, Lgrid, Lpp;
	int Lsize_7, addGrid; // These two numbers should be of integer type.
	double tau_p, tau_c, alpha1, p1;
	Matrix1D<double> L_1d_i, fL_i, tau_0, Ke1, scale_factor(pc.size);
	Matrix2D<double> pc_i, fLm;

	addGrid = 0;
	Lsize_7 = L.size;

	//////////
	// Steady state solution for any outer-L boundary
	// Create 1D grid at min energy and 90 deg alpha
	// First, find a closest index of 90 deg pitch-angle

	int ia_90 = alpha.size - 1;
	while (alpha.arr[L.size - 1][0][ia_90] > VC::pi / 2 && ia_90 > 0) ia_90--;

	// If maximum L value is less than 7, we add another points from Lmax(<7) to 7
	// and construct pc grid at each added point by moving adiabatically pc values at Lmax(<7)
	if (L.arr[L.size - 1][0][ia_90] < 7.0) {

		// construct new L grid
		Lmax = L.arr[L.size - 1][0][ia_90];
		Lmin = L.arr[0][0][ia_90];
		Lgrid = (Lmax - Lmin) / (L.size - 1);
		Lsize_7 = (7.0 - Lmin) / Lgrid + 1;
		addGrid = Lsize_7 - L.size; // grid points to add
		L_1d_i.AllocateMemory(L.size + addGrid); // construct new L grid
		fL_i.AllocateMemory(L.size + addGrid); // and psd grid

		// construct new pc grid at each added L point
		pc_i.AllocateMemory(L.size + addGrid, pc.size);

		for (il = 0; il < L.size; il++) {
			for (im = 0; im < pc.size; im++) {
				L_1d_i[il] = L.arr[il][0][ia_90];
				pc_i[il][im] = pc.arr[il][im][ia_90];
			}
		}

		// add pc and L values at each added point
		for (il = L.size; il < Lsize_7; il++) {
			for (im = 0; im < pc.size; im++) {
				L_1d_i[il] = Lmin + Lgrid * il;
				pc_i[il][im] = pc.arr[L.size - 1][im][ia_90] * sqrt(VF::B(L_1d_i[il]) / VF::B(L.arr[L.size - 1][im][ia_90]));
			}
		}
	} else {

		// construct new L grid
		addGrid = 0; // grid points to add
		L_1d_i.AllocateMemory(L.size); // construct new L grid
		fL_i.AllocateMemory(L.size); // and psd grid

		// construct new pc grid at each added L point
		pc_i.AllocateMemory(L.size, pc.size);

		for (il = 0; il < L.size; il++) {
			for (im = 0; im < pc.size; im++) {
				L_1d_i[il] = L.arr[il][0][ia_90];
				pc_i[il][im] = pc.arr[il][im][ia_90];
			}
		}

	}

	// solve steady state equation based on the newly constructed grid
	Lpp = (5.6 - 0.46 * Kp);

	tau_0.AllocateMemory(L.size + addGrid);
	Ke1.AllocateMemory(L.size + addGrid);
	fLm.AllocateMemory(L.size + addGrid, pc.size);

	for (im = 0; im < pc.size; im++) {
		for (il = 0; il < Lsize_7; il++) {

			Ke1[il] = VF::Kfunc(pc_i[il][im]); // Energy in MeV
			alpha1 = alpha.arr[L.size - 1][0][alpha.size - 1]; // ~90 deg alpha

			if(PSD_parameters.usetauLpp_SteadyState == "coulomb") {     // tau by coulomb scattering
				tau_0[il] = VF::Coulomb(L_1d_i[il],Ke1[il]);
			} else if(PSD_parameters.usetauLpp_SteadyState == "precipitation") {    // tau by precipitaion
				tau_0[il] = VF::Precipitation(Kp,L_1d_i[il],Ke1[il]);
			} else if (PSD_parameters.usetauLpp_SteadyState == "combined") {      // tau by combined scattering
				tau_c = VF::Coulomb(L_1d_i[il],Ke1[il]);
				tau_p = VF::Precipitation(Kp,L_1d_i[il],Ke1[il]);
				tau_0[il] = 1.0 / (1.0/tau_p+1.0/tau_c);
			} else {                                  // user specificed tau
				tau_0[il] = tauLpp;
			}

		}

		L_1d_i.writeToFile("L_1d_i.dat");
		Ke1.writeToFile("Ke1.dat");
		tau_0.writeToFile("tau_0.dat");

		steady_state_two_zone(fL_i, tau_0, Kp, alpha1, Ke1, Lsize_7, L_1d_i, fb_out, fb_in);

		fL_i.writeToFile("fL_i.dat");

		for (il = 0; il < Lsize_7; il++) {
			if (fL_i[il] < min_f) {
				fLm[il][im] = min_f;
			} else {
				fLm[il][im] = fL_i[il];
			}
		}
	}

	// estimate scale factor that would be used to scale down psd at the outer boundary (=Lmax(<7))
	for (im = 0; im < pc.size; im++) {
		scale_factor[im] = fb_out / fLm[L.size - 1][im];
	}

	for (im = 0; im < pc.size; im++) {
		for (il = 0; il < L.size; il++) {
			for (ia = 0; ia < alpha.size; ia++) {
				// filling for all pc
				// Why do we do it for alpha=90?:
				// Each point is calculated at the equator and multiplied by sin(pitch angle).
				// BUT! The equatorial point should have the SAME ENERGY.
				// So, we shouldn't use pc[...][ia_90], but should just use pc[...][ia]
				// From the other hand, LATER we use this boundary for the ortogonal grid WITHOUD CHANGES,
				// so, calculation with pc[...][ia_90] is surprisingly more correct at this point.
				// p1 = pc[il][im][alpha.size-1] * sqrt( VF::B(7.)/VF::B(L[il][im][alpha.size-1]));
				int ia_90 = alpha.size - 1;
				while (alpha.arr[il][im][ia_90] > VC::pi / 2 && ia_90 > 0)
					ia_90--;

				p1 = pc.arr[il][im][ia_90] * sqrt(VF::B(7.) / VF::B(
						L.arr[il][im][ia_90]));
				arr[il][im][ia] = sin(alpha.arr[il][im][ia]) * scale_factor[im]
						* fLm[il][im] * VF::Outer_Boundary_Flux_at_L7(
						VF::Kfunc(p1), J_L7_function) * pow(1.0 / p1, 2);

				// alpha dependence inside the loss cone
				if (alpha.arr[il][im][ia] < VF::alc(L.arr[il][im][ia])
						|| alpha.arr[il][im][ia] > VC::pi - VF::alc(
								L.arr[il][im][ia])) {
					arr[il][im][ia] = arr[il][im][ia] * pow(
							sinh(alpha.arr[il][im][ia] / VF::alc(
									L.arr[il][im][ia])), 2) / pow(sinh(1.0), 2);
				}
			}
		}
	}

#ifdef DEBUG_MODE
	//arr.writeToFile("./Debug/f_0.plt", L, pc, alpha);
#endif

}





/**
 * Calculate steady state for two zones.
 *
 * \param &f - function
 * \param tau - life time
 * \param Kp - Kp index value
 * \param alpha - pitch angle value
 * \param &Ke - kinetic energy
 * \param nx - number of points
 * \param &L - grid
 * \param f_bnd_out - value of outer boundary
 * \param f_bnd_in - value of inner boundary
 */
void steady_state_two_zone(Matrix1D<double> &f, Matrix1D<double> &tau, double Kp, double alpha, Matrix1D<double> &Ke, int nx, Matrix1D<double> &L, double f_bnd_out, double f_bnd_in) {//, int nit

	//Matrix1D<double> f(nx);
	//Matrix1D<double> L(nx);
	Matrix1D<double> C_1(nx);
	Matrix1D<double> C_2(nx);
	Matrix1D<double> C_3(nx);
	Matrix1D<double> alfa(nx);
	Matrix1D<double> A(nx);
	Matrix1D<double> B(nx);
	Matrix1D<double> C(nx);
	Matrix1D<double> D(nx);
	Matrix1D<double> Dxx(nx);

	int i;

    // calculate Dxx
	for (i = 0; i <= nx-1; i++) {
		//Dxx[i] = VF::Df(L[i], Kp) + VF::Dfe(alpha,Ke[i],L[i],Kp);
		Dxx[i] = pow(10, 0.506*Kp-9.325)*pow(L[i],10) + VF::Dfe(alpha,Ke[i],L[i],Kp);
	}

	Dxx.writeToFile("Dxx.dat");

	for (i = 1; i < nx-1; i++) {
		alfa[i] = 1.0/((L[i+1] - L[i-1])/2)*pow(L[i],2);
		//C_1[i] = (Dxx[i] + Dxx[i-1])/(pow(((L[i] + L[i-1])/2), 2));
		C_1[i] = ((Dxx[i] + Dxx[i-1])/2)/(L[i] - L[i-1])/(pow(((L[i] + L[i-1])/2), 2));
		C_2[i] = ((Dxx[i+1] + Dxx[i])/2)/(L[i+1] - L[i])/(pow(((L[i+1] + L[i])/2), 2));
		C_3[i] = 1./tau[i];
		//alfa[i] = pow(L[i],2)/(2.*pow((L[i+1] - L[i-1])/2, 2));
	}

    // compute values for the diagonals
	for (i = 1; i < nx-1; i++) {
		A[i] = alfa[i] * C_1[i];
		B[i] = -alfa[i] * (C_1[i] + C_2[i]) - C_3[i];
		C[i] = alfa[i] * C_2[i];
	}

	for (i = 0; i < nx; i++) D[i] = 0;;

    // setting upper and lower boundaries and values
	i = 0;
	A[i] = 0.;
	B[i] = 1;
	C[i] = 0;
	D[i] = f_bnd_in;

	i=nx-1;
	A[i] = 0;
	B[i] = 1;
	C[i] = 0;
	D[i] = f_bnd_out;

#ifdef DEBUG_MODE
		Dxx.writeToFile("./Debug/SS_DLL.plt");
		A.writeToFile("./Debug/SS_A.plt");
		B.writeToFile("./Debug/SS_B.plt");
		C.writeToFile("./Debug/SS_C.plt");
		D.writeToFile("./Debug/SS_D.plt");
#endif

	tridag(&A[0], &B[0], &C[0], &D[0], &f[0], nx);

	//	for (i = 0; i < nx; i++) f[i] = f[i]/f_bnd_out;

}


/**
 * \brief Calculate initial %PSD from steady state using boundary conditions.
 * Load initial %PSD from outer L
 *
 * \param &L - radial distance
 * \param &pc - MeV
 * \param &alpha - pitch angle
 * \param tau - life time
 * \param Kp - Kp index value
 * \param L_UpperBoundaryCondition - L's upper boundary condition
 * \param min_f - minimum value of vector
 * \param fb_out - value of outer boundary
 * \param fb_in - value of inner boundary
 */
void PSD::Load_initial_f_from_outer_L(GridElement &L, GridElement &pc, GridElement &alpha, double tau, double Kp, Matrix2D<double> L_UpperBoundaryCondition, double min_f, double fb_out, double fb_in) {
	int il, im, ia;
	double p1;
	Matrix1D<double> L_1d(L.size), fL(L.size);
	Matrix2D<double> fLm(L.size, pc.size);

	//	L_UpperBoundaryCondition.writeToFile("./Debug/L_UpperBoundaryCondition.dat");

	// steady state
	for (il = 0; il < L.size; il++) L_1d[il] = L.arr[il][0][alpha.size-1]; // create 1D grid at min energy and 90 deg alpha
	steady_state(fL, tau, Kp, L.size, L_1d, fb_out, fb_in);
	for (int i = 0; i < fL.size_x; i++) if (fL[i] < min_f) fL[i] = min_f;

#ifdef DEBUG_MODE
	fL.writeToFile("./Debug/fL.dat");
#endif

	// Steady state from L-upper boundary
	for (im = 0; im < pc.size; im++) {
		for (il = 0; il < L.size; il++) {
			for (ia = 0; ia < alpha.size; ia++) {
				p1 = sqrt( VF::B(7.)/VF::B(L.arr[il][im][ia])); // why 7?
				arr[il][im][ia] = sin(alpha.arr[il][im][ia]) * fL[il] * L_UpperBoundaryCondition[im][ia] * pow(1.0/p1,2);
			}
		}
	}

#ifdef DEBUG_MODE
	//arr.writeToFile("./Debug/f_0.plt", L, pc, alpha);
#endif

}



/**
*  Load initial %PSD from 2d-file (for 2d calculations).
*
* \param &L - grid element  L
* \param &pc - grid element pc
* \param &alpha - grid element alpha
* \param *filename - file name
*/

void PSD::Load_initial_f_2d(GridElement &L, GridElement &pc, GridElement &alpha, const char *filename) {
	int i, j, k;
	int il, im, ia;
	int jmax, lmax;
	// number of orbits
	lmax=8;
	// number of j on orbit
	jmax=16;

	Matrix2D<double> jperp(lmax+1, jmax+1);
	Matrix1D<double> energy(jmax+2), jcopy(jmax+2);
	Matrix1D<double> j_0m(pc.size), f_0m(pc.size);
	Matrix1D<double> dayno(lmax+1), eventtime(lmax+1), orbit(lmax+1);

	// Load j from input file
	ifstream input_f(filename);
	if (input_f != NULL && !input_f.eof()) {
		for (k = 0; k <= lmax; k++) {
			input_f >> orbit[k] >> dayno[k] >> eventtime[k];
			for (j = 0; j <= jmax; j++) {
				input_f >> i >> energy[j] >> jperp[k][j];
				energy[j] = energy[j]/1000;
			}
		}
	} else {
		throw error_msg("INITIAL_PSD_ERROR", "Error reading initial conditions input file %s.\n", filename);
		//cout << "Error reading initial conditions input file " << filename << endl;
		//getchar();
		//exit(-1);
	}
	input_f.close();

	// Copying j into 1D matrix_array for interpolation
	for (i = 0; i <= jmax; i++) jcopy[i] = jperp[2][i];

	// Adding upper border point and border condition
	energy[jmax+1] = VF::Kfunc(pc.GridElement_parameters.max);
	jcopy[jmax+1] = VC::zero_f;

#ifdef DEBUG_MODE
	jcopy.writeToFile("./Debug/jcopy.plt", energy);
#endif

	if (VF::Kfunc(pc.GridElement_parameters.min) < energy[0]) printf("Warning: Emin less than minimum data in input file!\n");

    // log all values in jcopy
//    for (i = 0; i < jmax+2; i++) {
//        jcopy[i] = log10(jcopy[i]);
//    }
    
	// Interpolating(old grid, new grid);
	Matrix1D<double> epc1D(pc.size); // temporary 1D energy-grid
    j_0m.Spline(jcopy, energy, epc1D, jcopy[0], log10(VC::zero_f), 0, 1); // spline into epc1D
    for (i = 0; i < pc.size; i++) {
		epc1D[i] = VF::Kfunc(pc.arr[0][i][0]);
	//	j_0m.Interpolate(jcopy, energy, epc1D);
//        j_0m[i] = pow(10, j_0m[i]);
//        
//        // Calculating from j_0m to f_0m.
//        if (j_0m[i] < 0)
//            j_0m[i] = VC::zero_f;
//        f_0m[i] = j_0m[i]/pow(pc.arr[0][i][0], 2);
    }
    for (i = 0; i < jmax+2; i++) {
        jcopy[i] = log10(jcopy[i]);
    }
    j_0m.Spline(jcopy, energy, epc1D, jcopy[0], log10(VC::zero_f), 0, 1); // spline into epc1D


	for (i = 0; i < pc.size; i++) {
		j_0m[i] = pow(10, j_0m[i]);
	}

	// Calculating from j_0m to f_0m.
	for (i = 0; i < pc.size; i++) {
		if (j_0m[i] < 0) j_0m[i] = VC::zero_f;
		f_0m[i] = j_0m[i]/pow(pc.arr[0][i][0], 2);
	}

#ifdef DEBUG_MODE
	j_0m.writeToFile("./Debug/j_0m.plt", epc1D);
	f_0m.writeToFile("./Debug/f_0m.plt", epc1D);
#endif

    /// \todo not exactly sure what this does
	for (il = 0; il < L.size; il++) {
		for (im = 0; im < pc.size; im++) {
			for (ia = 0; ia < alpha.size; ia++) {
                // copying over interpolated values
				arr[il][im][ia] = f_0m[im] * sin( alpha.arr[il][im][ia] );
				if (alpha.arr[il][im][ia] < VF::alc(L.arr[il][im][ia]) || alpha.arr[il][im][ia] > VC::pi - VF::alc(L.arr[il][im][ia])) {
					arr[il][im][ia] = arr[il][im][ia] * pow(sinh(alpha.arr[il][im][ia]/VF::alc(L.arr[il][im][ia])),2)/pow(sinh(1.0),2);
				}
			}
		}
	}
}



/**
*  Load initial %PSD from maxwell (for 2d calculations).
*
* \param &L - grid element  L
* \param &pc - grid element pc
* \param &alpha - grid element alpha
*/

void PSD::Load_initial_f_maxwell(GridElement &L, GridElement &pc, GridElement &alpha) {
	int il, im, ia;
	double Ke, alf, flux;

	for (il = 0; il < L.size; il++) {
		for (im = 0; im < pc.size; im++) {
			for (ia = 0; ia < alpha.size; ia++) {

				Ke = VF::Kfunc(pc.arr[0][im][0]);   // Energy in MeV
				alf = alpha.arr[il][im][ia];          // pitch angle in radian
				flux = exp(-(Ke-0.2)/0.1); //* ( sin(alf)-sin(VF::alc(L[il][im][ia])) );
				//flux = exp(-(Ke-0.2)/0.1)* ( sin(alf)-sin(VF::alc(L[il][im][ia])) ); */
				arr[il][im][ia] = flux / pow(pc.arr[0][im][0], 2);

			}
		}
	}
}



/**
* Calculate steady state.
*
* \param &f - function
* \param tau - life time
* \param Kp - Kp index value
* \param nx - number of points
* \param &L - grid
* \param f_bnd_out - the value on the outer boundary
* \param f_bnd_in - the value on the inner boundary
*/
void steady_state(Matrix1D<double> &f, double tau, double Kp, int nx, Matrix1D<double> &L, double f_bnd_out, double f_bnd_in) {//, int nit

	//Matrix1D<double> f(nx);
	//Matrix1D<double> L(nx);
	Matrix1D<double> C_1(nx);
	Matrix1D<double> C_2(nx);
	Matrix1D<double> C_3(nx);
	Matrix1D<double> alfa(nx);
	Matrix1D<double> A(nx);
	Matrix1D<double> B(nx);
	Matrix1D<double> C(nx);
	Matrix1D<double> D(nx);
	Matrix1D<double> Dxx(nx);

	int i;

	for (i = 0; i <= nx-1; i++) {
		// Radial diffusion coefficient approximation by Brautigam and Albert [2000]
		Dxx[i] = pow(10, 0.506*Kp-9.325)*pow(L[i],10);
	}

	for (i = 1; i < nx-1; i++) {
		alfa[i] = 1.0/((L[i+1] - L[i-1])/2)*pow(L[i],2);
		//C_1[i] = (Dxx[i] + Dxx[i-1])/(pow(((L[i] + L[i-1])/2), 2));
		C_1[i] = ((Dxx[i] + Dxx[i-1])/2)/(L[i] - L[i-1])/(pow(((L[i] + L[i-1])/2), 2));
		C_2[i] = ((Dxx[i+1] + Dxx[i])/2)/(L[i+1] - L[i])/(pow(((L[i+1] + L[i])/2), 2));
		C_3[i] = 1./tau;
		//alfa[i] = pow(L[i],2)/(2.*pow((L[i+1] - L[i-1])/2, 2));
	}

    // compute values for the diagonals
	for (i = 1; i < nx-1; i++) {
		A[i] = alfa[i] * C_1[i];
		B[i] = -alfa[i] * (C_1[i] + C_2[i]) - C_3[i];
		C[i] = alfa[i] * C_2[i];
	}

	for (i = 0; i < nx; i++) D[i] = 0;;

    // setting upper and lower boundary values
	i = 0;
	A[i] = 0.;
	B[i] = 1;
	C[i] = 0;
	D[i] = f_bnd_in;

	i=nx-1;
	A[i] = 0;
	B[i] = 1;
	C[i] = 0;
	D[i] = f_bnd_out;

#ifdef DEBUG_MODE
		Dxx.writeToFile("./Debug/SS_DLL.plt");
		A.writeToFile("./Debug/SS_A.plt");
		B.writeToFile("./Debug/SS_B.plt");
		C.writeToFile("./Debug/SS_C.plt");
		D.writeToFile("./Debug/SS_D.plt");
#endif

	tridag(&A[0], &B[0], &C[0], &D[0], &f[0], nx);

	//	for (i = 0; i < nx; i++) f[i] = f[i]/f_bnd_out;

}
