/**
* \file PSD.h
*
* \brief Makes operations with %PSD (Phase Space Density (%PSD).) (like, diffusion).
*
* \author Developed under supervision of the PI Yuri Shprits
* 
* \todo
* - What Interpolate function does
*/


#ifndef PhaseSpaceDensity_H
/// \def PhaseSpaceDensity_H
/// \brief Check if PhaseSpaceDensity_H is defined, to not include the class more then one time.
#define PhaseSpaceDensity_H

// General C++ headers
#include <fstream>
#include <string>
// Type of data (it could be deprecated)
//#include <map>

// Operations with Grid array
#include "../Grid/Grid.h"
// Operations with matrixes
#include "../Matrix/Matrix.h"
// Operations with coefficients
#include "../DiffusionCoefficient/DiffusionCoefficient.h"
// Pfrfmeters from ini file
#include "../Parameters/Parameters.h"
// Equation solver (main algoritm)
#include "MatrixSolver.h"

// class to operate by sources and losses processes
#include "../Grid/AdditionalSourcesAndLosses.h"


using namespace std;


/**
* \brief Phase Space Density class.
*
* Do diffusions and store the result in Matrix3D.
* Do all possible operations with %PSD (like output, loading, etc).
*/
class PSD {
public:

	
	/// \brief array of PSD values
	Matrix3D<double> arr;

	/// \brief Parameters from ini file
	Parameters_structure::PSD PSD_parameters;

	/**
	* \brief Create inital PSD (Steady State)
	*
	* \param parameters - interpolation parameters structure
	* \param &grid - grid
	* \param L_UpperBoundaryCondition - upper L boundary values, in case we need to create initial PSD as a steady state radial diffusion solution
	*/
	void Create_Initial_PSD(Parameters_structure::PSD parameters, Grid &grid, BoundaryCondition L_UpperBoundaryCondition);

	/**
	* \brief Interpolation function.
	*
	* \param &otherPSD - %PSD to interpolate to the values from
	* \param interpolationParameters - parameters for interpolation
	* \param &oldGrid - old grid
	* \param &newGrid - new grid
	* \param newGrid_pc_lowerBoundaryCondition - lower energy boundary VALUE
	* \param newGrid_pc_upperBoundaryCondition - upper energy boundary VALUE
	*/
	void Interpolate(PSD &otherPSD, Parameters_structure::Interpolation interpolationParameters, Grid &oldGrid, Grid &newGrid, Matrix2D<double> newGrid_pc_lowerBoundaryCondition, Matrix2D<double> newGrid_pc_upperBoundaryCondition);

	
	void Load_initial_f(GridElement &L, GridElement &pc, GridElement &alpha, double tau, double Kp, double min_f = 1.e-99, string J_L7_function = "J_L7", double fb_out = 1, double fb_in = 0, bool using_ip_90 = true);
	
	
	void Load_initial_f_two_zone(GridElement &L, GridElement &pc, GridElement &alpha, double tau, double tauLpp, double Kp, double min_psd = 1.e-99, string J_L7_function = "J_L7", double fb_out = 1, double fb_in = 0);
	
	
	void Load_initial_f_from_outer_L(GridElement &L, GridElement &pc, GridElement &alpha, double tau, double Kp, Matrix2D<double> L_UpperBoundaryCondition, double min_psd = 1.e-99, double fb_out = 1, double fb_in = 0);
	
	
	void Load_initial_f_2d(GridElement &L, GridElement &pc, GridElement &alpha, const char *filename);
	
	
	void Load_initial_f_file(GridElement &L, GridElement &pc, GridElement &alpha, const char *filename, bool with_grid);
	
	
	void Load_initial_f_maxwell(GridElement &L, GridElement &pc, GridElement &alpha);

	// Explicit mixed terms
	void DiffusionMixTermExplicit(double dt, double Lpp,
		DiffusionCoefficient &Dpca, DiffusionCoefficient &DpcaLpp,
		GridElement &L, GridElement &pc, GridElement &alpha, Matrix3D<double> Jacobian,
		Matrix2D<double> pc_lowerBoundaryCondition,
		Matrix2D<double> pc_upperBoundaryCondition,
		Matrix2D<double> alpha_lowerBoundaryCondition,
		Matrix2D<double> alpha_upperBoundaryCondition,
		string pc_lowerBoundaryCondition_calculationType,
		string pc_upperBoundaryCondition_calculationType,
		string alpha_lowerBoundaryCondition_calculationType,
		string alpha_upperBoundaryCondition_calculationType);


	void Diffusion_alpha(double dt, double Lpp,
		DiffusionCoefficient &Daa, DiffusionCoefficient &DaaLpp,
		GridElement &L, GridElement &pc, GridElement &alpha, Matrix3D<double> Jacobian,
		Matrix2D<double> alpha_lowerBoundaryCondition,
		Matrix2D<double> alpha_upperBoundaryCondition,
		string alpha_lowerBoundaryCondition_calculationType,
		string alpha_upperBoundaryCondition_calculationType);

	void Diffusion_pc(double dt, double Lpp,
		DiffusionCoefficient &Dpcpc, DiffusionCoefficient &DpcpcLpp,
		GridElement &L, GridElement &pc, GridElement &alpha, Matrix3D<double> Jacobian,
		Matrix2D<double> pc_lowerBoundaryCondition,
		Matrix2D<double> pc_upperBoundaryCondition,
		string pc_lowerBoundaryCondition_calculationType,
		string pc_upperBoundaryCondition_calculationType);

	void Diffusion_L(double dt, double Lpp,
		DiffusionCoefficient &DLL,
		GridElement &L, GridElement &pc, GridElement &alpha, Matrix3D<double> Jacobian,
		Matrix2D<double> lowerBoundaryCondition,
		Matrix2D<double> upperBoundaryCondition,
		string lowerBoundaryCondition_calculationType,
		string upperBoundaryCondition_calculationType,
		double tau, double tauLpp);


	void Diffusion_pc_alpha(double dt, double Lpp,
		DiffusionCoefficient &Dpcpc, DiffusionCoefficient &DpcpcLpp,
		DiffusionCoefficient &Daa, DiffusionCoefficient &DaaLpp,
		DiffusionCoefficient &Dpca, DiffusionCoefficient &DpcaLpp,
		GridElement &L, GridElement &pc, GridElement &alpha, Matrix3D<double> Jacobian,
		Matrix2D<double> pc_lowerBoundaryCondition,
		Matrix2D<double> pc_upperBoundaryCondition,
		Matrix2D<double> alpha_lowerBoundaryCondition,
		Matrix2D<double> alpha_upperBoundaryCondition,
		string pc_lowerBoundaryCondition_calculationType,
		string pc_upperBoundaryCondition_calculationType,
		string alpha_lowerBoundaryCondition_calculationType,
		string alpha_upperBoundaryCondition_calculationType);

	void Diffusion_pc_alpha_KC(double dt, double Lpp,
		DiffusionCoefficient &Dpcpc, DiffusionCoefficient &DpcpcLpp,
		DiffusionCoefficient &Daa, DiffusionCoefficient &DaaLpp,
		DiffusionCoefficient &Dpca, DiffusionCoefficient &DpcaLpp,
		GridElement &L, GridElement &pc, GridElement &alpha, Matrix3D<double> Jacobian,
		Matrix2D<double> pc_lowerBoundaryCondition,
		Matrix2D<double> pc_upperBoundaryCondition,
		Matrix2D<double> alpha_lowerBoundaryCondition,
		Matrix2D<double> alpha_upperBoundaryCondition,
		string pc_lowerBoundaryCondition_calculationType,
		string pc_upperBoundaryCondition_calculationType,
		string alpha_lowerBoundaryCondition_calculationType,
		string alpha_upperBoundaryCondition_calculationType);

    //@{
    /// calculation matrices A, B, and C
	CalculationMatrix matr_A, matr_B, matr_C;
    //@}


	//// scale boundary flux ////////
	//void ScaleBoundaryFlux(Parameters_structure::PSD parameters, GridElement &L, GridElement &pc, GridElement &alpha);
	/////////////////////////////////

	void SourcesAndLosses(
			Parameters_structure parameters,
			GridElement &L, GridElement &pc, GridElement &alpha,
			AdditionalSourcesAndLosses &SL,
			double dt,
			double Lpp, double tau, double tauLpp, double Kp);

	// output functions
	void Output_without_grid(double time);
	ofstream output_without_grid_file;  ///< output stream for PSD

};

/** Compute steady state solution for 1D diffusion
*/
void steady_state(Matrix1D<double> &f, double tau, double Kp, int nx, Matrix1D<double> &CL, double f_bnd_out = 1, double f_bnd_in = 0);
// kckim
void steady_state_two_zone(Matrix1D<double> &f, Matrix1D<double> &tau, double Kp, double alpha, Matrix1D<double> &Ke, int nx, Matrix1D<double> &CL, double f_bnd_out = 1, double f_bnd_in = 0);
//

#endif

