/**
 * Diffusion coefficients calculation, loading, activating, scaling etc code
 *
 * \file DiffusionCoefficient.cpp
 *
 * \todo comment this file if you understand the functions and math
 *
 * \author Developed by Yuri Shprits
 */

#include "DiffusionCoefficient.h"

#include <math.h>
#include <vector>
#include "rroots.h"
#include "../VariousFunctions/bisection.h"

#include "../Logging/Output.h"
#include "../Exceptions/error.h"

#include <ctime>

using namespace std;

#define double_zero 1.e-21
#define min_Dxx 1.e-21

#if defined(_WIN32) || defined(_WIN64)
#define strncasecmp _strnicmp
#define strcasecmp _stricmp
#endif


/** Function Get - return diffusion coefficient by loading or calculating it
 * Two way of call function:
 * With or without grid parameter
 * If it called without grid parameter, it takes grid stored in the class
 * \todo Remove grid from the class DiffusionCoefficients and remove functions, these assume it there
 * \param DxxParameters - diffusion coefficients param structure
 * \param &grid - optional grid parameter to use
 */
void DiffusionCoefficient::Create_Dxx(DxxParameters_structure  DxxParameters, Grid &grid) {
	bool Load_Failed = false;

	if (!arr.initialized) {
		// Allocating memory for the parent matrix class (the function is from Matrix class)
		arr.AllocateMemory(grid.L.size, grid.pc.size, grid.alpha.size);
	} else {
		arr = 0;
	}

	arr.name = DxxParameters.DxxName + " " + DxxParameters.waveName + "";
	this->DxxParameters = DxxParameters;

	this->type = DxxParameters.DxxType;
	this->useScale = DxxParameters.useScale;

	if (DxxParameters.loadOrCalculate != "LOC_LOAD" && DxxParameters.loadOrCalculate != "LOC_LOADORCALCULATE" && DxxParameters.loadOrCalculate != "LOC_CALCULATE") {
			throw error_msg("DXX_GET", "Undefined load or calculate choice.");
	}

	// loading or calculating diffusion coefficients
	// if DxxParameters.loadOrCalculate == LOC_LOADORCALCULATE - try to load first, and calculate in case of error
	if (DxxParameters.loadOrCalculate == "LOC_LOAD" || DxxParameters.loadOrCalculate == "LOC_LOADORCALCULATE") {
		try {
			// try to load
			LoadDiffusionCoefficient(grid.L, grid.epc, grid.alpha, DxxParameters.filename.c_str(), DxxParameters.filetype);
		} catch (error_msg err) {
			if (DxxParameters.loadOrCalculate == "LOC_LOADORCALCULATE") {
				Output::echo(2, "%s", err.what().c_str());
				Load_Failed = true;
			} else {
				err.add("UNKNOWN_ERROR", "Unknown error while loading diffusion coefficient file");
				throw err;
			}
		}
	}

	// Calculating if need to calculate or failed to load
	if (DxxParameters.loadOrCalculate == "LOC_CALCULATE" || (DxxParameters.loadOrCalculate == "LOC_LOADORCALCULATE" && Load_Failed)) {
		Output::echo(3, "Calculating %s...\n", arr.name.c_str());
		Calculate(grid.L, grid.epc, grid.alpha, DxxParameters);
		// Transformation to days^-1
		arr = arr * (60.0 * 60 * 24);
		// plotting the file, that we could not open to use it next time
		arr.writeToFile(DxxParameters.filename.c_str(), grid.L.arr, grid.epc.arr, grid.alpha.arr);
	}

	// MLT averaging
	if (DxxParameters.MLT_averaging != 1) {
		arr = arr * DxxParameters.MLT_averaging;
		Output::echo(3, "%s is MLT-averaged as %f %%.\n", arr.name.c_str(), DxxParameters.MLT_averaging*100);
	}
	// Multiplicator
	if (DxxParameters.multiplicator != 1) {
		arr = arr * DxxParameters.multiplicator;
		Output::echo(3, "%s is multiplied by %f.\n", arr.name.c_str(), DxxParameters.multiplicator);
	}
	is_active = false;

	// Normalizing diffusion coefficients to 'pc':
	// Version 1 (slowest):
	//if (this->type == "DCT_Dpp" || this->type == "DCT_DppLpp") arr = arr * grid.pc.arr * grid.pc.arr;
	//if (this->type == "DCT_Dpa" || this->type == "DCT_DpaLpp") arr = arr * grid.pc.arr;
	// Version 2 (faster):
	if (this->type == "DCT_Dpp" || this->type == "DCT_DppLpp") arr.times_equal(grid.pc.arr.times(grid.pc.arr)); 	// Dxx *= (pc * pc)
	if (this->type == "DCT_Dpa" || this->type == "DCT_DpaLpp") arr.times_equal(grid.pc.arr);						// Dxx *= pc

	arr.change_ind = clock();
}


/** Calculating the diffusion coefficients
 * \param &L - Grid element L
 * \param &epc - Grid element epc
 * \param &Alpha - Grid element Alpha
 * \param DxxParameters - Dxx parameters structure
 */
void DiffusionCoefficient::Calculate(GridElement &L, GridElement &epc, GridElement &Alpha, DxxParameters_structure  DxxParameters) {
	int il, im, ia;
	bool positivonly=true;
	double (*int_Dxx_loc)(double lambda, DxxParameters_structure DxxParameters);

	// Chousing the function by the type
	if (DxxParameters.DxxType == "DCT_Daa" || DxxParameters.DxxType == "DCT_DaaLpp") {
			int_Dxx_loc = int_Daa_loc;
	} else if (DxxParameters.DxxType == "DCT_Dpp" || DxxParameters.DxxType == "DCT_DppLpp") {
			int_Dxx_loc = int_Dpp_loc;
	} else if (DxxParameters.DxxType == "DCT_Dpa" || DxxParameters.DxxType == "DCT_DpaLpp") {
			int_Dxx_loc = int_Dpa_loc;
			positivonly = false;
	} else {
			throw error_msg("DXX_TYPE", "Unknown diffusion coefficient");
	}

	Output::echo(0, "Calculating %s for %s... ", DxxParameters.DxxName.c_str(), DxxParameters.waveName.c_str());

	for (il = 0; il < L.size ; il++) {
		for (im = 0; im < epc.size; im++) {
			for (ia = 0; ia < Alpha.size; ia++) {
				// caltulating the value in one grid point
				// /2 is because missing factor of 2 in Sammers formula
				arr[il][im][ia] = Dxx_ba(L.arr[il][im][ia], epc.arr[il][im][ia]/VC::mc2, Alpha.arr[il][im][ia], int_Dxx_loc, DxxParameters) / 2;
				// remove zeros and/or negative values
				if (positivonly && (arr[il][im][ia] < min_Dxx)) {
					arr[il][im][ia] = min_Dxx;
				} else if (!(arr[il][im][ia] >= 0) && !(arr[il][im][ia] <= 0)) { // if NAN
					if (positivonly) {
						arr[il][im][ia] = min_Dxx;
					} else {
						arr[il][im][ia] = 0;
					}
				}
			}
		}
	}

	Output::echo(2, "done!\n");
}

// LOADING PROCEDURES
/** Loading coefficient from file
 * runs loading function according file format (filetype)
 */
bool DiffusionCoefficient::LoadDiffusionCoefficient(GridElement &L, GridElement &pc, GridElement &alpha, string D_filename, string filetype) {
	if (!arr.initialized) {
		throw error_msg("DXX_UNINITIALIZED_MATRIX", "Matrix has not been initialized.");
	}
	Output::echo(2, "Loading %s... ", D_filename.c_str());
	// using different loading functions according to type of file
	// like, file with/without grid, different order of grid, etc
	if (filetype == "IFT_GRID") {
			LoadDiffusionCoefficientFromFileWithGrid(L, pc, alpha, D_filename);
	} else if (filetype == "IFT_GRID_LPA") {
			LoadDiffusionCoefficientFromFileWithGrid(L, pc, alpha, D_filename, "IFT_GRID_LPA");
	} else if (filetype == "IFT_GRID_APL") {
			LoadDiffusionCoefficientFromFileWithGrid(L, pc, alpha, D_filename, "IFT_GRID_APL");
	} else if (filetype == "IFT_PLANE") {
			// if it is "simple" (just matrix_array) - using corresponding function
			LoadDiffusionCoefficientFromPlaneFile(L, pc, alpha, D_filename);
	} else {
		throw error_msg("DXX_LOAD_TYPE_ERR", "Unknown coefficients file format %s", filetype.c_str());
	}
	Output::echo(2, "done!\n");
	return true;
}

/** Loading coefficient from plane file
 */
bool DiffusionCoefficient::LoadDiffusionCoefficientFromPlaneFile(GridElement &L, GridElement &pc, GridElement &alpha, string D_filename) {
	// iterators
	int il, im, ia;

	// making stresm
	ifstream input_D(D_filename.c_str());
	// checking if file was opened
	if (input_D != NULL && !input_D.eof()) {
		for (il = 0; il < L.size; il++)
			for (im = 0; im < pc.size; im++)
				for (ia = 0; ia < alpha.size; ia++) {
					// load file into D
					input_D >> arr[il][im][ia];
					if (input_D == NULL) {
						throw error_msg("LOAD_ERROR", "Wrong coefficient file %s (to few values)\n", D_filename.c_str());
					}
				}
	} else {
		throw error_msg("DXX_LOAD_GRID_ERR", "error reading input file %s\n", D_filename.c_str());
		//getchar();
		return false;
	}
	input_D.close();
	return true;

}


/** Loading coef from file with grid
 */
bool DiffusionCoefficient::LoadDiffusionCoefficientFromFileWithGrid(GridElement &L, GridElement &epc, GridElement &alpha, string D_filename, string gridOrder) {
	int il, im, ia;
	double Load_L, Load_epc, Load_alpha;// temporaraly variables for the grid point
	double err = 5.e-3;
	string tmp;

	// making stresm
	ifstream input_D(D_filename.c_str());
	// checking if file was opened
	if (input_D == NULL) {
		throw error_msg("DXX_FILE_OPEN_ERR", "Diffusion coefficient file %s not found.", D_filename.c_str());
	}

	// Skip header!
	input_D >> tmp;
	while (strcasecmp(tmp.c_str(), "ZONE") != 0 && !input_D.eof() ) {
			input_D >> tmp;
			Output::echo(2, "%s ", tmp.c_str());
			if (input_D.eof()) {
				throw error_msg("DXX_LOAD_GRID_ERR", "Keyword 'ZONE' not found in the diffusion coefficient file: %s\n", D_filename.c_str());
			}
	}
	// read to the end of the line with 'zone'
	input_D.ignore(9999, '\n');

	if (!input_D.eof()) {
		// load grid
		if (gridOrder == "IFT_GRID_LPA") {
				for (il = 0; il < L.size; il++)
					for (im = 0; im < epc.size; im++)
						for (ia = 0; ia < alpha.size; ia++) {
							input_D >> Load_L >> Load_epc >> Load_alpha;
							// load D
							//input_D >> (*this)[il][im][ia];
							input_D >> arr[il][im][ia];
							// check if grid is the same
							if (fabs(Load_L - L.arr[il][im][ia]) > err || fabs(Load_epc - epc.arr[il][im][ia]) > err || fabs(Load_alpha - alpha.arr[il][im][ia]) > err) {
								throw error_msg("DXX_LOAD_GRID_ERR", "Loading %s: grid mismatch.\nLoaded: L=%e, epc=%e, alpha=%e\nGrid:  L=%e, epc=%e, alpha=%e\n", D_filename.c_str(), Load_L, Load_epc, Load_alpha, L.arr[il][im][ia], epc.arr[il][im][ia], alpha.arr[il][im][ia]);
							}
						}
		} else if (gridOrder == "IFT_GRID_APL") {
				for (ia = 0; ia < alpha.size; ia++)
					for (im = 0; im < epc.size; im++)
						for (il = 0; il < L.size; il++) {
							input_D >> Load_alpha >> Load_epc >> Load_L ;
							// load D
							//input_D >> (*this)[il][im][ia];
							input_D >> arr[il][im][ia];
							// check if grid is the same
							if (fabs(Load_L - L.arr[il][im][ia]) > err || fabs(Load_epc - epc.arr[il][im][ia]) > err || fabs(Load_alpha - alpha.arr[il][im][ia]) > err) {
								throw error_msg("DXX_LOAD_GRID_ERR", "Loading %s: grid mismatch.\nLoaded: L=%e, epc=%e, alpha=%e\nGrid:  L=%e, epc=%e, alpha=%e\n", D_filename.c_str(), Load_L, Load_epc, Load_alpha, L.arr[il][im][ia], epc.arr[il][im][ia], alpha.arr[il][im][ia]);
							}
						}
		} else {
			throw error_msg("UNKNOWN_GRID_TYPE", "Unknown grid order: %s", gridOrder.c_str());
		}
	}
	input_D.close();
	return true;

}

/** Making DLL
 * take DLLType as input, do proper DLL creation method
 */
void DiffusionCoefficient::MakeDLL(GridElement &L, GridElement &pc, GridElement &alpha, double Kp, string DLLType) {
	if (DLLType == "DLLT_BE") {
			MakeDLL_BE(L, pc, alpha, Kp);
	} 
	else if (DLLType == "DLLT_B") {
			MakeDLL_B(L, pc, alpha, Kp);
	} 
	else if (DLLType == "DLLT_FAKE") {
			MakeDLL_FAKE(L, pc, alpha, Kp);
	} 
	else if (DLLType == "DLLT_Ozeke") {
			MakeDLL_Ozeke(L, pc, alpha, Kp);
	} 
	else if (DLLType == "DLLT_Ozeke_ME") {
			MakeDLL_Ozeke_ME(L, pc, alpha, Kp);
	} 
	else {
			throw error_msg("DLL_ERROR", "Unknown DLL type");
	}
	arr.change_ind = clock();

}

/**
 * initialize arr to local diffusion coefficients with Df()
 */
void DiffusionCoefficient::MakeDLL_B(GridElement &L, GridElement &pc, GridElement &alpha, double Kp) {
	int il, im, ia;
	for (il = 0; il < L.size; il++) {
		for (im = 0; im < pc.size; im++) {
			for (ia = 0; ia < alpha.size; ia++) {
				arr[il][im][ia] = Df(L.arr[il][im][ia], Kp);
				//(*this)[il][im][ia] = 0.0;
			}
		}
	}
}

/**
 * initialize arr to local diffusion coefficients with Df() + \f$(3000 * 10^{0.506*Kp - 9.325} * L^5)\f$
 */
void DiffusionCoefficient::MakeDLL_FAKE(GridElement &L, GridElement &pc, GridElement &alpha, double Kp) {
	int il, im, ia;
	for (il = 0; il < L.size; il++) {
		for (im = 0; im < pc.size; im++) {
			for (ia = 0; ia < alpha.size; ia++) {
				arr[il][im][ia] = Df(L.arr[il][im][ia], Kp) + 3e3*pow(10.0, 0.506*Kp-9.325)*pow(L.arr[il][im][ia],5);;
			}
		}
	}
}

/** Other version of DLL making procedure
 * adds electrostatic and magnetostatic radial diffusion coefficients to arr
 */
void DiffusionCoefficient::MakeDLL_BE(GridElement &L, GridElement &pc, GridElement &alpha, double Kp) {
	int il, im, ia;

	for (il = 0; il < L.size; il++) {
		for (im = 0; im < pc.size; im++) {
			for (ia = 0; ia < alpha.size; ia++) {

				// electrostatic radial diffusion coefficient in 1/day
				arr[il][im][ia] = VF::Dfe(alpha.arr[il][im][ia],VF::Kfunc(pc.arr[il][im][ia]), L.arr[il][im][ia],Kp);
				//(*this)[il][im][ia] = 0.2*VF::Dfe(alpha[il][im][ia],VF::Kfunc(pc[il][im][ia]),L[il][im][ia],Kp);

				// Add magnetostatic radial diffusion coefficient in 1/day
				arr[il][im][ia] = arr[il][im][ia] + Df(L.arr[il][im][ia], Kp);

			}
		}
	}
}


/** Other version of DLL making procedure
 * initializes arr to local diffusion coefficients using Df(), values no lower than 0.01
 */
void DiffusionCoefficient::MakeDLL100(GridElement &L, GridElement &pc, GridElement &alpha, double Kp) {
	int il, im, ia;

	for (il = 0; il < L.size; il++) {
		for (im = 0; im < pc.size; im++) {
			for (ia = 0; ia < alpha.size; ia++) {
				arr[il][im][ia] = VF::max(Df(L.arr[il][im][ia], Kp), 1.0/100);
			}
		}
	}

}

/** Making DLL Ozeke
 *  WARNING! This code uses prepublication values! WARNING!
 *  Previous version of DLL:
 *  Ozeke, L. G., I. R. Mann, K. R. Murphy, I. J. Rae, D. K. Milling, S. R. Elkington, A. A. A. Chan, and H. J. Singer (2012),  
 *  ULF Wave Derived Radiation Belt Radial Diffusion Coefficients, 
 *  J. Geophys. Res., doi:10.1029/2011JA017463, in press. (accepted 23 February 2012)
 *  Magnetic component
 */
void DiffusionCoefficient::MakeDLL_Ozeke(GridElement &L, GridElement &pc, GridElement &alpha, double Kp) {
	int il, im, ia;
	for (il = 0; il < L.size; il++) {
		for (im = 0; im < pc.size; im++) {
			for (ia = 0; ia < alpha.size; ia++) {
				arr[il][im][ia] = Df_Ozeke(L.arr[il][im][ia], Kp); // this.arr
			}
		}
	}
}

/** Making DLL Ozeke
 *  WARNING! This code uses prepublication values! WARNING!
 * Previous version of DLL:
 *  Ozeke, L. G., I. R. Mann, K. R. Murphy, I. J. Rae, D. K. Milling, S. R. Elkington, A. A. A. Chan, and H. J. Singer (2012),  
 *  ULF Wave Derived Radiation Belt Radial Diffusion Coefficients, 
 *  J. Geophys. Res., doi:10.1029/2011JA017463, in press. (accepted 23 February 2012)
 *  Magnetic and Electric component
 */
void DiffusionCoefficient::MakeDLL_Ozeke_ME(GridElement &L, GridElement &pc, GridElement &alpha, double Kp) {
	int il, im, ia;
	for (il = 0; il < L.size; il++) {
		for (im = 0; im < pc.size; im++) {
			for (ia = 0; ia < alpha.size; ia++) {
				arr[il][im][ia] = Df_Ozeke(L.arr[il][im][ia], Kp) + Df_Ozeke_E(L.arr[il][im][ia], Kp); // this.arr
			}
		}
	}
}


/** Activation (enabling/disabling) and scaling of the diffusion coefficients in one group.
 * It loops thru the list of the diffusion coefficients, enable which needs to be enabled, disable, which needs to be disabled and scale all enabled
 */
bool DiffusionCoefficientsGroup::ActivateAndScale(double time, double Kp) {
	// The last version of activated stuff.
	// Activating and scaling diffusion coefficients for different wave types
	double scaleCoeff = 0;
	unsigned int Dxx_it;
	// \todo check, if we need to make any changes, otherwise - just exit
	// check if we really need to do any changes
	bool need_changes = false;
	for (Dxx_it = 0; Dxx_it < DxxList.size(); Dxx_it++) {
		if (time >= DxxList[Dxx_it].DxxParameters.time_start && time < DxxList[Dxx_it].DxxParameters.time_end) {
			// if inactive - mark as active
			if (!DxxList[Dxx_it].is_active) {
				need_changes = true;
				break;
			}
			// if this Dxx needs to be scaled - scale it
			if (DxxList[Dxx_it].useScale) {
				need_changes = true;
				break;
			}
		} else {
			// if active - mark as inactive
			if (DxxList[Dxx_it].is_active) {
				need_changes = true;
				break;
			}
		}
	}
	if (need_changes == false) return false;

	CurrentDxx.arr = 0;
	// Indicator, that we changed the coefficient (and now we probably need to recalculate the model matrix)
	CurrentDxx.arr.change_ind = clock();
	// loop thrue list of the diffusion coefficients
	for (Dxx_it = 0; Dxx_it < DxxList.size(); Dxx_it++) {
		// if it is time to work - activate
		if (time >= DxxList[Dxx_it].DxxParameters.time_start && time < DxxList[Dxx_it].DxxParameters.time_end) {
			// if inactive - mark as active
			if (!DxxList[Dxx_it].is_active) {
				DxxList[Dxx_it].is_active = true;
				Output::echo(2, "Activated %s.\n", DxxList[Dxx_it].arr.name.c_str());
			}
			// if this Dxx needs to be scaled - scale it
			if (DxxList[Dxx_it].useScale) {
				// getting scale coefficient
				scaleCoeff = DxxList[Dxx_it].Scale(Kp);
				Output::echo(2, "Bw scaled %s by %f.\n", DxxList[Dxx_it].arr.name.c_str(), scaleCoeff);
				// scalling and adding value to the actualized Dxx which is used in calculations
				CurrentDxx.arr = CurrentDxx.arr + DxxList[Dxx_it].arr * scaleCoeff;
			} else {
				CurrentDxx.arr = CurrentDxx.arr + DxxList[Dxx_it].arr;
			}
		} else {
			// if active - mark as inactive
			if (DxxList[Dxx_it].is_active) {
				DxxList[Dxx_it].is_active = false;
				Output::echo(2, "Deactivated %s.", DxxList[Dxx_it].arr.name.c_str());
			}
		}
	}
	return false;
}

/** Scaling procedure
 * \todo Upgrade scaling diffusion coefficients procedure with scaling according to external array
 */
double DiffusionCoefficient::Scale(double Kp) {
	double Bw2=0, Bw2_0=0;
	if (!this->useScale) return 1;
	if (DxxParameters.waveType == "WT_CHORUS_DAY" || DxxParameters.waveType == "WT_CHORUS_NIGHT") {
		// scaling for chorus waves
		if (DxxParameters.DxxKp <= 2.333)	Bw2_0 = 2.0*pow(10, 0.73 + 0.91 * DxxParameters.DxxKp);
		if (DxxParameters.DxxKp > 2.333)	Bw2_0 = 2.0*pow(10, 2.5 + 0.18 * DxxParameters.DxxKp);
		if (Kp <= 2.333)	Bw2 = 2.0*pow(10, 0.73 + 0.91 * Kp);// * 1.e-12 * 1.e8;
		if (Kp > 2.333)		Bw2 = 2.0*pow(10, 2.5 + 0.18 * Kp);// * 1.e-12 * 1.e8;
		//////////////////////////////////////////////////
		/* Ughm? Commented by Dmitriy
		if(DxxParameters.waveType == "WT_CHORUS_DAY") {
			return (Bw2/Bw2_0)*2.5;
		} else {
			return (Bw2/Bw2_0)/2.5;
		}*/
		///////////////////////////////////////////////////
		return (Bw2/Bw2_0);
	} else if (DxxParameters.waveType == "WT_HISS_PP" || DxxParameters.waveType == "WT_HISS_PLUME") {
		// scale hiss diffusion coefficients is a square of Bw scale
		// return pow(Kp/DxxParameters.DxxKp, 2);
		return pow(Kp/DxxParameters.DxxKp, 2);
	}
	else if (DxxParameters.waveType == "WT_HISS_REAL") {
		// Daa(Kp)=Daa(Kp=4)*10^{g(Kp)-g(Kp=4)}, where g(Kp)=0.2607*Kp - 0.01547*Kp*Kp.
		//return pow( (Kp*3 + 26 ) / (DxxParameters.DxxKp*3 + 26), 2);	
		return pow(10, 0.2607*Kp - 0.01547*Kp*Kp - 0.2607*DxxParameters.DxxKp + 0.01547*DxxParameters.DxxKp*DxxParameters.DxxKp);		
    } else if (DxxParameters.waveType == "WT_EMIC_PLUME" ) {
		// scale EMIC diffusion coefficients as a Bw to the forth power
		return pow(Kp/DxxParameters.DxxKp, 4);
//////////////////////////kckim////////////////////////////////////
	} else if (DxxParameters.waveType == "WT_LIGHTNING") {
		return pow(Kp/DxxParameters.DxxKp, 2);
	//
///////////////////////////////////////////////////////////////////////////

	} else {
		//throw error_msg("DXX_SCALING", "No scaling for wave type %s", DxxParameters.waveType.c_str());
		return 1;
	}
}

/**
 * Return inverse of calculated parameters(?)
 *
 * \param pangle - pitch angle
 * \param lambda - lambda
 * \param L - L value
 *
 * \todo L is not used in this function. delete it or use it.
 *
 * \f$ \sin^{-1}( \sqrt{ f1(\lambda)} * \sin(pangle) ) \f$
 * where f1 is f1()
 */
double Alpha_ne(double pangle, double lambda, double L) {
	return asin( sqrt( f1(lambda)) * sin(pangle) );
}

/**
 * \f$ \sqrt{4.0 - 3.0 * \frac {\cos^2(\lambda)}{\cos^6(\lambda)}} \f$
 *
 */
double f1(double lambda) {
	return sqrt(4.0 - 3.0 * pow(cos(lambda), 2))/(pow(cos(lambda), 6));
}

/**
 * \f$ \frac{0.311 * f1(\lambda)}{L^3} \f$
 * where f1 is f1()
 */
double B(double lambda, double L) {
	return f1(lambda) / pow(L, 3) * VC::B_0;
}

/**
 * \f$ x^6 + ((3x)\sin^4(\alpha) - 4\sin^4(\alpha)) \f$
 */
double func_tmp (double x, double Alpha) {
	return pow(x, 6) + (3.0 * pow(sin(Alpha), 4)) * x - 4.0 * pow(sin(Alpha), 4);
}

/**
 * \f$ y*((x-s)^2)* \frac{(x+s*\epsilon)^2}{x*g} \f$
 * where g is polynomial sequence derived from b, s, and epsilon
 *
 * \todo Function is never called anywhere. Most likely replaced by F_cap2. Should be removed.
 */
double F_cap(double x, double y, double b, double s, double epsilon, DxxParameters_structure DxxParameters) {
	double c1 = 2.0*s*(-1+epsilon);
	double c2 = 1.0-4.0*epsilon+pow(epsilon,2);
	double c3 = -s*(-1+epsilon)*(b+4.0*epsilon)/2.0;
	double c4 = epsilon*(b+epsilon) ;
	double g  = pow(x,4) + c1*pow(x,3) + c2*pow(x,2) + c3*x + c4;
	return y*(pow((x-s),2))*(pow((x+s*epsilon),2))/(x*g);
}

/**
 * returns upper to lower ratio, using eta1, eta2, and eta3 from DxxParameters
 * 
 * \returns \f$ \frac{up}{down} \f$
 *
 * where up = \f$ \frac{2(x+a)}{\beta * \mu} \f$ and
 *
 * where down = \f$ 2x - (\frac{1}{\alpha^*} * \frac{1}{x-s}) + \frac{\eta1*\epsilon}{x + s*\epsilon} + \frac{\eta2*\epsilon}{4x + s*\epsilon} + \frac{\eta3*\epsilon}{16x + s*\epsilon} +
 (\frac{x}{\alpha^*} * \frac{1}{(x-s)^2}) + \frac{\eta1*\epsilon}{(x+s*\epsilon)^2} + \frac{4*\epsilon*\eta2}{(4x+s*\epsilon)^2} + \frac{16*\epsilon*\eta3}{(16x+s*\epsilon)^2} \f$
 */
double F_cap2(double x, double y, double a, double beta, double mu, double s, double epsilon, double Alpha_star, DxxParameters_structure DxxParameters) {
	double up = 2.0*(x+a)/(beta*mu);
	double down = 2.0*x-1.0/Alpha_star*(1.0/(x-s)+DxxParameters.eta1*epsilon/(x+s*epsilon)+DxxParameters.eta2*epsilon/(4.0*x+s*epsilon)+DxxParameters.eta3*epsilon/(16.0*x+s*epsilon))+
		x*1.0/Alpha_star*(1.0/(pow((x-s), 2))+DxxParameters.eta1*epsilon/(pow((x+s*epsilon), 2))+4.0*epsilon*DxxParameters.eta2/(pow((4.0*x+s*epsilon), 2))+16.0*epsilon*DxxParameters.eta3/(pow((16.0*x+s*epsilon), 2)));
	return up/down;
}

/**
 * \f$ h * \frac{func(a, DxxParameters) + func(b, DxxParameters)}{2} + h*s \f$
 *
 * returns quad integral of func
 */
double quad1(double (*func)(double lambda, DxxParameters_structure DxxParameters), double a, double b, int M, DxxParameters_structure DxxParameters) {
	//  f -integrated function
	//  a and b upper and lower limits
	//  M number of subintegrals
	double h = (b - a)/M;
	double s = func(a, DxxParameters);
	double x;
	int k;
	for (k = 1; k <= (M-1); k++) {
		x = a + h*k;
		s = s + func(x, DxxParameters);
	}
	s = h * (func(a, DxxParameters) + func(b, DxxParameters))/2.0 + h*s;
	return s;
}

/**
 * Return derivative of func_tmp
 *
 * func_tmp = \f$ x^6+(3*\sin^4(\alpha)) * x-4*\sin^4(\alpha) \f$
 */
double Dxx_ba(double L1, double EMeV, double Alpha, double int_Dxx_loc(double lambda, DxxParameters_structure DxxParameters), DxxParameters_structure DxxParameters) {
	// Searching for zero of function func_tmp=x.^6+(3*sin(DxxParameters.Alpha_eq_tmp)^4.)*x-4*sin(DxxParameters.Alpha_eq_tmp)^4.;
	// near the point 0.5
    /// \todo L and EMeV aren't used again in this function. Do we need it?
	DxxParameters.L = L1;
	DxxParameters.EMeV = EMeV;
	DxxParameters.Alpha = Alpha;

	double lambda_m;
	//lambda_m = acos(sqrt(fzero(func_tmp, 0.5)))*0.99;
	lambda_m = acos(sqrt(bisection (func_tmp, DxxParameters.Alpha, 0, 1)))*DxxParameters.mirror_point_coeff;
	if (lambda_m > DxxParameters.lam_max*VC::pi/180) lambda_m = DxxParameters.lam_max*VC::pi/180.;
	// quad1 - calculating integral
	return (1.0/(1.38-0.32*(sin(DxxParameters.Alpha) + pow(sin(DxxParameters.Alpha),0.5))) * quad1(int_Dxx_loc, DxxParameters.lam_min*VC::pi/180., lambda_m, DxxParameters.nint, DxxParameters));
}

/**
 * calls function Dxx_local() to get Daa
 *
 * returns \f$ Daa * \cos(Alpha\_ne(\alpha, \lambda, L)) * \frac{\cos^7(\lambda)}{\cos^2(\alpha)} \f$
 *
 * where Alpha_ne is Alpha_ne()
 */
double int_Daa_loc(double lambda, DxxParameters_structure DxxParameters) {
	//double Daa = Daa_local(lambda);
	double Daa = Dxx_local(lambda, Daa_root, DxxParameters);
	if (DxxParameters.Alpha*180/VC::pi < 0.001) Daa = 0;
	return Daa * cos(Alpha_ne(DxxParameters.Alpha, lambda, DxxParameters.L)) * pow(cos(lambda), 7) /(pow(cos(DxxParameters.Alpha), 2));
}

/**
 * calls function Dxx_local() to get Dpp
 *
 * returns \f$ Dpp * \frac{ \cos(\lambda) * \sqrt{1 + 3*\sin^2(\lambda)} }{\cos(Alpha\_ne(\alpha, \lambda, L))} \f$
 *
 * where Alpha_ne is Alpha_ne()
 */
double int_Dpp_loc(double lambda, DxxParameters_structure DxxParameters) {
	//double Dpp = Dpp_local(lambda);
	double Dpp = Dxx_local(lambda, Dpp_root, DxxParameters);
	if (DxxParameters.Alpha*180/VC::pi < 0.001) Dpp = 0;
	return Dpp * (cos(lambda) * pow((1 + 3*pow(sin(lambda), 2)), 0.5)) / cos(Alpha_ne(DxxParameters.Alpha, lambda, DxxParameters.L));
}

/**
 * calls function Dxx_local() to get Dpa
 *
 * returns \f$ Dpa * \frac{ \sin(Alpha\_ne(\alpha, \lambda, L)) * \cos^7(\lambda) }{\cos(\alpha) * \sin(\alpha)} \f$
 *
 * where Alpha_ne is Alpha_ne()
 */
double int_Dpa_loc(double lambda, DxxParameters_structure DxxParameters) {
	//double Dpa = Dpa_local(lambda);
	double Dpa = Dxx_local(lambda, Dpa_root, DxxParameters);
	if (DxxParameters.Alpha*180/VC::pi < 0.001) Dpa = 0;
	return Dpa * sin(Alpha_ne(DxxParameters.Alpha, lambda, DxxParameters.L)) * pow(cos(lambda), 7) / (cos(DxxParameters.Alpha) * sin(DxxParameters.Alpha));
}

/**
 * Obtain local derivative for a certain diffusion coefficient
 *
 * \param lambda - lambda
 * \param Dxx_root() - function parameter that returns partial derivative of a diffusion coefficient
 * \param DxxParameters - Dxx parameters struct to obtain grid element parameters
 *
 * \todo Get someone with physics background to document the Greek letter variables.
 */
double Dxx_local(double lambda, double Dxx_root(double Omega_e, double x, double mu, double su, double y, double beta, double a, double b, double Alpha_star, double s, double epsilon, double d_x, double x_m, double R, DxxParameters_structure DxxParameters), DxxParameters_structure DxxParameters ) {

	double Omega_e			= VC::qe * B(lambda, DxxParameters.L) / VC::m / VC::c_cgs;
	double Omega_e_eq		= VC::qe * B(0, DxxParameters.L) / VC::m / VC::c_cgs;           // Omega_e_eq = B_0/(L^3)
	double N_0				= 124.0 * pow((3.0 / DxxParameters.L), 4);
    double Omega_p			= sqrt(4.0 * VC::pi * N_0 * pow(VC::qe, 2) / VC::m);
    double Dxx_f            = DxxParameters.f; // to hold previous value
    DxxParameters.f         = Omega_p/Omega_e_eq;
	//double Omega_p;
	//if (DxxParameters.f > 0) {
	//double N_0;
	double LT;
    
    // rearranged if statements to be more concise
	if (DxxParameters.numberDensity == "ND_CONSTANT") {
			Omega_p				= Dxx_f * Omega_e_eq;
	} else if (DxxParameters.numberDensity == "ND_CHORUS") {
			// ignore, since parameters already set above
//			Output::echo(2, "%e \n", DxxParameters.f);
	} else if (DxxParameters.numberDensity == "ND_CHORUS_NIGHT" || DxxParameters.numberDensity == "ND_CHORUS_DAY") {
            if (DxxParameters.numberDensity == "ND_CHORUS_NIGHT")
                LT = 3;
            else
                LT = 9;
        
			N_0	+= 36.0*pow((3.0/DxxParameters.L),3.5) * cos((LT-(7.7*pow((3.0/DxxParameters.L),2.0)+12))*VC::pi/12);
//			if (DxxParameters.L >= 4.5) { Output::echo(2, "%e \n", DxxParameters.f); }
	} else if (DxxParameters.numberDensity == "ND_HISS" || DxxParameters.numberDensity == "ND_PL_CA_STARKS") {
			N_0	= pow(10, (-0.3145*DxxParameters.L + 3.9043));
            if (DxxParameters.numberDensity == "ND_PL_CA_STARKS" && DxxParameters.L <= 2)
            {
                double N1, N2, L1, L2, A, B;
                N1=10-6.;N2=8.3-6.;L1=1.2;L2=5.9;
                A=(N2-N1)/(L2-L1); B=N2-A*L2;
                N_0					= pow(10, A*DxxParameters.L+B);
            }
//			Output::echo(2, "%e \n", DxxParameters.f);
	} else if (DxxParameters.numberDensity == "ND_HISS2") {
			N_0	= 1e3*pow(4.0/DxxParameters.L, 4);
//			Output::echo(2, "%e \n", DxxParameters.f);
	} else {
		throw error_msg("DXX_DENSITY_ERROR", "Unknown number density type: %s\n", DxxParameters.numberDensity.c_str());
	}


	double epsilon			= 1.0/1837;
	double beta				= (pow((DxxParameters.EMeV * (DxxParameters.EMeV + 2)), 0.5)) / (DxxParameters.EMeV + 1);
	double mu				= cos(Alpha_ne(DxxParameters.Alpha, lambda, DxxParameters.L));
    
    // sin version of mu, returns sqrt( f1(lambda)) * sin(alpha)
	double su				= sin(Alpha_ne(DxxParameters.Alpha, lambda, DxxParameters.L));

	double lam_sp			= -1; // -1 corresponds to electrons
	double gamma			= DxxParameters.EMeV + 1;
	double a				= DxxParameters.s * lam_sp / gamma;

	double Alpha_star		= pow((Omega_e / Omega_p), 2);
	double b				= (1 + epsilon)/Alpha_star;

	//double d_omega			= d_omega_per * Omega_e_eq;
	//double Omega_m			= Omega_m_per * Omega_e_eq;
	double d_omega;
	double Omega_m;
	double Omega_1;     // lower cutoff
	double Omega_2;     // upper cutoff

    // to determine partial or absolute omega type
	if (DxxParameters.Omega_mType == "OT_PARTIAL") {
			Omega_m			= DxxParameters.Omega_m * Omega_e_eq;
			d_omega			= DxxParameters.d_omega * Omega_e_eq ;
			Omega_1			= DxxParameters.omega_lc * Omega_e_eq;
			Omega_2			= DxxParameters.omega_uc * Omega_e_eq;
	} else if (DxxParameters.Omega_mType == "OT_ABSOLUTE") {
			Omega_m			= DxxParameters.Omega_m * 2 * VC::pi;
			d_omega			= DxxParameters.d_omega * 2 * VC::pi;
			Omega_1			= DxxParameters.omega_lc * 2 * VC::pi;
			Omega_2			= DxxParameters.omega_uc * 2 * VC::pi;
	} else {
			throw error_msg("DXX_OMEGA_M_TYPE", "Unknown Omega_m type: ", DxxParameters.Omega_mType.c_str());
	}

	//	if (DxxParameters.Omega_mType == "partial") {
	//	} else if (DxxParameters.Omega_mType == "constant") {
	//	} else {
	//	}

	//double R				= pow((DxxParameters.dB / B(lambda, DxxParameters.L)), 2);
	double Bw;
	if (DxxParameters.BwFromLambda) {
		Bw = pow(10, 0.75 + 0.04*(lambda*180/VC::pi))*1.e-8 * DxxParameters.Bw; // should be gauss...
	} else {
		Bw = DxxParameters.Bw;
	}
	double R				= pow((Bw / B(lambda, DxxParameters.L)), 2);

	double x_m=0;
	double d_x=0;
	double x_1=0;
	double x_2=0;
    // determine particle type
	if (DxxParameters.particle == "PT_ELECTRON") {
		x_m				= Omega_m / Omega_e;
		d_x				= d_omega / Omega_e;
		x_1				= Omega_1 / Omega_e;
		x_2				= Omega_2 / Omega_e;
	} else if (DxxParameters.particle == "PT_PROTON") {
		double cfm = 16; // cubic feet per minute(?)
		x_m				= Omega_m / Omega_e * epsilon / cfm;
		d_x				= d_omega / Omega_e * epsilon / cfm;
		x_1				= Omega_1 / Omega_e * epsilon / cfm;
		x_2				= Omega_2 / Omega_e * epsilon / cfm;
	}

	std::vector<double> xx = rrouts(x_1, x_2, DxxParameters.eta1, DxxParameters.eta2, DxxParameters.eta3, epsilon, beta, mu, Alpha_star, a, DxxParameters);
	//
	double x, y;
	double Dxx = 0, Dxx_tmp = 0;
	for (unsigned int i = 1; i <= xx.size();i++) {
		x = xx[i-1];
		y = (x + a) / (beta * mu);
		Dxx_tmp = Dxx_root(Omega_e, x, mu, su, y, beta, a, b, Alpha_star, DxxParameters.s, epsilon, d_x, x_m, R, DxxParameters);
		Dxx = Dxx_tmp + Dxx;
	}
	if (xx.size() == 0)  {
		Dxx = 0;
	}

	return Dxx;
}

/**
 * Obtain roots for Daa coefficient
 *
 * \returns \f$ \frac{\pi}{2} * \frac{1}{\nu} * \Omega_e * \frac{1}{(EMeV + 1)^2} *
 \frac{R * (1 - \frac{x * \mu}{(y * \beta)})^2 * \left|fcap\right|} {
 (d_x * \left| \beta * \mu - fcap \right|} *
 e^{-(\frac{x - x_m}{d_x})^2} \f$
 *
 * where fcap = F_cap2()
 */
double Daa_root(double Omega_e, double x, double mu, double su, double y, double beta, double a, double b, double Alpha_star, double s, double epsilon, double d_x, double x_m, double R, DxxParameters_structure DxxParameters) {
    double fcap = F_cap2(x,y,a,beta,mu,s,epsilon,Alpha_star, DxxParameters);
	return		(VC::pi/2) * (1.0/DxxParameters.nu) * Omega_e * (1.0/(pow((DxxParameters.EMeV + 1), 2))) *
		((R * pow((1 - x * mu / (y * beta)), 2) * fabs(fcap))/
		(d_x * fabs(beta * mu - fcap)))*
		exp( - (pow(((x - x_m) / d_x), 2)));
}

/**
 * Obtain roots for Dpa coefficient
 *
 * \returns \f$ -\frac{\pi}{2} * \frac{1}{\nu} * \Omega_e * \frac{su}{\beta} * \frac{1}{(EMeV + 1)^2} *
 \frac{R * \frac{x}{y} * (1 - \frac{x * \mu}{y * \beta}) * \left|fcap\right|} {
 d_x * \left| \beta * \mu - fcap \right|} *
 e^{-(\frac{x - x_m}{d_x})^2} \f$
 *
 * where fcap = F_cap2()
 */
double Dpa_root(double Omega_e, double x, double mu, double su, double y, double beta, double a, double b, double Alpha_star, double s, double epsilon, double d_x, double x_m, double R, DxxParameters_structure DxxParameters) {
    double fcap = F_cap2(x,y,a,beta,mu,s,epsilon,Alpha_star,DxxParameters);
	return -(VC::pi/2)*(1.0/DxxParameters.nu)* Omega_e*su/beta*(1.0/(pow((DxxParameters.EMeV+1), 2)))*
		((R*(1-x/y*mu/beta)*(x/y)* fabs(fcap))/
		(d_x * fabs(beta * mu - fcap)))*
		exp( -(pow(((x - x_m) / d_x), 2)));
}

/**
 * Obtain roots for Dpp coefficient
 *
 * \returns \f$ \frac{\pi}{2} * \frac{1}{\nu} * \frac{1}{\beta^2} * (1 - \mu^2) * \Omega_e * \frac{1}{(EMeV + 1)^2} *
 \frac{R * (\frac{x}{y})^2 * \left|fcap\right|} {
 d_x * \left| \beta * \mu - fcap \right|} *
 e^{-(\frac{x - x_m}{d_x})^2} \f$
 *
 * where fcap = F_cap2()
 */
double Dpp_root(double Omega_e, double x, double mu, double su, double y, double beta, double a, double b, double Alpha_star, double s, double epsilon, double d_x, double x_m, double R, DxxParameters_structure DxxParameters) {
    double fcap = F_cap2(x,y,a,beta,mu,s,epsilon,Alpha_star,DxxParameters);
	return	(VC::pi/2) * (1.0/DxxParameters.nu) * (1./pow(beta, 2)) * (1-pow(mu, 2)) * Omega_e * (1.0/(pow((DxxParameters.EMeV + 1), 2))) *
		((R * (pow(x/y, 2)) * fabs(fcap))/
		(d_x*fabs(beta*mu - fcap)))*
		exp(-(pow(((x-x_m)/d_x),2)));
}


//std::vector<double> rrouts(double Omega_e_eq, double Omega_e, double eta1, double eta2, double eta3, double epsilon, double beta, double mu, double Alpha_star, double a, DxxParameters_structure DxxParameters) {
/**
 * rrouts implements roots() and get_quads() from rroots.cpp file to obtain real roots of polynomial
 */
std::vector<double> rrouts(double x_1, double x_2, double eta1, double eta2, double eta3, double epsilon, double beta, double mu, double Alpha_star, double a, DxxParameters_structure DxxParameters) {

	//double Omega_1			= DxxParameters.omega_lc * Omega_e_eq * epsilon / DxxParameters.cfm;
	//double Omega_2			= DxxParameters.omega_uc * Omega_e_eq * epsilon / DxxParameters.cfm;
	//double x_1				= Omega_1 / Omega_e;
	//double x_2				= Omega_2 / Omega_e;

    // bunch of coefficients of different orders
	double c0 = -DxxParameters.s;
	double c1 = DxxParameters.s * epsilon;
	double c2 = 1.0/4. * DxxParameters.s * epsilon;
	double c3 = 1.0/16. * DxxParameters.s * epsilon;

	double d0 = 1;
	double d1 = eta1*epsilon;
	double d2 = eta2/4.*epsilon;
	double d3 = eta3/16.*epsilon;

	double g3 = d0 + d1 + d2 + d3;
	double g2 = d0*(c1+c2+c3) + d1*(c0+c2+c3) + d2*(c0+c1+c3) + d3*(c0+c1+c2);
	double g1 = d0*(c1*c2 + c1*c3 + c2*c3) + d1*(c0*c2 + c0*c3 + c2*c3) + d2*(c0*c1 + c0*c3 + c1*c3) + d3*(c0*c1 + c0*c2 + c1*c2);
	double g0 = d0*c1*c2*c3 + d1*c0*c2*c3 + d2*c0*c1*c3 + d3*c0*c1*c2;

	double h4 = 1.;
	double h3 = c0 + c1 + c2 + c3;
	double h2 = c0*c1 + c0*c2 + c0*c3 + c1*c2 + c1*c3 + c2*c3;
	double h1 = c0*c1*c2 + c0*c1*c3 + c0*c2*c3 + c1*c2*c3;
	double h0 = c0*c1*c2*c3;

    // array of calculated coefficients of polynomial
	double A[7];
	A[0] = (pow(beta,2)*pow(mu,2)-1.)*h4;
	A[1] = (pow(beta,2)*pow(mu,2)-1.)*h3-2*a*h4;
	A[2] = (pow(beta,2)*pow(mu,2)-1.)*h2-2*a*h3-pow(beta,2)*pow(mu,2)*1./Alpha_star*g3-pow(a,2)*h4;
	A[3] = (pow(beta,2)*pow(mu,2)-1.)*h1-2*a*h2-pow(beta,2)*pow(mu,2)*1./Alpha_star*g2-pow(a,2)*h3;
	A[4] = (pow(beta,2)*pow(mu,2)-1.)*h0-2*a*h1-pow(beta,2)*pow(mu,2)*1./Alpha_star*g1-pow(a,2)*h2;
	A[5] = -2*a*h0-pow(beta,2)*pow(mu,2)*1./Alpha_star*g0-pow(a,2)*h1;
	A[6] = -pow(a,2)*h0;

	//a6=(beta^2*mu^2-1.)*h4;
	//a5=(beta^2*mu^2-1.)*h3-2*a*h4;
	//a4=(beta^2*mu^2-1.)*h2-2*a*h3-beta^2*mu^2*1./Alpha_star*g3-a^2*h4;
	//a3=(beta^2*mu^2-1.)*h1-2*a*h2-beta^2*mu^2*1./Alpha_star*g2-a^2*h3;
	//a2=(beta^2*mu^2-1.)*h0-2*a*h1-beta^2*mu^2*1./Alpha_star*g1-a^2*h2;
	//a1=-2*a*h0-beta^2*mu^2*1./Alpha_star*g0-a^2*h1;
	//a0=-a^2*h0;

	std::vector<double> xx;
    // r is to store quadratic factors, wr is real roots, wi is imaginary roots, quad used for quadratic calculation
	double r[7], wr[6],wi[6],quad[2];
	int n, numr; // number of roots
	n=6;
	// initialize estimate for 1st root pair
	quad[0] = 2.71828e-1;
	quad[1] = 3.14159e-1;
	// get roots
	get_quads(A, n, quad, r);
	numr = roots(r, n, wr, wi);

	for (int i = 0; i < numr; i++) {
		if (fabs(wi[i]) < double_zero && wr[i] < x_2 && wr[i] > x_1) {
			xx.push_back(wr[i]);
		}
	}

	return xx;

}


/** Radial Diffusion coefficient computed following [Brautigam and Albet , 2000]
 * \returns \f$ 10^{0.506*Kp - 9.325} * L^{10} \f$
 */
double Df(double L, double Kp) {
	return pow(10, 0.506*Kp-9.325)*pow(L,10);
}
/**
 * Radial Diffusion coefficient computed Ozeke,
 * Ozeke et al, 2013
 * Analytic Expressions for ULL Wave Radiation Belt Radial Diffusion Coefficient //  (before publishing)
 * DLL_B \f$ days^{-1} \f$
 *
 * \returns \f$ 6.62*10^{-13} * L^8 * 10^{L^2 * -0.0327 + L * 0.625 - Kp^2 * 0.0108 + Kp * 0.499} \f$
 */
double Df_Ozeke(double L, double Kp) {

	return 6.62e-13 * pow(L,8) * pow(10, (pow(L,2) * -0.0327 + L * 0.625 - pow(Kp,2) * 0.0108 + Kp * 0.499) );
}

/// DLL_E
/// \returns \f$ 2.16*10^{-8} * L^6 * 10^{0.217 * L + 0.461 * Kp} \f$
double Df_Ozeke_E(double L, double Kp) {

	return  2.16e-8 * pow(L,6) * pow(10,( 0.217 * L + 0.461 * Kp )); 
}



/* DEPRECATED
/// Radial Diffusion coefficient computed Ozeke, 2013 library (odc_DLL_Ozeke2013)
//  http://irbem.svn.sourceforge.net/viewvc/irbem/extras/opendc/
double Df_Ozeke(double L, double Kp) {

	return 1.94e-6 * pow(L,8) * pow(10, pow(L,2) * 0.112 + L * Kp * 0.0824 - L * 1.32 + pow(Kp,2) * 0.0291 - Kp * 0.270 );
}

double Df_Ozeke_E(double L, double Kp) {

	return  5.75e-9 * pow(L,6) * pow(10,( 0.208 * L + 0.457 * Kp )); 
}
*/

/**
 * Create all diffusion coefficients based on parameters
 * 
 * \param &DLL - Diffusion coefficients DLL
 * \param &Daa - see Main.cpp for diffusion coefficient definitions
 * \param &Dpcpc
 * \param &Dpca
 * \param &DaaLpp
 * \param &DpcpcLpp
 * \param &DpcaLpp
 * \param &parameters - parameters structure
 * \param &radialDiffusionGrid - grid to copy from
 * \param &localDiffusionsGrid - grid to copy coefficients into
 */
void CreateAllDiffusionCoefficients(DiffusionCoefficient &DLL,
		DiffusionCoefficientsGroup &Daa, DiffusionCoefficientsGroup &Dpcpc, DiffusionCoefficientsGroup &Dpca,
		DiffusionCoefficientsGroup &DaaLpp, DiffusionCoefficientsGroup &DpcpcLpp, DiffusionCoefficientsGroup &DpcaLpp,
		Parameters_structure &parameters, Grid &radialDiffusionGrid, Grid &localDiffusionsGrid) {

	// Diffusion coefficient group (Lists of similar diffusion coefficients, like Daa or Dpcpc...)
	DLL.arr.AllocateMemory(radialDiffusionGrid.L.size, radialDiffusionGrid.pc.size, radialDiffusionGrid.alpha.size);
	Daa.CurrentDxx.arr.AllocateMemory(radialDiffusionGrid.L.size, radialDiffusionGrid.pc.size, radialDiffusionGrid.alpha.size);
	DaaLpp.CurrentDxx.arr.AllocateMemory(radialDiffusionGrid.L.size, radialDiffusionGrid.pc.size, radialDiffusionGrid.alpha.size);
	Dpcpc.CurrentDxx.arr.AllocateMemory(radialDiffusionGrid.L.size, radialDiffusionGrid.pc.size, radialDiffusionGrid.alpha.size);
	DpcpcLpp.CurrentDxx.arr.AllocateMemory(radialDiffusionGrid.L.size, radialDiffusionGrid.pc.size, radialDiffusionGrid.alpha.size);
	Dpca.CurrentDxx.arr.AllocateMemory(radialDiffusionGrid.L.size, radialDiffusionGrid.pc.size, radialDiffusionGrid.alpha.size);
	DpcaLpp.CurrentDxx.arr.AllocateMemory(radialDiffusionGrid.L.size, radialDiffusionGrid.pc.size, radialDiffusionGrid.alpha.size);

	// Make it zero
	DLL.arr = 0;
	Daa.CurrentDxx.arr = 0;
	DaaLpp.CurrentDxx.arr = 0;
	Dpcpc.CurrentDxx.arr = 0;
	DpcpcLpp.CurrentDxx.arr = 0;
	Dpca.CurrentDxx.arr = 0;
	DpcaLpp.CurrentDxx.arr = 0;

	// Temporary diffusion coefficient
	DiffusionCoefficient Dxx_tmp;
	// Create the diffusion coefficients for each of the Dxx parameters files
	unsigned int Dxx_it;
	for (Dxx_it = 0; Dxx_it < parameters.DxxParametersList.size(); Dxx_it++) {

		Dxx_tmp.Create_Dxx(parameters.DxxParametersList[Dxx_it], localDiffusionsGrid);

		// Add the created coefficient to an appropriate group
		if (parameters.useAlphaDiffusion) {
			if (Dxx_tmp.type == "DCT_Daa")
				Daa.DxxList.push_back(Dxx_tmp);
			else if (Dxx_tmp.type == "DCT_DaaLpp")
				DaaLpp.DxxList.push_back(Dxx_tmp);
		}

		if (parameters.useEnergyDiffusion) {
			if (Dxx_tmp.type == "DCT_Dpp" || Dxx_tmp.type == "DCT_Dpcpc")
				Dpcpc.DxxList.push_back(Dxx_tmp);
			else if (Dxx_tmp.type == "DCT_DppLpp" || Dxx_tmp.type == "DCT_DpcpcLpp")
				DpcpcLpp.DxxList.push_back(Dxx_tmp);
		}

		if (parameters.useEnergyAlphaMixedTerms) {
			if (Dxx_tmp.type == "DCT_Dpca" || Dxx_tmp.type == "DCT_Dpa")
				Dpca.DxxList.push_back(Dxx_tmp);
			else if (Dxx_tmp.type == "DCT_DpcaLpp" || Dxx_tmp.type == "DCT_DpaLpp")
				DpcaLpp.DxxList.push_back(Dxx_tmp);

		}
	}
	// Check if Daa*Dpcpc > Dpca^2 for each Dpca
	if (parameters.useEnergyAlphaMixedTerms) {
		int Daa_it, Dpcpc_it, Dpca_it;
		// above Lpp
		for (Dpca_it = 0; Dpca_it < Dpca.DxxList.size(); Dpca_it++) {
			// For each Dpca...
			// Search for Daa with the same wave-type
			for (Daa_it = 0; Daa_it < Daa.DxxList.size(); Daa_it++) {
				if (Daa.DxxList[Daa_it].DxxParameters.waveType == Dpca.DxxList[Dpca_it].DxxParameters.waveType)
					break;
			}
			// Search for Dpp with the same wave-type
			for (Dpcpc_it = 0; Dpcpc_it < Dpcpc.DxxList.size(); Dpcpc_it++) {
				if (Dpcpc.DxxList[Dpcpc_it].DxxParameters.waveType == Dpca.DxxList[Dpca_it].DxxParameters.waveType)
					break;
			}
			// Check if Daa*Dpcpc > Dpca^2 for each Dpca
			int iL, im, ia;
			for (iL = 0; iL < localDiffusionsGrid.L.size; iL++)
				for (im = 0; im < localDiffusionsGrid.pc.size; im++)
					for (ia = 0; ia < localDiffusionsGrid.alpha.size; ia++)
						if (Daa.DxxList[Daa_it].arr[iL][im][ia] * Dpcpc.DxxList[Dpcpc_it].arr[iL][im][ia] * 0.99 <= Dpca.DxxList[Dpca_it].arr[iL][im][ia] * Dpca.DxxList[Dpca_it].arr[iL][im][ia])
							Output::echo(0, "WARNING: Diffusion coefficient error. %s: Dpp * Daa * 0.99 <= Dpa^2. \nAt (%d,%d,%d): L=%f, epc=%f, alpha=%f; Dpp=%e, Daa=%e, Dpa=%e\n",
									Dpca.DxxList[Dpca_it].DxxParameters.waveType.c_str(),
									iL, im, ia,
									localDiffusionsGrid.L.arr[iL][im][ia], localDiffusionsGrid.epc.arr[iL][im][ia], localDiffusionsGrid.alpha.arr[iL][im][ia],
									Dpcpc.DxxList[Dpcpc_it].arr[iL][im][ia], Daa.DxxList[Daa_it].arr[iL][im][ia], Dpca.DxxList[Dpca_it].arr[iL][im][ia]);
		}

		// Below Lpp
		for (Dpca_it = 0; Dpca_it < DpcaLpp.DxxList.size(); Dpca_it++) {
			// For each Dpca...
			// Search for Daa with the same wave-type
			for (Daa_it = 0; Daa_it < DaaLpp.DxxList.size(); Daa_it++) {
				if (DaaLpp.DxxList[Daa_it].DxxParameters.waveType == DpcaLpp.DxxList[Dpca_it].DxxParameters.waveType)
					break;
			}
			// Search for Dpp with the same wave-type
			for (Dpcpc_it = 0; Dpcpc_it < DpcpcLpp.DxxList.size(); Dpcpc_it++) {
				if (DpcpcLpp.DxxList[Dpcpc_it].DxxParameters.waveType == DpcaLpp.DxxList[Dpca_it].DxxParameters.waveType)
					break;
			}
			// Check if Daa*Dpcpc > Dpca^2 for each Dpca
			int iL, im, ia;
			for (iL = 0; iL < localDiffusionsGrid.L.size; iL++)
				for (im = 0; im < localDiffusionsGrid.pc.size; im++)
					for (ia = 0; ia < localDiffusionsGrid.alpha.size; ia++)
						if (DaaLpp.DxxList[Daa_it].arr[iL][im][ia] * DpcpcLpp.DxxList[Dpcpc_it].arr[iL][im][ia] * 0.99 <= DpcaLpp.DxxList[Dpca_it].arr[iL][im][ia] * DpcaLpp.DxxList[Dpca_it].arr[iL][im][ia])
							Output::echo(0, "WARNING: Diffusion coefficient error. %s: Dpp * Daa * 0.99 <= Dpa^2. \nAt (%d,%d,%d): L=%f, epc=%f, alpha=%f; Dpp=%e, Daa=%e, Dpa=%e\n",
									DpcaLpp.DxxList[Dpca_it].DxxParameters.waveType.c_str(),
									iL, im, ia,
									localDiffusionsGrid.L.arr[iL][im][ia], localDiffusionsGrid.epc.arr[iL][im][ia], localDiffusionsGrid.alpha.arr[iL][im][ia],
									DpcpcLpp.DxxList[Dpcpc_it].arr[iL][im][ia], DaaLpp.DxxList[Daa_it].arr[iL][im][ia], DpcaLpp.DxxList[Dpca_it].arr[iL][im][ia]);
		}

	}

	// calculating DLL If we need to do Radial diffusion
	if (parameters.useRadialDiffusion) {
		DLL.MakeDLL(radialDiffusionGrid.L, radialDiffusionGrid.pc, radialDiffusionGrid.alpha, parameters.Kp[0], parameters.DLLType);
//			DLL.writeToFile("./Debug/DLL.plt", radialDiffusionGrid.L, radialDiffusionGrid.epc, radialDiffusionGrid.alpha);
	}

}



void Output1DHeaders(ofstream &output1D,
		DiffusionCoefficientsGroup &Daa, DiffusionCoefficientsGroup &Dpcpc, DiffusionCoefficientsGroup &Dpca,
		DiffusionCoefficientsGroup &DaaLpp, DiffusionCoefficientsGroup &DpcpcLpp, DiffusionCoefficientsGroup &DpcaLpp) {

	// Title of file (first line)
	output1D << "Variables = \"time\", \"Kp\", \"Boundary_fluxes\", \"Lpp\", \"tau\" ";

	// Writing headers for all diffusion coefficients scaling coefficients
	unsigned int Dxx_it;
	for (Dxx_it = 0; Dxx_it < Daa.DxxList.size(); Dxx_it++) {
		output1D << "\"" << Daa.DxxList[Dxx_it].arr.name << "\"" << ", ";
	}
	for (Dxx_it = 0; Dxx_it < Dpcpc.DxxList.size(); Dxx_it++) {
		output1D << "\"" << Dpcpc.DxxList[Dxx_it].arr.name << "\"" << ", ";
	}
	for (Dxx_it = 0; Dxx_it < Dpca.DxxList.size(); Dxx_it++) {
		output1D << "\"" << Dpca.DxxList[Dxx_it].arr.name << "\"" << ", ";
	}

	for (Dxx_it = 0; Dxx_it < DaaLpp.DxxList.size(); Dxx_it++) {
		output1D << "\"" << DaaLpp.DxxList[Dxx_it].arr.name << "\"" << ", ";
	}
	for (Dxx_it = 0; Dxx_it < DpcpcLpp.DxxList.size(); Dxx_it++) {
		output1D << "\"" << DpcpcLpp.DxxList[Dxx_it].arr.name << "\"" << ", ";
	}
	for (Dxx_it = 0; Dxx_it < DpcaLpp.DxxList.size(); Dxx_it++) {
		output1D << "\"" << DpcaLpp.DxxList[Dxx_it].arr.name << "\"" << ", ";
	}

	output1D << endl;
	output1D << "ZONE T=\"1d-output\" " << endl;
}

void Output1DValues(ofstream &output1D,
		DiffusionCoefficientsGroup &Daa, DiffusionCoefficientsGroup &Dpcpc, DiffusionCoefficientsGroup &Dpca,
		DiffusionCoefficientsGroup &DaaLpp, DiffusionCoefficientsGroup &DpcpcLpp, DiffusionCoefficientsGroup &DpcaLpp,
		double &time, Parameters_structure &parameters, int iteration) {
	
	// Title of file (first line)
	output1D << time << "\t" << parameters.Kp[iteration] << "\t" << parameters.Bf[iteration] << "\t" << parameters.Lpp[iteration] << "\t" << parameters.tau[iteration] << "\t"; //parameters.dBw2[iteration] << "\t" <<

	unsigned int Dxx_it;
	for (Dxx_it = 0; Dxx_it < Daa.DxxList.size(); Dxx_it++) {
		output1D << Daa.DxxList[Dxx_it].Scale(parameters.Kp[iteration]) << "\t";
	}
	for (Dxx_it = 0; Dxx_it < Dpcpc.DxxList.size(); Dxx_it++) {
		output1D << Dpcpc.DxxList[Dxx_it].Scale(parameters.Kp[iteration]) << "\t";
	}
	for (Dxx_it = 0; Dxx_it < Dpca.DxxList.size(); Dxx_it++) {
		output1D << Dpca.DxxList[Dxx_it].Scale(parameters.Kp[iteration]) << "\t";
	}

	for (Dxx_it = 0; Dxx_it < DaaLpp.DxxList.size(); Dxx_it++) {
		output1D << DaaLpp.DxxList[Dxx_it].Scale(parameters.Kp[iteration]) << "\t";
	}
	for (Dxx_it = 0; Dxx_it < DpcpcLpp.DxxList.size(); Dxx_it++) {
		output1D << DpcpcLpp.DxxList[Dxx_it].Scale(parameters.Kp[iteration]) << "\t";
	}
	for (Dxx_it = 0; Dxx_it < DpcaLpp.DxxList.size(); Dxx_it++) {
		output1D << DpcaLpp.DxxList[Dxx_it].Scale(parameters.Kp[iteration]) << "\t";
	}

	output1D << endl;
}
