/**
 * Diffusion coefficients calculation, loading, activating, scaling etc. Header file.
 *
 * \file DiffusionCoefficient.h
 *
 * \author Developed by Yuri Shprits
 */

#ifndef DiffusionCoefficient_H
#define DiffusionCoefficient_H

#include <string>
#include <vector>
#include "../Matrix/Matrix.h"
#include "../Grid/Grid.h"

using namespace std;

/** \brief Holds diffusion coefficient matrix and routines to load and calculate it.
 *
 * It gets DiffusionCoefficientParameters_structure as an input with all waves parameters.
 */
class DiffusionCoefficient {
public:
	Matrix3D<double> arr; ///< array of diffusion coefficients

	/// flag, indicated that the initialization was passed
	bool Dxx_initialized;

	/// Derivative parameters
	DxxParameters_structure DxxParameters;

	// Variables

	/// Type of the diffusion coefficient: Daa, Dpp, Dpa etc... Described in types.h file as an enumeration.
	string type;

	/// flag, if is enabled right now
	bool is_active;

	/// flag, if scale is applyable
	bool useScale;

	/// Function, that scale diffusion coefficients
	double Scale(double Kp);

	/// Create diffusion coefficients
	void Create_Dxx(DxxParameters_structure DxxParameters, Grid &grid);

	/// Loads diffusion coefficients
	/// Call other functions depends on filetype - file type
	/// \param &L - Grid element - L
	/// \param &pc - Grid element - pc
	/// \param &alpha - One more grid element - alpha
	/// \param D_filename - filename to load diffusion coefficient from
	/// \param filetype = IFT_GRID grid type inside the file (yes, it can be different)
	bool LoadDiffusionCoefficient(GridElement &L, GridElement &pc, GridElement &alpha, string D_filename, string filetype = "IFT_GRID");

	/// Load Dxx from the file with Grid
	bool LoadDiffusionCoefficientFromFileWithGrid(GridElement &L, GridElement &pc, GridElement &alpha, string D_filename, string gridOrder = "IFT_GRID_LPA");

	/// Load Dxx from file without grid
	bool LoadDiffusionCoefficientFromPlaneFile(GridElement &L, GridElement &pc, GridElement &alpha, string D_filename);

	/// function calculate Dxx
	void Calculate(GridElement &L, GridElement &epc, GridElement &alpha, DxxParameters_structure DxxParameters);

	/// Making DLL
	void MakeDLL(double Kp);
	/// Making DLL
	void MakeDLL(GridElement &L, GridElement &pc, GridElement &alpha, double Kp, string DLLType = "DLLT_B");
	/// Making DLL
	void MakeDLL_B(GridElement &L, GridElement &pc, GridElement &alpha, double Kp);
	/// Making DLL
	void MakeDLL_FAKE(GridElement &L, GridElement &pc, GridElement &alpha, double Kp);
	/// Making DLL
	void MakeDLL_BE_res(GridElement &L, GridElement &pc, GridElement &alpha, double Kp);
	/// Making DLL
	void MakeDLL_BE(GridElement &L, GridElement &pc, GridElement &alpha, double Kp);
	/// Making DLL
	void MakeDLL100(GridElement &L, GridElement &pc, GridElement &alpha, double Kp);
	/// Making DLL_M
	void MakeDLL_Ozeke(GridElement &L, GridElement &pc, GridElement &alpha, double Kp);
	/// Making DLL_ME
	void MakeDLL_Ozeke_ME(GridElement &L, GridElement &pc, GridElement &alpha, double Kp);

};

/**
 * It has DiffusionCoefficient class as a parent class, so it stores there summation of all the coefficients to use in diffusion.
 * So, it is DiffusionCoefficient in common sense.
 * \todo change DiffusionCoefficientGroup name to the DiffusionCoefficient name and DiffusionCoefficient name to something else.
 */
 /// Holds list of instances of DiffusionCoefficient class of same type (like Daa, Dpp, etc), but produced by different waves (Daa_chorus, Daa_EMIC, etc).
class DiffusionCoefficientsGroup {
private:
	/// flag, indicated that the initialization was passed
	//bool DxxGroup_initialized;
public:
	DiffusionCoefficient CurrentDxx;
	// variables

	/// List of diffusion coefficients in that group. Actually, it's a list of waves used in the diffusion coefficient that this group represent.
	vector <DiffusionCoefficient> DxxList;

	/// Function activate (enable/disable) diffusion coefficients according time events and scale diffusion coefficients.
	/// It would be cool to call it "Actualize", but unfortunately it would be too complicated
	/// \param time - time
	/// \param Kp - Kp
	bool ActivateAndScale(double time, double Kp);

	// constructors
	DiffusionCoefficientsGroup() {
		//DxxGroup_initialized = false;
		//*this = 0;
	}

};

// Calculations...
	double Dxx_ba(double L, double epc, double alpha, double int_Dxx_loc(double lambda, DxxParameters_structure DxxParameters), DxxParameters_structure DxxParameters);	//
	double Dxx_local(double lambda, double Dxx_root(double Omega_e, double x, double mu, double su, double y, double beta, double a, double b, double alpha_star, double s, double epsilon, double d_x, double x_m, double R, DxxParameters_structure DxxParameters), DxxParameters_structure DxxParameters);

	double Daa_root(double Omega_e, double x, double mu, double su, double y, double beta, double a, double b, double alpha_star, double s, double epsilon, double d_x, double x_m, double R, DxxParameters_structure DxxParameters);
	double Dpa_root(double Omega_e, double x, double mu, double su, double y, double beta, double a, double b, double alpha_star, double s, double epsilon, double d_x, double x_m, double R, DxxParameters_structure DxxParameters);
	double Dpp_root(double Omega_e, double x, double mu, double su, double y, double beta, double a, double b, double alpha_star, double s, double epsilon, double d_x, double x_m, double R, DxxParameters_structure DxxParameters);

	double int_Daa_loc(double lambda, DxxParameters_structure DxxParameters);
	double int_Dpp_loc(double lambda, DxxParameters_structure DxxParameters);
	double int_Dpa_loc(double lambda, DxxParameters_structure DxxParameters);

	// various functions
//	double B(double lambda, double L);
	double f1(double lambda);

	double Alpha_ne(double pangle, double lambda, double L);
	double func_tmp (double x, double Alpha);

	double F_cap(double x, double y, double b, double s, double epsilon, DxxParameters_structure DxxParameters);
    // F_cap2 replaced F_cap. F_cap is no longer used.
    double F_cap2(double x, double y, double a, double beta, double mu, double s, double epsilon, double Alpha_star, DxxParameters_structure DxxParameters);
	//std::vector<double> rrouts(double Omega_e_eq, double Omega_e, double yida1, double yida2, double yida3, double epsilon, double beta, double mu, double alpha_star, double a, DiffusionCoefficientParameters_structure DxxParameters_structure);

	/// routs finding routine
	std::vector<double> rrouts(double x_1, double x_2, double yida1, double yida2, double yida3, double epsilon, double beta, double mu, double alpha_star, double a, DxxParameters_structure DxxParameters);

	double Df (double L, double Kp) ;

	double Df_Ozeke (double L, double Kp) ;

	double Df_Ozeke_E (double L, double Kp) ;

	void CreateAllDiffusionCoefficients(DiffusionCoefficient &DLL,
			DiffusionCoefficientsGroup &Daa, DiffusionCoefficientsGroup &Dpcpc, DiffusionCoefficientsGroup &Dpca,
			DiffusionCoefficientsGroup &DaaLpp, DiffusionCoefficientsGroup &DpcpcLpp, DiffusionCoefficientsGroup &DpcaLpp,
			Parameters_structure &parameters, Grid &radialDiffusionGrid, Grid &localDiffusionsGrid);

		/**
		* \brief Function used in main() to print header in output file. Header needs to understand the output file.
		*
		* Writing headers for all diffusion coefficients scaling coefficients
		* \param output1D - output stream to file
		* \param Daa - Pitch-angle D. all other parameters - is a type of diffusion coefficients (DiffusionCoefficientsGroup). To file - Dxx.DxxList[Dxx_it].arr.name
		* \param Dpcpc - Energy D
		* \param Dpca - Mix pitch-angle and energy.
		* \param DaaLpp - Pitch-angle with plasma pause
		* \param DpcpcLpp - Energy with plasma pause
		* \param DpcaLpp - Pitch-angle and energy with plasma pause
		*/
		void Output1DHeaders(ofstream &output1D,
		DiffusionCoefficientsGroup &Daa, DiffusionCoefficientsGroup &Dpcpc, DiffusionCoefficientsGroup &Dpca,
		DiffusionCoefficientsGroup &DaaLpp, DiffusionCoefficientsGroup &DpcpcLpp, DiffusionCoefficientsGroup &DpcaLpp);

		/**
		* \brief Function used in main() to print values of diffusion coefficients in output file.
		*
		* Write information about 1d variables into the file with 1d variables
		* \param output1D - output stream to file
		* \param Daa - Pitch-angle D. all other parameters - is a type of diffusion coefficients (DiffusionCoefficientsGroup). To file - Dxx.DxxList[Dxx_it].arr.name
		* \param Dpcpc - Energy D
		* \param Dpca - Mix pitch-angle and energy.
        * \param DaaLpp - Pitch-angle with plasma pause
        * \param DpcpcLpp - Energy with plasma pause
        * \param DpcaLpp - Pitch-angle and energy with plasma pause
		* \param time - current simulation time
		* \param parameters - (Parameters_structure) structure with all parameters from ini files and loaded/calcaulated parameters such as Kp and UpperBoundary?
		* \param iteration - Number of current iteration of the simulation
		*/
		void Output1DValues(ofstream &output1D,
		DiffusionCoefficientsGroup &Daa, DiffusionCoefficientsGroup &Dpcpc, DiffusionCoefficientsGroup &Dpca,
		DiffusionCoefficientsGroup &DaaLpp, DiffusionCoefficientsGroup &DpcpcLpp, DiffusionCoefficientsGroup &DpcaLpp,
		double &time, Parameters_structure &parameters, int iteration);


#endif


