/**
 * \file BoundaryConditions.cpp
 * Boundary condition class.
 *
 * \author Developed under supervision of the PI Yuri Shprits
 */

#include "BoundaryConditions.h"
#include <math.h>

// Function 'strncasecmp' called '_strnicmp' in windows
#if defined(_WIN32) || defined(_WIN64)
#define strncasecmp _strnicmp
#define strcasecmp _stricmp
#endif



/**
 * Making boundary conditions.
 *
 * Fill-out 2D array of boundary conditions. (in general, 3D grid case) And save boundary condition type for future use.
 *
 * \param parameters - boundary condition parameters
 * \param psd2DSlice - 2D array, slice of PSD on a boundary
 * \param gridElement1 - one side of the grid
 * \param gridElement2 - second side of the grid
 */
void BoundaryCondition::MakeBoundaryCondition(Parameters_structure::BoundaryCondition parameters, Matrix2D<double> psd2DSlice, Matrix2D<double> gridElement1, Matrix2D<double> gridElement2) {
	int ix, iy;

	if (!arr.initialized) {
		// Allocate memory
		arr.AllocateMemory(psd2DSlice.size_x, psd2DSlice.size_y);
	}

	// store type in a class variable, for the ease of future use
	this->type = parameters.type;
	if (parameters.type == "BCT_CONSTANT_VALUE") {
			// Saving boundary value inside the class.
			// If boundary is on constant value of the function - saving this type and the value.
			this->calculationType = "BCT_CONSTANT_VALUE";

			// If value is zero, we don't put zero to PSD, but put some small value
			//(*this) = VF::max(value, VC::zero_f);
			arr = VF::max(parameters.value, VC::zero_f);

	} else if (parameters.type == "BCT_CONSTANT_PSD") {
			// If boundary is on constant PSD value  - saving this type and we need to store the initial values of PSD later, after initialization of PSD.
			this->calculationType = "BCT_CONSTANT_VALUE";

			// So, this line store psd slice as a boundry, and change all zeros to some small value
			for (ix = 0; ix < arr.size_x; ix++) {
				for (iy = 0; iy < arr.size_y; iy++) {
					arr[ix][iy] = VF::max(psd2DSlice[ix][iy], VC::zero_f);
				}
			}
	} else if (parameters.type == "BCT_CONSTANT_DERIVATIVE") {
			// If boundary is on constant derivative of the function - saving this type and the value of the derivative (usually - zero derivative).
			this->calculationType = "BCT_CONSTANT_DERIVATIVE";

			// boundary condition is a constant derivative - so, storing derivative value for boundary conditions for each boundary point
			// storing it inside the parent class Matrix.
			// this - is a pointer to current class. '=' - operator.
			arr = parameters.value;
	} else if (parameters.type == "BCT_CONSTANT_SECOND_DERIVATIVE") {
			// If boundary is on constant second derivative of the function
			this->calculationType = "BCT_CONSTANT_SECOND_DERIVATIVE";

			// boundary condition is a constant derivative - so, storing derivative value for boundary conditions for each boundary point
			// storing it inside the parent class Matrix.
			// this - is a pointer to current class. '=' - operator.
			arr = parameters.value;
	} else if (parameters.type == "BCT_FILE" || parameters.type == "BCT_FILE_GRID" || parameters.type == "BCT_MULTIPLE_FILES") {
			// If boundary are loaded from file... see BoundaryCondition::LoadBoundaryCondition
			this->calculationType = "BCT_CONSTANT_VALUE";
			// this->filename = parameters.filename; // don't need this i guess... but it does not do any bad.
	} else {
			throw error_msg("BCT_ERROR", "Unknown boundary condition type: %d", parameters.type.c_str());
	}
}


/**
 * Load boundary conditions from file.
 *
 * Here we actually fill-out 2D array of boundary conditions. (in general, 3D grid, case)
 *
 * \param parameters - boundary condition parameters
 * \param gridElement1 - one side of the grid
 * \param gridElement2 - second side of the grid
 */
void BoundaryCondition::LoadBoundaryCondition(Parameters_structure::BoundaryCondition parameters, Matrix2D<double> gridElement1, Matrix2D<double> gridElement2) {
	int ix, iy;
	double err = 1.e-3;
	string inBuf;

	if (!arr.initialized) {
		// Allocating memory
		arr.AllocateMemory(gridElement1.size_x, gridElement1.size_y);
	}

	Output::echo(0, "Loading boundary conditions from files... ");
	// store type in a class variable, for the ease of future use
	this->type = parameters.type;
	//if (initialType == "BCT_FILE" || initialType == "BCT_FILE_GRID") {
	if (parameters.type == "BCT_FILE" || parameters.type == "BCT_FILE_GRID") {
			this->calculationType = "BCT_CONSTANT_VALUE";
			//			this->readFromFile(filename);
			// making stream
			{
				ifstream input(parameters.filename.c_str());
				// checking if file was opened
				if (input == NULL) {
					throw error_msg("BC_FILE_OPEN_ERR", "Boundary conditions file %s not found.", parameters.filename.c_str());
				}
				// Skip header!
				input >> inBuf;
				while (strcasecmp(inBuf.c_str(), "ZONE") != 0 && !input.eof() ) input >> inBuf;
				input.ignore(9999, '\n'); // read to the end of the line with 'zone'
				double loadedTime, loadedAx1, loadedAx2, loadedValue;
				// size_x - time steps, size_y and size_z - two dimensions of PSD slice
				for (ix = 0; ix < arr.size_x; ix++) {
					for (iy = 0; iy < arr.size_y; iy++) {
						if (parameters.type == "BCT_FILE_GRID") input >> loadedAx1 >> loadedAx2;
						input >> loadedValue;
						// check if grid is the same
						if (parameters.type == "BCT_FILE_GRID" && (fabs(loadedAx1 - gridElement1[ix][iy]) > err || fabs(loadedAx2 - gridElement2[ix][iy]) > err)) {
							if (input.eof()) {
								throw error_msg("BC_LOAD_GRID_ERR", "Loading %s: not enough data points in the boundary condition file.\n", parameters.filename.c_str());
							} else {
								throw error_msg("BC_LOAD_GRID_ERR", "Loading %s: grid mismatch.\nLoaded: ax1=%e, ax2=%e\nGrid:  ax1=%e, ax2=%e\n(i1=%d, i2=%d)\n", parameters.filename.c_str(), loadedAx1, loadedAx2, gridElement1[ix][iy], gridElement2[ix][iy], ix, iy);
							}
						} else {
								arr[ix][iy] = loadedValue;
						}
						// skip till the end of the line
						input.ignore(9999, '\n');
					}
				}
				input.close();
			}
			arr.writeToFile("./Debug/" + arr.name + "_BC.dat");
			Output::echo(0, "done.\n");
	} else if (parameters.type == "BCT_MULTIPLE_FILES") {
		// In that case we need to change boundary conditions (BC) with time.
		// Now we load the list of boundary conditions (files with boundary values) and corresponding times, when we need to apply the BC
		this->calculationType = "BCT_CONSTANT_VALUE";

		Output::echo(0, "Loading a list of boundary conditions... ");
		ifstream input(parameters.filename.c_str());
		// checking if file was opened
		if (input == NULL) {
			throw error_msg("BC_FILE_OPEN_ERR", "Boundary conditions file %s not found.", parameters.filename.c_str());
		}

		// Skip header of the file
		input >> inBuf;
		while (strcasecmp(inBuf.c_str(), "ZONE") != 0 && !input.eof() ) input >> inBuf;
		input.ignore(9999, '\n'); // read to the end of the line with 'zone'

		// reading the file
		double time_tmp;
		string filename_tmp;
		while (!input.eof()) {
			// reading time values and function values from the file into vectors
			input >> time_tmp >> filename_tmp;
			BC_change_time.push_back(time_tmp);
			BC_filename.push_back(filename_tmp);
		}

		// Load the first on
		ifstream input2(BC_filename[0].c_str());
		// checking if file was opened
		if (input2 == NULL) {
			throw error_msg("BC_FILE_OPEN_ERR", "Boundary conditions file %s not found.", parameters.filename.c_str());
		}
		Output::echo(3, "Updating a boundary condition: %s", BC_filename[0].c_str());
		// Skip header of the file
		string inBuf;
		input2 >> inBuf;
		while (strcasecmp(inBuf.c_str(), "ZONE") != 0 && !input2.eof() ) input2 >> inBuf;
		input2.ignore(9999, '\n'); // read to the end of the line with 'zone'
		double tmp;
		for (ix = 0; ix < arr.size_x; ix++) {
			for (iy = 0; iy < arr.size_y; iy++) {
				input2 >> arr[ix][iy];
			}
		}

	}

}
/**
 * Update boundary condition for current time step
 *
 * \param iteration - iteration number
 * \param PSD_2D_Slice - 2d slice of PSD for that BC
 * \param time - total time
 * \param dt - time step
 */
void BoundaryCondition::Update(int iteration, Matrix2D<double> PSD_2D_Slice, double time, double dt) {
	int ix, iy;

	if (type == "BCT_MULTIPLE_FILES") {
		if (time == -1 || dt == -1)
			throw error_msg("BC_ERR", "To use BCT_MULTIPLE_FILES, time and dt need to be specified when the function is called.");
		// Find the next time to change BC
		int i;
		for (i = 0; i < BC_change_time.size() && BC_change_time[i] < time; i++) {};
		// Check if it's time to load next BC, i.e. the time to change is between the previous and current time steps
		if (time - dt < BC_change_time[i] && time >= BC_change_time[i]) {
			// Load next BC
			ifstream input(BC_filename[i].c_str());
			// checking if file was opened
			if (input == NULL) {
				throw error_msg("BC_FILE_OPEN_ERR", "Boundary conditions file %s not found.", BC_filename[i].c_str());
			}
			Output::echo(3, "Updating a boundary condition: %s", BC_filename[i].c_str());
			// Skip header of the file
			string inBuf;
			input >> inBuf;
			while (strcasecmp(inBuf.c_str(), "ZONE") != 0 && !input.eof() ) input >> inBuf;
			input.ignore(9999, '\n'); // read to the end of the line with 'zone'
			double tmp;
			for (ix = 0; ix < arr.size_x; ix++) {
				for (iy = 0; iy < arr.size_y; iy++) {
					input >> tmp;
					arr[ix][iy] = tmp;
				}
			}
		}
	}

}
