/**
 * \file Grid.cpp
 * Grids, grid elements and boundary conditions source.
 *
 * \author Developed under supervision of the PI Yuri Shprits
 */

#include <iostream>
#include <iomanip>

#include "Grid.h"
#include <math.h>
#include "../VariousFunctions/variousFunctions.h"
#include "../Exceptions/error.h"

using namespace std;

#define maxERR 1.e-6

#if defined(_WIN32) || defined(_WIN64)
#define strncasecmp _strnicmp
#define strcasecmp _stricmp
#endif

/**
 * Convert pc to epc (kinetic energy).
 *
 * \param pc - pc.
 */
void GridElement::Kfunc_equal(GridElement pc) {
	arr.AllocateMemory(pc.arr.size_x, pc.arr.size_y, pc.arr.size_z);
	int x, y, z;
	for (x = 0; x < pc.arr.size_x; x++)
		for (y = 0; y < pc.arr.size_y; y++)
			for (z = 0; z < pc.arr.size_z; z++)
				arr[x][y][z] = VF::Kfunc(pc.arr[x][y][z]);
}

/**
 * Function return the value for the specified point on regular grid.
 *
 * Very useful function, allows to avoid a lot of typos in creation of regular grids everywhere.
 *
 * Just don't have to repeat regular grid creation code a lot of times.
 * The function create logarithmic grid if useLogScale == true as: 10^(log10(min) + x * (log10(max) - log10(min))/(size-1);
 * if useLogScale == false, then grid is: min + x * (max - min)/(size-1).
 * If grid size == 1 the function returns max.
 *
 * PS: size, min and max stored in the gridElement already.
 * PPS: useLogScale is also there.
 *
 * \param iteratorL - index on L-grid.
 * \param iteratorPc - index on pc-grid.
 * \param iteratorAlpha - index on alpha-grid.
 * \param gridElementDirection - index on the grid, which we are making.
 */
void GridElement::SetRegularGridValue(int iteratorL, int iteratorPc, int iteratorAlpha, int gridElementDirection) {
	if (GridElement_parameters.size > 1)
		// in case of 3d
		if (GridElement_parameters.useLogScale) // in case of log scale
			arr[iteratorL][iteratorPc][iteratorAlpha] = pow(10, log10(GridElement_parameters.min) + gridElementDirection*(log10(GridElement_parameters.max) - log10(GridElement_parameters.min))/(GridElement_parameters.size - 1));
		else // in case of regular scale
			arr[iteratorL][iteratorPc][iteratorAlpha] = GridElement_parameters.min + gridElementDirection*(GridElement_parameters.max - GridElement_parameters.min)/(GridElement_parameters.size - 1);
	else // in case of 2d
		arr[iteratorL][iteratorPc][iteratorAlpha] = GridElement_parameters.max;
}

/**
 * create a grid based on each grid element's parameters
 *
 * \param parameters_L - parameters of grid element L
 * \param parameters_pc - parameters of grid element pc
 * \param parameters_alpha - parameters of grid element alpha
 * \param parameters_epc - parameters of grid element epc
 * \param grid_filename - filename of grid
 * \param gridType - grid type
 * \param SecondGrid - grid parameter in case creating copy of this grid
 */
void Grid::Create_Grid(Parameters_structure::GridElement parameters_L,
						Parameters_structure::GridElement parameters_pc,
						Parameters_structure::GridElement parameters_alpha,
						Parameters_structure::GridElement parameters_epc,
						string grid_filename, string gridType,
						Grid SecondGrid) {

	//if (!L.initialized) {
		//throw error_msg("GRID_INITIALIZATION_ERROR", "Grid was not initialized");

		// allocating memory for all member classes
		L.arr.AllocateMemory(parameters_L.size, parameters_pc.size, parameters_alpha.size);
		pc.arr.AllocateMemory(parameters_L.size, parameters_pc.size, parameters_alpha.size);
		alpha.arr.AllocateMemory(parameters_L.size, parameters_pc.size, parameters_alpha.size);
		epc.arr.AllocateMemory(parameters_L.size, parameters_pc.size, parameters_alpha.size);
		Jacobian.AllocateMemory(parameters_L.size, parameters_pc.size, parameters_alpha.size);

		//L.Initialize(parameters_L);
		L.GridElement_parameters = parameters_L;
		L.arr.name = parameters_L.name;
		L.size = parameters_L.size;

		//pc.Initialize(parameters_pc);
		pc.GridElement_parameters = parameters_pc;
		pc.arr.name = parameters_pc.name;
		pc.size = parameters_pc.size;

		//alpha.Initialize(parameters_alpha);
		alpha.GridElement_parameters = parameters_alpha;
		alpha.arr.name = parameters_alpha.name;
		alpha.size = parameters_alpha.size;

		//epc.Initialize(parameters_epc);
		epc.GridElement_parameters = parameters_epc;
		epc.arr.name = parameters_epc.name;
		epc.size = parameters_epc.size;

		type = gridType; // save the grid type for future use

		//LSize = L.size;
		//pcSize = pc.size;
		//alphaSize = alpha.size;
		//epcSize = epc.size;

	//}

	// iterators, cold be useful later
	int il, im, ia;
	// temparary matrix_arrays
	Matrix3D<double> mu_arr(L.size, pc.size, alpha.size);
	// temporary variables
	double mu, Jc, mu_check, Jc_check;
	if (gridType == "GT_FILE") {
		int il, im, ia;
		string inBuf;
		// making stresm
		ifstream input(grid_filename.c_str());
		if (input != NULL && !input.eof()) {
			// Skipping header.
			input >> inBuf;
			while (strcasecmp(inBuf.c_str(), "ZONE") != 0 && !input.eof() ) {
				input >> inBuf;
			}
			// read to the end of the line with 'zone'
			input.ignore(9999, '\n');
			for (il = 0; il < L.size; il++) {
				for (im = 0; im < epc.size; im++) {
					for (ia = 0; ia < alpha.size; ia++) {
						input >> L.arr[il][im][ia] >> epc.arr[il][im][ia] >> alpha.arr[il][im][ia] >> pc.arr[il][im][ia];
						// skip till the end of the line
						input.ignore(9999, '\n');
						if (input.eof()) {
							throw error_msg("GRID_INITIALIZATION_ERROR", "Not enough grid points in file %s - check grid resolution.", grid_filename.c_str());
						}
					}
				}
			}
			input.ignore(1, '\n');
			if (!input.eof()) {
				Output::echo(0, "Warning: probably too many points in the grid file %s - check grid resolution.", grid_filename.c_str());
			}

			//  calculating Jacobian for the grid
			// \todo Calculate Jacobian from the grid data
			int il, im, ia;
			for (il = 0; il < L.size; il++) {
				for (im = 0; im < pc.size; im++) {
					for (ia = 0; ia < alpha.size; ia++) {
						// remove all constant - they don't matter
						Jacobian[il][im][ia] = pow(pc.arr[il][im][ia],2) * pow(L.arr[il][im][ia],2) * VF::G(alpha.arr[il][im][ia]);
						// Jacobian[il][im][ia] = -4.0 * pow(pc[il][im][ia],2) * pow(L[il][im][ia],2) * VF::G(alpha[il][im][ia])/VF::B(L[il][im][ia]);
						//psd.G[q0][q1][q2]  = -4.0 * pow(psd.pc[q0][q1][q2],2)  * pow(psd.L[q0][q1][q2],2)  * G(psd.alpha[q0][q1][q2])/B(psd.L[q0][q1][q2]);
					}
				}
			}

		} else {
			throw error_msg("GRID_INITIALIZATION_ERROR", "Grid file %s not found.", grid_filename.c_str());
		}
	} else if (gridType == "GT_L_PC_ALPHA") {
			// making grid
			for (il = 0; il < L.size ; il++){
				for (im = 0; im < pc.size; im++){
					for (ia = 0; ia < alpha.size; ia++){
						// You see, instead of that multiline formula for grid, appears everywherem we have just one.
						L.SetRegularGridValue(il, im, ia, il);
						pc.SetRegularGridValue(il, im, ia, im);
						alpha.SetRegularGridValue(il, im, ia, ia);
					}
				}
			}
			//epc = pc.Kfunc();
			epc.Kfunc_equal(pc);

			//  calculating Jacobian for the grid
			for (il = 0; il < L.size; il++) {
				for (im = 0; im < pc.size; im++) {
					for (ia = 0; ia < alpha.size; ia++) {
						// remove all constant - they don't matter
						// !!!!!!!!!!!!!Jacobian[il][im][ia] = pow(pc.arr[il][im][ia],2) * pow(L.arr[il][im][ia],2) * VF::G(alpha.arr[il][im][ia]);
						// Jacobian[il][im][ia] = -4.0 * pow(pc[il][im][ia],2) * pow(L[il][im][ia],2) * VF::G(alpha[il][im][ia])/VF::B(L[il][im][ia]);
						Jacobian[il][im][ia] = pow(pc.arr[il][im][ia],2) * VF::G(alpha.arr[il][im][ia]);

					}
				}
			}
			//Jacobian = 1;
			//Output::echo(0, "Jacobian == 1 !!!!!!!!!!!!!!!!\n");


	} else if (gridType == "GT_L_MU_J") {
			// Make L regular grid
			for (il = 0; il < L.size ; il++) {
				for (im = 0; im < pc.size; im++) {
					for (ia = 0; ia < alpha.size; ia++) {
						// You see, instead of that multiline formula for grid, appears everywherem we have just one.
						L.SetRegularGridValue(il, im, ia, il);
					}
				}
			}
			// Making a 2-D grid on the top level at L=Lmax
			il = L.size-1;
			for (im = 0; im < pc.size; im++){
				for (ia = 0; ia < alpha.size; ia++) {
					// You see, instead of that multiline formula for grid, appears everywherem we have just one.
					pc.SetRegularGridValue(il, im, ia, im);
					alpha.SetRegularGridValue(il, im, ia, ia);
				}
			}
			// Making a grid which conserves 1st J1 and second J2 adiabatic invariant
			if (L.size > 1) {
				for (im = 0; im < pc.size; im++) {
					for (ia = 0; ia < alpha.size; ia++) {
						mu = VF::mu_calc(L.arr[L.size-1][im][ia], pc.arr[L.size-1][im][ia], alpha.arr[L.size-1][im][ia]);
						Jc = VF::Jc_calc(L.arr[L.size-1][im][ia], pc.arr[L.size-1][im][ia], alpha.arr[L.size-1][im][ia]);
						for (il = 0; il < L.size-1 ; il++) {
							if (alpha.arr[L.size-1][im][ia] < VC::pi/2)
								alpha.arr[il][im][ia] = find_alpha(Jc * sqrt(L.arr[il][im][ia]) / sqrt(8.0 * VF::B(1) * VC::mc2 * mu), alpha.GridElement_parameters.min, VC::pi/2);
							else if (alpha.arr[L.size-1][im][ia] > VC::pi/2)
								alpha.arr[il][im][ia] = find_alpha(Jc * sqrt(L.arr[il][im][ia]) / sqrt(8.0 * VF::B(1) * VC::mc2 * mu), VC::pi/2, alpha.GridElement_parameters.max);
							else
								alpha.arr[il][im][ia] = VC::pi/2;
							pc.arr[il][im][ia] = sqrt(2.0 * mu * VC::mc2 * VF::B(L.arr[il][im][ia])) / sin(alpha.arr[il][im][ia]);
							mu_check = VF::mu_calc(L.arr[il][im][ia], pc.arr[il][im][ia], alpha.arr[il][im][ia]);
							Jc_check = VF::Jc_calc(L.arr[il][im][ia], pc.arr[il][im][ia], alpha.arr[il][im][ia]);
							if (fabs(mu - mu_check) > maxERR || fabs(Jc - Jc_check) > maxERR) {
								throw error_msg("L_MU_J_GRID_ERROR", "Grid computation error (L, pc, a) = (%f, %f, %f) :\n mu=%e, Jc=%e\n mu_chk=%e, Jc_chk=%e\n mu_err=%e, Jc_err=%e.\n", L.arr[il][im][ia], pc.arr[il][im][ia], alpha.arr[il][im][ia], mu, Jc, mu_check, Jc_check, fabs(mu - mu_check), fabs(Jc - Jc_check));
							}
						}
					}
				}
			}
			epc.Kfunc_equal(pc);

			//  calculating Jacobian for the grid
			for (il = 0; il < L.size; il++) {
				for (im = 0; im < pc.size; im++) {
					for (ia = 0; ia < alpha.size; ia++) {
						// remove all constant - they don't matter

//						Jacobian[il][im][ia] = pow(pc[il][im][ia],2) * pow(L[il][im][ia],-2) * VF::G(alpha[il][im][ia]);;
						Jacobian[il][im][ia] = pow(L.arr[il][im][ia],-2); // <--
//						Jacobian[il][im][ia] = 1;
//						Jacobian[il][im][ia] = pow(L[il][im][ia],2);

						// Jacobian[il][im][ia] = -4.0 * pow(pc[il][im][ia],2) * pow(L[il][im][ia],2) * VF::G(alpha[il][im][ia])/VF::B(L[il][im][ia]);
						//psd.G[q0][q1][q2]  = -4.0 * pow(psd.pc[q0][q1][q2],2)  * pow(psd.L[q0][q1][q2],2)  * G(psd.alpha[q0][q1][q2])/B(psd.L[q0][q1][q2]);
					}
				}
			}

	} else if (gridType == "GT_L_MU_ALPHA") {
			// Make L regular grid
			for (il = 0; il < L.size ; il++) {
				for (im = 0; im < pc.size; im++) {
					for (ia = 0; ia < alpha.size; ia++) {
						// You see, instead of that multiline formula for grid, appears everywherem we have just one.
						L.SetRegularGridValue(il, im, ia, il);
					}
				}
			}
			// Making a 2-D grid on the top level at L=Lmax
			il = L.size-1;
			for (im = 0; im < pc.size; im++){
				for (ia = 0; ia < alpha.size; ia++) {
					pc.SetRegularGridValue(il, im, ia, im);
					alpha.SetRegularGridValue(il, im, ia, ia);
//					mu_arr[il][im][ia] = VF::pc2mu(L[il][im][ia], pc[il][im][ia], alpha[il][im][ia]);
				}
			}
			for (im = 0; im < pc.size; im++) {
				for (ia = 0; ia < alpha.size; ia++) {
					for (il = 0; il < L.size-1 ; il++) {
						alpha.SetRegularGridValue(il, im, ia, ia);
						// mu on the top
						mu = VF::pc2mu(L.arr[L.size-1][im][ia], pc.arr[L.size-1][im][ia], alpha.arr[L.size-1][im][ia]);
						pc.arr[il][im][ia] = VF::mu2pc(L.arr[il][im][ia], mu, alpha.arr[il][im][ia]);
					}
				}
			}
			epc.Kfunc_equal(pc);

			//  calculating Jacobian for the grid
			for (il = 0; il < L.size; il++) {
				for (im = 0; im < pc.size; im++) {
					for (ia = 0; ia < alpha.size; ia++) {
						// remove all constant - they don't matter
						// L^-2?? If it's on (L, pc, alpha) grid, I think it should be L^2. It does not matter cause it cancels anyway, but still.
						Jacobian[il][im][ia] = pow(pc.arr[il][im][ia],2) * pow(L.arr[il][im][ia],-2) * VF::G(alpha.arr[il][im][ia]);
						// Jacobian[il][im][ia] = -4.0 * pow(pc[il][im][ia],2) * pow(L[il][im][ia],2) * VF::G(alpha[il][im][ia])/VF::B(L[il][im][ia]);
						//psd.G[q0][q1][q2]  = -4.0 * pow(psd.pc[q0][q1][q2],2)  * pow(psd.L[q0][q1][q2],2)  * G(psd.alpha[q0][q1][q2])/B(psd.L[q0][q1][q2]);
					}
				}
			}

	} else if (gridType == "GT_PERP_TO_L_MU_J") {
			if (!SecondGrid.Grid_initialized) {
				throw error_msg("GRID_INIT_ERR", "GT_PERP_TO_L_MU_J with empty second grid");
			}
			L = SecondGrid.L;
			alpha = SecondGrid.alpha;

			// make new pc
			double pc_local_max, pc_local_min;
			for (il = 0; il < L.size ; il++){
				// maximum from the upper ENERGY boundary
				pc_local_max = SecondGrid.pc.arr[il][SecondGrid.pc.size-1][alpha.size-1];
				for (ia = 0; ia < alpha.size; ia++) if (pc_local_max < SecondGrid.pc.arr[il][SecondGrid.pc.size-1][ia]) pc_local_max = SecondGrid.pc.arr[il][SecondGrid.pc.size-1][ia];
				// MAXIMUM from the LOWER energy boundary
				pc_local_min = SecondGrid.pc.arr[il][0][alpha.size-1];
				for (ia = 0; ia < alpha.size; ia++) if (pc_local_min < SecondGrid.pc.arr[il][0][ia]) pc_local_min = SecondGrid.pc.arr[il][0][ia];
				for (im = 0; im < pc.size; im++){
					for (ia = 0; ia < alpha.size; ia++){
						if (pc.size > 1) {
							if (!pc.GridElement_parameters.useLogScale) pc.arr[il][im][ia] = pc_local_min + im*(pc_local_max - pc_local_min)/(pc.size - 1);
							else pc.arr[il][im][ia] = pow(10, log10(pc_local_min) + im*(log10(pc_local_max) - log10(pc_local_min))/(pc.size - 1));
						} else pc.arr[il][im][ia] = pc.GridElement_parameters.max;
					}
				}
			}

			epc.Kfunc_equal(pc);

			//  calculating Jacobian for the grid
			for (il = 0; il < L.size; il++) {
				for (im = 0; im < pc.size; im++) {
					for (ia = 0; ia < alpha.size; ia++) {
						// remove all constant - they don't matter
						Jacobian[il][im][ia] = pow(pc.arr[il][im][ia],2) * pow(L.arr[il][im][ia],2) * VF::G(alpha.arr[il][im][ia]);
						// Jacobian[il][im][ia] = -4.0 * pow(pc[il][im][ia],2) * pow(L[il][im][ia],2) * VF::G(alpha[il][im][ia])/VF::B(L[il][im][ia]);
						//psd.G[q0][q1][q2]  = -4.0 * pow(psd.pc[q0][q1][q2],2)  * pow(psd.L[q0][q1][q2],2)  * G(psd.alpha[q0][q1][q2])/B(psd.L[q0][q1][q2]);
					}
				}
			}

	} else {
			throw error_msg("GRID_UNKNOWN_TYPE", "Unknown grid type: %s", gridType.c_str());
	}

	Jacobian.writeToFile("./Debug/Jacobian.plt");


	Grid_initialized = true;

}


/**
 *  Output grid to a file.
 *
 * \param filename - file name
 */
void Grid::Output(string filename) {
	ofstream output(filename.c_str());
	int il, im, ia;

	// output header of the file
	output << "VARIABLES = ";
	output << "\"L, Earth radius\", \"Energy, MeV\", \"Pitch-angle, deg\", \"pc\" " << endl;

	// output values...
	output << "ZONE T=\"Grid\" " << " I=" << alpha.size << ", J=" << epc.size << ", K=" << L.size << endl;
	output.setf(ios_base::scientific, ios_base::floatfield);
	for (il = 0; il < L.size; il++) {
		for (im = 0; im < epc.size; im++) {
			for (ia = 0; ia < alpha.size; ia++) {
				output << L.arr[il][im][ia] << "\t" << epc.arr[il][im][ia] << "\t" << alpha.arr[il][im][ia] << "\t" << VF::pfunc(epc.arr[il][im][ia]) << endl;
			}
		}
	}
}

/**
 *  Function finds alpha-root of equation Y(alpha)/sin(alpha) = RHS.. Used in grid-building.
 *
 * \param RHS - right hand side
 * \param alpha_min - min alpha value to look for root
 * \param alpha_max - max alpha value to look for root
 * \param ERR - max error (min accuracy)
 * \param max_it - max number of iterations
 * \param it - current iteration number (for recursion)
 */
double find_alpha(double RHS, double alpha_min, double alpha_max, double ERR, int max_it, int it) {

	double alpha_root;
	do {
		it++;

		alpha_root = (alpha_min + alpha_max) / 2;
		double y0 = VF::Y(alpha_min)/sin(alpha_min) - RHS, yy = VF::Y(alpha_root)/sin(alpha_root) - RHS;

		if (yy * y0 < 0) alpha_max = alpha_root;
		else if (yy * y0 > 0) alpha_min = alpha_root;
		else return alpha_root;

	} while (fabs(alpha_max - alpha_min) > ERR);

	return alpha_root;

}
