/**
 * \file Grid.h
 * Grids, grid elements and boundary conditions headers.
 *
 * \author Developed under supervision of the PI Yuri Shprits
 */

#ifndef Grid_H
#define Grid_H

#include <string>
#include "../Matrix/Matrix.h"
#include "../Parameters/Parameters.h"
#include "../VariousFunctions/variousFunctions.h"
#include "Grid.h"
#include "BoundaryConditions.h"

using namespace std;


/**
 * \brief Array of values of coordinate axes.
 *
 * Class for %GridElement in parent class Matrix3D holds values of one coordinate from 3-dimensional (usually) coordinate system.
 * Has two BoundaryCondition members: lowerBoundaryCondition and upperBoundaryCondition.
 * Has functions to create grids of different type.
 */
class GridElement  {
public:
	Matrix3D<double> arr; ///< array of grid points

	/// initialized flag
	bool GridElement_initialized;

	/// parameters for grid element
    Parameters_structure::GridElement GridElement_parameters;

	/// size of grid element
	int size;

	// other functions
	//GridElement Kfunc(); // return NEW Grid Element equal to Kfunc of present
	void Kfunc_equal(GridElement arg); // make THIS element equal to Kfunc(arg). Preferred to use this one, works faster
	void SetRegularGridValue(int il, int im, int ia, int gridElementDirection); // return value for specified parameters for regular grid
};


/**
 * \brief Computational grid composed of 3 different GridElement
 *
 * Combined from 3 grid elements: L, pc, Alpha and additional array of epc values for convenience.
 */
class Grid {
private:
public:
	// members
    //@{
    /// grid elements to be used
	GridElement L, pc, alpha, epc;
    //@}
	//Matrix3D<double> L, pc, alpha, epc;
	//int LSize, pcSize, alphaSize, epcSize;

    /// grid type
	string type;
    /// check for initialized status
	bool Grid_initialized;

	void Create_Grid(Parameters_structure::GridElement parameters_L,
					Parameters_structure::GridElement parameters_pc,
					Parameters_structure::GridElement parameters_alpha,
					Parameters_structure::GridElement parameters_epc,
					string grid_filename, string gridType,
					Grid SecondGrid = Grid()); // the last parameter is needed in case we create the grid based on another grid

	void Output(string filename);

	/** Jacobian of the diffusion equation.
	 * In 1d_universal_solver there is a parameter 'n' represented power of variable in the equation
	 * \f$ x^n \frac{\partial}{\partial x} x^{-n} D_{xx} \frac{\partial f}{\partial x} \f$.
	 * More general case is using Jacobian:
	 * \f$ Jacobian \frac{\partial}{\partial x} Jacobian^{-1} D_{xx} \frac{\partial f}{\partial x} \f$.
	 */
	Matrix3D<double> Jacobian;

};

/** Root finding function, used in grid construction.
 */
double find_alpha(double RHS, double alpha_min, double alhpa_max, double ERR = 1e-12, int max_it = 100, int it = 0);

#endif
