/**
 * \mainpage
 *
 * The document <a href="../Algorithms.pdf">Algorithms.pdf</a> may be useful for understanding the structure of the code.
 * <a href="../AllFunctions.pdf">AllFunctions.pdf</a> and <a href="../Description.pdf">Description.pdf</a> may also be helpful for understanding the code better.
 *
 * This document was generated by the Doxygen system. It has a description of the API, and used my main program's code, which runs the  main() function. 
 * If you want to understand how the VERB code works, please, read the overview and after read the code and comments themselves.
 *
 * The VERB code was developed by the Radiation Belt Modeling Group, a research group under the Earth, Planetary, and Space Sciences department at UCLA. Their website can be found here: http://rbm.epss.ucla.edu
 *
 * The head researchers of this project are Yuri Shprits, Adam Kellerman, Alexander Drozdov, Ksenia Orlova, and Tatiana Podladchikova. The code and documentation have been updated by this team as well as other researchers, last updated in 2015.
 *
 * # Overview.
 * 
 * See Main.cpp documentation for general algorithm explanation.
 *
 * The code is designed to solve the Fokker-Planck diffusion equation: \f[
 \frac{\partial f}{\partial t} = \frac{1}{p^2} \frac{\partial}{\partial p} p^2 D_{pp} \frac{\partial f}{\partial p}\Big|_{\textrm{const }\alpha, L} +
 \frac{1}{\cos(\alpha) \sin(\alpha) T(\alpha)} \frac{\partial}{\partial \alpha} \cos(\alpha) \sin(\alpha) T(\alpha) D_{\alpha \alpha} \frac{\partial f}{\partial \alpha}\Big|_{\textrm{const } p, L} +
 L^2 \frac{\partial}{\partial L} \frac{1}{L^2} D_{LL} \frac{\partial f}{\partial L}\Big|_{\textrm{const } J_1, J_2} + Q - S
 * \f]
 *
 * This equation describes the evolution of phase space density (PSD) in time and phase space.
 * It has three diffusion terms: _Radial_, _Energy_, and _Pitch-angle_. _Radial_ diffusion needs a different grid than the two other (local, _Energy_, and _Pitch-angle_) diffusions. Q - local sources. S - local losses.
 * In the code we split the equation into \b three parts and solve them one after another with interpolation between radial and local grids:
 * 		- \e Radial diffusion
 * 		- Interpolation from radial grid to local grid.
 * 		- \e Energy diffusion.
 * 		- \e Pitch-angle diffusion.
 * 		- Interpolation from local to radial grid.
 *
 * We have \b two grids in the code and phase space density arrays on each of them. %PSD values and grids are stored in the class PSD. The structure of the classes is following:
 *
 * - Matrix3D - basic class describe matrix type of variable to manipulate them. This class has to be considered to understand the logic of the code. 
 * - PSD - class, holds phase space density values and does diffusions.
 * - Grid - grid, in which all that diffusions are (This class stored grid variables). Operate with variables of GridElement class.
 * 		- GridElement. In the code you may find next variables:
 *			- L ( radial distance),
 *			- pc ( momentum * speed of light in units of MeV)
 *			- &alpha; (pitch angle in rad)
 *			- epc ( energy in MeV )
 *		- three GridElement%s (since we have a 3D grid), and fourth one, epc, is a function of pc and is just useful enough to keep it.
 * - BoundaryCondition - upper, lower - upper and lower boundary conditions for diffusion on each GridElement%s
 *
 * PSD as general class, contains methods to compute diffusion. It keeps PSD values in the parent class Matrix3D.
 * Grid has 3-dimensions, so it has 3 GridElement%s, and one additional element epc which is just useful for the output.
 * Each GridElement%s is a 3D-Array of values of coordinates. We have to do it this way, because grids are irregular.
 * BoundaryCondition class is used for the boundary conditions for diffusion calculation.
 *
 * PSD needs diffusion coefficients, grid, and boundary conditions as an input parameter for diffusion functions PSD::Diffusion_L(), PSD::Diffusion_pc(), PSD::Diffusion_alpha(), PSD::Diffusion_pc_alpha().
 * Diffusion coefficients for each diffusion are stored inside DiffusionCoefficientsGroup class. The class structure:
 *
 *  - DiffusionCoefficientsGroup Daa, Dpcpc etc. - summary of diffusion coefficients for one direction. For example, when we have different waves and different Daa coefficients respectively we have to combine them into one diffusion coefficient for the calculation.
 * 		- DiffusionCoefficient *_chorus, *_hiss, *_EMIC etc - diffusion coefficients by each type of wave separately. The coefficients load from file or calculate and store in array of DiffusionCoefficient elements which combine into DiffusionCoefficientsGroup.

 * DiffusionCoefficient%s, produced by each wave type, are grouped by directions in the DiffusionCoefficientsGroup class. That class can turn the waves on and off and scale their effect,
 * and store the summation as Daa or Dpp.
 * 
 * Initial conditions defined by PSD::Create_Initial_PSD.
 *
 * ## Classes constructions.
 * The code has empty constructor and functions, like `AllocateMemory()` and `Initialize()` which define an object. We also have a constructor, which call these functions during execution,
 * just for shortening of the code. And it is like that for almost every class.
 *
 * ## Abbreviations
 * - PSD    - Phase space density.
 * - RCM    - Rice convection model (http://rocco.rice.edu/~toffo/research/rcm/rcm.html - electron seed population)
 * - GMRES  - Generalized minimal residual method. For matrix inversion.
 * - LAPACK - Linear Algebra PACKage. For linear algebra methods. External computation methods that are not to be modified.
 *
 * ## Physics Variables
 * - DLL - Diffusion coefficient for radial
 * - Daa - Pitch-angle D
 * - Dpcpc - Energy D
 * - Dpca - Mix pitch-angle and energy
 * - Lpp - Location of plasma pause
 * - DaaLpp - Pitch-angle with plasma pause
 * - DpcpcLpp - Energy with plasma pause
 * - DpcaLpp - Pitch-angle and energy with plasma pause
 * - Df - radial diffusion coefficient
 * - Kp - scale diffusion coefficient
 * - tau - Lifetime outside of plasmasphere
 * - tauLpp - Lifetime inside of plasmasphere
 * 
 * \todo
 * - specify page on link for parameters
 * - why we need load boundary in two different sections? LoadBoundaryCondition | MakeBoundaryCondition
 * - is MixTerms in a AM_Block method? - Yes
 * - why psdRadialDiffusion - doesn't check for == 0?
 * - why local->radial interpolation after write the Output? -(for debug)
 * - what to exclude from Doxygen about VERB code??
 */

/** 
* \file Main.cpp
* \author Developed by Dmitry Subbotin under supervision of the PI Yuri Shprits.
* \brief This is the main program file.
*
* \details Code below corresponds to the commented part. It works only in uncommented state.
*
* \code
* // \def DEBUG 
* // \brief  Reserved for debug purposes if defined
* // #define DEBUG
* 
* // \def DEBUG_MODE
* // \brief  define DEBUG_MODE value. Checks if memory was alocated for a matrix at each use.
* // \warning Additional checks and outputs for debbugging if defined. MUCH SLOWER, FOR DEBUG ONLY!
* // #define DEBUG_MODE
*
* \endcode
*/

/**
* \def _CRT_SECURE_NO_DEPRECATE 
* \brief  No "depreciated" warnings
*
* \def __VERB_VERSION_NUMBER__ 
* \brief  Number of current VERB code version
*/

#define _CRT_SECURE_NO_DEPRECATE
#define __VERB_VERSION_NUMBER__ "2.3.0.0" // Daniel Choi version
// presious __VERB_VERSION_NUMBER__  2.03.000 is deprecated. New version starts with 2.2.3
// 2.2.0   - Ozeke Dll coefficient [version title never used]
// 2.2.1   - new BF [version title never used]
// 2.2.2   - magnetopause [version title never used]
// 2.2.3   - parametrization of Hiss
// 2.2.3.3 - New expiration date for the code



#include <valarray>

// for using specific mathematical functions and constants
#include <math.h>

// for using time in the code
#include <ctime>
#include <time.h>

// class to make output to the log file and to the screen
#include "../Logging/Output.h"

// various functions (converting, calculation of magnetic field, etc)
#include "../VariousFunctions/variousFunctions.h"

// class operating matrix of phase space density. Also, do diffusion
#include "../Diffusion/PSD.h"

// class initializes the calculation grid and allows to use it
#include "../Grid/Grid.h"

// class load and give as access to boundary conditions
#include "../Grid/BoundaryConditions.h"

// class to operate by sources and losses processes
#include "../Grid/AdditionalSourcesAndLosses.h"

// class load and give an access to parameters from parameter.ini file
#include "../Parameters/Parameters.h"


// Memory leaks ???
#if defined(_WIN32) || defined(_WIN64)
#define _CRTDBG_MAP_ALLOC
#include <stdlib.h>
#include <crtdbg.h>

/** \code
* //* #ifdef _DEBUG
* //* #define DEBUG_NEW new(_NORMAL_BLOCK, __FILE__, __LINE__)
* //* #define new DEBUG_NEW
* //* #endif
* \endcode
*/

#endif


using namespace std;



/// \brief Parameters structure, with all parameters from the parameters.ini file. The default parameters defined in the constructor
Parameters_structure parameters;

/**
 * \brief Main code.
 * \details NOTE: The code bodies displayed below are abbreviations of the functions and variable names for simplicity's sake. To see the full code, look at source code.
 * \param[in] argv[] The name of the parameter file
 * \param argc Number of arguments, mandatory automatic parameter (does not need to be specified).
 */
int main(int argc, char* argv[]) {
	
	//* \deprecated:
	//* memory leaks
	//* _CrtSetDbgFlag ( _CRTDBG_ALLOC_MEM_DF | _CRTDBG_LEAK_CHECK_DF );
	//* _crtBreakAlloc = 24800; // use the number from memory leak debug output to find the leak, if any exists

	// \deprecated:
	// try-catch block for exception handling. Any exception from the code are handled at the end of the block
	try {

		// extern - once again we define the variable parameters (see above)
		// The default parameters will be defined and the constructor of this class
		extern Parameters_structure parameters;

		// code uses 2 grids for calculations Radial and Local diffusion
		Grid
			radialDiffusionGrid, localDiffusionsGrid;

		// for each calculation code uses different PSD objects for Radial and Local
		PSD
			psdRadialDiffusion, psdLocalDiffusions;

		// This is a basic matrix used for check purposes calculation
		// (to check the solution at each time step)
		Matrix3D<double> psdPrevious;

		// Each boundary condition used special class object
		BoundaryCondition
			L_LowerBoundaryCondition, L_UpperBoundaryCondition,
			pc_LowerBoundaryCondition, pc_UpperBoundaryCondition,
			alpha_LowerBoundaryCondition, alpha_UpperBoundaryCondition;

		// Diffusion coefficient for Radial diffusion
		DiffusionCoefficient DLL;

		// Diffusion coefficients for Local diffusion
		DiffusionCoefficientsGroup
			Dpcpc, DpcpcLpp,
			Daa, DaaLpp,
			Dpca, DpcaLpp;

		// Extra losses which could be included in the simulation (t/tau)
		AdditionalSourcesAndLosses SL;

		// Global counter for simulation ???
		double simulation_time = 0;

        //----------------------------------------------------------------------------------------------------
		// Setting up and loading parameters
		//----------------------------------------------------------------------------------------------------

		//=== timer, to measure wall clock time ===
		// This time required for checking expiration time for using the code.
			std::clock_t zero_time, start_time, current_time;
			zero_time = clock();
			time_t now = time(NULL);

			printf("%d\n", localtime(&now)->tm_year);
            // if current year is 2016 or later, needs to be updated
			if (localtime(&now)->tm_year + 1900 >= 2016) {
				cout << "The program version is outdated, please update." << endl;
				return 0;
			}
		//=== end of timer section ===

		// Initialize log-file, save log-file name for future use
		// open log file from [namespace Output defined in Output.h]			
		Output::open_log_file("logfile.log");

		time(&now); // current time into variable now
		Output::echo(0, "Version: %s\n", __VERB_VERSION_NUMBER__);
		Output::echo(0, "Compiled time: %s %s\n", __DATE__, __TIME__); // Predefined compiler macros used to save compilation time in the file - very useful to figure out later which version of the code was used
		Output::echo(0, "Starting at %s\n", ctime(&now) ); // ctime - return char for time

		// if we have more than one argument in calling of main program (that means we have some arguments)
		if (argc > 1) {
			// in this case use first argument as the name of file for loading different variables and parameters
			// Using procedure (function) "Load_parameters", that is described in file "Parameters.cpp" for loading parameters from file
			parameters.Load_parameters(argv[1]);
		} else {
			// in other case try using Parameters.ini as the name of file with variables and constants to load
			parameters.Load_parameters("Parameters.ini");
		}

		// Output debug info or not.
		// parameters.outputLvl - contain a level of debug information
		Output::set_output_lvl(parameters.outputLvl);

		// output1D - file for writing 1d stuff (like Kp, Lpp etc). Open new stream for this filename
		ofstream output1D((parameters.general_Output_parameters.folderName + parameters.general_Output_parameters.fileName1D).c_str());

		//----------------------------------------------------------------------------------------------------
		// Creating grids. Parameters should be loaded from ini file
		//----------------------------------------------------------------------------------------------------
		
        /**
         * \details ## Step 1: Creating Grids
         *
         * Create the radial and local diffusions grids from the given parameters specified by Parameters.ini
         *
         * \code{.cpp}
         * radialDiffusionGrid.Create_Grid(...);
         * localDiffusionsGrid.Create_Grid(...);
         * \endcode
         */
		// Parameters.ini define the Radial grid 
		Output::echo(1, "Creating radial diffusion grid... ");
		radialDiffusionGrid.Create_Grid(parameters.radialDiffusionGrid_L,
					parameters.radialDiffusionGrid_pc,
					parameters.radialDiffusionGrid_alpha,
					parameters.radialDiffusionGrid_epc,
					parameters.radialDiffusionGrid_filename,
					parameters.radialDiffusionGrid_type);

		Output::echo(1, "done.\n");

		// Parameters.ini define the Local grid 
		Output::echo(1, "Creating local diffusions grid... ");
		localDiffusionsGrid.Create_Grid(parameters.localDiffusionsGrid_L,
					parameters.localDiffusionsGrid_pc,
					parameters.localDiffusionsGrid_alpha,
					parameters.localDiffusionsGrid_epc,
					parameters.localDiffusionsGrid_filename,
					parameters.localDiffusionsGrid_type, radialDiffusionGrid);
		Output::echo(1, "done!\n");


		// Output grids
		// After grids defined they write down to file.
		// See Grid::Output
        /**
         * \details Output the grids to plt files for MATLAB to read
         *
         * \code{.cpp}
         * localDiffusionsGrid.Output("./perp_grid.plt");
         * radialDiffusionGrid.Output("./rad_grid.plt");
         * \endcode
         */
		localDiffusionsGrid.Output((parameters.general_Output_parameters.folderName + "perp_grid.plt").c_str());
		radialDiffusionGrid.Output((parameters.general_Output_parameters.folderName + "rad_grid.plt").c_str());


	    //----------------------------------------------------------------------------------------------------
		//  (LoadBoundaryCondition) Setting up boundary conditions that should be loaded from file
		//----------------------------------------------------------------------------------------------------

        /**
         * \details ## Step 2: Set Boundary Conditions 
         * 
         * Loading the boundaries that should be loaded from a file\n
         * Check the parameters, if the type of boundary is defined we can load the boundary.\n
         * Else the default value for boundary type is BCT_CONSTANT_PSD\n
         * The boundary defined only on the slice of the grid\n
         * See Matrix3D::xSlice and Matrix3D::ySlice
         *
         * \code{.cpp}
         * if (parameters.L_UpperBoundaryCondition.type == "BCT_FILE" || "BCT_FILE_GRID" || "BCT_MULTIPLE_FILES") {
         *      L.UpperBoundaryCondition.LoadBoundaryCondition(
         *          parameters.L_UpperBoundaryCondition,
         *          radialDiffusionGrid.epc.arr.xSlice(radialDiffusionGrid.L.size-1),
         *          radialDiffusionGrid.alpha.arr.xSlice(radialDiffusionGrid.L.size-1));
         * }
         * if (parameters.pc_LowerBoundaryCondition.type == "BCT_FILE" || "BCT_FILE_GRID" || "BCT_MULTIPLE_FILES") {
         *      pc_LowerBoundaryCondition.LoadBoundaryCondition(
         *          parameters.pc_LowerBoundaryCondition,
         *          localDiffusionsGrid.L.arr.ySlice(0),
         *          localDiffusionsGrid.alpha.arr.ySlice(0));
         * }
         * \endcode
         */

		if (parameters.L_UpperBoundaryCondition.type == "BCT_FILE" || parameters.L_UpperBoundaryCondition.type == "BCT_FILE_GRID" || parameters.L_UpperBoundaryCondition.type == "BCT_MULTIPLE_FILES") {
			L_UpperBoundaryCondition.LoadBoundaryCondition(
				parameters.L_UpperBoundaryCondition,
				radialDiffusionGrid.epc.arr.xSlice(radialDiffusionGrid.L.size-1),
				radialDiffusionGrid.alpha.arr.xSlice(radialDiffusionGrid.L.size-1));
		}
		if (parameters.pc_LowerBoundaryCondition.type == "BCT_FILE" || parameters.pc_LowerBoundaryCondition.type == "BCT_FILE_GRID" || parameters.pc_LowerBoundaryCondition.type == "BCT_MULTIPLE_FILES") {
			pc_LowerBoundaryCondition.LoadBoundaryCondition(
				parameters.pc_LowerBoundaryCondition,
				localDiffusionsGrid.L.arr.ySlice(0),
				localDiffusionsGrid.alpha.arr.ySlice(0));
		}

        //----------------------------------------------------------------------------------------------------
		// (Radial) Initializing Phase Space Densities AND loading initial values
		//----------------------------------------------------------------------------------------------------

		// see PSD::Create_Initial_PSD for details
		Output::echo(1, "Radial diffusion initialization... ");
        /**
         * \details create initial radial diffusion PSD
         *
         * \code{.cpp}
         * psdRadialDiffusion.Create_Initial_PSD();
         * \endcode
         */
		psdRadialDiffusion.Create_Initial_PSD(parameters.psdRadialDiffusion, radialDiffusionGrid, L_UpperBoundaryCondition);
		Output::echo(1, "done!\n");

	    //----------------------------------------------------------------------------------------------------
		//  (MakeBoundaryCondition) Setting up boundary conditions that should be set up using PSD values
		//----------------------------------------------------------------------------------------------------


		// Making boundary conditions. Same as before
		// Parameters are PSD and grid values at the boundary, made by 'Slice' procedure
        /**
         * \details making boundary conditions using corresponding slice method for L, pc, and alpha
         * where GridElement represents L, pc, or alpha, and GridElementSlice represents xSlice, ySlice, and zSlice respectively
         *
         * \code{.cpp}
         * GridElement_LowerBoundaryCondition.MakeBoundaryCondition(..., GridElementSlice(0));
         * GridElement_UpperBoundaryCondition.MakeBoundaryCondition(..., GridElementSlice(size-1));
         * \endcode
         */
		Output::echo(1, "Making boundary conditions... ");
		L_LowerBoundaryCondition.MakeBoundaryCondition(
			parameters.L_LowerBoundaryCondition,
			psdRadialDiffusion.arr.xSlice(0),
			radialDiffusionGrid.epc.arr.xSlice(0),
			radialDiffusionGrid.alpha.arr.xSlice(0));
		L_UpperBoundaryCondition.MakeBoundaryCondition(
			parameters.L_UpperBoundaryCondition,
			psdRadialDiffusion.arr.xSlice(radialDiffusionGrid.L.size-1),
			radialDiffusionGrid.epc.arr.xSlice(radialDiffusionGrid.L.size-1),
			radialDiffusionGrid.alpha.arr.xSlice(radialDiffusionGrid.L.size-1));

		// psdRadialDiffusion everywhere because we don't have local PSD yet
		pc_LowerBoundaryCondition.MakeBoundaryCondition(
			parameters.pc_LowerBoundaryCondition,
			psdRadialDiffusion.arr.ySlice(0),
			localDiffusionsGrid.L.arr.ySlice(0),
			localDiffusionsGrid.alpha.arr.ySlice(0));
		pc_UpperBoundaryCondition.MakeBoundaryCondition(
			parameters.pc_UpperBoundaryCondition,
			psdRadialDiffusion.arr.ySlice(radialDiffusionGrid.pc.size-1),
			localDiffusionsGrid.L.arr.ySlice(radialDiffusionGrid.pc.size-1),
			localDiffusionsGrid.alpha.arr.ySlice(radialDiffusionGrid.pc.size-1));

		alpha_LowerBoundaryCondition.MakeBoundaryCondition(
			parameters.alpha_LowerBoundaryCondition,
			psdRadialDiffusion.arr.zSlice(0),
			localDiffusionsGrid.L.arr.zSlice(0),
			localDiffusionsGrid.epc.arr.zSlice(0));
		alpha_UpperBoundaryCondition.MakeBoundaryCondition(
			parameters.alpha_UpperBoundaryCondition,
			psdRadialDiffusion.arr.zSlice(radialDiffusionGrid.alpha.size-1),
			localDiffusionsGrid.L.arr.zSlice(radialDiffusionGrid.alpha.size-1),
			localDiffusionsGrid.epc.arr.zSlice(radialDiffusionGrid.alpha.size-1));
		Output::echo(1, "done!\n");

		//----------------------------------------------------------------------------------------------------
		// (Local) Initializing Phase Space Densities AND loading initial values
		//----------------------------------------------------------------------------------------------------

        /**
         * \details create local diffusions PSD, and interpolate them with radial PSD
         *
         * \code{.cpp}
         * psdLocalDiffusions.Create_Initial_PSD();
         * psdLocalDiffusions.Interpolate();
         * \endcode
         */
		Output::echo(1, "Local diffusions initialization... ");
		// Create local diffusion PSD
		// see PSD::Create_Initial_PSD for details
		psdLocalDiffusions.Create_Initial_PSD(parameters.psdLocalDiffusions, localDiffusionsGrid, L_UpperBoundaryCondition);
		// Radial to Local grids initial PSD interpolation. It is necessary to have same PSD on different grids
		psdLocalDiffusions.Interpolate(psdRadialDiffusion, parameters.interpolation, radialDiffusionGrid, localDiffusionsGrid, pc_LowerBoundaryCondition.arr, pc_UpperBoundaryCondition.arr);
		Output::echo(1, "done!\n");

		//----------------------------------------------------------------------------------------------------
		// Initializing PSD complete
		//----------------------------------------------------------------------------------------------------

		// After grids defined we can work with Matrix3d::psdPrevious required for check calculation
		psdPrevious.AllocateMemory(localDiffusionsGrid.L.size, localDiffusionsGrid.pc.size, localDiffusionsGrid.alpha.size);

		// First data writing for output files.
		// output initial conditions into the file as "picture" at the 0-step
		psdLocalDiffusions.Output_without_grid(0);
		psdRadialDiffusion.Output_without_grid(0);

		// Create any additional sources and losses (load from file or something).
		// Right now it's empty, but is useful for future development of the code.
		SL.Create_SL(parameters.SL, localDiffusionsGrid, parameters.totalIterationsNumber, parameters.timeStep);

	    //----------------------------------------------------------------------------------------------------
		// Creating diffusion coefficients
		//----------------------------------------------------------------------------------------------------
		
        /**
         * \details create and obtain all diffusion coefficients, stored into function arguments
         *
         * \code{.cpp}
         * CreateAllDiffusionCoefficients(DLL, Daa, Dpcpc, Dpca, DaaLpp, DpcpcLpp, DpcaLpp, parameters, radialDiffusionGrid, localDiffusionsGrid);
         * \endcode
         */
		// see DiffusionCoefficient.h
		CreateAllDiffusionCoefficients(DLL, Daa, Dpcpc, Dpca, DaaLpp, DpcpcLpp, DpcaLpp, parameters, radialDiffusionGrid, localDiffusionsGrid);

		//----------------------------------------------------------------------------------------------------
		// Writing header into the 1d output file
		//----------------------------------------------------------------------------------------------------
		// Writing headers for all diffusion coefficients scaling coefficients
		Output1DHeaders(output1D, Daa, Dpcpc, Dpca, DaaLpp, DpcpcLpp, DpcaLpp);
		// write information about 1d variables into the file with 1d variables
		Output1DValues(output1D, Daa, Dpcpc, Dpca, DaaLpp, DpcpcLpp, DpcaLpp, simulation_time, parameters, 0);

	    //----------------------------------------------------------------------------------------------------
		//
		//  Main loop. Making loop from 1 to total number of iterations
		//
		//----------------------------------------------------------------------------------------------------
		
        
		// You don't believe, but there are counters for loops
		// il - for L, im - for pc, ia - for alpha
		int il, im, ia;

		Output::echo(0, "Starting...\n");
		// totalIterationsNumber - calculated while parameters loaded
		for (int iteration = 1; iteration < parameters.totalIterationsNumber; iteration++) {
			// calculating time (in days) by multiplication time step (dt) to iteration number (it)
			simulation_time = parameters.timeStep * iteration;

			// Output information about current calculations on screen
			Output::echo(1, "Step %d of %d. Time = %d days, %.0f hours.\n", iteration, parameters.totalIterationsNumber, int(simulation_time), (simulation_time-int(simulation_time))*24);

			//----------------------------------------------------------------------------------------------------
			// Radial diffusion
			//----------------------------------------------------------------------------------------------------
			if (radialDiffusionGrid.L.size > 1) {

				// Update boundary conditions, if needed see BoundaryCondition::Update
                /**
                 * \details Update the boundary conditions for L if need be
                 *
                 * \code{.cpp}
                 * L_LowerBoundaryCondition.Update();
                 * L_UpperBoundaryCondition.Update();
                 * \endcode
                 */
				Output::echo(3, "Updating L-boundaries...\n");
				L_LowerBoundaryCondition.Update(iteration, psdRadialDiffusion.arr.xSlice(0));
				L_UpperBoundaryCondition.Update(iteration, psdRadialDiffusion.arr.xSlice(radialDiffusionGrid.L.size-1), simulation_time, parameters.timeStep);
				Output::echo(3, "Done.\n");

                /**
                 * \details Update radial diffusion coefficients if using radial diffusion
                 *
                 * \code{.cpp}
                 * if (parameters.useRadialDiffusion) {
                 *      DLL.MakeDLL(L, pc, alpha, Kp, DLLType);
                 * }
                 * \endcode
                 */
				// Update DLL. DLL calculates on each step by defined formula or other method (see DiffusionCoefficients::MakeDLL)
				if (parameters.useRadialDiffusion) {
					Output::echo(3, "Updating radial diffusion coefficient...\n");
						DLL.MakeDLL(radialDiffusionGrid.L, radialDiffusionGrid.pc, radialDiffusionGrid.alpha, parameters.Kp[iteration], parameters.DLLType);
					Output::echo(3, "Done (Max(DLL) = %e).\n", DLL.arr.max());
				}

				// If Debug folder exists - put calculated DLL into file
				// Why it moved away of the if(parameters.useRadialDiffusion) ???
				DLL.arr.writeToFile("./Debug/DLL.dat", radialDiffusionGrid.L.arr, radialDiffusionGrid.epc.arr, radialDiffusionGrid.alpha.arr);

				if (parameters.useRadialDiffusion) Output::echo(3, "Calculating radial diffusion... \n");

				//---------------------------------
				// Radial diffusion PSD calculation
                //---------------------------------
                
                /**
                 * \details ## Step 3: Calculate Radial Diffusion PSD
                 *
                 * Calculate diffusion for radial distance, given diffusion coefficients, grid elements, and boundary conditions (including boundary flux)\n
                 * check PSD::Diffusion_L for implementation
                 * 
                 * \code{.cpp}
                 * psdRadialDiffusion.Diffusion_L(timeStep, Lpp[iteration], DLL, L, pc, alpha, Jacobian, LowerBoundaryCondition.arr, UpperBoundaryCondition.arr * Bf[iteration], LowerBoundaryCondition.calculationType, UpperBoundaryCondition.calculationType, tau[iteration], tauLpp[iteration]);
                 * \endcode
                 */
				psdRadialDiffusion.Diffusion_L(parameters.timeStep, parameters.Lpp[iteration],
					DLL,
					radialDiffusionGrid.L, radialDiffusionGrid.pc, radialDiffusionGrid.alpha,  radialDiffusionGrid.Jacobian,
					L_LowerBoundaryCondition.arr,
					L_UpperBoundaryCondition.arr * parameters.Bf[iteration],
					L_LowerBoundaryCondition.calculationType,
					L_UpperBoundaryCondition.calculationType,
					parameters.tau[iteration],
					parameters.tauLpp[iteration]);			
					/**
                     * \details
					 * How to calculate BF - it is a boundary condition file
                     *
					 * μ = μ(L\_boundary)
                     *
					 * \f$ Bf = \frac{Flux-data(L\_boundary, μ)}{ pc(L\_boundary, μ) } / [ \frac{J\_L7}{pc(L=7, μ)} * \frac{steady\_state(L\_boundary)}{steady\_state(L=7)} ] \f$
					 */
				if (parameters.useRadialDiffusion) Output::echo(3, "Done. \n");

				// Interpolation from radial diffusion grid onto local diffusion grid
				// radial->local
				Output::echo(3, "Interpolation radial->local... \n");

				//* For time-dependent boundary conditions (used for RCM coupling) use next
				//* *.arr.xSlice(iteration) create a 2D array of boundary values at the current time
				//* psdLocalDiffusions.Interpolate(psdRadialDiffusion, parameters.interpolation, radialDiffusionGrid, localDiffusionsGrid, pc_LowerBoundaryCondition.arr.xSlice(iteration), pc_UpperBoundaryCondition.arr.xSlice(iteration));

				//* Constant in time boundary conditions (a typical simulation)
				//* psdLocalDiffusions.Interpolate(psdRadialDiffusion, parameters.interpolation, radialDiffusionGrid, localDiffusionsGrid, pc_LowerBoundaryCondition.arr.xSlice(0), pc_UpperBoundaryCondition.arr.xSlice(0));

				// Constant in time low-energy boundary conditions, but flexible high-Energy boundary.
				// Created to fix interpolation problems.
                /**
                 * \details Interpolate from radial to local diffusion grids so that captures upper boundary condition with ySlice max size
                 *
                 * \code{.cpp}
                 * psdLocalDiffusions.Interpolate(..., psdRadialDiffusion.arr.ySlice(radialDiffusionGrid.pc.size-1));
                 * \endcode
                 */
				psdLocalDiffusions.Interpolate(psdRadialDiffusion, parameters.interpolation, radialDiffusionGrid, localDiffusionsGrid, pc_LowerBoundaryCondition.arr, psdRadialDiffusion.arr.ySlice(radialDiffusionGrid.pc.size-1));
				Output::echo(3, "Done. \n");
			} // end if for L-grid size checking > 1;

            //----------------------------------------------------------------------------------------------------
			// Local Diffusion
            //----------------------------------------------------------------------------------------------------

			// Update everything (analogy of radial diffusion)
			// Update boundary conditions (see BoundaryCondition::Update)
            /**
             * \details Update everything: pc-boundaries, alpha-boundaries, diffusion coefficients
             *
             * \code{.cpp}
             * if (localDiffusionsGrid.pc.size > 1) {
             *      pc_LowerBoundaryCondition.Update(iteration, psdLocalDiffusions.arr.ySlice(0));
             *      pc_UpperBoundaryCondition.Update(iteration, psdLocalDiffusions.arr.ySlice(radialDiffusionGrid.pc.size-1));
             * }
             * if (localDiffusionsGrid.alpha.size > 1) {
             *      alphaLower...Update(zSlice(0));
             *      alphaUpper...Update(zSlice(...);
             * }
             * if (parameters.useEnergyDiffusion) {
             *      Dpcpc.ActivateAndScale(simulation_time, parameters.Kp[iteration]);
             *      DpcpcLpp.ActivateAndScale(...);
             * }
             * if (parameters.useAlphaDiffusion) {
             *      Daa...;
             *      DaaLpp...;
             * }
             * if (parameters.useAlphaDiffusion && parameters.useEnergyDiffusion) {
             *      Dpca...;
             *      DpcaLpp...;
             * }
             * \endcode
             */
			if (localDiffusionsGrid.pc.size > 1) {
				Output::echo(3, "Updating pc-boundaries... \n");
				pc_LowerBoundaryCondition.Update(iteration, psdLocalDiffusions.arr.ySlice(0));
				pc_UpperBoundaryCondition.Update(iteration, psdLocalDiffusions.arr.ySlice(radialDiffusionGrid.pc.size-1));
				Output::echo(3, "Done. \n");
			}
			if (localDiffusionsGrid.alpha.size > 1) {
				Output::echo(3, "Updating alpha-boundaries... \n");
				alpha_LowerBoundaryCondition.Update(iteration, psdLocalDiffusions.arr.zSlice(0));
				alpha_UpperBoundaryCondition.Update(iteration, psdLocalDiffusions.arr.zSlice(radialDiffusionGrid.alpha.size-1));
				Output::echo(3, "Done. \n");
			}

			// Update (scale) diffusion coefficients (see DiffusionCoefficientsGroup::ActivateAndScale)
			// ??? Dpcpc.CurrentDxx.arr.max(), DpcpcLpp.CurrentDxx.arr.max() -- %%
			if (parameters.useEnergyDiffusion) {				
				Output::echo(3, "Updating momentum diffusion coefficients...\n");
				Dpcpc.ActivateAndScale(simulation_time, parameters.Kp[iteration]);
				DpcpcLpp.ActivateAndScale(simulation_time, parameters.Kp[iteration]);
				Output::echo(3, "Done (Max(Dpp) = %e / Max(DppLpp) = %e).\n", Dpcpc.CurrentDxx.arr.max(), DpcpcLpp.CurrentDxx.arr.max());
			}
			if (parameters.useAlphaDiffusion) {
				Output::echo(3, "Updating pitch-angle diffusion coefficients...\n");
				Daa.ActivateAndScale(simulation_time, parameters.Kp[iteration]);
				DaaLpp.ActivateAndScale(simulation_time, parameters.Kp[iteration]);
				Output::echo(3, "Done (Max(Daa) = %e / Max(DaaLpp) = %e).\n", Daa.CurrentDxx.arr.max(), DaaLpp.CurrentDxx.arr.max());
			}
			if (parameters.useAlphaDiffusion && parameters.useEnergyDiffusion) {
				Output::echo(3, "Updating mixed diffusion coefficients... \n");
				Dpca.ActivateAndScale(simulation_time, parameters.Kp[iteration]);
				DpcaLpp.ActivateAndScale(simulation_time, parameters.Kp[iteration]);
				Output::echo(3, "Done (Max(Dpa) = %e / Max(DpaLpp) = %e).\n", Dpca.CurrentDxx.arr.max(), DpcaLpp.CurrentDxx.arr.max());
			}


			//---------------------------------
			// Local diffusion PSD calculation
            //---------------------------------
			// Local diffusions according to derivatives approximation method (block or split simply)
            /**
             * \details ## Step 4: Calculate Local Diffusion PSD
             *
             * calculate Diffusion for pc and alpha if approximating with Block method\n
             * checks to see that grid is 3D by checking pc and alpha size\n
             * see PSD::Diffusion_pc_alpha for details on matrix calculations
             * 
             * \code{.cpp}
             * if (parameters.psdLocalDiffusions.approximationMethod == "AM_Block_C" || "AM_Block_LR") {
             *      if (localDiffusionsGrid.pc.size > 1 && localDiffusionsGrid.alpha.size > 1) {
             *          psdLocalDiffusions.Diffusion_pc_alpha(...);
             *      }
             * }
             * \endcode
             */
			if (parameters.psdLocalDiffusions.approximationMethod == "AM_Block_C" || parameters.psdLocalDiffusions.approximationMethod == "AM_Block_LR") {
			// "_C" and "_LR" are "central" and "left-right" approximation methods

					if (localDiffusionsGrid.pc.size > 1 && localDiffusionsGrid.alpha.size > 1) {
						Output::echo(3, "Calculating local diffusions using BLOCK method... \n");
						// local diffusions Momentum + Pitch angle
						// ??? mixterms dosn't work here at all ???
						psdLocalDiffusions.Diffusion_pc_alpha(parameters.timeStep,
								parameters.Lpp[iteration],
								Dpcpc.CurrentDxx, DpcpcLpp.CurrentDxx,
								Daa.CurrentDxx, DaaLpp.CurrentDxx,
								Dpca.CurrentDxx, DpcaLpp.CurrentDxx,
								localDiffusionsGrid.L, localDiffusionsGrid.pc, localDiffusionsGrid.alpha, localDiffusionsGrid.Jacobian,
								pc_LowerBoundaryCondition.arr,
								pc_UpperBoundaryCondition.arr,
								alpha_LowerBoundaryCondition.arr,
								alpha_UpperBoundaryCondition.arr,
								pc_LowerBoundaryCondition.calculationType,
								pc_UpperBoundaryCondition.calculationType,
								alpha_LowerBoundaryCondition.calculationType,
								alpha_UpperBoundaryCondition.calculationType);
						Output::echo(3, "Done. \n");

						if (iteration == 1 && parameters.outputModelMatrix) { // Id Debug folder exists and outputModelMatrix - write down the Matrix from the Local diffusion computation
							psdLocalDiffusions.matr_A.writeToFile("./Debug/Matr_A.dat");
							psdLocalDiffusions.matr_B.writeToFile("./Debug/Matr_B.dat");
							psdLocalDiffusions.matr_C.writeToFile("./Debug/Matr_C.dat");
						}
					} // grid check

			} // approximationMethod
            /**
             * \details calculate Diffusion for pc, alpha, or MixedTerms if approximating with Split method
             * checks to see that grid has dimension in respective element
             * see PSD::Diffusion_pc, PSD::Diffusion_alpha, and PSD::DiffusionMixTermExplicit for details on matrix calculations
             *
             * \code{.cpp}
             * if (parameters.psdLocalDiffusions.approximationMethod == "AM_Split_C" || "AM_Split_LR" || "AM_Split_C_oldway" || "AM_Split_LR_oldway") {
             *      if (parameters.useEnergyDiffusion && localDiffusionsGrid.pc.size > 1) {
             *          psdLocalDiffusions.Diffusion_pc(...);
             *      }
             *      if (useAlphaDiffusion && alpha.size > 1) {
             *          Diffusion_alpha(...);
             *      }
             *      if (useEnergyAlphaMixedTerms && pc.size > 1 && alpha.size > 1) {
             *          DiffusionMixTermExplicit(...);
             *      }
             * }
             * \endcode
             */
            else if (parameters.psdLocalDiffusions.approximationMethod == "AM_Split_C"
					|| parameters.psdLocalDiffusions.approximationMethod == "AM_Split_LR"
					|| parameters.psdLocalDiffusions.approximationMethod == "AM_Split_C_oldway"
					|| parameters.psdLocalDiffusions.approximationMethod == "AM_Split_LR_oldway") {

					// local diffusion
					// Momentum diffusion					
					if (parameters.useEnergyDiffusion && localDiffusionsGrid.pc.size > 1) {
						Output::echo(3, "Calculating energy diffusions using SPLIT method... \n");
						psdLocalDiffusions.Diffusion_pc(parameters.timeStep,
												parameters.Lpp[iteration],
												Dpcpc.CurrentDxx, DpcpcLpp.CurrentDxx,
												localDiffusionsGrid.L, localDiffusionsGrid.pc, localDiffusionsGrid.alpha, localDiffusionsGrid.Jacobian,
												pc_LowerBoundaryCondition.arr,
												pc_UpperBoundaryCondition.arr,
												pc_LowerBoundaryCondition.calculationType,
												pc_UpperBoundaryCondition.calculationType);
						Output::echo(3, "Done. \n");

						if (iteration == 1 && parameters.outputModelMatrix) { /* TBD: Model matrix output here */ }
					} // grid check

					// local diffusion
					// Pitch angle diffusion
					if (parameters.useAlphaDiffusion && localDiffusionsGrid.alpha.size > 1) {
						Output::echo(3, "Calculating pitch angle diffusions using SPLIT method... \n");
						psdLocalDiffusions.Diffusion_alpha(parameters.timeStep,
												parameters.Lpp[iteration],
												Daa.CurrentDxx, DaaLpp.CurrentDxx,
												localDiffusionsGrid.L, localDiffusionsGrid.pc, localDiffusionsGrid.alpha, localDiffusionsGrid.Jacobian,
												alpha_LowerBoundaryCondition.arr,
												alpha_UpperBoundaryCondition.arr,
												alpha_LowerBoundaryCondition.calculationType,
												alpha_UpperBoundaryCondition.calculationType);
						Output::echo(3, "Done. \n");

						if (iteration == 1 && parameters.outputModelMatrix) { /* TBD: Model matrix output here */ }
					} // grid check

					// Explicit mixed terms
					if (parameters.useEnergyAlphaMixedTerms && localDiffusionsGrid.pc.size > 1 && localDiffusionsGrid.alpha.size > 1) {

						Output::echo(3, "Calculating mixed terms EXPLICITLY ... \n");
						psdLocalDiffusions.DiffusionMixTermExplicit(parameters.timeStep,
							parameters.Lpp[iteration],
							Dpca.CurrentDxx, DpcaLpp.CurrentDxx,
							localDiffusionsGrid.L, localDiffusionsGrid.pc, localDiffusionsGrid.alpha, localDiffusionsGrid.Jacobian,
							pc_LowerBoundaryCondition.arr,
							pc_UpperBoundaryCondition.arr,
							alpha_LowerBoundaryCondition.arr,
							alpha_UpperBoundaryCondition.arr,
							pc_LowerBoundaryCondition.calculationType,
							pc_UpperBoundaryCondition.calculationType,
							alpha_LowerBoundaryCondition.calculationType,
							alpha_UpperBoundaryCondition.calculationType);
						Output::echo(3, "Done. \n");
					} // grid check

			}  // approximationMethod
			else if (parameters.psdLocalDiffusions.approximationMethod == "AM_Kyung-Chan") { // Undocumented Kyung-Chan method

				if (localDiffusionsGrid.pc.size > 1 && localDiffusionsGrid.alpha.size > 1) {
					Output::echo(3, "Calculating local diffusions using Kyung-Chan's approximation... \n");
					// do local diffusions
					psdLocalDiffusions.Diffusion_pc_alpha_KC(parameters.timeStep,
							parameters.Lpp[iteration],
							Dpcpc.CurrentDxx, DpcpcLpp.CurrentDxx,
							Daa.CurrentDxx, DaaLpp.CurrentDxx,
							Dpca.CurrentDxx, DpcaLpp.CurrentDxx,
							localDiffusionsGrid.L, localDiffusionsGrid.pc, localDiffusionsGrid.alpha, localDiffusionsGrid.Jacobian,
							pc_LowerBoundaryCondition.arr,
							pc_UpperBoundaryCondition.arr,
							alpha_LowerBoundaryCondition.arr,
							alpha_UpperBoundaryCondition.arr,
							pc_LowerBoundaryCondition.calculationType,
							pc_UpperBoundaryCondition.calculationType,
							alpha_LowerBoundaryCondition.calculationType,
							alpha_UpperBoundaryCondition.calculationType);
					Output::echo(3, "Done. \n");

					if (iteration == 1 && parameters.outputModelMatrix) { // Id Debug folder exists and outputModelMatrix - write down the Matrix from the Local diffusion computation
						psdLocalDiffusions.matr_A.writeToFile("./Debug/Matr_A.dat");
						psdLocalDiffusions.matr_B.writeToFile("./Debug/Matr_B.dat");
						psdLocalDiffusions.matr_C.writeToFile("./Debug/Matr_C.dat");
					}
				} // grid check

			} else {
					throw error_msg("APPROX_ERROR", "Approximation unknown");
			} // end of  if(approximationMethod)

			
			//-----------------
			// Sources & Losses
            //-----------------
			// Loss-Cone losses, electron lifetime losses (if any), sources from coupling with RCM, etc
			
            /**
             * \details calculate and update sources and losses
             *
             * \code{.cpp}
             * SL.Update(iteration, parameters.SL, localDiffusionsGrid);
             * psdLocalDiffusions.SourcesAndLosses(...);
             * \endcode
             */
			SL.Update(iteration, parameters.SL, localDiffusionsGrid); // update array of SL

			psdLocalDiffusions.SourcesAndLosses(
					parameters,
					localDiffusionsGrid.L, localDiffusionsGrid.pc, localDiffusionsGrid.alpha,
					SL,
					parameters.timeStep,
					parameters.Lpp[iteration],
					parameters.tau[iteration],
					parameters.tauLpp[iteration], parameters.Kp[iteration]);


			//----------------------------------------------------------------------------------------------------
			//  Make PSD positive
			//----------------------------------------------------------------------------------------------------
            /**
             * \details checks for negative PSD values, and sets them to floating point zero
             * 
             * \code{.cpp}
             * if (psdLocalDiffusions.arr[il][im][ia] < VC::zero_f) {
             *      psdLocalDiffusions.arr[il][im][ia] = VC::zero_f;
             * }
             * \endcode
             */
			if (parameters.NoNegative) // This check needs for correct strange calculation when PSD === 0
				for (il = 0; il < localDiffusionsGrid.L.size; il++)
					for (im = 0; im < localDiffusionsGrid.pc.size; im++)
						for (ia = 0; ia < localDiffusionsGrid.alpha.size; ia++)
							if (psdLocalDiffusions.arr[il][im][ia] < VC::zero_f)
									psdLocalDiffusions.arr[il][im][ia] = VC::zero_f; // keep negative value to see where the negative values appears...

			//----------------------------------------------------------------------------------------------------
			//  Checking the answer (probably pretty meaningless, since we check the inversion of the model matrix)
			//----------------------------------------------------------------------------------------------------
			// ??? \deprecated ?
			if (false) {
				// check the local solution
				// error = f(t+1) - f(t) - dt* d/dp Dpp d/dp f - dt*d/da Daa d/da f - dt* d/dp Dpa d/da f - dt*d/da Dap d/dp f - dt*f/tau
				static Matrix3D<double> PSD_Der_a_L(localDiffusionsGrid.L.size, localDiffusionsGrid.pc.size, localDiffusionsGrid.alpha.size);
				static Matrix3D<double> PSD_Der_a_R(localDiffusionsGrid.L.size, localDiffusionsGrid.pc.size, localDiffusionsGrid.alpha.size);
				static Matrix3D<double> PSD_Der_m_L(localDiffusionsGrid.L.size, localDiffusionsGrid.pc.size, localDiffusionsGrid.alpha.size);
				static Matrix3D<double> PSD_Der_m_R(localDiffusionsGrid.L.size, localDiffusionsGrid.pc.size, localDiffusionsGrid.alpha.size);

				static Matrix3D<double> error(localDiffusionsGrid.L.size, localDiffusionsGrid.pc.size, localDiffusionsGrid.alpha.size);

				// Some variables
				double LaaLR, LaaRL, LmmLR, LmmRL;
				double LmaLL, LmaLR, LmaRL, LmaRR;
				double LamLL, LamLR, LamRL, LamRR;
				double taulc;

				error = 0;
				double maxerror = 0, aveerror = 0;

				for (il = 0; il < localDiffusionsGrid.L.size; il++) {
					for (im = 0; im < localDiffusionsGrid.pc.size; im++) {
						ia = 0;
						PSD_Der_a_R[il][im][ia] = localDiffusionsGrid.Jacobian[il][im][ia] * Dpca.CurrentDxx.arr[il][im][ia] * (psdLocalDiffusions.arr[il][im][ia+1] - psdLocalDiffusions.arr[il][im][ia])/(localDiffusionsGrid.alpha.arr[il][im][ia+1] - localDiffusionsGrid.alpha.arr[il][im][ia]);
						for (ia = 1; ia < localDiffusionsGrid.alpha.size-1; ia++) {
							PSD_Der_a_L[il][im][ia] = localDiffusionsGrid.Jacobian[il][im][ia] * Dpca.CurrentDxx.arr[il][im][ia] * (psdLocalDiffusions.arr[il][im][ia] - psdLocalDiffusions.arr[il][im][ia-1])/(localDiffusionsGrid.alpha.arr[il][im][ia] - localDiffusionsGrid.alpha.arr[il][im][ia-1]);
							PSD_Der_a_R[il][im][ia] = localDiffusionsGrid.Jacobian[il][im][ia] * Dpca.CurrentDxx.arr[il][im][ia] * (psdLocalDiffusions.arr[il][im][ia+1] - psdLocalDiffusions.arr[il][im][ia])/(localDiffusionsGrid.alpha.arr[il][im][ia+1] - localDiffusionsGrid.alpha.arr[il][im][ia]);
						}
						ia = localDiffusionsGrid.alpha.size-1;
						PSD_Der_a_L[il][im][ia] = localDiffusionsGrid.Jacobian[il][im][ia] * Dpca.CurrentDxx.arr[il][im][ia] * (psdLocalDiffusions.arr[il][im][ia] - psdLocalDiffusions.arr[il][im][ia-1])/(localDiffusionsGrid.alpha.arr[il][im][ia] - localDiffusionsGrid.alpha.arr[il][im][ia-1]);
					}
					for (ia = 0; ia < localDiffusionsGrid.alpha.size; ia++) {
						im = 0;
						PSD_Der_m_R[il][im][ia] = localDiffusionsGrid.Jacobian[il][im][ia] * Dpca.CurrentDxx.arr[il][im][ia]*(psdLocalDiffusions.arr[il][im+1][ia] - psdLocalDiffusions.arr[il][im][ia])/(localDiffusionsGrid.pc.arr[il][im+1][ia] - localDiffusionsGrid.pc.arr[il][im][ia]);
						for (im = 1; im < localDiffusionsGrid.pc.size-1; im++) {
							PSD_Der_m_L[il][im][ia] = localDiffusionsGrid.Jacobian[il][im][ia] * Dpca.CurrentDxx.arr[il][im][ia]*(psdLocalDiffusions.arr[il][im][ia] - psdLocalDiffusions.arr[il][im-1][ia])/(localDiffusionsGrid.pc.arr[il][im][ia] - localDiffusionsGrid.pc.arr[il][im-1][ia]);
							PSD_Der_m_R[il][im][ia] = localDiffusionsGrid.Jacobian[il][im][ia] * Dpca.CurrentDxx.arr[il][im][ia]*(psdLocalDiffusions.arr[il][im+1][ia] - psdLocalDiffusions.arr[il][im][ia])/(localDiffusionsGrid.pc.arr[il][im+1][ia] - localDiffusionsGrid.pc.arr[il][im][ia]);
						}
						im = localDiffusionsGrid.pc.size-1;
						PSD_Der_m_L[il][im][ia] = localDiffusionsGrid.Jacobian[il][im][ia] * Dpca.CurrentDxx.arr[il][im][ia]*(psdLocalDiffusions.arr[il][im][ia] - psdLocalDiffusions.arr[il][im-1][ia])/(localDiffusionsGrid.pc.arr[il][im][ia] - localDiffusionsGrid.pc.arr[il][im-1][ia]);
					}
				}

				for (il = 0; il < localDiffusionsGrid.L.size; il++) {
					for (im = 1; im < localDiffusionsGrid.pc.size-1; im++) {
						for (ia = 1; ia < localDiffusionsGrid.alpha.size-1; ia++) {

							// dda D d/da f
							LaaLR = 1.0/(localDiffusionsGrid.Jacobian[il][im][ia]) *
								(PSD_Der_a_R[il][im][ia] - PSD_Der_a_R[il][im][ia-1])/
								(localDiffusionsGrid.alpha.arr[il][im][ia] - localDiffusionsGrid.alpha.arr[il][im][ia-1]);
							LaaRL = 1.0/(localDiffusionsGrid.Jacobian[il][im][ia]) *
								(PSD_Der_a_L[il][im][ia+1] - PSD_Der_a_L[il][im][ia])/
								(localDiffusionsGrid.alpha.arr[il][im][ia+1] - localDiffusionsGrid.alpha.arr[il][im][ia]);

							// dda D d/da f
							LmmLR = 1.0/(localDiffusionsGrid.Jacobian[il][im][ia]) *
								(PSD_Der_m_R[il][im][ia] - PSD_Der_m_R[il][im-1][ia])/
								(localDiffusionsGrid.pc.arr[il][im][ia] - localDiffusionsGrid.pc.arr[il][im-1][ia]);
							LmmRL = 1.0/(localDiffusionsGrid.Jacobian[il][im][ia]) *
								(PSD_Der_m_L[il][im+1][ia] - PSD_Der_m_L[il][im][ia])/
								(localDiffusionsGrid.pc.arr[il][im+1][ia] - localDiffusionsGrid.pc.arr[il][im][ia]);

							// d/dp D d/da f
							LmaRR = 1.0/(localDiffusionsGrid.Jacobian[il][im][ia]) *
								(PSD_Der_a_R[il][im+1][ia] - PSD_Der_a_R[il][im][ia])/
								(localDiffusionsGrid.pc.arr[il][im+1][ia] - localDiffusionsGrid.pc.arr[il][im][ia]);
							LmaRL = 1.0/(localDiffusionsGrid.Jacobian[il][im][ia]) *
								(PSD_Der_a_L[il][im+1][ia] - PSD_Der_a_L[il][im][ia])/
								(localDiffusionsGrid.pc.arr[il][im+1][ia] - localDiffusionsGrid.pc.arr[il][im][ia]);
							LmaLR = 1.0/(localDiffusionsGrid.Jacobian[il][im][ia]) *
								(PSD_Der_a_R[il][im][ia] - PSD_Der_a_R[il][im-1][ia])/
								(localDiffusionsGrid.pc.arr[il][im][ia] - localDiffusionsGrid.pc.arr[il][im-1][ia]);
							LmaLL = 1.0/(localDiffusionsGrid.Jacobian[il][im][ia]) *
								(PSD_Der_a_L[il][im][ia] - PSD_Der_a_L[il][im-1][ia])/
								(localDiffusionsGrid.pc.arr[il][im][ia] - localDiffusionsGrid.pc.arr[il][im-1][ia]);


							// d/da D d/dp f
							LamRR = 1.0/(localDiffusionsGrid.Jacobian[il][im][ia]) *
								(PSD_Der_m_R[il][im][ia+1] - PSD_Der_m_R[il][im][ia])/
								(localDiffusionsGrid.alpha.arr[il][im][ia+1] - localDiffusionsGrid.alpha.arr[il][im][ia]);
							LamRL = 1.0/(localDiffusionsGrid.Jacobian[il][im][ia]) *
								(PSD_Der_m_L[il][im][ia+1] - PSD_Der_m_L[il][im][ia])/
								(localDiffusionsGrid.alpha.arr[il][im][ia+1] - localDiffusionsGrid.alpha.arr[il][im][ia]);
							LamLR = 1.0/(localDiffusionsGrid.Jacobian[il][im][ia]) *
								(PSD_Der_m_R[il][im][ia] - PSD_Der_m_R[il][im][ia-1])/
								(localDiffusionsGrid.alpha.arr[il][im][ia] - localDiffusionsGrid.alpha.arr[il][im][ia-1]);
							LamLL = 1.0/(localDiffusionsGrid.Jacobian[il][im][ia]) *
								(PSD_Der_m_L[il][im][ia] - PSD_Der_m_L[il][im][ia-1])/
								(localDiffusionsGrid.alpha.arr[il][im][ia] - localDiffusionsGrid.alpha.arr[il][im][ia-1]);

							error[il][im][ia] = psdLocalDiffusions.arr[il][im][ia] - psdPrevious[il][im][ia]
								- (LaaLR + LaaRL) / 2 * parameters.timeStep // d/da D d/da f
								- (LmmLR + LmmRL) / 2 * parameters.timeStep // d/da D d/da f
								- (LmaRR + LmaRL + LmaLR + LmaLL) / 4 * parameters.timeStep // d/dm D d/da f
								- (LamRR + LamRL + LamLR + LamLL) / 4 * parameters.timeStep; // d/dm D d/da f

							if (parameters.useLossCone && ((localDiffusionsGrid.alpha.arr[il][im][ia] <= VF::alc(localDiffusionsGrid.L.arr[il][im][0])) || (localDiffusionsGrid.alpha.arr[il][im][ia] >= (VC::pi - VF::alc(localDiffusionsGrid.L.arr[il][im][0])))) ) {
								taulc = 0.25 * VF::bounce_time_new(localDiffusionsGrid.L.arr[il][im][ia], localDiffusionsGrid.pc.arr[il][im][ia], localDiffusionsGrid.alpha.arr[il][im][ia]);
								error[il][im][ia] += psdLocalDiffusions.arr[il][im][ia]/taulc * parameters.timeStep;
							}

							aveerror += fabs(error[il][im][ia]);
							maxerror = (fabs(maxerror) > fabs(error[il][im][ia]))?maxerror:error[il][im][ia];
						}
					}
				}
				aveerror = aveerror / (localDiffusionsGrid.L.size) / (localDiffusionsGrid.epc.size-2) / (localDiffusionsGrid.alpha.size-2);
				Output::echo(20, "Max calculation error = %e; ave calculation error = %e\n", maxerror, aveerror);

				stringstream filename;
				filename << "./Debug/Calc_error_" << iteration << ".plt";

				error.writeToFile(filename.str().c_str(), localDiffusionsGrid.L.arr, localDiffusionsGrid.epc.arr, localDiffusionsGrid.alpha.arr);

				psdPrevious = psdLocalDiffusions.arr;
			}


	    //----------------------------------------------------------------------------------------------------
		//  Writing output
		//----------------------------------------------------------------------------------------------------
			// Write PSD into a file
			// namespace VF defined in variousFunctions.cpp
			// If this calculation step correspond to write time - do write to output file
			if (VF::check_time(iteration, parameters.general_Output_parameters.iterStep)) {
				Output::echo(0, "Step %d of %d. Time = %d days, %.0f hours. ", iteration, parameters.totalIterationsNumber, int(simulation_time), (simulation_time-int(simulation_time))*24);
				Output::echo(0, "Writing data files.\n");

				// Write information about 1d variables into the file with 1d variables
				Output1DValues(output1D, Daa, Dpcpc, Dpca, DaaLpp, DpcpcLpp, DpcaLpp, simulation_time, parameters, iteration);

				//
				// Most important function with write down the calculated PSD
				//
				psdLocalDiffusions.Output_without_grid(simulation_time);
				if (radialDiffusionGrid.L.size > 1) psdRadialDiffusion.Output_without_grid(simulation_time);

				// 'Flushing' of log files (saving from memory to disk). Otherwise it saves it only at the end, which is not convenient.
				Output::flush_log_file();

				// Save the last PSD so it can be used as an initial condition to continue the calculation
				psdRadialDiffusion.arr.writeToFile(parameters.general_Output_parameters.folderName + "/PSD0.dat", radialDiffusionGrid.L.arr, radialDiffusionGrid.epc.arr, radialDiffusionGrid.alpha.arr);
			}

            /**
             * \details interpolate from local diffusions to radial diffusion grid
             *
             * \code{.cpp}
             * psdRadialDiffusion.Interpolate();
             * \endcode
             */
			if (radialDiffusionGrid.L.size > 1) {
				// Interpolation from Local diffusions grid to Radial diffusion grid
				// local->radial
				Output::echo(3, "Interpolation local->radial... ");

				//* For time-dependent boundary conditions (used for RCM coupling) use next
				//* *.arr.xSlice(iteration) create a 2D array of boundary values at the current time
				//* psdRadialDiffusion.Interpolate(psdLocalDiffusions, parameters.interpolation, localDiffusionsGrid, radialDiffusionGrid, pc_LowerBoundaryCondition.arr.xSlice(iteration), pc_UpperBoundaryCondition.arr.xSlice(iteration));

				//* Constant in time boundary conditions (a typical simulation)
				//* psdRadialDiffusion.Interpolate(psdLocalDiffusions, parameters.interpolation, localDiffusionsGrid, radialDiffusionGrid, pc_LowerBoundaryCondition.arr.xSlice(0), pc_UpperBoundaryCondition.arr.xSlice(0));

				// Constant in time low-energy boundary conditions, but flexible high-Energy boundary.
				// Created to fix interpolation problems.
				psdRadialDiffusion.Interpolate(psdLocalDiffusions, parameters.interpolation, localDiffusionsGrid, radialDiffusionGrid, pc_LowerBoundaryCondition.arr, psdLocalDiffusions.arr.ySlice(radialDiffusionGrid.pc.size-1));
				Output::echo(3, "done!\n");
			}

		} // end if for Main loop.

		//----------------------------------
		// End of simulation
		//----------------------------------

		// Last write the log
		std::time(&now);
		current_time = std::clock();
		Output::echo(0, "End of simulation %s", ctime(&now));
		Output::echo(0, "Total time %0.2f sec\n", ( current_time - zero_time ) / (double)CLOCKS_PER_SEC );

		// Close the log
		Output::close_log_file();
		return 0; // end program if there wasn't any exeption


	//----------------------------------
	// Exceptions
	//----------------------------------
	// \deprecated?

	// try block ends.
	} catch (error_msg err) {
		Output::echo(0, "\nCRITICAL ERROR:\n%s\n", err.what().c_str());
		Output::close_log_file();
		return 1;
	} catch (exception exp) {
		Output::echo(0, "\nCRITICAL ERROR:\n%s\n", exp.what());
		Output::close_log_file();
		return 1;
	}

	//* #ifndef DEBUG_MODE
	//*	catch (...) {
	//*		Output::echo(0, "Unknown exception!");
	//*	}
	//* #endif

	Output::echo(0, "UNEXPECTED EXIT\n");
	Output::close_log_file();

	//*  Memory leaks
	//* _CrtDumpMemoryLeaks();
	return 0;
}
