/**
 * Spline interpolation
 * \file spline.cpp
 */

//#include <iostream>
#include <math.h>
#include "../../Exceptions/error.h"

//using namespace std;

/**
 * Spline interpolation - calculation of an array with second derivatives (called only once for each interpolation of an array)
 *
 * Given arrays x[1..n] and y[1..n] containing a tabulated function, i.e., yi = f (xi ), with
 * x1 < x2 < . . . < xN , and given values yp1 and ypn for the first derivative of the interpolating
 * function at points 1 and n, respectively, this routine returns an array y2[1..n] that contains
 * the second derivatives of the interpolating function at the tabulated points xi . If yp1 and/or
 * ypn are equal to 1 - 10^30 or larger, the routine is signaled to set the corresponding boundary
 * condition for a natural spline, with zero second derivative on that boundary.
 *
 * \param *x - old x
 * \param *y - old function
 * \param n - number of points
 * \param yp1 - derivative on boundary / flag if > 10^30
 * \param ypn - derivative on boundary / flag if > 10^30
 * \param *y2 - array with second derivatives of the function
 *
 */
void spline(double *x, double *y, int n, double yp1, double ypn, double *y2)
{
	int i,k;
	double p,qn,sig,un,*u,*vector();
	void free_vector();

	u = new double[n];
	if (yp1 > 0.99e30) {
		y2[0]=0.0;
		u[0]=0.0;
	} else {
		y2[0] = -0.5;
		u[0]=(3.0/(x[1]-x[0]))*((y[1]-y[0])/(x[1]-x[0])-yp1);
	}
	for (i=1;i<=n-2;i++) {
		sig=(x[i]-x[i-1])/(x[i+1]-x[i-1]);
		p=sig*y2[i-1]+2.0;
		y2[i]=(sig-1.0)/p;
		u[i]=(y[i+1]-y[i])/(x[i+1]-x[i]) - (y[i]-y[i-1])/(x[i]-x[i-1]);
		u[i]=(6.0*u[i]/(x[i+1]-x[i-1])-sig*u[i-1])/p;
	}
	if (ypn > 0.99e30) {
		qn=0.0;
		un=0.0;
	} else {
		qn=0.5;
		un=(3.0/(x[n-1]-x[n-2]))*(ypn-(y[n-1]-y[n-2])/(x[n-1]-x[n-2]));
	}
	y2[n-1]=(un-qn*u[n-2])/(qn*y2[n-2]+1.0);
	for (k=n-2;k>=0;k--) {
		y2[k]=y2[k]*y2[k+1]+u[k];
/*		if (!(y2[k] >= 0 || y2[k] <= 0) || y2[k] > 0.99e30) {
			cout << "error in spline() - inf" << endl;
			getchar();
			exit(-1);
		}*/
	}

	delete u;
}


/**
 * Spline interpolation
 *
 * Given the arrays xa[1..n] and ya[1..n], which tabulate a function (with the xai -s in order),
 * and given the array y2a[1..n], which is the output from spline above, and given a value of
 * x, this routine returns a cubic-spline interpolated value y.
 *
 * \param *xa - old grid array
 * \param *ya - old function array
 * \param *y2a - old function second derivatives array
 * \param n - size of arrays
 * \param x - new grid point
 * \param *y - new grid value
 */
void splint(double *xa, double *ya, double *y2a, int n, double x, double *y)
{
	int klo,khi,k;
	double h,b,a;
	void nrerror();

	klo=0;//1;
	khi=n-1;//n;
	while (khi-klo > 1) {
		k=(khi+klo) >> 1;
		if (xa[k] > x) khi=k;
		else klo=k;
	}
	h=xa[khi]-xa[klo];
	if (h == 0.0) //nrerror("Bad XA input to routine SPLINT");
		throw error_msg("SPLINE_ERROR", "Bad XA input to routine SPLINT");
	a=(xa[khi]-x)/h;
	b=(x-xa[klo])/h;
	*y=a*ya[klo]+b*ya[khi]+((a*a*a-a)*y2a[klo]+(b*b*b-b)*y2a[khi])*(h*h)/6.0;
/*	if (!(*y >= 0 || *y <= 0) || *y > 0.99e30) {
		cout << "error in splint() - inf" << endl;
		getchar();
		exit(-1);
	}*/

}
