/** Matrix 1D, 2D and 3D and operations with them
 *
 * \file Matrix.h
 *
 * File has 1D-class, 2D-class and 3D-class of matrixes and various functions to work with them.
 *
 * \author Developed under supervision of the PI Yuri Shprits
 *
 */

#include <assert.h>
#include <string>
#include <string.h>
#include <fstream>
#include <memory.h>
#include <math.h>
#include <map>

#ifndef matrix_array_MATRIX_H
#define matrix_array_MATRIX_H

#include "./Interpolation/linear.h"
#include "./Interpolation/spline.h"
#include "./Interpolation/polilinear.h"
#include "./Interpolation/polint.h"
#include "./Interpolation/ratint.h"
#include "./Interpolation/akima.h"



#include "../Exceptions/error.h"

using namespace std;

/**
 * \brief one dimensional matrix class
 *
 * Matrixes and operations.
 */
template <typename T>
class Matrix1D {
public:
	T *matrix_array;						///< Array to keep the values

	bool initialized;						///< Flag, equal true if initialized
	int size_x;								///< size x
	string name;							///< name of the Matrix

	// constructors and destructors
	Matrix1D() { initialized = false; };
	Matrix1D( int x_size ); //CC changed size_x to x_size for consistency
	Matrix1D( int x_size , string name);
	Matrix1D( const Matrix1D<T> &M );
	~Matrix1D();

	void AllocateMemory( int x_size ); //CC changed size_x to x_size for consistency

	// Operators
	inline T& operator[](int i);													///< Return the i-th value of matrix
	inline T& operator[](int i) const;
	inline T& operator()(int x);													///< Return the x-th value of matrix
	inline T& Value (int x) { return operator()(x); }								///< Return the (x,y) value of matrix
	inline Matrix1D<T>& MatrixArray () { return *this; }							///< Return pointer to the instance of the class.
	inline T* MatrixArrayPointer () { return matrix_array; }							///< Return pointer to the instance of the class.

	// unary
	inline const Matrix1D& operator+() const { return *this; }                      ///< Return itself as positive version of values
	inline const Matrix1D operator-() const { return ((*this)*(-1)); }              ///< Return negative version of values

	// The following operators modify the matrix they applied to
	inline Matrix1D& operator= (const Matrix1D<T> &M);
	inline Matrix1D& operator= (const T val);

	// \todo Some of the matrix operators still need to be implemented
	// I didn't have time yet to write these functions - these are matrix opearations
	//inline Matrix1D& operator*= (const Matrix1D<T> &M); 					// reserved for something good
	//inline Matrix1D& operator/= (const Matrix1D<T> &M); 					// reserved for something good
	//inline Matrix1D& operator+= (const Matrix1D<T> &M);
	//inline Matrix1D& operator-= (const Matrix1D<T> &M);
	//inline Matrix1D& operator*= (const T Val);
	//inline Matrix1D& operator/= (const T Val);
	//inline Matrix1D& operator+= (const T Val); 								///< Add the Val to each matrix element, stores result in the matrix it's applied to
	//inline Matrix1D& operator-= (const T Val); 								///< Substract the Val from each matrix element, stores result in the matrix it's applied to

	//inline Matrix1D& times_equal (const Matrix1D<T> &M); 					///< Arraywise multiplication (A.*B), stores result in the matrix it's applied to
	//inline Matrix1D& divide_equal (const Matrix1D<T> &M); 					///< Arraywise division (A./B), stores result in the matrix it's applied to

	// The following operators save the result to a new matrix
	//inline Matrix1D operator* (const Matrix1D<T> &M) const; 				// reserved for something good
	//inline Matrix1D operator/ (const Matrix1D<T> &M) const;  				// reserved for something good
	//inline Matrix1D operator+ (const Matrix1D<T> &M) const;
	//inline Matrix1D operator- (const Matrix1D<T> &M) const;
	inline Matrix1D operator* (const T Val) const;
	inline Matrix1D operator/ (const T Val) const;
	//inline Matrix1D operator+ (const T Val) const; 							///< Add the Val to each matrix element, stores result in a new matrix
	//inline Matrix1D operator- (const T Val) const; 							///< Substract the Val from each matrix element, stores result in a new matrix

	inline Matrix1D times (const Matrix1D<T> &M) const; 					///< Arraywise multiplication (A.*B), stores result in a new matrix
	inline Matrix1D divide (const Matrix1D<T> &M) const; 					///< Arraywise division (A./B), stores result in a new matrix

	inline T dot (const Matrix1D<T> &M) const; 					///< Dot product
	inline T norm () const; 					///< Norm


	// interpolations
	void Akima_interpolation(Matrix1D<T> &old_function, Matrix1D<T> &old_grid, Matrix1D<T> &new_grid, int extrapolation_type = 0, double lb = 0, double ub = 0);
	void Akima_interpolation2(Matrix1D<T> &old_function, Matrix1D<T> &old_grid, Matrix1D<T> &new_grid, int extrapolation_type = 0, double lb = 0, double ub = 0);
	void Interpolate(Matrix1D<T> &old_function, Matrix1D<T> &old_grid, Matrix1D<T> &new_grid);
	void Spline(Matrix1D<T> &old_function, Matrix1D<T> &old_grid, Matrix1D<T> &new_grid, double lb, double ub, double lin_spline_coef = 0, double max_second_der = 0);
	void Polint(Matrix1D<T> &old_function, Matrix1D<T> &old_grid, Matrix1D<T> &new_grid);
	void Ratint(Matrix1D<T> &old_function, Matrix1D<T> &old_grid, Matrix1D<T> &new_grid);
	void Polilinear(Matrix1D<T> &old_function, Matrix1D<T> &old_grid, Matrix1D<T> &new_grid, double lb, double ub);

	void Spline2(Matrix1D<T> &old_function, Matrix1D<T> &old_grid, Matrix1D<T> &new_grid, double lb, double ub, double lin_spline_coef = 0, double max_second_der = 0);


	//T max();
	//T maxabs();

	// writeToFileting
	void writeToFile(string filename);
	void writeToFile(string filename, Matrix1D<T> &grid_x);
	void readFromFile(string filename);
	void readFromFile(string filename, Matrix1D<T> &grid_x);
};

/**
 * \brief two dimensional matrix class
 *
 * Matrixes and operations.
 */
template <typename T> class Matrix2D {
private:
	/// Matrix array (array of links to other arrays). Final likns pointed to the memory addresses of the plane array. Matrix[x][y] can be used.
	/// Also, all rows saved in the memory one after anouther as a big array. So Matrix[x+x_size*y] can be also used.
	T **matrix_array;
public:
	bool initialized;														///< Flag, equal true if initialized
    //@{
    /// size x, size y
	int size_x, size_y;
    //@}
	string name;															///< name of the Matrix

	// Constructors and destructors
	Matrix2D() { initialized = false; };
	Matrix2D( const Matrix2D<T> &M );
	Matrix2D( int size_x, int size_y );
	~Matrix2D();

	void AllocateMemory(int size_x, int size_y);

	// Operators

	inline T* operator[](int i) { return matrix_array[i]; }						///< Return the i-th pointer to 1d-array. Next [j] can be applied, so we have regular [i][j].
	inline T* operator[](int i) const { return matrix_array[i]; }                   ///< Return the i-th pointer to 1d-array as const.
	inline T& operator()(int x, int y) { return matrix_array[0][x*size_y + y]; }	///< Return the (x,y)-th value of matrix
	inline T& Value (int x, int y) { return operator()(x, y); }				///< Return the (x,y) value of matrix
	inline Matrix2D<T>& MatrixArray () { return *this; }					///< Return pointer to the instance of the class.

	// unary
    inline const Matrix2D& operator+() const { return *this; }              ///< Return itself as positive version of values
	inline const Matrix2D operator-() const { return ((*this)*(-1)); }      ///< Return negative version of values

	// The following operators modify the matrix they applied to
	inline Matrix2D& operator= (const Matrix2D<T> &M);
	inline Matrix2D& operator= (const T val);

	//inline Matrix2D& operator*= (const Matrix2D<T> &M); 					// reserved for something good
	//inline Matrix2D& operator/= (const Matrix2D<T> &M); 					// reserved for something good
	//inline Matrix2D& operator+= (const Matrix2D<T> &M);
	//inline Matrix2D& operator-= (const Matrix2D<T> &M);
	//inline Matrix2D& operator*= (const T Val);
	//inline Matrix2D& operator/= (const T Val);
	//inline Matrix2D& operator+= (const T Val); 								///< Add the Val to each matrix element, stores result in the matrix it's applied to
	//inline Matrix2D& operator-= (const T Val); 								///< Substract the Val from each matrix element, stores result in the matrix it's applied to

	//inline Matrix2D& times_equal (const Matrix2D<T> &M); 					///< Arraywise multiplication (A.*B), stores result in the matrix it's applied to
	//inline Matrix2D& divide_equal (const Matrix2D<T> &M); 					///< Arraywise division (A./B), stores result in the matrix it's applied to

	// The following operators save the result to a new matrix
	//inline Matrix2D operator* (const Matrix2D<T> &M) const; 				// reserved for something good
	//inline Matrix2D operator/ (const Matrix2D<T> &M) const;  				// reserved for something good
	//inline Matrix2D operator+ (const Matrix2D<T> &M) const;
	//inline Matrix2D operator- (const Matrix2D<T> &M) const;
	inline Matrix2D operator* (const T Val) const;
	inline Matrix2D operator/ (const T Val) const;
	//inline Matrix2D operator+ (const T Val) const; 							///< Add the Val to each matrix element, stores result in a new matrix
	//inline Matrix2D operator- (const T Val) const; 							///< Substract the Val from each matrix element, stores result in a new matrix

	inline Matrix2D times (const Matrix2D<T> &M) const; 					///< Arraywise multiplication (A.*B), stores result in a new matrix
	inline Matrix2D divide (const Matrix2D<T> &M) const; 					///< Arraywise division (A./B), stores result in a new matrix


	// It returns maximum between values from class psd2DSlice and argument (VC::zero_f in that case)
	Matrix2D max_of(T val);

	void Interpolate(Matrix2D<T> &old_function, Matrix2D<T> &old_grid_x, Matrix2D<T> &old_grid_y, Matrix2D<T> &new_grid_x, Matrix2D<T> &new_grid_y);

	// Return corresponding index of 1d array
	inline int index1d(int x, int y) const;

	// writeToFileting
	void writeToFile(string filename);
	void writeToFile(string filename, Matrix2D<T> &grid_x, Matrix2D<T> &grid_y);
	void readFromFile(string filename);
	void readFromFile(string filename, Matrix2D<T> &grid_x, Matrix2D<T> &grid_y);
};

/**
 * \brief three dimensional matrix class
 *
 * Matrixes and operations.
 */
template <typename T> class Matrix3D {
private:
	/// Plane array of values. All rows saved in the memory one after another as a big array.
	T *plane_array;
	/// Matrix array (array of links to other arrays). Final links pointed to the memory addresses of the plane array. Matrix[x][y] can be used.
	T ***matrix_array;
public:
	bool initialized;														///< Flag, equal true if initialized
    //@{
    /// size x, size y, size z
	int size_x, size_y, size_z;
    //@}
	string name;															///< name of the Matrix

	// constructors and destructors
	/// Default constructor. Do nothing.
	Matrix3D() { initialized = false; };
	Matrix3D( const Matrix3D<T> &M );
    /// Constructor with initializing x y z values
	Matrix3D( int size_x, int size_y, int size_z );
	~Matrix3D();

	void AllocateMemory(int size_x, int size_y, int size_z);

	// Operators
	inline T** operator[] (int i); 											///< Return the i-th pointer to 2d-array. Next [j][k] can be applied, so we have regular [i][j][k].
	inline T** operator[] (int i) const { return matrix_array[i]; }         ///< Return the i-th pointer to 2d-array as const.
	inline T& operator() (int x, int y, int z); 							///< Return the (x,y,z) value of matrix
	inline T& Value (int x, int y, int z) { return operator()(x, y, z); }	///< Return the (x,y,z) value of matrix
	inline Matrix3D<T>& MatrixArray () { return *this; }					///< Return pointer to the instance of the class.

	// The following operators modify the matrix they applied to
	inline Matrix3D& operator= (const Matrix3D<T> &M);
//	inline Matrix3D& operator= (const Matrix2D<T> &M); 						// confusing function
	inline Matrix3D& operator= (const T Val);

	// unary
	inline const Matrix3D& operator+() const { return *this; }              ///< Return itself as positive version of values
	inline const Matrix3D operator-() const { return ((*this)*(-1)); }      ///< Return negative version of values

	//inline Matrix3D& operator*= (const Matrix3D<T> &M); 					// reserved for something good
	//inline Matrix3D& operator/= (const Matrix3D<T> &M); 					// reserved for something good
	inline Matrix3D& operator+= (const Matrix3D<T> &M);
	inline Matrix3D& operator-= (const Matrix3D<T> &M);
	inline Matrix3D& operator*= (const T Val);
	inline Matrix3D& operator/= (const T Val);
	inline Matrix3D& operator+= (const T Val); 								///< Add the Val to each matrix element, stores result in the matrix it's applied to
	inline Matrix3D& operator-= (const T Val); 								///< Substract the Val from each matrix element, stores result in the matrix it's applied to

	inline Matrix3D& times_equal (const Matrix3D<T> &M); 					///< Arraywise multiplication (A.*B), stores result in the matrix it's applied to
	inline Matrix3D& divide_equal (const Matrix3D<T> &M); 					///< Arraywise division (A./B), stores result in the matrix it's applied to

	// The following operators save the result to a new matrix
	//inline Matrix3D operator* (const Matrix3D<T> &M) const; 				// reserved for something good
	//inline Matrix3D operator/ (const Matrix3D<T> &M) const;  				// reserved for something good
	inline Matrix3D operator+ (const Matrix3D<T> &M) const;
	inline Matrix3D operator- (const Matrix3D<T> &M) const;
	inline Matrix3D operator* (const T Val) const;
	inline Matrix3D operator/ (const T Val) const;
	//inline Matrix3D operator+ (const T Val) const; 							///< Add the Val to each matrix element, stores result in a new matrix
	//inline Matrix3D operator- (const T Val) const; 							///< Substract the Val from each matrix element, stores result in a new matrix

	inline Matrix3D times (const Matrix3D<T> &M) const; 					///< Arraywise multiplication (A.*B), stores result in a new matrix
	inline Matrix3D divide (const Matrix3D<T> &M) const; 					///< Arraywise division (A./B), stores result in a new matrix

	// Saving (loading) of a matrix into (from) file
	void writeToFile(string filename); 										///< Save matrix to a file
	void writeToFile(string filename,
			Matrix3D<T> &grid_x, Matrix3D<T> &grid_y, Matrix3D<T> &grid_z); ///< Save matrix to a file, including grid
	void readFromFile(string filename);  									///< Load matrix from a file
	void readFromFile(string filename,
			Matrix3D<T> &grid_x, Matrix3D<T> &grid_y, Matrix3D<T> &grid_z); ///< Load matrix to a file

	// Some other stuff
	string change_ind;														///< Variables useful for tracking of changes (time of change can be stored here)

	inline int index1d(int x, int y, int z);								///< Returns index of the element (x,y,z) in 1d array

	T max();
	T maxabs();
	Matrix3D<T> abs();

	// slices - get 2D slice from 3D array
	Matrix2D<T> xSlice(int p_x) const;
	Matrix2D<T> ySlice(int p_y) const;
	Matrix2D<T> zSlice(int p_z) const;

};

// written this way for documentation to work
/** 
 * \class DiagMatrix
 * Diagonal matrix.
 * \brief This method of storage for matrices is convenient for diagonal (spread) matrices.
 * Stored as map (diagonal number, 1d diagonal array)
 * The USED diagonals of the matrix are stored in 1d arrays.
 * Used in CalculationMatrix
 *
 * \typedef DiagMatrix
 * Diagonal matrix.
 * This method of storage for matrices is convenient for diagonal (spread) matrices.
 * Stored as map (diagonal number, 1d diagonal array)
 * The USED diagonals of the matrix are stored in 1d arrays.
 * Used in CalculationMatrix
 */
typedef map <int , Matrix1D<double> > DiagMatrix;

/** Model matrix (or related matrices)
 * It is based on DiagMatrix and have methods for conversion from 3D or 2D PSD (and related) arrays into 1d array of unknown elements
 *
 * \brief This matrix calculates the diagonal values and index given parameters for x, y, and z
 */
class CalculationMatrix : public DiagMatrix {
public:
    ///
	bool initialized;                       ///< Flag, equal true if initialized
    //@{
    /** length of x, length of y, total size */
	int x_size, y_size, total_size;
    //@}
	// flag, if needs to be recalculated
	string change_ind;						///< Variables useful for changes tracking (store here time when changed)

	// Constructors
	CalculationMatrix() { this->initialized = false; }
	// !!! CalculationMatrix(int x_size, int y_size = 0, int z_size = 0, int n_of_diags = 1);
	CalculationMatrix(int x_size, int y_size = 1, int z_size = 1, int n_of_diags = 1);

	// Initialization
	// !!! void Initialize(int x_size, int y_size = 0, int z_size = 0, int n_of_diags = 1);
	void Initialize(int x_size, int y_size = 1, int z_size = 1, int n_of_diags = 1);

	// Returns 1d index for multiple dimension array (2D or 3D)
	int index1d(int x, int y = 0, int z = 0);

	// Save to a file
	void writeToFile(string filename);

	// Operators
	Matrix1D<double> operator* (Matrix1D<double> &V) const;

};


#endif
