/**
 *  \file Parameters.cpp
 * Loads parameters.
 * Loads parameters from file into map, then use parameters-structure related functions to get parameters values from the map and store them into the parameters structures.
 *
 * \author Developed under supervision of the PI Yuri Shprits
 */

#include "Parameters.h"

#include "../VariousFunctions/variousFunctions.h"
#include "../Exceptions/error.h"
#include "../Logging/Output.h"
#include "../VariousFunctions/erf.h"
#include <iostream>
#include <iomanip> // for setpersidion

using namespace std;


/**
 * Read each parameter from parameters file
 */
template <typename T>
bool ReadFromFile(T &variable, ifstream &parameters_file, string variable_name, string default_value = "") {

	Output::echo(0, "%s = ", variable_name.c_str());

	// Reset read pointer to beginning of file
	parameters_file.clear();
	parameters_file.seekg(0, ios::beg);

	string word1="", word2="";
	string string_value;
	if (!parameters_file.eof()) {
		while (parameters_file) {
			word2 = word1;
			// read one word
			parameters_file >> word1;
			// if last word (word2 equals previous word1) is "variableName" then read next word for the variable value
			if (word1 == "=") {
				if (word2 == variable_name) {
					parameters_file >> string_value;
					break;
				}
			}
		}
	}

	ostringstream ostr; // temporary
	if (string_value != "") {
		StrToVal(string_value, variable);
		// output whatever was loaded
		ostr << variable;
		Output::echo(0, "%s\n", ostr.str().c_str());
		return true;
	} else if (default_value != "") {
		StrToVal(default_value, variable);
		// output whatever was loaded
		ostr << variable;
		Output::echo(0, "%s (default value)\n", ostr.str().c_str());
		return true;
	} else {
		throw error_msg("ERROR", "Parameter value %s was not found.\n", variable_name.c_str());
	}
}


/**
 * Loads parameters from file
 */
bool Parameters_structure::Load_parameters(string filename) {

	ifstream parameters_file;
	parameters_file.open(filename.c_str());
	if (parameters_file == NULL) {
		throw error_msg("CONST_FILE_OPEN_ERR", "Error reading %s for parameters.", filename.c_str());
	}

	ReadFromFile(outputLvl, parameters_file, "outputLvl", "0");

	// look into map for value for "nDays" and store it into nDays
	ReadFromFile(nDays, parameters_file, "nDays");
	// look into map for value for "timeStep" and store it into timeStep
	ReadFromFile(timeStep, parameters_file, "timeStep");
	// convert hours into days
	timeStep = timeStep/24.0;
	// calculating total number of iterations
	totalIterationsNumber = int(nDays/timeStep)+1;

	ReadFromFile(useRadialDiffusion, parameters_file, "useRadialDiffusion", "true");
	ReadFromFile(useAlphaDiffusion, parameters_file, "useAlphaDiffusion", "true");
	ReadFromFile(useEnergyDiffusion, parameters_file, "useEnergyDiffusion", "true");
	ReadFromFile(useEnergyAlphaMixedTerms, parameters_file, "usePcAlphaMixedTerms", "true");

	if (useRadialDiffusion) {
		// if 4-th parameter exists - it is default value which is used if we didn't find any value inside the map
		ReadFromFile(DLLType, parameters_file, "DLLType", "DLLT_B");
	}

	Kp.AllocateMemory(totalIterationsNumber);
	ReadFromFile(useKp, parameters_file, "useKp", string("file"));
	if (useKp == "file") {
		ReadFromFile(fileKp, parameters_file, "fileKp");
		load_1d(Kp, fileKp, timeStep);
	} else {
		ReadFromFile(constKp, parameters_file, "constKp");
		Kp = constKp;
	}

	ReadFromFile(useBf, parameters_file, "useBf", string("file"));
	Bf.AllocateMemory(totalIterationsNumber);
	// check for useBf parameter
	if (useBf == "file") {
		// we need to read values from file. Getting filename from parameters
		ReadFromFile(fileBf, parameters_file, "fileBf");
		// loading that 1d file and interpolating in on out time-axis
		load_1d(Bf, fileBf, timeStep);
	} else if (useBf == "constant") {
		// if Bf is constant - using constBf parameter from map
		ReadFromFile(constBf, parameters_file, "constBf");
		Bf = constBf;
	} else {
		Bf = 1;
	}
	for (int i = 0; i < totalIterationsNumber; i++) {
		if (Bf[i] <= 0) {
			throw error_msg("BF_ERROR", "Negative or zero Bf[time[%d]=%f]=%f after loading and interpolation.\n", i, timeStep*i, Bf[i]);
		}
	}


	ReadFromFile(useLpp, parameters_file, "useLpp", string("No"));
	Lpp.AllocateMemory(totalIterationsNumber);
	if (useLpp == "file") {
		ReadFromFile(fileLpp, parameters_file, "fileLpp", string(""));
		load_1d(Lpp, fileLpp, timeStep);
	} else if (useLpp == "constant") {
		ReadFromFile(constLpp, parameters_file, "constLpp", "0.0");
		Lpp = constLpp;
	} else if (useLpp == "calculate" || str2bool(useLpp) == true) {
		// using Kp24 - the max Kp index past 24 hours for calculate the location of the plasma pause
		double Kp24 = 0;
		int iteration, back_it;
		for (iteration = 0; iteration < totalIterationsNumber; iteration++) {
			for (back_it = iteration; (fabs(timeStep * back_it - timeStep * iteration) <= 1.0) && (back_it >= 0); back_it--) {
				Kp24 = (Kp24 > Kp[back_it]) ? Kp24 : Kp[back_it];
			}
			Lpp[iteration] = (5.6 - 0.46*Kp24);
			Kp24 = 0;
		}
	} else {
		Lpp = 0;
	}

	//string usetau, usetauLpp;

	ReadFromFile(usetau, parameters_file, "usetau", string("constant"));
	tau.AllocateMemory(totalIterationsNumber);
	if (usetau == "x/Kp") {
		double tmp_tau;
		ReadFromFile(tmp_tau, parameters_file, "tau");
		for (int i = 0; i < totalIterationsNumber; i++) {
			tau[i] = tmp_tau/max(Kp[i], 0.01);
		}
		// tau = Kp/Kp/Kp * tmp_tau;
	} if (str2bool(usetau) == true ||  usetau == "constant") {
		double tmp_tau;
		ReadFromFile(tmp_tau, parameters_file, "tau");
		tau = tmp_tau;
	} else {
		tau = 1e99;
	}
	ReadFromFile(usetauLpp, parameters_file, "usetauLpp", string("constant"));
	tauLpp.AllocateMemory(totalIterationsNumber);
	if (str2bool(usetauLpp) == true || usetauLpp == "constant") {
		double tmp_tauLpp;
		ReadFromFile(tmp_tauLpp, parameters_file, "tauLpp");
		tauLpp = tmp_tauLpp;
	} else {
		tauLpp = 1e99;
	}

	ReadFromFile(outputModelMatrix, parameters_file, "outputModelMatrix", "false");
	ReadFromFile(NoNegative, parameters_file, "NoNegative", "false");
	ReadFromFile(useLossCone, parameters_file, "useLossCone", "true");

	// Read output parameters from map by calling function, which read output parameters from the map
	ReadFromFile(general_Output_parameters.timeStep, parameters_file, "general_Output_parameters.timeStep", "1");
	ReadFromFile(general_Output_parameters.iterStep, parameters_file, "general_Output_parameters.iterStep", "1");
	// change dt from hours into days
	general_Output_parameters.timeStep = general_Output_parameters.timeStep/24.0;
	ReadFromFile(general_Output_parameters.logFileName, parameters_file, "general_Output_parameters.logFileName", "logfile.log");
	ReadFromFile(general_Output_parameters.folderName, parameters_file, "general_Output_parameters.folderName", "./");
	ReadFromFile(general_Output_parameters.fileName1D, parameters_file, "general_Output_parameters.fileName1D", "Out1D.plt");

	// Grid
	ReadFromFile(radialDiffusionGrid_type, parameters_file,					"radialDiffusionGrid_type");
	ReadFromFile(localDiffusionsGrid_type, parameters_file,					"localDiffusionsGrid_type");

	if (radialDiffusionGrid_type == "GT_FILE")
		ReadFromFile(radialDiffusionGrid_filename, parameters_file,					"radialDiffusionGrid_filename");
	if (localDiffusionsGrid_type == "GT_FILE")
		ReadFromFile(localDiffusionsGrid_filename, parameters_file,					"localDiffusionsGrid_filename");

	ReadFromFile(localDiffusionsGrid_L.name, parameters_file,					"localDiffusionsGrid_L.name", string("L"));
	ReadFromFile(localDiffusionsGrid_L.size, parameters_file,					"localDiffusionsGrid_L.size");
	ReadFromFile(localDiffusionsGrid_L.useLogScale, parameters_file,				"localDiffusionsGrid_L.useLogScale");
	ReadFromFile(localDiffusionsGrid_L.min, parameters_file,						"localDiffusionsGrid_L.min");
	ReadFromFile(localDiffusionsGrid_L.max, parameters_file,						"localDiffusionsGrid_L.max");


	ReadFromFile(localDiffusionsGrid_epc.name, parameters_file,					"localDiffusionsGrid_epc.name", string("Energy, MeV"));
	ReadFromFile(localDiffusionsGrid_epc.size, parameters_file,					"localDiffusionsGrid_epc.size");
	ReadFromFile(localDiffusionsGrid_epc.useLogScale, parameters_file,			"localDiffusionsGrid_epc.useLogScale");
	ReadFromFile(localDiffusionsGrid_epc.min, parameters_file,					"localDiffusionsGrid_epc.min");
	ReadFromFile(localDiffusionsGrid_epc.max, parameters_file,					"localDiffusionsGrid_epc.max");

	localDiffusionsGrid_pc = localDiffusionsGrid_epc;
	localDiffusionsGrid_pc.min = VF::pfunc(localDiffusionsGrid_epc.min);
	localDiffusionsGrid_pc.max = VF::pfunc(localDiffusionsGrid_epc.max);

	ReadFromFile(localDiffusionsGrid_alpha.name, parameters_file,					"localDiffusionsGrid_alpha.name", string("Pitch angle, deg"));
	ReadFromFile(localDiffusionsGrid_alpha.size, parameters_file,					"localDiffusionsGrid_alpha.size");
	ReadFromFile(localDiffusionsGrid_alpha.useLogScale, parameters_file,			"localDiffusionsGrid_alpha.useLogScale");
	ReadFromFile(localDiffusionsGrid_alpha.min, parameters_file,					"localDiffusionsGrid_alpha.min");
	ReadFromFile(localDiffusionsGrid_alpha.max, parameters_file,					"localDiffusionsGrid_alpha.max");

	// change alpha from graduss to radians
	localDiffusionsGrid_alpha.min = localDiffusionsGrid_alpha.min*VC::pi/180;
	localDiffusionsGrid_alpha.max = localDiffusionsGrid_alpha.max*VC::pi/180;


	// copy grid parameters from the grid for the local proceses to the grid for the radial diffusion
	// cause most of them are the same.
	radialDiffusionGrid_L = localDiffusionsGrid_L;
	radialDiffusionGrid_epc = localDiffusionsGrid_epc;
	radialDiffusionGrid_alpha = localDiffusionsGrid_alpha;
	radialDiffusionGrid_pc = localDiffusionsGrid_pc;

	// It's repeated 6 times, cause Yuri doesn't understand it the other way
	ReadFromFile(L_LowerBoundaryCondition.type, parameters_file, "L_LowerBoundaryCondition.type", "BCT_CONSTANT_PSD");
	if (L_LowerBoundaryCondition.type == "BCT_FILE"
		|| L_LowerBoundaryCondition.type == "BCT_FILE_GRID"
		|| L_LowerBoundaryCondition.type == "BCT_MULTIPLE_FILES") {
			ReadFromFile(L_LowerBoundaryCondition.filename, parameters_file, "L_LowerBoundaryCondition.filename");
	} else {
		ReadFromFile(L_LowerBoundaryCondition.value, parameters_file, "L_LowerBoundaryCondition.value", "1e-9");
	}

	ReadFromFile(L_UpperBoundaryCondition.type, parameters_file, "L_UpperBoundaryCondition.type", "BCT_CONSTANT_PSD");
	if (L_UpperBoundaryCondition.type == "BCT_FILE"
		|| L_UpperBoundaryCondition.type == "BCT_FILE_GRID"
		|| L_UpperBoundaryCondition.type == "BCT_MULTIPLE_FILES") {
			ReadFromFile(L_UpperBoundaryCondition.filename, parameters_file, "L_UpperBoundaryCondition.filename");
	} else {
		ReadFromFile(L_UpperBoundaryCondition.value, parameters_file, "L_UpperBoundaryCondition.value", "1e-9");
	}

	ReadFromFile(pc_LowerBoundaryCondition.type, parameters_file, "pc_LowerBoundaryCondition.type", "BCT_CONSTANT_PSD");
	if (pc_LowerBoundaryCondition.type == "BCT_FILE"
		|| pc_LowerBoundaryCondition.type == "BCT_FILE_GRID"
		|| pc_LowerBoundaryCondition.type == "BCT_MULTIPLE_FILES") {
			ReadFromFile(pc_LowerBoundaryCondition.filename, parameters_file, "pc_LowerBoundaryCondition.filename");
	} else {
		ReadFromFile(pc_LowerBoundaryCondition.value, parameters_file, "pc_LowerBoundaryCondition.value", "1e-9");
	}

	ReadFromFile(pc_UpperBoundaryCondition.type, parameters_file, "pc_UpperBoundaryCondition.type", "BCT_CONSTANT_PSD");
	if (pc_UpperBoundaryCondition.type == "BCT_FILE"
		|| pc_UpperBoundaryCondition.type == "BCT_FILE_GRID"
		|| pc_UpperBoundaryCondition.type == "BCT_MULTIPLE_FILES") {
			ReadFromFile(pc_UpperBoundaryCondition.filename, parameters_file, "pc_UpperBoundaryCondition.filename");
	} else {
		ReadFromFile(pc_UpperBoundaryCondition.value, parameters_file, "pc_UpperBoundaryCondition.value", "1e-9");
	}

	ReadFromFile(alpha_LowerBoundaryCondition.type, parameters_file, "alpha_LowerBoundaryCondition.type", "BCT_CONSTANT_PSD");
	if (alpha_LowerBoundaryCondition.type == "BCT_FILE"
		|| alpha_LowerBoundaryCondition.type == "BCT_FILE_GRID"
		|| alpha_LowerBoundaryCondition.type == "BCT_MULTIPLE_FILES") {
			ReadFromFile(alpha_LowerBoundaryCondition.filename, parameters_file, "alpha_LowerBoundaryCondition.filename");
	} else {
		ReadFromFile(alpha_LowerBoundaryCondition.value, parameters_file, "alpha_LowerBoundaryCondition.value", "1e-9");
	}

	ReadFromFile(alpha_UpperBoundaryCondition.type, parameters_file, "alpha_UpperBoundaryCondition.type", "BCT_CONSTANT_PSD");
	if (alpha_UpperBoundaryCondition.type == "BCT_FILE"
		|| alpha_UpperBoundaryCondition.type == "BCT_FILE_GRID"
		|| alpha_UpperBoundaryCondition.type == "BCT_MULTIPLE_FILES") {
			ReadFromFile(alpha_UpperBoundaryCondition.filename, parameters_file, "alpha_UpperBoundaryCondition.filename");
	} else {
		ReadFromFile(alpha_UpperBoundaryCondition.value, parameters_file, "alpha_UpperBoundaryCondition.value", "1e-9");
	}

	ReadFromFile(psdRadialDiffusion.initial_PSD_Kp0, parameters_file, "psdRadialDiffusion.initial_PSD_Kp0", "0.0");
	// write down first Kp0 for initializing psd
	if (psdRadialDiffusion.initial_PSD_Kp0 <= 1e-99) {
		psdRadialDiffusion.initial_PSD_Kp0 = Kp[0];
		Output::echo(0, "psdRadialDiffusion.initial_PSD_Kp0 set to %f.\n", psdRadialDiffusion.initial_PSD_Kp0);
	}

	// Radial diffusion parameters
	ReadFromFile(psdRadialDiffusion.initial_PSD_Type, parameters_file,	"psdRadialDiffusion.initial_PSD_Type", "IPSDT_STEADY_STATE");
	if (psdRadialDiffusion.initial_PSD_Type == "IPSDT_STEADY_STATE" || psdRadialDiffusion.initial_PSD_Type == "IPSDT_STEADY_STATE_L7"
		|| psdRadialDiffusion.initial_PSD_Type == "IPSDT_STEADY_STATE_FROM_BOUNDARY") {
		ReadFromFile(psdRadialDiffusion.initial_PSD_tauSteadyState, parameters_file,	"psdRadialDiffusion.initial_PSD_tauSteadyState", "0.0");
		if (psdRadialDiffusion.initial_PSD_tauSteadyState <= 1.e-98) {
			psdRadialDiffusion.initial_PSD_tauSteadyState = 4.0/psdRadialDiffusion.initial_PSD_Kp0;
			Output::echo(0, "psdRadialDiffusion.initial_PSD_tauSteadyState set to %f.\n", psdRadialDiffusion.initial_PSD_tauSteadyState);
		}

	///////////////////////////////// kckim ////////////////////////////////////////
	} else if (psdRadialDiffusion.initial_PSD_Type == "IPSDT_STEADY_STATE_TWO_ZONE") {
		// Lpp for two zone steady state
		ReadFromFile(psdRadialDiffusion.initial_PSD_tauSteadyState, parameters_file,	"psdRadialDiffusion.initial_PSD_tauSteadyState", "0.0");
		if (psdRadialDiffusion.initial_PSD_tauSteadyState <= 1.e-98) {
			psdRadialDiffusion.initial_PSD_tauSteadyState = 4.0/psdRadialDiffusion.initial_PSD_Kp0;
			Output::echo(0, "psdRadialDiffusion.initial_PSD_tauSteadyState set to %f.\n", psdRadialDiffusion.initial_PSD_tauSteadyState);
		}
		// tauLpp for two zone steady state
		ReadFromFile(psdRadialDiffusion.usetauLpp_SteadyState, parameters_file, "psdRadialDiffusion.usetauLpp_SteadyState", string("combined"));
		if (psdRadialDiffusion.usetauLpp_SteadyState == "constant") {
			ReadFromFile(psdRadialDiffusion.initial_PSD_tauLppSteadyState, parameters_file, "psdRadialDiffusion.initial_PSD_tauLppSteadyState");
		}
	////////////////////////////////////////////////////////////////////////////////
	} else if (psdRadialDiffusion.initial_PSD_Type == "IPSDT_FILE"
		|| psdRadialDiffusion.initial_PSD_Type == "IPSDT_FILE_GRID"
		|| psdRadialDiffusion.initial_PSD_Type == "IPSDT_ORBIT_FLUX_2D"){
		//////////////////////////// kckim ////////////////////////////////////
		//|| psdRadialDiffusion.initial_PSD_Type == "IPSDT_FLUX_2D_MAWELLIAN"){
		///////////////////////////////////////////////////////////////////////
		ReadFromFile(psdRadialDiffusion.initial_PSD_fileName, parameters_file, "psdRadialDiffusion.initial_PSD_fileName");
	}


	ReadFromFile(psdRadialDiffusion.initial_PSD_some_constant_value, parameters_file, "psdRadialDiffusion.initial_PSD_some_constant_value", "1e-99"); // value_psd
	ReadFromFile(psdRadialDiffusion.initial_PSD_J_L7_function, parameters_file, "psdRadialDiffusion.initial_PSD_J_L7_function", "J_L7_corrected");

	ReadFromFile(psdRadialDiffusion.initial_PSD_outer_psd, parameters_file, "psdRadialDiffusion.initial_PSD_outer_psd", "1.0");
	ReadFromFile(psdRadialDiffusion.initial_PSD_inner_psd, parameters_file, "psdRadialDiffusion.initial_PSD_inner_psd", "0.0");

	ReadFromFile(psdRadialDiffusion.output_PSD_folderName, parameters_file,	"psdRadialDiffusion.output_PSD_folderName", general_Output_parameters.folderName );
	ReadFromFile(psdRadialDiffusion.output_PSD_fileName4D, parameters_file,	"psdRadialDiffusion.output_PSD_fileName4D", "OutPSD_rad.dat");

	ReadFromFile(psdRadialDiffusion.output_PSD_timeStep, parameters_file,   "psdRadialDiffusion.output_PSD_timeStep", "1");
	// convert hours into days
	psdRadialDiffusion.output_PSD_timeStep = psdRadialDiffusion.output_PSD_timeStep/24.0;

	psdRadialDiffusion.approximationMethod = "AM_Split_C";
	psdRadialDiffusion.solutionMethod = "SM_Tridiag";

	// Local diffusions parameters
	psdLocalDiffusions.initial_PSD_Type = "IPSDT_CONSTANT";
	psdLocalDiffusions.initial_PSD_some_constant_value = psdRadialDiffusion.initial_PSD_some_constant_value; // value_psd

	ReadFromFile(psdLocalDiffusions.output_PSD_folderName, parameters_file,	"psdLocalDiffusions.output_PSD_folderName", general_Output_parameters.folderName);
	ReadFromFile(psdLocalDiffusions.output_PSD_fileName4D, parameters_file,	"psdLocalDiffusions.output_PSD_fileName4D", string("OutPSD.dat"));

	ReadFromFile(psdLocalDiffusions.output_PSD_timeStep, parameters_file,   "psdLocalDiffusions.output_PSD_timeStep", "1");
	// convert hours into days
	psdLocalDiffusions.output_PSD_timeStep = psdLocalDiffusions.output_PSD_timeStep/24.0;

	ReadFromFile(psdLocalDiffusions.approximationMethod, parameters_file, "psdLocalDiffusions.approximationMethod", "AM_Split_C");
	if (psdLocalDiffusions.approximationMethod == "AM_Split_C" || psdLocalDiffusions.approximationMethod == "AM_Split_LR") {
		ReadFromFile(psdLocalDiffusions.solutionMethod, parameters_file, "psdLocalDiffusions.solutionMethod", "SM_Tridiag");
	}  else  {
		ReadFromFile(psdLocalDiffusions.solutionMethod, parameters_file, "psdLocalDiffusions.solutionMethod", "SM_Relaxation");

		if (psdLocalDiffusions.solutionMethod == "SM_GMRES" || psdLocalDiffusions.solutionMethod == "SM_GMRES2") {
			ReadFromFile(psdLocalDiffusions.GMRES_parameters.SOL_maxiter, parameters_file, "psdLocalDiffusions.GMRES_parameters.SOL_maxiter", "100");
			ReadFromFile(psdLocalDiffusions.GMRES_parameters.SOL_i_max, parameters_file,   "psdLocalDiffusions.GMRES_parameters.SOL_i_max", "25");
			ReadFromFile(psdLocalDiffusions.GMRES_parameters.SOL_max_iter_err, parameters_file,   "psdLocalDiffusions.GMRES_parameters.SOL_max_iter_err", "1e-1");
			ReadFromFile(psdLocalDiffusions.GMRES_parameters.preconditioner_type, parameters_file,   "psdLocalDiffusions.GMRES_parameters.preconditioner_type", "none");
			ReadFromFile(psdLocalDiffusions.GMRES_parameters.use_normalization, parameters_file,   "psdLocalDiffusions.GMRES_parameters.use_normalization", "Yes");
		}
	}

	// Sources and losses parameters
	ReadFromFile(SL.SL_L_top, parameters_file, "SL.SL_L_top", "false");
	if (SL.SL_L_top) {
		ReadFromFile(SL.SL_L_top_filename, parameters_file, "SL.SL_L_top_filename");
	}
	ReadFromFile(SL.SL_E_min, parameters_file, "SL.SL_E_min", "false");
	if (SL.SL_E_min) {
		ReadFromFile(SL.SL_E_min_filename, parameters_file, "SL.SL_E_min_filename");
	}

	// Interpolation parameters
	ReadFromFile(interpolation.type, parameters_file, "interpolation.type", "IT_SPLINE");
	ReadFromFile(interpolation.useLog, parameters_file,	"interpolation.useLog", "Log_10");
	ReadFromFile(interpolation.linearSplineCoef, parameters_file, "interpolation.linearSplineCoef", "1");
	ReadFromFile(interpolation.maxSecondDerivative, parameters_file, "interpolation.maxSecondDerivative", "1e2");

	//ReadFromFile(DxxParameters_structureList, parameters_file, "CoeffFileName");

	// Checks

	// Check outer L boundry condition creation parameters
	if (psdRadialDiffusion.initial_PSD_Type == "IPSDT_STEADY_STATE_FROM_BOUNDARY"
		&& (L_UpperBoundaryCondition.type != "BCT_FILE"
		&& L_UpperBoundaryCondition.type != "BCT_FILE_GRID"
        && L_UpperBoundaryCondition.type != "BCT_MULTIPLE_FILES")
		) throw error_msg("PARAMETERS_ERROR", "Outer L boundary can not be used for PSD steady state, cause it is set to be created from PSD itself.");

	if ((psdLocalDiffusions.approximationMethod == "AM_Block_LR" || psdLocalDiffusions.approximationMethod == "AM_Block_C") && !(localDiffusionsGrid_pc.size > 1 && localDiffusionsGrid_alpha.size > 1)) {
		throw error_msg("BLOCK_WITH_1D", "Block method can not be used with 1D");
	}

	if (localDiffusionsGrid_pc.size < 3 && useEnergyDiffusion)
		throw error_msg("ERROR", "Not enough grid points for energy diffusion (localDiffusionsGrid.pc.size = %d and useEnergyDiffusion = Yes)", localDiffusionsGrid_pc.size);

	if (localDiffusionsGrid_alpha.size < 3 && useAlphaDiffusion)
		throw error_msg("ERROR", "Not enough grid points for pitch angle diffusion (localDiffusionsGrid.alpha.size = %d and useAlphaDiffusion = Yes)", localDiffusionsGrid_alpha.size);

	if (localDiffusionsGrid_L.size < 3 && useRadialDiffusion)
		throw error_msg("ERROR", "Not enough grid points for radial diffusion (localDiffusionsGrid.L.size = %d and useRadialDiffusion = Yes)", localDiffusionsGrid_L.size);

	// Read diffusion coefficients
	parameters_file.seekg(0,ios::beg);
	string word1="", word2="";
	string string_value, dxx_parameters_file_name;
	if (parameters_file != NULL) {
		while (parameters_file) {
			word2 = word1;
			// read one word
			parameters_file >> word1;
			// if last word (word2 equals previous word1) is "variableName" then read next word for the variable value
			if (word1 == "=") {
				if (word2 == "CoeffFileName") {
					parameters_file >> dxx_parameters_file_name;

					// add a value to the vector of diffusion coefficients parameters structures (well, a structure for a diffusion coefficient)
					DxxParametersList.push_back(DxxParameters_structure());
					// Read that diffusion coefficient parameters from it's file
					// ReadFromFile(DxxParameters_structureList.back());

					// Load dxx parameters into the just added dxx parameter structure
					DxxParametersList.back().Load_dxx_parameters(dxx_parameters_file_name);

				}
			}
		}
	}


	ReadFromFile(SL.useLmp, parameters_file, "sourcesAndLosses.useLmp", "false");
	if (SL.useLmp) {
		ReadFromFile(SL.fileLmp, parameters_file, "sourcesAndLosses.fileLmp");
	}
	
	if (SL.useLmp == true){

		// check localDiffusionsGrid_alpha.size
		if (localDiffusionsGrid_alpha.size < 1 ){
			throw error_msg("ERROR", "Not enough alpha grid points for magnitopause shadowing (localDiffusionsGrid_alpha.size = %d)", localDiffusionsGrid_alpha.size);
		}
			
		// load file Matrix2D Lmp
		SL.Lmp.AllocateMemory(totalIterationsNumber, localDiffusionsGrid_alpha.size);
		load_2d(SL.Lmp, SL.fileLmp, timeStep);
		
		// // // load_2d(); // // //


	}


	return true;
}


/**
 * Loads parameters from file
 */
bool DxxParameters_structure::Load_dxx_parameters(string dxx_parameters_file_name) {
	ifstream dxx_parameters_file;
	dxx_parameters_file.open(dxx_parameters_file_name.c_str());
	if (dxx_parameters_file == NULL) {
		throw error_msg("DXX_PARAM_FILE_OPEN_ERR", "Error reading diffusion coefficients parameters file %s.", dxx_parameters_file_name.c_str());
	}

	ReadFromFile(DxxType, dxx_parameters_file,	"DxxType");
	ReadFromFile(DxxName, dxx_parameters_file,			"DxxType");

	ReadFromFile(waveType, dxx_parameters_file,			"waveType");
	ReadFromFile(waveName, dxx_parameters_file,			"waveType");
    
	ReadFromFile(time_start, dxx_parameters_file,			"time_start");
	ReadFromFile(time_end, dxx_parameters_file,			"time_end");

	ReadFromFile(useScale, dxx_parameters_file, "useScale", "false");
	if (useScale) {
		ReadFromFile(DxxKp , dxx_parameters_file, "DxxKp");
	}

	ReadFromFile(loadOrCalculate, dxx_parameters_file, "loadOrCalculate");

	ReadFromFile(filetype, dxx_parameters_file,			"filetype");
	ReadFromFile(filename, dxx_parameters_file,			"filename");

	ReadFromFile(multiplicator, dxx_parameters_file,	"multiplicator", "1.0");
	ReadFromFile(MLT_averaging, dxx_parameters_file,	"MLT_averaging", "1.0");

	if (loadOrCalculate == "LOC_CALCULATE" || loadOrCalculate == "LOC_LOADORCALCULATE") {
		ReadFromFile(numberDensity, dxx_parameters_file,	"numberDensity");
		ReadFromFile(Omega_mType, dxx_parameters_file,		"Omega_mType");
		ReadFromFile(Omega_m, dxx_parameters_file,			"Omega_m");
		ReadFromFile(d_omega, dxx_parameters_file,			"d_omega");
		ReadFromFile(omega_lc, dxx_parameters_file,			"omega_lc");
		ReadFromFile(omega_uc, dxx_parameters_file,			"omega_uc");
        ReadFromFile(eta1, dxx_parameters_file,             "eta1");
        ReadFromFile(eta2, dxx_parameters_file,             "eta2");
        ReadFromFile(eta3, dxx_parameters_file,             "eta3");
		ReadFromFile(Bw, dxx_parameters_file,				"Bw");
		ReadFromFile(BwFromLambda, dxx_parameters_file,		"BwFromLambda");

		ReadFromFile(nint, dxx_parameters_file,				"nint");
		ReadFromFile(mirror_point_coeff, dxx_parameters_file,				"mirror_point_coeff", "0.999");

		ReadFromFile(lam_min, dxx_parameters_file,			"lam_min", "0.0");
		ReadFromFile(lam_max, dxx_parameters_file,			"lam_max");
		ReadFromFile(nu, dxx_parameters_file,				"nu", "0.0");
		ReadFromFile(s, dxx_parameters_file,				"sign");
		if (numberDensity == "ND_CONSTANT") {
			ReadFromFile(f, dxx_parameters_file,				"f_ratio");
		}
		ReadFromFile(particle, dxx_parameters_file,			"particle");

		double sigma = (omega_uc - omega_lc) / (2.0 * d_omega);
		Output::echo(0, "sigma = %f\n", sigma);
		nu = erf(sigma)*sqrt(VC::pi);
		Output::echo(0, "nu = %f\n", nu);

		if (Bw <= 0 && BwFromLambda == false) {
			throw error_msg("DXX_ERROR", "BwFromLambda is false and Bw = %f\n", Bw);
		} else if (BwFromLambda == true && Bw <= 0) {
			Bw = 1;
		}

	}

	return true;
}


/**
 * Converting string to double.
 */
void StrToVal(string input, double &place) {
	istringstream(input) >> place;
//	Output::echo("%e", place);
}

/**
 * Converting string to int.
 */
void StrToVal(string input, int &place) {
	istringstream(input) >> place;
//	Output::echo(place);
}

/**
 * Converting string to string. Function StrToVal is used in template, so we need thad function to make template function works in case of string.
 */
void StrToVal(string input, string &place) {
	place = input;
//	Output::echo(place);
}

/**
 * Converting string to bool.
 */
void StrToVal(string input, bool &place) {
	place = str2bool(input);;
//	Output::echo(place);
}


/**
 * Converting string to boolean.
 */
bool str2bool(string str) {
	if (str == "Yes" || str == "yes" || str == "1" || str == "True" || str == "true") return true;
	else return false;
}

/**
 * Converting boolean to string.
 */
string bool2str(bool b) {
	return (b==true)?"Yes":"No";
}


/** Read 1d matrix_array from txt-file and interpolate to out time-axis.
 */
void load_1d(Matrix1D<double> &var, string filename, double dt, int var_size) {
	// var_size is a flag, if it is equal to zero, then we assume memory 1d-matrix allocated
	// if it is equal to some value, then we need to allocate the memory
	if (var_size != 0) {
		// allocate memory
		var.AllocateMemory(var_size);
	}  else  {
		// if var_size==0, assume memory already allocated
		var_size = var.size_x;
	}

	ifstream input_1d(filename.c_str());
	if (input_1d != NULL) {
		vector <double> time1d, value1d;
		double time_tmp, value_tmp;
		while (!input_1d.eof()) {
			// reading time values and function values from the file into vectors
			input_1d >> time_tmp >> value_tmp;
			time1d.push_back(time_tmp);
			value1d.push_back(value_tmp);
		}
		double time;
		double a;
		// interpolating loaded values into out time-axis
		for (int it = 0; it < var.size_x; it++) {
			time = dt*it;

			unsigned int i = 0;
			while (((++i) < (time1d.size()-1)) && time > time1d[i]) {
				// Not enough Kp error here
			};

			a = (time - time1d[i-1]) / (time1d[i] - time1d[i-1]);
			var[it] = value1d[i-1] + a*(value1d[i] - value1d[i-1]);
		}
	} else {
		throw error_msg("LOAD_1D_ERROR", "Load 1d file error: file not found %f\n", filename.c_str());
	}
}


/** Read 2d matrix_array from txt-file and interpolate to out time-axis.
 * NEED TO BE TESTED
 */
void load_2d(Matrix2D<double> &var, string filename, double dt, int var_size_x, int var_size_y) {


	// vector of time value
	vector <double> time1d;


	// vectors of 2d values
	T_2d_double_vector value2d;

	// var_size is a flag, if it is equal to zero, then we assume memory 1d-matrix allocated
	// if it is equal to some value, then we need to allocate the memory
	if (var_size_x != 0 & var_size_y != 0) {
		// allocate memory
		var.AllocateMemory(var_size_x,var_size_y);
	} /* else  {
		// if var_size==0, assume memory already allocated
		var_size = var.size_x * var.size_y;
	}*/

	ifstream input_2d(filename.c_str());

	if (input_2d != NULL) {

		double time_tmp, value_tmp;
		int it = 0;
		while (input_2d >> time_tmp) { // reading time values and function values from the file into vectors
									   // This method of reading prevent emty lines reading
			time1d.push_back(time_tmp);

			for(int f=0; f < var.size_y; f++){
				input_2d >> value_tmp;
				
				// Add value value_tmp to value2d (vector of changing size)
				T_2d_double_vector_add(value2d, it, f) = value_tmp;
			}
			it++;
			
		}
		double time;
		double a;
		// interpolating loaded values into out time-axis
		for (it = 0; it < var.size_x; it++) {
			time = dt*it;

			unsigned int i = 0;
			while (((++i) < (time1d.size()-1)) && time > time1d[i]) {
				// Not enough Lmp error here (stupid check)
			};

		
		#ifdef DEBUG_MODE
			cout << "\n" << setprecision(2) << setw(5) << time << ' ';
		#endif


			for (int ia = 0; ia < var.size_y; ia++)
			{
				a = (time - time1d[i-1]) / (time1d[i] - time1d[i-1]);
				var[it][ia] = value2d[i-1][ia] + a*(value2d[i][ia] - value2d[i-1][ia]);

				#ifdef DEBUG_MODE
					if (var[it][ia] > 6) cout << 0 << ' ';
					else cout << 1 << ' ';
				#endif
				
			}
		}
	} else {
		throw error_msg("LOAD_2D_ERROR", "Load 2d file error: file not found %f\n", filename.c_str());
	}
}


    /** function needs to operate with 2d vectors (add value)
     * resizes vector to parameter length, return value at proper indices
     *
     * \param &v - typedef 2D vector of vector of doubles
     * \param x - length x
     * \param y - length y
     */
	double& T_2d_double_vector_add(T_2d_double_vector &v, size_t x, size_t y)
	{
		if(v.size() <= x)     v.resize(x + 1);
		if(v[x].size() <= y)  v[x].resize(y + 1, -1);
		return v[x][y];    
	}


