/**
 * \file Parameters.h
 * All parameters, loaded from .ini file are described here in one structure.
 * Each branch of the structure holds parameters for one object (class) initialization
 * So, we need to pass one structure (branch of the main structure) to the classes constructors
 *
 * \author Developed under supervision of the PI Yuri Shprits
 */

#ifndef LOAD_CONSTANCE_H		// continue if have not load this file yet
#define LOAD_CONSTANCE_H

#include <string>
#include <vector>
#include <map>
#include "../Matrix/Matrix.h"
#include "../VariousFunctions/variousConstants.h"
#include "../Logging/Output.h"
#include <sstream>

using namespace std;

/** Diffusion coefficient parameters
 * \brief Struct that holds various parameters to be used for Dxx.
 *
 */
struct DxxParameters_structure {
public:
	string DxxType;					///< Dxx type. Check StrToVal(string input, DiffusionCoefficientTypes &place) for known values.
	string DxxName;													///< Dxx name.

	string waveType;											///< Wave type (chorus, hiss etc). Check StrToVal(string input, WaveTypes &place) for known values.
	string waveName;													///< Wave name.

	string filetype;										///< Type of the file (w/wo grid etc)
	string filename;														///< File name for loading or saving diffusion coefficients array.

	double time_start;												///< Time, when diffusion coefficient starts.
	double time_end;													///< Time, when diffusion coefficient ends.

	bool useScale;														///< Flag, use scaling if equal to true.
	double DxxKp;														///< Kp value associated with parameters (for Kp-scaling)

	string loadOrCalculate;			///<Load diffusion coefficient or calculate flag.

	string numberDensity;						///< Number density model. Check StrToVal(string input, NumberDensities &place) for known values.
	double MLT_averaging;										///<MLT averaging part of the orbit.

	double multiplicator;				///< Dxx is multiplied to this number after all other operations. Useful for fast Dxx modifications.

	string Omega_mType;						///<Omega_m type. Check StrToVal(string input, Omega_mTypes &place) for known values.
	double Omega_m;												///< Omega_m value.
	double d_omega;													///< d_omega value.
	double omega_uc;												///< omega upper cutoff
	double omega_lc;													///< omega lower cutoff

	double Bw;															///< Bw value.
	bool BwFromLambda;											///< Bw lambda dependance flag.
	int nint; ///< Number of points in integral
	double mirror_point_coeff; ///< Coefficient for mirror point (like 0.999, cause we can not integrate all the way to mirror point)
	double lam_min;												///< Minimum latitude.
	double lam_max;												///< Maximum latitude.

    //@{
    /// eta values, calculation parameters
    double eta1, eta2, eta3;
    //@}
    
	double nu;      ///< \f$ erf(\sigma)*\sqrt{\pi} \f$
	double s;       ///< sign
	double f;       ///< f_ratio

	//double cfm;
	string particle;											///< Type of particles, produced wave? Ions or electrons. Check StrToVal(string input, ParticleTypes &place) for known values.

    //@{
	/// global
	double L, EMeV, Alpha;
    //@}

	bool Load_dxx_parameters(string filename);

};

/** List of diffusion coefficient parameters.
 */
//typedef std::vector<DxxParameters> List_of_AllDxxParameters;

/** 
 * \brief Main parameters structure that holds smaller structures for individual parameters.
 *
 * The structure was modified without multiple substructures according to request of project PI
 */
struct Parameters_structure  {
	// General parameters Days
	int outputLvl;			///< Detalization level of screen output.
	double nDays;						///< Simulation total time, days.
	double timeStep;					///< Time step, hours.
	int totalIterationsNumber;	///<Total number of iterations.
	bool useRadialDiffusion;				///< Using diffusions flags.
	bool useAlphaDiffusion; 				///< Using diffusions flags.
	bool useEnergyDiffusion;				///< Using diffusions flags.
	bool useEnergyAlphaMixedTerms;			///< Using mixed terms flags.
	/// DLL type (which method to use to calculate). Check StrToVal(string input, DLLTypes &place) for known values.
	string DLLType;
	Matrix1D<double> Kp;			///< Kp array
	string useKp; 							///<Flag, using Kp index. constant - use constant value, file - read from file.
	double constKp; 					///< constant Kp value (if it's constant).
	string fileKp; 							///< Kp file name.
	Matrix1D<double> Bf; 			///< Boundary flux array
	string useBf; 							///<Flag, using boundary flux dependences if equal true.
	double constBf; 						///< constant Bf value (if it's constant).
	string fileBf;								///< Bf file name.
	Matrix1D<double> Lpp;		///< Lpp array
	string useLpp; 						///<Flag, using plasma pause location dependences if equal true.
	double constLpp;					///< constant Lpp value (if it's constant).
	string fileLpp; 							///< Lpp file name.
    //@{
    /// key word to define if and how to use tau/tauLpp
	string usetau, usetauLpp;
    //@}
	Matrix1D<double> tau;							///<Lifetime out of the plasmasphere.
	Matrix1D<double> tauLpp;						///<Lifetime inside of the plasmasphere.

	bool outputModelMatrix; 							///< Output model matrix.
	bool NoNegative;	///< Artificially no negative PSD

	bool useLossCone;   ///< specify whether using loss cone or not

	/** 
     * \brief General program output parameters structure.
	 */
	struct General_Output_parameters {
		double timeStep;				///< Time step for output.
		int iterStep;				///< Number of iterations between outputs
		string logFileName;			///< log-file name.
		string folderName;			///< output folder name.
		string fileName1D;			///< 1d-output file name.
	} general_Output_parameters;    ///< instance of the struct

	string
		radialDiffusionGrid_type, ///< Local diffusions grid type
		localDiffusionsGrid_type; ///< Radial diffusions grid type

	string radialDiffusionGrid_filename; ///< filename, if grid load from file
	string localDiffusionsGrid_filename; ///< filename, if grid load from file

	/**
     * \brief %Grid element parameters structure.
	 */
	struct GridElement  {
		/// Grid element name. Optional.
		string	name;
		/// Grid size.
		int		size;
		/// Flag, use logarithmic scale.
		bool	useLogScale;
		double min; ///< min value.
		double max; ///< max value.

	}   //@{
        /// grid elements for local and radial diffusion, and corresponding types of elements
		radialDiffusionGrid_L, radialDiffusionGrid_pc, radialDiffusionGrid_alpha, radialDiffusionGrid_epc,
		localDiffusionsGrid_L, localDiffusionsGrid_pc, localDiffusionsGrid_alpha, localDiffusionsGrid_epc;
        //@}


	/** 
     * \brief Boundary conditions parameters structure.
	 */
	struct BoundaryCondition {
		/// Type. Check StrToVal(string input, BoundaryConditionTypes &place) for known values.
		string	type;
		/// Value (if it is constant value on the boundary).
		double	value;
		/// File name to load boundary conditions from.
		string filename;

	}
		L_LowerBoundaryCondition,
		L_UpperBoundaryCondition,
		pc_LowerBoundaryCondition,
		pc_UpperBoundaryCondition,
		alpha_LowerBoundaryCondition,
		alpha_UpperBoundaryCondition;

	/** 
     * \brief %PSD  parameters structure.
	 */
	struct PSD {
		/** Initial %PSD  parameters structure.
		 */
		string initial_PSD_Type;		///< Tells us where to get initial %PSD values. Check StrToVal(string input, InitialPSDTypes &place) for known values.
		string initial_PSD_fileName;								///< File name, if we need to load initial values from file.
		double initial_PSD_tauSteadyState;					///< Tau for steady state, if we are calculation initial values as steady state solution.

		///////////// kckim //////////////////////////
		string usetauLpp_SteadyState;                           ///< Tau within plasmasphere for steady state
		double initial_PSD_tauLppSteadyState;					///< Tau for steady state, if we are calculation initial values as steady state solution.
		//////////////////////////////////////////////


		double initial_PSD_Kp0;										///< Kp for steady state?
		double initial_PSD_some_constant_value;						///< Some %psd value. Used as initial PSD value everywhere if initial_PSD_Type = IPSDT_CONSTANT or as minimal initial PSD value in other cases. Also used for some inner belt imitation if initial_PSD_Type = IPSDT_STEADY_STATE.
		string initial_PSD_J_L7_function;							///< Name of the function for flux at L=7. Can be 'J_L7' or 'J_L7_corrected'.
		double initial_PSD_outer_psd;			///< Outer PSD boundary value for steady state, default 1
		double initial_PSD_inner_psd;			///< Inner PSD boundary value for steady state, default 0

		/** %PSD output parameters structure
		 */
		string output_PSD_folderName;							///< Folder name for psd output.
		string output_PSD_fileName4D;							///< File name for psd output.
		double output_PSD_timeStep;								///< Time step for psd output.

		/// Approximation method. Check StrToVal(string input, ApproximationMethods &place) for known values.
		string approximationMethod;
		/// Solution method. Check StrToVal(string input, SolutionMethods &place) for known values.
		string solutionMethod;

        /**
         * \brief GMRES parameters structure
         */
		struct GMRES_parameters_structure {

			int SOL_maxiter; ///< Maximum number of steps, for iteration methods
			int SOL_i_max; ///< For GMRES - number of iterations before restart
			double SOL_max_iter_err; ///< maximum error, for iteration methods
			string preconditioner_type; ///< Type of preconditioner to use.
			bool use_normalization; ///< To normalize (or not) matrix A from Ax=B system.
		} GMRES_parameters; ///< Parameters structure for GMRES

	} 	psdRadialDiffusion, 			///< %PSD parameters for radial diffusion.
		psdLocalDiffusions;			///< %PSD parameters for local diffusions.

	/**
	 * \brief Sources and losses parameters structure.
	 */
	struct SL_structure {
        /// SL parameters
		bool SL_L_top;              ///< losses upper bound
		string SL_L_top_filename;   ///< filename for losses upper bound
		bool SL_E_min;              ///< sources lower bound
		string SL_E_min_filename;   ///< filename for sources lower bound

		Matrix2D<double> Lmp;	///< magnetopause position Lmp (time, alpha)
		bool useLmp;			///< magnetopause position flag
		string fileLmp;			///< magnetopause position file name
	} SL;                       ///< instance of the struct

	/** 
     * \brief %Interpolation parameters structure
	 */
	struct Interpolation {
		/// Interpolation type: spline, linear etc.
		string	type;
		/// Flag, if we should interpolate logarithm of the values.
		string	useLog;

		/** Parameters for spline interpolation - defined the ratio between spline and linear. 0 - pure spline, 1 - pure linear. */
		double				linearSplineCoef;

		/** Maximum second derivative value, to start using linearSplineCoef. */
		double				maxSecondDerivative;
	} interpolation; ///< %Interpolation parameters.


	// diffusion coefficients
	// List_of_AllDxxParameters AllDxxParameters; ///< List of diffusion coefficients parameters.
	vector<DxxParameters_structure> DxxParametersList; ///< List of diffusion coefficients parameters.


	bool Load_parameters(string filename);
};


void StrToVal(string input, double &place);
void StrToVal(string input, int &place);
void StrToVal(string input, string &place);
void StrToVal(string input, bool &place);

void load_1d(Matrix1D<double> &var, string filename, double dt, int var_size = 0);

void load_2d(Matrix2D<double> &var, string filename, double dt, int var_size_x = 0, int var_size_y = 0);
// vectors of 2d values
typedef vector<double> T_double_vector;                 ///< typedef vector of doubles
typedef vector<T_double_vector> T_2d_double_vector;     ///< typedef vector of vector of doubles
	

// add value to the vector of 2d values
double& T_2d_double_vector_add(T_2d_double_vector &v, size_t x, size_t y);


bool str2bool(string str);
string bool2str(bool b);
#endif
