/**
 * \file variousFunctions.cpp
 * \brief Various functions description, see namespace VF for details
 *
 * \ref deprecated
 *
 * \author Developed by Yuri Shprits, Dmitry Subbotin, Kyung-Chan Kim
 */

#include "variousFunctions.h"
#include <math.h>
#include <valarray>
#include <iostream>
#include <sstream>

// namespace described in header file
namespace VF {
	
	/**
	 *  Loss cone in rad.
	 * \f[ sin(\alpha)^2 = \frac{L^{-3}}{\sqrt{4 - \frac{3}{L}}} \f]
	 */
	double alc(double L) {
		return asin(pow(L, -1.5)*pow(4-3./L,-0.25));
	}

	/** T(&alpha;) - function, related to bounce time <br>
	 * \f[ y = sin(\alpha) \f]
	 * \f[ T = 1.38 + 0.055 \times y^{1/3} - 0.32 \times y^{1/2} - 0.037 \times y^{2/3} - 0.394 \times y + 0.056 \times y^{4/3} \f]
	 * [Orlova and Shprits, On the bounce-averaging of scattering rates and the calculation of bounce period ,2011]
	 */
	double T(double alpha) {
		double y = sin(alpha);
		return (1.38 + 0.055 * pow(y, 1.0/3) - 0.32 * pow(y, 1.0/2) - 0.037 * pow(y, 2.0/3) - 0.394 * y + 0.056 * pow(y, 4.0/3));
	}

	/** This function used in Fokker-Plank equation. \f$ G = \cos(\alpha) \times \sin(\alpha) \f$ from Jacobian from %mu;,J to %alpha;, p and bounce avarage.
	 * G-function \f$ G = \cos(\alpha) \times \sin(\alpha) \times T(\alpha)\f$ 
	 */
	double G(double alpha) {
		// cos(x)*sin(x)*T(y);
		return cos(alpha)*sin(alpha)*T(alpha);
	}

	/**
	* Bounce time, in days.
	* \f[ \tau = \frac{4.0 \times L \times R_E \times mc^2}{pc \times c \times T(\alpha) \times ( 60 \times 60 \times 24 )}\f]
	*/
	double bounce_time_new(double L, double pc, double alpha) {
		return 4.0 * L * VC::RE * VC::mc2 / pc / VC::c_si * T(alpha) / 60 / 60 / 24;
	}

	/**
	 * Y(&alpha;) function. More accurate approximation.
	 * \f[ y = sin(\alpha); T_0, T_1 = constant; \f]
	 * \f[ Y = 2(1 - y) \times T_0 + (T_0 - T_1) \times (y \log{y} + 2y - 2 \sqrt{y}) \f]
	 */
	double Y(double alpha) {
		double T0 = 1.3802;
		double T1 = 0.7405;
		double y = sin(alpha);
		return 2.0*(1.0 - y)*T0 + (T0 - T1)*(y*log(y) + 2.0*y - 2.0*sqrt(y));
	}


	////////////////////////////

	/**
	 * Computation of dipole magnetic field in Gauss. 
	 * \f[ B = \frac{0.31}{L^3} \f] 
	 */
	double B (double Lparam) {
		return 0.31/pow(Lparam, 3);
	}

	/**
	 * ---------------------------------------------------------------------------------- <br>
     *  Electrostatic radial diffusion coefficient [see Brautigam and Albert(2000),JGR]   <br>
	 * ---------------------------------------------------------------------------------- <br>
	 *  This model seems to overestimate radial diffusion rate, so we need to scale down  <br>
	 *  by adjusting a parameter "scale_factor".                                          <br>
	 * ---------------------------------------------------------------------------------- <br>
	 *  (1) Inputs  : alpha in radian, Ke in MeV, L and Kp in dimensionless unit          <br>
	 *  (2) Outputs : DLLe in 1/day or Re^2/day                                           <br>
	 * ---------------------------------------------------------------------------------- <br>
	 */
	double Dfe (double alpha, double Ke, double L, double Kp) {
		double E_kp,DLLe;
		double fac,dy,ty,factor,gamma,beta,wd;
		double Re = 6.378*1e8;    // Earth radius(cm)

		// Sometime E-field effect is too strong, so we need to scale down
	    // by adjusting this parameter until more accurate DLLe model is found
		double scale_factor = 0.65;

		// exponenetial decay time in sec
		double T0 = 2700.0;

		// convective electric field in mV/m, Kp=1 to 6;
		if(Kp > 6) Kp=6;
		E_kp = 0.26 * (Kp-1) + 0.1;

		// convert energy(MeV) to mu(MeV/G)
		//pc = Ke*(Ke+2.0*VC::mc2);    // pc in MeV
		//mu = (pc*sin(alpha)*sin(alpha)) / (VF::B(L)*VC::mc2);

		// angular drift velocity wd (rad/s) averaged over a bounce in a dipole
		// [see Schulz and Lanzerotti, 1974 and Schulz, 1991]

		fac = pow(sin(alpha),(3.0/4.0));
		dy = (5.520692-2.357194*sin(alpha)+1.279385*fac) / 12.0;
        ty = 1.380173-(0.639693*fac);
		factor = dy/ty;              // pitch angle dependency

		gamma = (Ke + VC::mc2)/VC::mc2;
		beta = sqrt(1.0-1.0/(pow(gamma,2.0)));
		//gamma = sqrt(1.0+2.0*VF::B(L)*mu/VC::mc2);
		//wd = (1.475*1e-3*mu) / gamma / pow(L,2.0) * factor;  // wd(rad/s),mu(MeV/G)
		wd = 1.216 * 1e-3 * gamma * pow(beta,2.0) * L * factor;

		//wd = (1.475*1e-3*Ke) / VF::B(L) / pow(L,2.0) * factor; // wd(rad/s),Ke(MeV),B(G)

		// electrostatic radial diffusion coefficient in cm^2/sec
		DLLe = (T0*pow(E_kp,2.0)*1e6) / (4.0*pow(VC::B_0,2)) / (1.0+pow((wd*T0/2.0),2.0)) * pow(L,6); // E_kp(mV/m),wd(1/s),T0(s),B_0(G)

		return scale_factor * DLLe * (60.0*60.0*24.0)/pow(Re,2);  // DLLe in Re^2/day or 1/day
	}

	/**
     * -------------------------------------------------------------------- <br>
	 * Coulomb scattering electron lifetime [Lyons and Thorne, 1973, JGR]   <br>
	 * -------------------------------------------------------------------- <br>
	 * (1) Inputs : L in dimensionless, energy in MeV                       <br>
	 * (2) Outputs: tau in days                                             <br>
	 * -------------------------------------------------------------------- <br>
	 */
	double Coulomb(double L,double energy) {
	    double  n,tau,scale_factor;

		scale_factor = 1.0;   // 1.8
	    // n is the equatorial electron density in #/cm3
		n=1e3*pow((4.0/L),4.0);   // original model

		// Carpenter and Anderson model, 1992, JGR
		// If we apply the following model, tau model also should be changed.
	    //n=pow(10, (-0.3145*L + 3.9043));

	    // electron lifetime due to coulomb scattering in sec
        tau=( 3e8*pow((energy*1e3),1.5) ) / n;

        return scale_factor * tau/(60.0*60.0*24.0);    // in days
	}

	/**
	 * --------------------------------------------------------------------- <br>
	 *  Empirical electron lifetime related to the resonant scattering by    <br>
	 *  plasmaspheric hiss, lightining-generated whistlers                   <br>
	 *  and VLF transmitter signals, which is valid only in the slot region  <br>
	 *  between 2.5 and Lpp (plasmapause location) [see Su et al, 2010, JGR] <br>
	 * --------------------------------------------------------------------- <br>
	 *  (1) Inputs  : energy in MeV, Kp and L in dimensionless               <br>
	 *  (2) Outputs : tau in days                                            <br>
	 * --------------------------------------------------------------------- <br>
	 * This model is valid up to Lpp, so class "PSD::SourcesAndLosses" checks if input L is larger than Lpp or not.
	 * For L<2.5 \f$ \tau = 10^99 \f$, for L>=2.5
	 * \f[ \tau = \frac{2}{Kp}^2 \times 10^{(L-3.5)^4}+0.4 \f]
	 */
	double Precipitation(double Kp,double L,double energy) {
	    double  n,tau;
		if(L>=2.5) {      // this model is valid up to Lpp
			tau = pow((2.0/Kp),2.0) * pow(10.0,pow((L-3.5),4.0)+0.4);  // in days
		} else {
			tau =1e99;
		}

		return tau;  // in days
	}

	/**
	 * -------------------------------------------------------------------- <br>
	 *  Subroutine to call various flux model of the outer boundary at L=7  <br>
	 * -------------------------------------------------------------------- <br>
	 *  (1) Inputs  : EL - energy(MeV) at L=7                               <br>
	 *                J_L7_function - name of boundary flux function        <br>
	 *  (2) Outputs : JL - flux at L=7                                      <br>
	 * -------------------------------------------------------------------- <br>
	 */
	double Outer_Boundary_Flux_at_L7(double E7,string J_L7_function) {

		double J7;

		// boundary flux model at L=7
		if (J_L7_function == "J_L7") {
			J7 = VF::J_L7( E7,0.2,2.e3,1,7 );
		} else if (J_L7_function == "J_L7_old") {
			J7 = VF::J_L7_old( E7 );
		} else if (J_L7_function == "J_L7_corrected") {
			J7 = VF::J_L7_corrected( E7 );
		} else if (J_L7_function == "J_L7_corrected_0") {
			J7 = VF::J_L7_corrected_0( E7 );
		} else {
			throw error_msg("INITIAL_PSD", "Unknown boundary function J_L7: %s. Should be 'J_L7' or 'J_L7_corrected'.", J_L7_function.c_str());
		}

		return J7;
	}

	/**
	 * Computation of moumentum from Kinetic energy.
	 * \f[  pc = (\sqrt{ { \frac{K}{mc^2} + 1}^2 } - 1) \times mc^2 \f] 
	 */
	double pfunc(double K) {//, double mc2) {
		return sqrt( pow( K / VC::mc2 + 1, 2) - 1) * VC::mc2;
	}

	/**
	 * Computation of Kinetic energy from given momentum.
	 * \f[  K = (\sqrt{ 1 + ({\frac{pc}{mc^2}})^2 } - 1) \times mc^2 \f] 
	 */
	double Kfunc(double pc) {// , double mc2) {
		return ( sqrt( 1.0 + pow( pc / VC::mc2, 2)  ) - 1) * VC::mc2;
	}

	/**
	 * Interpolate outer boundary.
	 * This function used by VF::J_L7_corrected, VF::J_L7_corrected_0 ... The output is interpolation of the initial spectrum for specific point.
	 */
	double J_L7 (double K, double x1, double y1, double x2, double y2) {
		//double x1=0.2;
		//double y1=2.e3;
		//double x2=1.;
		//double y2=7.;
		//   x1=0.2
		//   y1=2.e3
		//   x2=1.
		//   y2=7.
		double bcoef = log(y1/y2)/(x2-x1);
		double acoef = y2*exp(bcoef*x2);
		return acoef*exp(-bcoef*K);
	}

	/**
	 * Specifying outer boundary, more accurate function. 
	 */
	double J_L7_corrected (double K) {
		double x[10] = {1e-5, 1e-4, 1e-3, 1e-2, 1e-1, 2e-1, 1e0, 3e0, 1e1, 2e1};
		double y[10] = {3e9,  2e7,  9e6,  3e5,  1e4,  2e3, 7e0, 3e-3, 1e-3, 1e-3};

		unsigned i=0;

		while (x[++i] < K && i < sizeof(x)/8-1);
		double result = J_L7 (K, x[i-1], y[i-1], x[i], y[i]);
		return result;
	}

	/**
	 * Specifying outer boundary, more accurate function, zero at infinity
	 */
	double J_L7_corrected_0 (double K) {
		double x[10] = {1e-5, 1e-4, 1e-3, 1e-2, 1e-1, 2e-1, 1e0, 3e0, 1e1, 1e2};
		double y[10] = {3e9,  2e7,  9e6,  3e5,  1e4,  2e3, 7e0, 3e-3, 1e-3, VC::zero_f};

		unsigned i=0;

		while (x[++i] < K && i < sizeof(x)/8-1);
		double result = J_L7 (K, x[i-1], y[i-1], x[i], y[i]);
		return result;
	}

	/** old version \f$ J = 8222.6 \times \exp(-7.068 \times K);  \f$ */
	double J_L7_old (double K) {
		double flux;

		flux=8222.6*exp(-7.068*K); // K(kinetic energy) in MeV

		return flux;
	}

	/**
	 * Calculating pc from L, &mu;, alpha.
	 * \f[ pc = \frac{\sqrt{ \mu \times 2 \times B(L) \times mc^2 }}{\sin(\alpha)} \f]
	 */
	double mu2pc (double L, double mu, double alpha) {
		return sqrt( mu * 2 * B(L) * VC::mc2 ) / sin(alpha);
		//return sqrt(mu*B(L)); // alpha == 90'
	}

	/**
	 * Calculating &mu; from L, pc, alpha.
	 * \f[ \mu = \frac{pc^2 \times \sin(\alpha)^2}{B(L) \times 2 \times mc^2 } \f]
	 */
	double pc2mu (double L, double pc, double alpha) {
		return pow(pc, 2) * pow(sin(alpha), 2) / (B(L) * 2 * VC::mc2);
	}

	/**
	 * Caltulating J*c from L, pc, &alpha;. 
	 * \warning Determined up to a constant. 
	 * \f[ J = 2 \times pc \times L \times Y(\alpha) \f]
	 */
	 double Jc_calc(double L, double pc, double Alpha) {
		return 2*pc*L*Y(Alpha);
	}

	//	double Jc2Alpha (double L, double pc, double Jc) {
	//		double a = 1;
	//		return sqrt(acos(Jc/2/0.7405/a/L/pc));
	//	}

	/**
	 * Getting f on specific energy. Linear interpolation of logarithm.
	 */
	double f_interp(double E, double f1, double E1, double f2, double E2) {
		double a, b, f_epc;
		if (E < E1 || E > E2) {
			return -1;
		}
		if (f1 <= 0 || f2 <= 0) return 0;
		b = log(f2/f1)/(E1 - E2);
		a = f1/exp(-b*E1);
		f_epc = a * exp(-b*E);
		return f_epc;
	}

	/**
	 * Caltulating &mu; from L, pc, &alpha;.
	 * \f[ \mu = \frac{pc^2}{mc^2} \times  \frac{ \sin(\alpha)^2 } { 2 \times B(L) }; \f]
	 */
	double mu_calc(double L, double pc, double Alpha) {
		return pow(pc,2)/VC::mc2 * pow(sin(Alpha),2) / (2 * B(L));
	}

	/**
	 * Chorus density model? (deprecated?)
	 * \f$ D = 10^{-0.3145L + 3.9043} \f$
	 */
	double density(double L) {
		return pow(10, (-0.3145*L + 3.9043));
	}

	/*
	 * Chorus density model?
	 */
	/*Matrix3D<double> density(Matrix3D<double> &L) {
		Matrix3D<double> tmp;
		tmp.AllocateMemory(L.size_x, L.size_y, L.size_z);
		for (int x = 0; x < tmp.size_x; x++)
			for (int y = 0; y < tmp.size_y; y++)
				for (int z = 0; z < tmp.size_z; z++)
					tmp[x][y][z] = pow(10, (-0.3145)*L[x][y][z] + 3.9043);
		return tmp;
	}*/

	/// Return maximum.
	double max(double v1, double v2) {
		return (v1>v2)?v1:v2;
	}

	/// Convert double to string
	string dtostr(double n) {
		std::ostringstream o;
		o << n;
		return o.str();
	}

	/// Small routine for checking if it is right iteration for the next output
	bool check_time(int iter, int d_iter) {
		if ((iter % d_iter) == 0) {
			return true;
		}
		return false;
	}
}
