/**
 * \file variousFunctions.h
 * \brief Various functions definition, see namespace VF for details
 *
 * \author Developed by Yuri Shprits
 */

#include "../Matrix/Matrix.h"
#include "variousConstants.h"

/**
 * \namespace VF
 * \brief Various functions
 *
 * This namespace presented in variousFunctions.h and variousFunctions.cpp.
 * - &alpha; is equatorial pitch-angle
 * - pc is momentum
 * - L - L-shell <br>
 */
namespace VF {

	/// \brief G-function
	/// \param alpha - &alpha;
	double G(double alpha);

	
	/// \brief Loss cone calculations
	/// \param L - L-shell
	double alc(double L);

	/// \brief Function for bounce time
	/// \param alpha - &alpha;
	double T(double alpha);

	/// \brief Y(&alpha;) function
	/// \param alpha - &alpha;
	double Y(double alpha);

	/** \brief Interpolation f on specific energy. 
	* \param E  - target E
	* \param f1 - left f(E1)
	* \param E1 - left E1
	* \param f2 - right f(E2)
	* \param E2 - right E2
	*/ 
	double f_interp(double E, double f1, double E1, double f2, double E2);

	/** \brief Caltulating &mu;
	* \param L - L-shell
	* \param pc - momentum
	* \param Alpha - &alpha; */
	double mu_calc(double L, double pc, double Alpha);

	/** \brief Caltulating J*c
	* \param L - L-shell
	* \param pc - momentum
	* \param Alpha - &alpha; */
	double Jc_calc(double L, double pc, double Alpha);

	/// \brief Chorus density model
	/// \param L - L-shell
	double density(double L);

	//Matrix3D<double> density(Matrix3D<double> &L);

	/// \brief Simple maximum
	/// \param v1 - first input double variable
	/// \param v2 - second input double variable
	double max(double v1, double v2);

	/// \brief Double to string.
	/// \param n - convert this variable to string
	string dtostr(double n);

	/// \brief Function determs is it right iteration for the next output.
	/// \param iter - base iteration to check
	/// \param d_iter  - second iteration to check
	bool check_time(int iter, int d_iter);

	
	// --------------------------------------------------------------

	/// \brief Electrostatic radial diffusion coefficient
	/// \param alpha - &alpha;
	/// \param Ke - Kinetic energy
	/// \param L - L-shell
	/// \param Kp - Kp-Index
	double Dfe (double alpha, double Ke, double L, double Kp);

	/// \brief Coulomb scattering electron lifetime
	/// \param L - L-shell
	/// \param energy - Energy
	double Coulomb(double L,double energy);

	/// \brief Empirical electron lifetime
	/// \param Kp - Kp-Index
	/// \param L - L-shell
	/// \param energy - Energy
	double Precipitation(double Kp,double L,double energy);
	// --------------------------------------------------------------

	/// \name Computation of dipole magnetic field
	// --------------------------------------------------------------
	
	/// \brief Dipole magnetic field
	/// \param Lparam - L-shell
	double B (double Lparam);
	
	// \deprecated Function defined in DiffusionCoefficient.h
	//double Df (double L, double Kp);


	/// \name Converting pc to energy
	// --------------------------------------------------------------

	/// \brief Computation of Kinetic energy from given momentum .
	/// \param pc - momentum
	double Kfunc(double pc);// , double mc2 = 0.511);

	/// \brief Computation of moumentum from Kinetic energy .
	/// \param K - Energy (MeV)
	double pfunc(double K);//, double mc2); 

	/// \name function for bounce
	// --------------------------------------------------------------
	
	/** \brief bounce time
	* \param L - L-shell
	* \param pc - momentum
	* \param alpha - &alpha; */
	double bounce_time_new(double L, double pc, double alpha);

	///\name Specifying outer boundary
	// --------------------------------------------------------------

	/** \brief Call various flux model of the outer boundary
	* \param E7 - Energy in MeV at L=7
	* \param J_L7_function - name of boundary flux function  */
	double Outer_Boundary_Flux_at_L7(double E7,string J_L7_function);

	/// \brief Outer boundary spectrum (accurate)
	/// \param K - Energy
	double J_L7_corrected (double K);

	/// \brief Outer boundary spectrum (old)
	/// \param K - Energy
	double J_L7_old (double K);

	/// \brief Interpolate spectrum
	/// \param K - Energy
	/// \param x1 - Left energy point
	/// \param y1 - Left flux point
	/// \param x2 - Right energy point
	/// \param y2 - Right flux point
	double J_L7 (double K, double x1, double y1, double x2, double y2);

	/// \brief Outer boundary spectrum (accurate with 0 end the end)
	/// \param K - Energy
	double J_L7_corrected_0 (double K);


	/// \name Converting &mu; to pc
	// --------------------------------------------------------------
	
	/** \brief Convert &mu; to pc
	* \param L - L-shell
	* \param mu - &mu;
	* \param alpha - &alpha; */
	double mu2pc (double L, double mu, double alpha);
	/** \brief Convert pc to &mu;
	* \param L - L-shell
	* \param pc - momentum
	* \param alpha - &alpha; */
	double pc2mu (double L, double pc, double alpha);
}
